/* common.h - contains defines and should be included everywhere.

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifndef COMMON_H
#define COMMON_H

/* System include files */

#include <stdio.h>		/* printf, FILE */
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */
#include <errno.h>
#include <sys/types.h>		/* *_t */
#ifdef WITH_PTHREAD
#include <pthread.h>
#endif
#include <string.h>		/* strcasecmp */
#include <stddef.h>		/* #define offsetof */

#include <glib.h>

#ifdef HAVE_ASSERT
#include <assert.h>
#else
#ifndef assert
#define assert(arg)  \
if (0 == (long)(arg)) \
{ \
  (void)fprintf(stderr, "Assertion Failed in %s at line %d\n", \
  __FILE__, __LINE__ ); \
  (void)fflush(stderr); \
  (void)abort(); \
}
#endif
#endif

/* Needed for DEC Alpha, Digital Unix */
#ifdef TGL_ISALPHA
typedef unsigned char u_int8_t;
typedef unsigned short u_int16_t ;
typedef unsigned int u_int32_t ;
typedef unsigned long long u_int64_t ;
typedef char int8_t ;
typedef short int16_t ;
typedef int int32_t ;
typedef long long int64_t ;
typedef u_int8_t uint8_t;
typedef u_int16_t uint16_t;
typedef u_int32_t uint32_t;
typedef u_int64_t uint64_t;
#endif

/* DENTS include files */
#include "setup.h"		/* Bring in the compile time #define's */

/* Just because it's so common */
#include "err.h"

/* Since autoconf doesn't set a variable */
/*
   #ifndef HAVE_STRINGIZE
   #error Compiler does not support stringification
   #endif
 */

/* Kludge since insure++ redefines all of the types to 4 byte values */
/*  VERY weird. Thanks to Greg Rumple */
#ifdef __INSURE__
typedef unsigned char __insure_guint8;
typedef unsigned short __insure_guint16;
typedef unsigned long __insure_guint32;
typedef unsigned long long __insure_u_int64_t;

#define guint8        __insure_guint8
#define guint16       __insure_guint16
#define guint32       __insure_guint32
#define u_int64_t       __insure_u_int64_t
#endif

/* Some common typedefs */
typedef int (*func_i_v) (void);
typedef void (*func_v_v) (void);

/*                                                                              
   ** This is shamelessly "borrowed" from the ircd source code
   ** Not all systems define it, so if they don't, we will - jerdfelt
   **
   ** 'offsetof' is defined in ANSI-C. The following definition
   ** is not absolutely portable (I have been told), but so far
   ** it has worked on all machines I have needed it. The type
   ** should be size_t but...  --msa
 */
#ifndef offsetof
#define offsetof(t,m) (int)((&((t *)0L)->m))
#endif

/* Some commonly used macro's */
#define stricmp strcasecmp
#define strnicmp strncasecmp
#define strncpyzt(d, s, l)			\
	do {					\
	  strncpy((char *)(d), (char *)(s), l);	\
	  (d)[l]=0;				\
	} while(0)
#define strncatzt(d, s, l)			\
	do {					\
	  strncat((char *)(d), (char *)(s), l-strlen((char *)(d))); \
	  (d)[l]=0;				\
	} while(0)

/* If we have debugging compiled in, display the messages, else noop them */
#ifdef DENTS_DEBUG
#define DPRINTF(x) do { printf x; } while (0)
#else
#define DPRINTF(x) do { ; } while(0)
#endif

/*
 * Generic mutex locking routines, prints errors when things go wrong
 */
#ifdef WITH_PTHREAD

#define LOCK(func, name) \
  do {			\
    int rc;		\
    if ((rc = pthread_mutex_lock(&name)) != 0)	\
      err_fatal("%s: couldn't lock mutex @ %p: %m", func, &name);	\
  } while(0)

#define UNLOCK(func, name) \
  do {			\
    int rc;		\
    if ((rc = pthread_mutex_unlock(&name)) != 0)	\
      err_fatal("%s: couldn't unlock mutex @ %p: %m", func, &name);	\
  } while (0);

#else

#define LOCK(func, name) do { ; } while(0)
#define UNLOCK(func, name) do { ; } while(0)

#endif

/*
 * Generic macro's to put and get 16 and 32 bit integers (endian independant)
 *
 * GRABINT{16,32} will grab an integer pointed to by pointer p and put into
 *  interger x
 * PUTINT{16,32} will put integer x into the memory pointed to by pointer p
 * PUTINTP{16,32} will put an integer pointed to by pointer x into the memory
 *  pointed to by pointer p
 *
 * I wonder if people buy computers because htons and htonl are noop's on
 *  their architecture?
 */
#define GRABINT16(p, x)						\
	do {							\
	  guint16 tempu16;					\
	  memcpy((void *)&tempu16, p, sizeof(guint16));	\
	  p += sizeof(guint16);				\
	  x = ntohs(tempu16);					\
        } while(0)
#define GRABINT32(p, x)						\
	do {							\
	  guint32 tempu32;					\
	  memcpy((void *)&tempu32, p, sizeof(guint32));	\
	  p += sizeof(guint32);				\
	  x = ntohl(tempu32);					\
	} while(0)

#define PUTINT16(p, x)						\
	do {							\
	  guint16 tempu16;					\
	  tempu16 = htons(x);					\
	  memcpy(p, (void *)&tempu16, sizeof(guint16));	\
	  p += sizeof(guint16);				\
	} while(0)
#define PUTINT32(p, x)						\
	do {							\
	  guint32 tempu32;					\
	  tempu32 = htonl(x);					\
	  memcpy(p, (void *)&tempu32, sizeof(guint32));	\
	  p += sizeof(guint32);				\
	} while(0)

/* Big endian is easy because it's same as network byte order */
#if (G_BYTE_ORDER == G_BIG_ENDIAN)

#define PUTINTP16(p, x)						\
	do {							\
	  memcpy((void *)p, (void *)x, sizeof(guint16));	\
	  p += sizeof(guint16);				\
	} while(0)
#define PUTINTP32(p, x)						\
	do {							\
	  memcpy((void *)p, (void *)x, sizeof(guint32));	\
	  p += sizeof(guint32);				\
	} while(0)

#elif (G_BYTE_ORDER == G_LITTLE_ENDIAN)

#define PUTINTP16(p, x)						\
	do {							\
	  guint16 tempu16, tempu16b;				\
	  memcpy((void *)&tempu16b, (void *)x, sizeof(guint16)); \
	  tempu16 = htons(tempu16b);				\
	  memcpy((void *)p, (void *)&tempu16, sizeof(guint16)); \
	  p += sizeof(guint16);				\
	} while(0)
#define PUTINTP32(p, x)						\
	do {							\
	  guint32 tempu32, tempu32b;				\
	  memcpy((void *)&tempu32b, (void *)x, sizeof(guint32)); \
	  tempu32 = htonl(tempu32b);				\
	  memcpy((void *)p, (void *)&tempu32, sizeof(guint32)); \
	  p += sizeof(guint32);				\
	} while(0)

#endif /* G_BYTE_ORDER == G_LITTLE_ENDIAN */

#endif
