/* link.h - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

/*
 * link.h
 *
 *  Implements some generic routines for managing linked list's. Also
 *  implements a generic block allocator for performance.  These routines
 *  are also threadsafe.
 *
 * There are two types of lists handled by this library.
 *
 *  A linkedlist is a doubly linked list.   You are responsible for allocating
 *    entries in the list.
 *  A blocklist is built on top of the linkedlist.  Use 
 *    block_allocandadd() or block_allocandaddtolink() to get memory.
 *    Performance gains are made by allocating chunks of entries at a time
 *
 *
 * Generic Routines (or macros):
 *
 * entry = link_first(head);        # returns the first entry in the list 
 * entry = 
 *   link_next(entry, link_member); # returns the next entry in the list 
 * entry =
 *   link_prev(entry, link_member); # returns the prev entry in the list
 *
 *
 * linkedlist only:
 * 
 * head = LINK_INITIALIZER(structname, llentrystructmember)
 *   or
 * link_reset(&head, offsetof(structname, llentrystructmember));
 *                                  # Initialize a linkedlist structure
 * link_add(head, llentry)          # add an entry to the end of the list
 * link_delete(head, llentry)       # delete an entry from the list
 *
 *
 *
 * blocklist only:
 *
 * blocklisthead = BLOCK_INITIALIZER(structname, chunksize)
 *   or
 * block_setup (&blocklisthead, sizeof(structname), chunksize);
 *                                  # Initialize a blockalloc list structure
 * entry = 
 *   block_allocandadd(blockhead)   # Add an entry to a blockedlist 
 *                                  # returns a pointer to a place for memory
 * entry =
 *   block_allocandaddtolink (blockhead, llentry) 
 *                                  # Allocate a record from the pool and add
 *                                  # it to a (different) linked list
 * block_delete (blockhead, llentry)
 *                                  # remove an entry from the list
 *                                  # and add the entry back to the free pool
 * 
 * Quick tutorial on how to use the Linked List API can be found
 * in the main() of the test driver in link.c
 *
 */

#ifndef LINK_H
#define LINK_H

#include "common.h"

#include <string.h>		/* memset */

/*
 * Generic Linked List Entry structure
 */
typedef struct llentry
{
  struct llentry *next, *prev;
}
llentry;

/*
 * Linked List
 */
typedef struct
{
  void *begin, *end;		/* Beginning and end */

  guint16 num;			/* Number of entries in the linked list */

  guint16 offset;		/* Offset of llentry from beginning of */
  /*  structure */
#ifdef WITH_PTHREAD
  pthread_mutex_t lock;		/* Mutex */
#endif
}
linkedlist;

/*
 * Generic structure used for Linked List's.
 *
 * Anything can use this, but the first 2 entries of the structure, MUST BE
 * a pointer for next and prev:
 *
 * void *next, *prev;
 */
typedef struct
{
  guint32 structsize;		/* How much mem to allocate for one struct */
  guint16 numalloc;		/* Number of entries to allocate at once */

  linkedlist used;		/* Used linked list */
  linkedlist free;		/* Free linked list */

#ifdef WITH_PTHREAD
  pthread_mutex_t lock;		/* Mutex */
#endif
}
blocklist;

/*
 * Macro/functions
 */
#if 0
#define link_first(l, s, e)			\
	((l->begin == NULL) ? l->begin :	\
	(void *)((char *)l->begin - ((char *)&(e) - (char *)(s))))
#endif

#define block_allocandadd(b) block_allocandaddtolink(b, &(b)->used)
#define link_first(l)				\
	(((l)->begin == NULL) ? (l)->begin :	\
	(void *)((char *)(l)->begin - (l)->offset))
#define link_next(s, e)			\
	((e.next == NULL) ? e.next :	\
	(void *)((char *)e.next - ((char *)&(e) - (char *)(s))))
#define link_prev(s, e)			\
	((e.prev == NULL) ? e.prev :	\
	(void *)((char *)e.prev - ((char *)&(e) - (char *)(s))))

#ifdef WITH_PTHREAD

#if 0
#define LINK_INITIALIZER				\
    { NULL, NULL, 0, PTHREAD_MUTEX_INITIALIZER }
#endif

#define LINK_INITIALIZER(s, f)				\
    { NULL, NULL, 0, offsetof(s, f), PTHREAD_MUTEX_INITIALIZER }
#define BLOCK_INITIALIZER(t, num)			\
  { sizeof(t), num,					\
    LINK_INITIALIZER(t, ll), LINK_INITIALIZER(t, ll),	\
  PTHREAD_MUTEX_INITIALIZER }

#else /* WITH_PTHREAD */

#if 0
#define LINK_INITIALIZER { NULL, NULL, 0 }
#endif

#define LINK_INITIALIZER(s, f)				\
    { NULL, NULL, 0, offsetof(s, f)}
#define BLOCK_INITIALIZER(t, num)			\
  { sizeof(t), num, LINK_INITIALIZER(t, ll), LINK_INITIALIZER(t, ll) }

#endif

/*
 * Prototypes
 */
/* link.c */
void link_add (linkedlist * link, llentry * entry);
void *block_allocandaddtolink (blocklist * block, linkedlist * link);
void link_delete (linkedlist * link, llentry * entry);
void block_delete (blocklist * block, llentry * entry);
void link_reset (linkedlist * link, guint16 offset);
void block_setup (blocklist * block, guint16 structsize, guint16 numalloc);

#endif
