/* mod_frl.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/*
 * mod_frl.c
 *
 *  Module for computed forward/reverse lookups
 *
 * FIXME: Hardcoding domain names, etc is lame. Config file support needed!
 */
#include <string.h>
#include <ctype.h>

#include "dns.h"
#include "addr_tools.h"

/* Basic info on the module, used later */
#define MODULE_NAME	mod_frl

#define MODULE_VER_MAJOR 0
#define MODULE_VER_MINOR 0
#define MODULE_VER_MICRO 1

#include "module_syms.h"
#include "dns_module_syms.h"

#define DOMAIN_NAME "frl.test."
#define INADDR_DOMAIN ".159.69.207"

#define DEFAULT_TTL (60 * 60 * 24)

/*
 * mod_frl_answer
 *
 *  Called when we've received a packet from a client for a zone we're
 * answering for
 *
 * user-37kb7vl.dialup.mindspring.com
 */
int
mod_frl_answer (dnspktrec * dnsqpkt, dnsbuildrec * build)
{
#define ROUTINE "mod_frl_answer ()"
  guchar name[DNS_MAXNAMELENGTH + 1];
  guchar basestr[] = "0123456789abcdefghijklmnopqrstuv";
  guchar *p, *p1, *addr_string, ipv4_addr[16], *rdata_ptr;
  guint32 l = 0, ip_addr = 0;
  int i;

  DPRINTF (("mod_frl_dns_answer: looking up: %s\n", dnsqpkt->query.name));

  /* We don't know anything other than the Internet class */
  if ((dnsqpkt->query.class != DC_IN) && (dnsqpkt->query.class != DC_ANY))
    return (DE_NXDOMAIN);

  /* Copy over the name */
  strncpyzt (name, dnsqpkt->query.name, DNS_MAXNAMELENGTH);

  rdata_ptr = NULL;
  /* We handle A and PTR types, everything else we don't know about */
  switch (dnsqpkt->query.type) {
  case DT_A:
    {

      /* Strip off the domain and period */
      name[strlen (name) - strlen (dnsqpkt->qru.query.zone->zonename) - 1] =
	0;

      /* Must start with user- */
      if (strnicmp (name, "user-", 5)) {
	DPRINTF (
		 ("%s: this module expects user- to be the prefix\n",
		  ROUTINE));
	return (DE_NXDOMAIN);
      }

      /* It can be at most 12 characters */
      if (strlen (name) > 12) {
	DPRINTF (("%s: name is '%s' (not 12 characters)\n", ROUTINE, name));
	return (DE_NXDOMAIN);
      }

      /* Go through the last 7 characters */
      for (i = 0, p = name + 5; i < 7; i++, p++) {
	/* If it's not a known character, it's an invalid name */
	if ((p1 = strchr (basestr, *p)) == NULL) {
	  DPRINTF (("%s: found an invalid character at position %d\n",
		    ROUTINE, i));
	  return (DE_NXDOMAIN);
	}

	l = (l * 32) + (p1 - basestr);
      }

      ip_addr = htonl (l);
      ipv4_addr[0] = 0;

      addr_string = at_inet4_ntop ((guint8 *) & ip_addr, ipv4_addr,
				   sizeof (ipv4_addr));

      if (!addr_string) {
	DPRINTF (("%s: Invalid IPv4 address!\n", ROUTINE));
	return (DE_NXDOMAIN);
      }

      if (dns_types_pack_A (build, &rdata_ptr, addr_string) == -1) {
	DPRINTF (("%s: Couldn't pack address %s\n", ROUTINE, addr_string));
	return (DE_NXDOMAIN);
      }

      /* Then record it */
      dns_build_add (build, DNS_BUILD_ANSWER, name, DEFAULT_TTL, DC_IN, DT_A,
		     rdata_ptr);
    }
    break;

  case DT_PTR:
    {
      int byte0, byte1, byte2, byte3;
      char name_buf[DNS_MAXNAMELENGTH];

      DPRINTF (("%s: converting %s into an IP addr\n", ROUTINE, name));

      /* name currently looks like 000.111.222.333.IN-ADDR.ARPA */
      /* in reverse order (333 is the highest order byte)      */
      if (!isdigit (name[0])) {
	DPRINTF (("%s: Expected name to start with a digit\n", ROUTINE));
	return (DE_NXDOMAIN);
      }

      byte0 = byte1 = byte2 = byte3 = -1;

      if (sscanf
	  (name, "%03d.%03d.%03d.%03d.", &byte0, &byte1, &byte2, &byte3) != 4) {
	DPRINTF (
		 ("%s: Expected reverse dotted decimal notation IP addr at the front of the name\n",
		  ROUTINE));
	return (DE_NXDOMAIN);
      }

      if ((byte0 < 0 || byte0 > 255)
	  || (byte1 < 0 || byte1 > 255)
	  || (byte2 < 0 || byte2 > 255)
	  || (byte3 < 0 || byte3 > 255)) {
	DPRINTF (("%s: Expected all portions to be in range 0-255\n",
		  ROUTINE));
	return (DE_NXDOMAIN);
      }

      l = (guchar) byte0 + ((guchar) byte1 << 8) + ((guchar) byte2 << 16) +
	((guchar) byte3 << 24);

      strncpyzt (name_buf, "user-", DNS_MAXNAMELENGTH);
      p = name_buf + 5;
      for (i = 0; i < 7; i++) {
	p[6 - i] = basestr[l % 32];
	l /= 32;
      }
      name_buf[13] = 0;

      /* FIXME: when we get config file support, take out hard coded
         domain name!
       */
      if (dns_types_pack_PTR (build, &rdata_ptr, DOMAIN_NAME, name_buf) == -1) {
	DPRINTF (("%s: couldn't pack domain name %s (%s) for PTR lookup\n",
		  ROUTINE, DOMAIN_NAME, name_buf));
	return (DE_NXDOMAIN);
      }

      /* Then record it */
      dns_build_add (build, DNS_BUILD_ANSWER, name, DEFAULT_TTL, DC_IN,
		     DT_PTR, rdata_ptr);
      break;
    }
  default:
    return (DE_NXDOMAIN);
  }

  return (DE_NOERROR);
#undef ROUTINE
}

/*
 * Exported functions
 */

/*
 * module_dns_zone_add
 *
 *  Add's a zone
 */
MODULE_FUNC (int, dns_zone_add, (dnszonerec * zone))
{
  DPRINTF (("mod_frl_dns_zone_add: Added %s\n", zone->zonename));

  return (0);
}

/*
 * module_dns_zone_delete
 *
 *  Delete's a zone
 */
MODULE_FUNC (int, dns_zone_delete, (dnszonerec * zone))
{
  printf ("mod_frl_dns_zone_delete: Called\n");

  return (DE_NOTIMP);
}

/*
 * No module_dns_zone_load
 */

/*
 * No module_dns_zone_reload
 */

/*
 * module_init
 *
 *  All we do is just print a banner
 */
MODULE_FUNC (int, init, (void))
{
  DPRINTF (("mod_frl_init: version %d.%d.%d by Johannes Erdfelt\n",
	    MODULE_VER_MAJOR, MODULE_VER_MINOR, MODULE_VER_MICRO));

  return (0);
}

/*
 * No module_destroy
 */

/*
 * Module compile support code
 */

/*
 * Define some symbols
 */
MODULE_DNS_ANSWER = {
mod_frl_answer, NULL};

/*
 * Exported symbols
 */
MODULE_SYMS (MODULE_SYMS_INIT MODULE_SYMS_DNS)
