/* mod_msg_tgl.c - a message-passing module allowing external processes
 * to answer DNS queries.
 * 
 * by Todd Lewis, Copyright 1999.
 * 
 * Taken from mod_mysql_eza.c 
 * Copyright (C) 1999  Eric Z. Ayers  (Eric.Ayers@mindspring.com)
 * 
 * Dents - a threaded modular dns-server.
 * Copyright (C) 1999  Dents Team  (http://www.dents.org)
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.  This program is distributed in the
 * hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.  
 * 
 * See DESCRIPTION for more information about using this module.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define WITH_MSG_TGL
#ifdef WITH_MSG_TGL

/* Basic info on the module, used later */
/* the following macros must be defined before including moudle_syms.h */

#define MODULE_NAME		mod_msg_tgl
#define MODULE_NAME_STRING	"mod_mysg_tgl"

#define MODULE_VER_MAJOR 0
#define MODULE_VER_MINOR 0
#define MODULE_VER_MICRO 1

/***************************************************************************
 *			Include Files
 ***************************************************************************/
/* UNIX includes */
#include <stdio.h>		/* printf, etc. */
#include <unistd.h>		/* exec, fork */
#include <sys/time.h>		/* select */
#include <sys/types.h>		/* select */
#include <glib.h>
#include <common.h>

/* DENTS includes */
#include "dns.h"
#include "module_syms.h"
#include "dns_module_syms.h"
#include "link.h"
#include "addr_tools.h"

#include "mod_msg_tgl.h"
#include "pipeutil.h"
#include "stoi.h"

/***************************************************************************
 *			Macro Definitions
 ***************************************************************************/

#define NOT_IMPL(arg) DPRINTF(("%s: NOT IMPLEMENTED %s", ROUTINE, (arg)))


/***************************************************************************
 *			Static Module Data
 ***************************************************************************/

/* This is here for early development.  I hope soon to remove this and
 * let the per-dproc locking take effect. */
GMutex *global_mutex;

/* This is the master list of driver processes.  Used to find the responsible
 * dproc when a query comes in and to see if the requested dproc already
 * exists when a new zone is requested to be loaded. */
static dproc *dprocs;

/***************************************************************\
 **			Internal Functions                    **
\***************************************************************/

static  gboolean
is_finis(char **answer)
{
    gboolean rval = TRUE;

    if (!answer) {
	return (FALSE);
    }
    if (pipeutil_count_elems(answer) != 2) {
	rval = FALSE;
    }
    else if (strcmp("F", answer[1])) {
	rval = FALSE;
    }
    return (rval);
}

static  gboolean
is_err(char **answer)
{
    if (!answer) {
	return (FALSE);
    }
    if (pipeutil_count_elems(answer) < 3) {
	return (FALSE);
    }
    if (strcmp("E", answer[1])) {
	return (FALSE);
    }
    return (TRUE);
}

static  gboolean
is_soa(char **answer)
{
    if (!answer) {
	return (FALSE);
    }
    if (pipeutil_count_elems(answer) != 10) {
	return (FALSE);
    }
    else if (strcmp("SOA", answer[1])) {
	return (FALSE);
    }
    return (TRUE);
}

/* returns 0 on success, -1 if answer is not an SOA message, -2 if the domain
 * is not found in the dproc, -10 through -19 on out of memory, -20 through
 * -29 if a numerical parameter can not be scanf'd.  On error, all original
 * SOA values are retained. */
int
remember_soa(char **answer, dproc * dp)
{
    dproc_zone *c = NULL;
    char   *mkeeper = NULL;
    char   *rkeeper = NULL;
    u_int32_t ikeeper[5];
    int     rval = 0;
    GList  *gl;

    fprintf(stderr, "remember_soa() called!\n");
    /* Is this an SOA? */
    if (is_soa(answer) != TRUE) {
	fprintf(stderr, "remember_soa() called on a message not SOA!\n");
	fflush(stderr);
	return (-1);
    }

    /* Where is the dproc_zone struct? */
    if (!(c = find_dz_by_name(answer[2], dp))) {
	fprintf(stderr, "remember_soa() called, but domain %s not found!\n", answer[2]);
	fflush(stderr);
	return (-2);
    }

    mkeeper = c->soa_mname;
    rkeeper = c->soa_rname;
    ikeeper[0] = c->soa_serial;
    ikeeper[1] = c->soa_refresh;
    ikeeper[2] = c->soa_retry;
    ikeeper[3] = c->soa_expire;
    ikeeper[4] = c->soa_minimum;

    if (!(c->soa_mname = strdup(answer[3]))) {
	rval = -10;
    }
    else if (!(c->soa_rname = strdup(answer[4]))) {
	rval = -11;
    }
    else if (!(sscanf(answer[5], "%i", &c->soa_serial))) {
	rval = -20;
    }
    else if (!(sscanf(answer[6], "%i", &c->soa_refresh))) {
	rval = -21;
    }
    else if (!(sscanf(answer[7], "%i", &c->soa_retry))) {
	rval = -22;
    }
    else if (!(sscanf(answer[8], "%i", &c->soa_expire))) {
	rval = -23;
    }
    else if (!(sscanf(answer[9], "%i", &c->soa_minimum))) {
	rval = -24;
    }

    if (rval) {
	/* Error; roll back */
	if (c->soa_mname != mkeeper) {
	    free(c->soa_mname);
	}
	if (c->soa_rname != rkeeper) {
	    free(c->soa_rname);
	}
	c->soa_mname = mkeeper;
	c->soa_rname = rkeeper;
	c->soa_serial = ikeeper[0];
	c->soa_refresh = ikeeper[1];
	c->soa_retry = ikeeper[2];
	c->soa_expire = ikeeper[3];
	c->soa_minimum = ikeeper[4];
    }
    else {
	/* free the now-unused strings from the previous setting. */
	if (mkeeper) {
	    free(mkeeper);
	}
	if (rkeeper) {
	    free(rkeeper);
	}
    }
    return (rval);
}

gboolean
is_options(char **answer)
{
    if (!answer) {
	return (FALSE);
    }
    if (strcmp(answer[1], "O")) {
	return (FALSE);
    }
    return (TRUE);
}

/* sets options for that zone; returns number of options which experienced
 * an error. */
int
remember_options(char **answer, dproc * dp)
{
    dproc_zone *c = NULL;
    int     i, rval = 0;

    if (is_options(answer) != TRUE) {
	fprintf(stderr, "remember_options() called on a message not options!\n");
	fflush(stderr);
	return (-1);
    }
    /* Where is the dproc_zone struct? */
    if (!(c = find_dz_by_name(answer[2], dp))) {
	fprintf(stderr, "remember_options() called, but domain %s not found!\n", answer[2]);
	fflush(stderr);
	return (-2);
    }

    c->options = 0;
    for (i = 3; i < pipeutil_count_elems(answer); i++) {
	if (!(strcmp(answer[i], "NUMERIC"))) {
	    c->options |= DPROC_NUMERIC;
	}
	else if (!(strcmp(answer[i], "EXTENDEDQ"))) {
	    c->options |= DPROC_EXTENDEDQ;
	}
	else {
	    fprintf(stderr, "remember_options() called for %s; unrecognized option, %s.\n", dp->name, answer[i]);
	    fflush(stderr);
	    rval++;
	}
    }
    return (rval);
}

/***************************************************************\
 **			Exported Functions                    **
\***************************************************************/

/*
 * mod_msg_tgl_answer
 *
 * Description: This function takes a query packet, queries the
 *              mysql database, and creates a reply in the build
 *              structure.
 * 
 * Input:       dnsqpkt - Query packet
 * Output:      build   - Response to return
 * 
 */
int
mod_msg_tgl_answer(dnspktrec * dnsqpkt, dnsbuildrec * build)
{
#define ROUTINE "module_msg_tgl_answer()"
    char    sbuf[255];
    dproc  *dp = NULL;
    dproc_zone *dz = NULL;
    char  **answer;
    char   *zone_name;
    int     i, j, error_return = DE_NXDOMAIN, ttl, type, class, section = DNS_BUILD_ANSWER,
            opos;
    guint16 pri;
    guchar *rdata_buf;		/* buffer to store RDATA for RR */
    struct timeval tv;
    char    ipbuf[16];
    char   *k, *v;

    zone_name = dnsqpkt->qru.query.zone->zonename;

    fprintf(stderr, "%s called for %s.\n", ROUTINE, zone_name);
    dp = dnsqpkt->qru.query.zone->mdata;
    dz = find_dz_by_name(zone_name, dp);

    /* First, we send the query message to the client, modulo his options. */
    if (dz->options & DPROC_EXTENDEDQ) {
	if (dz->options & DPROC_NUMERIC) {
	    g_snprintf((char *) &sbuf, 255, "%u EQ %s %u %u %u %s %u %u %u\n", \
		       dp->snum++, dnsqpkt->query.name, dnsqpkt->query.type, \
		       dnsqpkt->query.class, dnsqpkt->id, \
		       at_inet4_ntop((guchar *) & dnsqpkt->ip.s_addr, (char *)&ipbuf, sizeof(ipbuf)), \
		   dnsqpkt->port, dnsqpkt->flags.opcode, dnsqpkt->flags.rd);
	}
	else {
	    g_snprintf((char *) &sbuf, 255, "%u EQ %s %s %u %u %s %u %u %u\n", \
		       dp->snum++, dnsqpkt->query.name, itos(dnsqpkt->query.type, stoi_rrtypes), \
		     itos(dnsqpkt->query.class, stoi_ctypes), dnsqpkt->id, \
		       at_inet4_ntop((guchar *) & dnsqpkt->ip.s_addr, (char *)&ipbuf, sizeof(ipbuf)), \
		   dnsqpkt->port, dnsqpkt->flags.opcode, dnsqpkt->flags.rd);
	}
    }
    else if (dnsqpkt->query.class != 1 && dnsqpkt->query.class != 255) {
	/* dproc did not request extended queries, and this is
	 * a strange class, so since that class is not implemented
	 * in this case, return NOTIMP. */
	return (DE_NOTIMP);
    }
    else {
	if (dz->options & DPROC_NUMERIC) {
	    g_snprintf((char *) &sbuf, 255, "%u Q %s %u\n", \
		       dp->snum++, dnsqpkt->query.name, dnsqpkt->query.type);
	}
	else {
	    g_snprintf((char *) &sbuf, 255, "%u Q %s %s\n", \
		       dp->snum++, dnsqpkt->query.name, itos(dnsqpkt->query.type, stoi_rrtypes));
	}
    }
    write(dp->wfile, (void *) &sbuf, strlen((char *) &sbuf));

    /* The we get back a bunch of messages from the client and process them. */
    while (1) {
	tv.tv_sec = 1;
	tv.tv_usec = 0;

	answer = pipeutil_getline(dp, &tv);
	if (!answer) {
		error_return = 2;
	    break;
	} else {
		fprintf(stderr, "Got %s from dproc.\n", answer[1]);
		fflush(stderr);
	}

	ttl = dz->soa_minimum;
	type = 0;
	class = 1;

	if (is_finis(answer)) {
	    break;
	}
	else if (is_options(answer) == TRUE) {
	    remember_options(answer, dp);
	}
	else if (is_soa(answer)) {
	    remember_soa(answer, dp);
	    ttl = dz->soa_minimum;
	}
	else if (is_err(answer) == TRUE) {
	    /* FIXME: log what we get here. */
	    sscanf(answer[2], "%u", &i);
		fprintf(stderr, "dproc %s returned an error of %s.\n", dp->name, answer[2]);
		fflush(stderr);
	    if (i > 1 && i < 5) {
		/* It's an RFC1035:4.1.1 rcode */
		return (i);
	    }
	    else {
		return (2);
	    }
	}
	else {
	    /* It's an answer, so let's process it. */
	    error_return = 0;
	    if (strcmp("A", answer[1])) {
		/* Answers must be answers. */
		fprintf(stderr, "%s: Operator %s != A!  Breaking.\n", ROUTINE, answer[1]);
		break;
	    }
	    i = pipeutil_count_elems(answer);
	    if (i < 5) {
		/* Answers must have at least 5 elems. */
		fprintf(stderr, "%s: Received answer from %s, num elems (%i) < 5!  Breaking.\n", ROUTINE, dp->name, i);
		break;
	    }
	    /* Ok, we have a good answer candidate. */

	    if (*answer[3] < 48 || *answer[3] > 57) {
		/* it's textual; convert. */
		type = stoi(answer[3], stoi_rrtypes);
	    }
	    else {
		if ((sscanf(answer[3], "%d", &type) != 1) || (type < 1 || type > 255)) {
		    err_warn("%s: %s; zone %s: type field not valid (%s) expected integer between 1 & 255 for query %s.\n",
			     MODULE_NAME_STRING, ROUTINE, zone_name,
			     answer[3],
			     (char *) &sbuf);
		    error_return = DE_NXDOMAIN;
		    break;
		}
	    }
		fprintf(stderr, "Switching on type %u from %s.\n", type, answer[3]);

	    switch (type) {
	    case DT_A:
		/* FIXME: addition of <section> throws everything off by 1. */
		/* FORMAT: <seq#> A <name> <type> <data> (<options>) */
		/* E.g.: 1 A www.todd.com. A 188.134.10.39 ttl=14625 */
		{
		    if (dns_types_pack_A(build, &rdata_buf, answer[4]) == -1) {
			err_warn("%s: %s; zone %s IPv4 address for A Record (%s) not valid.\n",
				 MODULE_NAME_STRING, ROUTINE, zone_name,
				 answer[4]);

			error_return = DE_NXDOMAIN;
			break;
		    } else {
			fprintf(stderr, "%s: Packed address: %s\n", ROUTINE, answer[4]);
		    }
		    opos = 5;
		    break;	/* From case, a true language turd. */
		}

	    case DT_CNAME:
	    case DT_PTR:
	    case DT_NS:
	    case DT_MD:
	    case DT_MF:
	    case DT_MB:
	    case DT_MG:
	    case DT_MR:

		{
		    /* TYPE: NS|CNAME|PTR|MD|MF|MB|MG|MR
		     * INHERITS FROM: answer
		     * FORMAT: <seq#> A <name> <type> <name> (<options>) */
		    if (i < 5) {
			err_warn("Mis-formatted message from %s.\n", dp->name);
			error_return = DE_NXDOMAIN;
			break;
		    }
		    if (dns_types_pack_1_domain_name(build, &rdata_buf, zone_name, answer[4]) == -1) {
			err_warn("%s: %s; zone %s domain for for %s Record (%s) not valid.\n",
			  MODULE_NAME_STRING, ROUTINE, zone_name, answer[5],
				 answer[4]);

			error_return = DE_NXDOMAIN;
			break;
		    }
		    opos = 5;
		    break;
		}

	    case DT_TXT:
	    case DT_NSAP:
		{
		    /* FORMAT: <seq#> A <name> <type> <string> (<options>) */
		    if (i < 5) {
			err_warn("Mis-formatted message from %s.\n", dp->name);
			error_return = DE_NXDOMAIN;
			break;
		    }
		    if (dns_types_pack_1_character_string(build, &rdata_buf, answer[4]) == -1) {
			err_warn("%s: %s; zone %s domain for for %s Record (%s) not valid.\n",
			  MODULE_NAME_STRING, ROUTINE, zone_name, answer[5],
				 answer[4]);

			error_return = DE_NXDOMAIN;
			break;
		    }
		    opos = 5;
			break;
		}
	    case DT_MX:
	    case DT_AFSDB:
		{
		    /* FORMAT: <seq#> A <name> MX <prio> <exchange> (<options>) */
		    /* FORMAT: <seq#> A <name> AFSDB <prio> <exchange> (<options>) */
		    if (i < 6) {
			err_warn("Mis-formatted MX|AFSDB message from %s (%s, %u<6).\n", dp->name, answer[3], i);
			error_return = DE_NXDOMAIN;
			break;
		    }
		    sscanf(answer[4], "%hu", &pri);
		    if (dns_types_pack_uint16_domain(build, &rdata_buf, pri, zone_name, answer[5]) == -1) {
			err_warn("%s: %s; zone %s domain for for %s Record (%s) not valid.\n",
			  MODULE_NAME_STRING, ROUTINE, zone_name, answer[5],
				 answer[5]);

			error_return = DE_NXDOMAIN;
		    }
		    opos = 6;
		    break;
		}
	    case DT_SOA:
		{
		    /* It's a biggy:
		     * FORMAT: <seq#> A <name> <section> <class> SOA <mname> <rname> <serial> <refresh> <retry> <expire> <minimum> (<ttl>) */
		    opos = 11;
		    break;
		}
		/*
		 * case DT_AAAA:
		 * case DT_HINFO:
		 * case DT_ISDN:
		 * case DT_MINFO:
		 * case DT_NULL:
		 * case DT_RP:
		 * case DT_WKS:
		 */
	    default:
		{
		    opos = 512;
		    error_return = DE_NOTIMP;
		}
	    }
	    /* Here's where we process any optional args. */

	    for (j = opos; j < i; j++) {
		k = answer[j];
		v = answer[j];
		while (v++) {
		    if (*v == 61) {
			*v = 0;
			v++;
			break;
		    }
		}
		if (*v) {
		    /* We have a value, so process. */
		    if (!(strcmp(k, "ttl"))) {
			    if (sscanf(v, "%u", &ttl) != 1) {
				err_warn("Scan for ttl in \"%s\" failed!\n", v);
				error_return = DE_NXDOMAIN;
				break;
			    } else {
				fprintf(stderr, "OPTION: ttl=%u.\n", ttl);
			}
		    }
		    else if (!(strcmp(k, "section"))) {

			if (*v < 48 || *v > 57) {
			    /* it's textual; convert. */
			    section = stoi(v, stoi_sections);
			    if (section < 0) {
				section = 0;
				err_warn("%s: Could not translate section \"%s\"; using default of %u.\n", v, section);
			    }
			}
			else {
			    /* it's numeric; scan. */
			    if (sscanf(v, "%hi", &section) != 1) {
				err_warn("Scan for section failed!\n");
				error_return = DE_NXDOMAIN;
				break;
			    }
			}
				fprintf(stderr, "OPTION: section=%u.\n", section);

		    }
		    else if (!(strcmp(k, "class"))) {
			if (*v < 48 || *v > 57) {
			    /* it's textual; convert. */
			    class = stoi(v, stoi_ctypes);
			    if (class < 0) {
				section = 1;
				err_warn("%s: Could not translate class\"%s\"; using default of %u.\n", v, class);
			    }
			}
			else {
			    /* it's numeric; scan. */
			    if (sscanf(v, "%hi", &class) != 1) {
				err_warn("Scan for section failed!\n");
				error_return = DE_NXDOMAIN;
				break;
			    }
			}
				fprintf(stderr, "OPTION: class=%u.\n", class);
		    }
		    else if (!(strcmp(k, "authority"))) {
			if (!(strcmp(v, "1"))) {
			    dnsqpkt->flags.aa = 1;
			}
			else if (!(strcmp(v, "0"))) {
			    dnsqpkt->flags.aa = 0;
			}
			else {
			    err_warn("%s: do not understand argument \"%s\" to key \"authority\".\n", ROUTINE, v);
			}

		    }
		    else {
			err_warn("%s: do not understand option \"%s\".\n", ROUTINE, v);
		    }
		}
	    }
	}
	dns_build_add(build, section,
		      answer[2], ttl, class,
		      type, (void *) rdata_buf);
	error_return = DE_NOERROR;

	pipeutil_free_bird(answer);
    }

    pipeutil_free_bird(answer);

    pipeutil_release_dproc(dp);
    return (error_return);
#undef ROUTINE
}				/* end mod_msg_tgl_answer() */


/*
 * module_dns_zone_delete
 *
 *  Delete's a zone
 *
 * FIXME: what are the valid return values from this function???
 *
 */
MODULE_FUNC(int, dns_zone_delete, (dnszonerec * zone))
{
#define ROUTINE "mod_msg_tgl_zone_delete ()"
    int     rval = 0;
    dproc  *dp = NULL;
    char    wbuf[255];
    char  **answer;
    struct timeval tv;

    g_mutex_lock(global_mutex);
    DPRINTF(("%s: Called\n", ROUTINE));

    if (!(dp = pipeutil_find_and_grab_dproc(zone->zonename))) {
	return (-1);
    }

    g_snprintf((char *) &wbuf, 255, "%i ZD %s\n", dp->snum++, zone->zonename);
    fprintf(stderr, "delete(): Sending message to %s: \"%s\".\n", dp->name, &wbuf);
    write(dp->wfile, (void *) &wbuf, strlen((char *) &wbuf));

    tv.tv_sec = 1;
    tv.tv_usec = 0;

    answer = pipeutil_getline(dp, &tv);
    if (!answer) {
	/* FIXME: do something exciting here. */
	return(-1);
    }
    pipeutil_display(pipeutil_count_elems(answer), answer);

    if (is_finis(answer) == 0) {
	rval = -1;
    }

    pipeutil_demand_unload_dproc(dp);
    pipeutil_release_dproc(dp);

    g_mutex_unlock(global_mutex);
    return (rval);

#undef ROUTINE
}

/*
 * module_mysql_zone_load
 *
 * A new zone is requested to be loaded.
 */

MODULE_FUNC(int, dns_zone_load, (dnszonerec * zone))
{
#define ROUTINE "module_msg_tgl_zone_load ()"
    dproc  *dp = NULL;
    char  **answer;
    char   *name;
    char    msgbuf[255];
    char    wordbuf[64];
    int     i;
    struct timeval tv;
    dproc_zone *zd;

    g_mutex_lock(global_mutex);
    DPRINTF(("%s: Starting %d args", ROUTINE, zone->argc));

    if (dp = pipeutil_find_and_grab_dproc(zone->zonename)) {
	pipeutil_release_dproc(dp);
	return (-1);
    }

    if (!(dp = pipeutil_demand_load_dproc(zone->argv[0]))) {
	return (-1);
    }

    if (!(zd = malloc(sizeof(dproc_zone)))) {
	fprintf(stderr, "Out of memory!\n");
	fflush(stderr);
	exit(-1);
    }

    if (!(zd->name = malloc(strlen(zone->zonename) + 1))) {
	fprintf(stderr, "Out of memory!\n");
	fflush(stderr);
	exit(-1);
    }
	zd->soa_rname=NULL;
	zd->soa_mname=NULL;

    zd->soa_rname = NULL;
    zd->soa_mname = NULL;

    strcpy(zd->name, zone->zonename);
    dp->domains = g_list_append(dp->domains, zd);
    if (!(zd->args = malloc(zone->argc * sizeof(char *)))) {
	fprintf(stderr, "Out of memory!\n");
	fflush(stderr);
	exit(-1);
    }

    g_snprintf((char *) &msgbuf, sizeof(msgbuf), "%u ZL %s ", dp->snum++, zone->zonename);

    for (i = 0; i < zone->argc; i++) {
	if (!(zd->args[i] = strdup(zone->argv[i]))) {
	    fprintf(stderr, "Out of memory!\n");
	    fflush(stderr);
	    exit(-1);
	}
	g_snprintf(wordbuf, sizeof(wordbuf), "%s ", zone->argv[i], 64);
	strncat(msgbuf, wordbuf, (sizeof(msgbuf) - strlen(msgbuf)));
    }
    strncat(msgbuf, "\n", (sizeof(msgbuf) - strlen(msgbuf)));
    fprintf(stderr, "load(): Sending message to %s: \"%s\".\n", dp->name, &msgbuf);
    write(dp->wfile, (void *) &msgbuf, strlen((char *) &msgbuf));

    while (1) {
	/* It can take a while for perl to load. */
	tv.tv_sec = 5;
	tv.tv_usec = 0;

	answer = pipeutil_getline(dp, &tv);
	if (!answer) {
	    /* FIXME: do something exciting here. */
		return(-1);
	}

	if (is_soa(answer) == TRUE) {

	    if (dproc_has_domain(dp, zone->zonename) == TRUE) {
		fprintf(stderr, "Joy!  %s has %s\n", dp->name, zone->zonename);
	    }
	    else {
		fprintf(stderr, "Aiee!  %s doesn't have %s\n", dp->name, zone->zonename);
		return (-1);
	    }
	    remember_soa(answer, dp);
	    break;
	}
	else if (is_options(answer) == TRUE) {
	    remember_options(answer, dp);
	}
	else {
	    fprintf(stderr, "load(): Expected O or SOA from %s, got:", dp->name);
	    pipeutil_display(pipeutil_count_elems(answer), answer);
	    fprintf(stderr, "returning(-1).\n");
	    return (-1);
	}
    }

    dp->refcount++;
    zone->mdata = dp;

    g_mutex_unlock(global_mutex);
    return (ERR_DNS_RDB_NOERROR);

#undef ROUTINE
}

/*
 * module_mysql_zone_add
 *
 * I thought this was deprecated?
 */
MODULE_FUNC(int, dns_zone_add, (dnszonerec * zone))
{
    return (module_dns_zone_load(zone));
}

/*
 * module_dns_zone_reload
 *
 *  Reload's a zone.
 *
 * Note that the names of the return values from this function are 
 * leftover from the * mod_stdddb implementation (ERR_DNS_RDB_XXXX)
 * which are due to be renamed.
 */
MODULE_FUNC(int, dns_zone_reload, (dnszonerec * zone))
{
#define ROUTINE "mod_msg_tgl_zone_reload()"
    int     rval = 0;
    dproc  *dp = NULL;
    char  **answer;
    char    wbuf[255];
    struct timeval tv;

    g_mutex_lock(global_mutex);

    if (!(dp = pipeutil_find_and_grab_dproc(zone->zonename))) {
	return (-1);
    }

    g_snprintf((char *) &wbuf, 255, "%i ZR %s\n", dp->snum++, zone->zonename);
    fprintf(stderr, "reload(): Sending message to %s: \"%s\".\n", dp->name, &wbuf);
    write(dp->wfile, (void *) &wbuf, strlen((char *) &wbuf));

    tv.tv_sec = 1;
    tv.tv_usec = 0;

    answer = pipeutil_getline(dp, &tv);
    if (!answer) {
	/* FIXME: do something exciting here. */
	return(-1);;
    }
    pipeutil_display(pipeutil_count_elems(answer), answer);
    if (is_finis(answer) == 0) {
	rval = -1;
    }
    pipeutil_release_dproc(dp);
    pipeutil_free_bird(answer);
    g_mutex_unlock(global_mutex);
    return (rval);
}				/* end dns_zone_reload() */


/*
 * module_init
 *
 *  All we do is just print a banner
 */
MODULE_FUNC(int, init, (void))
{

/* FIXME: nothing much to do; we do all at module load time. */

    pipeutil_init();
    global_mutex = g_mutex_new();
    DPRINTF(("mod_msg_tgl_init: version %d.%d.%d by Todd G. Lewis\n",
	     MODULE_VER_MAJOR, MODULE_VER_MINOR, MODULE_VER_MICRO));

    dprocs = NULL;

    return 0;
}

/*
 * module_destroy
 *
 * Called when the module is unloaded. Always after all zones have
 * stopped using it
 */
MODULE_FUNC(void, destroy, (void))
{
    /* If all zones have really stopped, then we have nothing to do. */
}				/* end mod_msg_tgl_destroy () */

/***************************************************************************
 ***************************************************************************
 ** 		DENTS module black voodoo magic
 ***************************************************************************
 ***************************************************************************/

/*
 * Module compile support code
 */

/* Each answering function in order of desired execution. The list MUST */
/*  be NULL terminated */
MODULE_DNS_ANSWER = {
    mod_msg_tgl_answer, NULL
};

/*
 * Exported symbols
 */
MODULE_SYMS(MODULE_SYMS_INIT	/* module_init() */
	    MODULE_SYMS_DESTROY	/* module_destroy() */
	    MODULE_SYMS_DNS	/* Mandatory DNS symbols */
    /* module_dns_zone_add() */
    /* module_dns_zone_delete() */
    /* module_dns_answer() */
	    MODULE_SYMS_DNS_ZONE_RELOAD		/* module_dns_zone_reload() */
	    MODULE_SYMS_DNS_ZONE_LOAD	/* module_dns_zone_load() */
)
#endif /* WITH_MSG_TGL */
