

/* headers */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <errno.h>		/* perror */
#include <signal.h>		/* kill, signal, sigprocmask */
#include <stdio.h>		/* printf */
#include <stdlib.h>		/* malloc */
#include <string.h>		/* strncpy */
#include <unistd.h>		/* fork, getpid, read, gettimeofday */
#include <sys/time.h>		/* gettimeofday */
#include <sys/types.h>		/* waitpid, kill */
#include <sys/stat.h>		/* stat */
#include <sys/wait.h>		/* waitpid */
#include <sys/mman.h>		/* mprotect */
#include <glib.h>		/* GList */
#include "mod_msg_tgl.h"	/* dproc */
#include "pipeutil.h"		/* prototypes for all public functions herein */
#include "signal_handlers.h"	/* reg_chld_handler */

GMutex *uberlock;

/* convenience macros */
#define DELIM 32		/* ASCII 32, space   */
#define EOL 10			/* ASCII 10, newline */
#define EOS  0			/* ASCII  0, null    */
#define NDEAD 255

/* globals */
static int count;
static dproc *sys_dproc_list;

/* Misc. functions */
#ifdef PIPEUTIL_TEST
static void usage(char *name);
#endif /*PIPEUTIL_TEST */
static void show_count(int sig);
static void catch_chld(int sig);

/* Child-process stuff */
static dproc *make_child(int argc, char *argv[]);
static gboolean is_alive(pid_t pid);
static dproc *find_child_by_pid(pid_t pid);
static dproc *find_child_by_name(char *name);
static void reclaim_child(dproc * child_dproc);
static void reap_any_dead(void);
static void grab_uber(void);
static void free_uber(void);

/* Function Implementations */

#ifdef PIPEUTIL_TEST
static void
usage(char *name)
{
    fprintf(stderr, "Usage:   %s <program>\n", name);
}

#endif /*PIPEUTIL_TESPIPEUTIL_TESTT */

void
walk_list(void)
{
    dproc  *t = sys_dproc_list;

    if (!t) {
	return;
    }
    fprintf(stderr, "System dproc list: ");
    for (t = sys_dproc_list; t; t = t->next) {
	fprintf(stderr, "%s(%i), ", t->name, t->pid);
    }
    fprintf(stderr, "(end).\n");
    fflush(stderr);
}

/* Returns a NULL-terminated, argv-style array of words.  If there's an
 * error, then returns NULL.  If "timeout==NULL", then it waits forever
 * for return from the child process.  Be sure to call pipeutil_free_bird() on 
 * the resulting data when you're done. */
extern char **
pipeutil_getline(dproc * tdproc, struct timeval *timeout)
{
    char  **words = tdproc->index;
    int     i, j, gcount, wantword, hasline;
    struct timeval scratch, target, mwait;
    fd_set  rfds;
    size_t  mys;

    gcount = 0;
    wantword = 1;
    *words = NULL;

    /* fprintf(stderr, "pipeutil_getline() called.\n"); */

    /* First, we shift any leftovers over to left of our buffer. */
    for (i = 0; i < ((tdproc->lof) - (tdproc->los)); i++) {
	tdproc->rbuf[i] = tdproc->rbuf[i + tdproc->los];
    }
    tdproc->lof -= tdproc->los;
    tdproc->los = 0;

    if (timeout) {
	/* We set up our target time for later reference. */
	if (gettimeofday(&scratch, NULL)) {
	    /* Careful programmers check everything, and I'm too
	     * bad a programmer to be hobbled by unchecked errors. */
	    /* fprintf(stderr, "Can't gettimeofday; bailing.\n"); */
	    return (NULL);
	}
	target.tv_sec = scratch.tv_sec + timeout->tv_sec;
	target.tv_usec = scratch.tv_usec + timeout->tv_usec;
	/* Overflow is cool. */
	if (target.tv_usec >= 1000000) {
	    target.tv_sec++;
	    target.tv_usec -= 1000000;
	}
    }
    /* fprintf(stderr, "Entering hasline, target time is (%i.%i).\n", target.tv_sec, target.tv_usec); */
    hasline = 0;
    for (j = 0; j < 1000 && hasline != 1; i++) {
	/* fprintf(stderr, "Examining buffer from %i to %i...\n", tdproc->los, tdproc->lof); */
	for (i = tdproc->los; i < tdproc->lof; i++) {
	    /* Do we have a line's worth of data? */
	    if (tdproc->rbuf[i] == EOL || tdproc->rbuf[i] == EOS || i == RBUF_SIZE) {
		hasline = 1;
		break;
	    }
	    else {
		/* fprintf(stderr, "hasline: character %u is not %u or %u.\n", (char) tdproc->rbuf[i], (char) EOL, (char) EOS); */
	    }
	}
	/* fprintf(stderr, "Done checking for newline; moving on to read.\n"); */
	if (hasline == 0) {
	    /* Apparently not!  Grab some more data, subject to
	     * timeout. */
	    i = 0;
	    FD_ZERO(&rfds);
	    FD_SET(tdproc->rfile, &rfds);
	    if (timeout) {
		/* fprintf(stderr, "Has timeout.\n"); */
		if (gettimeofday(&scratch, NULL)) {
		    fprintf(stderr, "Can't gettimeofday; bailing.\n");
		    return (NULL);
		}
		if ((scratch.tv_sec > target.tv_sec) || \
		    (scratch.tv_sec == target.tv_sec && \
		     scratch.tv_usec > target.tv_usec)) {
		    /* We've exceeded our timeout. */
		    fprintf(stderr, "Timeout exceeded; bailing.\n");
		    return (NULL);
		}
		else {
		    /* fprintf(stderr, "present time (%i.%i) not yet target time (%i.%i).\n", scratch.tv_sec, scratch.tv_usec, target.tv_sec, target.tv_usec); */
		}
		mwait.tv_sec = target.tv_sec - scratch.tv_sec;
		mwait.tv_usec = target.tv_usec - scratch.tv_usec;
		/* fprintf(stderr, "mwait is (%i.%i).\n", mwait.tv_sec, mwait.tv_usec); */
		if (mwait.tv_usec < 0) {
		    /* Underflow is cool. */
		    mwait.tv_usec += 1000000;
		    mwait.tv_sec--;
		}
		fprintf(stderr, "mwait with correction is (%i.%i).\n", mwait.tv_sec, mwait.tv_usec);
		mwait.tv_sec = 0;
		if (mwait.tv_usec > 200) {
		    mwait.tv_usec = 200;
		}
		fprintf(stderr, "selecting with timeout %i.%i...\n", mwait.tv_sec, mwait.tv_usec);
		i = select((tdproc->rfile + 1), &rfds, NULL, NULL, &mwait);
	    }
	    else {
		fprintf(stderr, "selecting without timeout...\n");
		i = select((tdproc->rfile + 1), &rfds, NULL, NULL, NULL);
	    }
	    fprintf(stderr, "select() returned with %i.\n", i);
	    fflush(stderr);
	    if (FD_ISSET(tdproc->rfile, &rfds)) {
		/* FIXME: read returns the number of bytes read, needs to be adjusted before assigned to lof. */
		if (is_alive(tdproc->pid) == FALSE) {
		    fprintf(stderr, "process %s(%i) is dead!  returning.\n", tdproc->name, tdproc->pid);
		    return (NULL);
		}
		else {
		    /* fprintf(stderr, "process %s(%i) is alive.  Continuing.\n", tdproc->name, tdproc->pid); */
		}
		fprintf(stderr, "reading from process %s.\n", tdproc->name);
		mys = read(tdproc->rfile, \
			   &tdproc->rbuf[tdproc->lof], \
			   (RBUF_SIZE - tdproc->lof));

		fprintf(stderr, "%i bytes read from process %s on %i.\n", mys, tdproc->name, j);
		if (mys < 0 && errno != EINTR) {
		    /* This was a real, honest-to-goodness error. */
		    fprintf(stderr, "Error in read; killing child.  (rval=%i, i=%i, errno=%i).\n", tdproc->lof, i, errno);
		    kill(tdproc->pid, SIGINT);
		    return (NULL);
		}
		if (mys > 0) {
		    tdproc->lof += mys;
		    /* fprintf(stderr, "Got %i bytes on try %i; should loop.\n", mys, j); */
		    break;
		}
		if (mys == 0) {
		    /* fprintf(stderr, "Got no input on try %i; should loop.\n", j); */
		}
	    }
	}
	/* fprintf(stderr, "Did hasline loop...\n"); */
    }
    /* fprintf(stderr, "Left hasline...\n"); */

    /* Allright, we finally have a line of text to process! */
    for (i = tdproc->los; i < tdproc->lof; i++) {
	if ((tdproc->rbuf[i] == EOS) || (tdproc->rbuf[i] == EOL)) {
	    tdproc->rbuf[i] = EOS;
	    words[gcount++] = NULL;
	    /* fprintf(stderr, "Setting los to %i.\n", i + 1); */
	    tdproc->los = i + 1;
	    break;
	}
	else if (tdproc->rbuf[i] == DELIM) {
	    tdproc->rbuf[i] = EOS;
	    wantword = 1;
	}
	else {
	    /* It's a real character. */
	    if (wantword == 1) {
		words[gcount++] = &tdproc->rbuf[i];
		if (gcount == (INDEX_SIZE - 2)) {
		    words[gcount + 1] = NULL;
		    break;
		}
		wantword = 0;
	    }
	}
    }
    if (i == RBUF_SIZE || gcount == (INDEX_SIZE - 2)) {
	if (tdproc->rbuf[i] != EOL) {
	    tdproc->rbuf[i] = EOS;
	    i = 0;
	    /* drain the lizard */
	    while (i != EOL) {
		read(tdproc->rfile, &i, 1);
	    }
	}
    }
    return (words);
}

extern void
pipeutil_free_bird(char **bird)
{
    /* If we move back to dynamic allocation, free it here. */
}

/* num elems, or -1 on error. */
extern int
pipeutil_count_elems(char **bird)
{
    int     i = 0;

    if (!bird) {
	return (-1);
    }

    while (bird[i] != NULL) {
	i++;
    }
    return (i);
}

static void
show_count(int sig)
{
    printf("Caught %u: query count is %u\n", sig, count);
    fflush(stdout);
}

static void
handle_child(pid_t child, int status)
{
    dproc  *kid;

    fprintf(stderr, "Child is %i, status %i\n", child, status);
    kid = find_child_by_pid(child);
    kid->status = status;
    kid->is_dead = TRUE;
}

/* This is our SIGCHLD handler; it sticks the pid on a list for later cleaning. */
static void
catch_chld(int sig)
{
    int     status;
    pid_t   child;

    fprintf(stderr, "catch_child() called on %i.  ", getpid());
    while (1) {
	child = waitpid(-1, &status, WNOHANG);
	if (child < 1) {
	    fprintf(stderr, "\n");
	    break;
	}
	else {
	    fprintf(stderr, "Child is %i, status %i\n", child, status);
	    handle_child(child, status);
	}
    }
    walk_list();
    fflush(stderr);
}

/* You need the uberlock for this. */
static void
reap_any_dead(void)
{
    dproc  *child_dproc = NULL, *needs_reaping = NULL;

    /* fprintf(stderr, "Reaping any dead.\n"); */
    for (child_dproc = sys_dproc_list; child_dproc; child_dproc = child_dproc->next) {
	if (child_dproc->is_dead == TRUE) {
	    if (needs_reaping != NULL) {
		reclaim_child(needs_reaping);
	    }
	    needs_reaping = child_dproc;
	}
    }
    if (needs_reaping != NULL) {
	reclaim_child(needs_reaping);
    }
    /* fprintf(stderr, "Done reaping.\n"); */
}

/* You need the uberlock to do this. */
static void
reclaim_child(dproc * child_dproc)
{
    /* WARNING: If you are in the middle of walking the system dproc list and
     * you call this routine, and this routine frees child_dproc, then
     * will your for loop survive its next round? */

    /* This should almost certainly call back into Dents and delete
     * any domains still registered under this dproc. */

    /* FIXME: Store the configs and then respawn the sucker.
     * This would be more in line with The Dents Way. */

    fprintf(stderr, "Reclaiming child: %s(%u), status %i.\n", child_dproc->name, child_dproc->pid, child_dproc->status);
    /* order is important here; don't change without thinking. */
    if (sys_dproc_list == child_dproc) {
	sys_dproc_list = child_dproc->next;
    }
    if (child_dproc->prev) {
	child_dproc->prev->next = child_dproc->next;
    }
    if (child_dproc->next) {
	child_dproc->next->prev = child_dproc->prev;
    }
    g_list_free(child_dproc->domains);
    close(child_dproc->wfile);
    close(child_dproc->rfile);
    free(child_dproc);
}

static void
init_uber(void)
{
    uberlock = g_mutex_new();
}

void
pipeutil_init(void)
{
    init_uber();
}

/* Either works or never returns! */
static void
grab_uber(void)
{
    g_mutex_lock(uberlock);
    reap_any_dead();
}

/* Either works or never returns! */
static void
free_uber(void)
{
    reap_any_dead();
    g_mutex_unlock(uberlock);
}

/* argv[i] must = NULL. */
extern void
pipeutil_display(int argc, char *argv[])
{
    int     i;

    fprintf(stderr, "pipeutil_display called (%i).\n", argc);
    if (argc < 0) {
	return;
    }
    for (i = 0; i < argc; i++) {
	fprintf(stderr, "(%i, %s), ", i, argv[i]);
    }
    fprintf(stderr, "(%i, %s).\n", i, argv[i]);
}

/* If a dproc of this name already exists, then this procedure returns
 * a pointer thereto.  Otherwise, it spawns a new one.  It is advised that
 * one only pass a fully-qualified pathname as "name". */
extern dproc *
pipeutil_demand_load_dproc(char *name)
{
    dproc  *rval;
    char   *argv[3];

    if ((rval = find_child_by_name(name))) {
	return (rval);
    }
    argv[0] = name;
    argv[1] = NULL;
    return (make_child(1, argv));
}

/* This routine lowers the reference count on the passed dproc and,
 * if refcount reaches zero, it reclaims the dproc. */
extern void
pipeutil_demand_unload_dproc(dproc * tdproc)
{
    if (!--tdproc->refcount) {
	reclaim_child(tdproc);
    }
}
/* This should be a replacement for dochild.  It allocates a
 * readpair and writepair along with the dproc.  It records
 * the pid and adds the dproc to the global dproc list. 
 *
 * Returns pointer on success, NULL on failure.  The dproc is
 * locked when returned and needs to be unlocked. */
static dproc *
make_child(int argc, char *argv[])
{
    dproc  *rval;
    static int readpair[2];
    static int writepair[2];
    int     i = 0, cpid;

    fprintf(stderr, "make_child() called on %s.\n", argv[0]);

    rval = malloc(sizeof(dproc));

    if ((pipe(readpair) != 0) || (pipe(writepair) != 0)) {
	perror("Pipe failed");
	return (NULL);
    }
    cpid = fork();
    if (cpid == -1) {
	perror("Fork failed");
	free(rval);
	return (NULL);
    }

    if (cpid == 0) {
	/* We're the child. */
	close(0);
	close(1);
	close(2);

	if (dup2(writepair[0], 0) == -1) {
	    perror("dup2() failed");
	    exit(-1);
	}
	if (dup2(readpair[1], 1) == -1) {
	    perror("dup2() failed");
	    exit(-1);
	}
	/* if (dup2(readpair[1], 2) == -1) {
	 * exit(-1);
	 * } */

	for (i = 3; i < 1024; i++) {
	    close(i);
	}

	/* In perl doing this the right way would be easy... */
	if (argc == 2) {
	    execlp(argv[0], argv[0], NULL);
	}
	else if (argc == 3) {
	    execlp(argv[0], argv[0], argv[1], NULL);
	}
	else if (argc == 4) {
	    execlp(argv[0], argv[0], argv[1], argv[2], NULL);
	}
	else {
	    execlp(argv[0], argv[0], argv[1], argv[2], argv[3], NULL);
	}
	fprintf(stderr, "execlp() failed!\n");
	exit(-1);
    }
    else {
	/* We're the parent */
	i = (strlen(argv[0]) + 1);
	rval->name = malloc(i);
	strncpy(rval->name, argv[0], i);
	rval->pid = cpid;
	fprintf(stderr, "make_child(): creating child %s, %i at %p.\n", rval->name, rval->pid, rval);
	rval->rfile = readpair[0];
	rval->wfile = writepair[1];
	close(readpair[1]);
	close(writepair[0]);
	rval->los = 0;
	rval->lof = 0;
	/* rval->rbuf needs no initialization */
	rval->domains = NULL;
	rval->refcount = 0;
	rval->is_dead = FALSE;
	rval->lock = g_mutex_new();
	rval->prev = NULL;

	if(is_alive(rval->pid)==FALSE){
		fprintf(stderr, "Process %s alread dead!  Aborting.\n", rval->name);
		free(rval->name);
		free(rval);
		free_uber();
		return(NULL);
	}

	fprintf(stderr, "Adding process %s to system list.\n", rval->name);
	if (sys_dproc_list == NULL) {
	    walk_list();
	    rval->next = NULL;
	    grab_uber();
	}
	else {
	    grab_uber();
	    sys_dproc_list->prev = rval;
	    rval->next = sys_dproc_list;
	}
	sys_dproc_list = rval;
	reg_chld_handler(handle_child, rval->pid);
	fprintf(stderr, "Process %s added to system list.\n", rval->name);
	walk_list();
	free_uber();
	return (rval);
    }
}

/* You should own uberlock before calling this. */
static dproc *
find_child_by_pid(pid_t pid)
{
    dproc  *retval;

    for (retval = sys_dproc_list; retval; retval = retval->next) {
	if (retval->pid == pid) {
	    return (retval);
	}
    }
    return (NULL);
}

/* finds the dproc according to the name of the dproc executable.
 * NULL on failure. */
static dproc *
find_child_by_name(char *name)
{
    dproc  *retval;

    grab_uber();
    for (retval = sys_dproc_list; retval; retval = retval->next) {
	if (!(strcmp(retval->name, name))) {
	    free_uber();
	    return (retval);
	}
    }
    free_uber();
    return (NULL);
}

/* Will return a dz for which name is the same as or a child of dz->name;
 * NULL on error. */
dproc_zone *
find_dz_by_name(char *name, dproc * dp)
{
    GList  *gl;
    dproc_zone *dz;
    size_t  a, b;
    int     i;
    char   *spot;

    a = strlen(name);
    for (gl = dp->domains; gl; gl = gl->next) {
	dz = gl->data;
	b = strlen(dz->name);
	if (b < a) {
	    i = a - b;
	    spot = (char *) &name[i];
	}
	else {
	    spot = name;
	}
	if (!(strcmp(spot, dz->name))) {
	    return (dz);
	}
    }
    return (NULL);
}

static  gboolean
is_alive(pid_t pid)
{
    /* Works on most modern unices; feel free to add your own implementation 
     * and #def it accordingly. */
    char    buf[18];		/* "/proc/4294967296" +1 for good measure */
    struct stat sbuf;

    snprintf((char *) &buf, (size_t) 17, "/proc/%u", pid);
    if (stat((char *) &buf, &sbuf)) {
	return (FALSE);
    }
    return (TRUE);
}

extern  gboolean
dproc_has_domain(dproc * dp, char *domainname)
{
    GList  *domainl;
    dproc_zone *dz;

    for (domainl = dp->domains; domainl; domainl = domainl->next) {
	dz = domainl->data;
	if (!strcmp(dz->name, domainname)) {
	    return (TRUE);
	}
    }
    return (FALSE);
}

extern dproc *
pipeutil_find_and_grab_dproc(char *domainname)
{
    dproc  *retval = NULL;

    grab_uber();
    for (retval = sys_dproc_list; retval; retval = retval->next) {
	if (dproc_has_domain(retval, domainname) == TRUE) {
	    g_mutex_lock(retval->lock);
	}
    }
    fflush(stderr);
    free_uber();
    return (NULL);
}

extern void
pipeutil_release_dproc(dproc * rdproc)
{
    g_mutex_unlock(rdproc->lock);
}

#ifdef PIPEUTIL_TEST
int
main(int argc, char *argv[])
{
    int     i = 0;
    char  **buf;
    int     do_continue = 1;
    dproc  *child = NULL;
    char   *button[2];
    char    wbuf[256];
    struct timeval timeout;

    if (argc < 2) {
	usage(argv[0]);
	exit(-1);
    }
    sys_dproc_list = NULL;
    count = 0;
    signal(SIGHUP, show_count);
    signal(SIGCHLD, catch_chld);

    for (i = 1; i < argc; i++) {
	button[0] = argv[i];
	button[1] = NULL;
	/* button[1] = argv[i]; */
	child = make_child(2, button);
	if (child == NULL) {
	    fprintf(stderr, "Aiee!  make_child() on %s failed!  Aborting.\n", argv[i]);
	}
    }

    while (do_continue) {
	for (child = sys_dproc_list; child; child = child->next) {
	    if (is_alive(child->pid) == FALSE) {
		fprintf(stderr, "My baby(%u) is dead!  I can't go on!\n", child->pid);
		do_continue = 0;
		break;
	    }
	    snprintf((char *) &wbuf, sizeof(wbuf), "%u How Are you\n", count++);
	    if (!(write(child->wfile, (void *) &wbuf, strlen((char *) &wbuf)))) {
		fprintf(stderr, "Aiee!  snprintf() failed.\n");
		exit(-1);
	    }
	    timeout.tv_sec = 1;
	    timeout.tv_usec = 500;
	    /* printf("Getting line on %s.\n", child->name); */
	    if (1) {
		buf = pipeutil_getline(child, &timeout);
	    }
	    else {
		buf = pipeutil_getline(child, NULL);
	    }
	    i = pipeutil_count_elems(buf);
	    if (!i || (buf == NULL)) {
		fprintf(stderr, "Error in input on %s; breaking.\n", child->name);
		sleep(1);
		/* do_continue = 0; */
		break;
	    }
	    /* printf("Line gotten on %s.\n", child->name); */
	    printf("%s(%i): ", child->name, child->pid);
	    pipeutil_display(i, buf);
	    pipeutil_free_bird(buf);
	}
    }

    return (0);
}
#endif /* PIPEUTIL_TEST */
