/* mod_recursive.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/*
 * mod_recursive.c
 *
 *  Module for recursive queries
 */
#include "dns.h"

/* Basic info on the module, used later */
#define MODULE_NAME	mod_recursive

#define MODULE_VER_MAJOR 0
#define MODULE_VER_MINOR 0
#define MODULE_VER_MICRO 1

#include "addr_tools.h"
#include "module_syms.h"
#include "dns_module_syms.h"

/* Information saved for recursive queries so we can send a response back */
typedef struct
{
  llentry ll;			/* Blocklist */

  llentry link;			/* Linked together duplicates */

  dnsosqueryrec *osq;		/* OSQ for query we sent */

  dnsflagsrec flags;		/* Flags from original query */
  guint8 opcode;		/* Opcode from original query */

  guint16 id;			/* id packet was received with */
  struct in_addr ip;		/* IP of sending client */
  guint16 port;			/* Port of sending client */

  socketrec *sock;		/* Socket we received the packet from */

  dnsdbrrrec *referrer;		/* Node that caused the referrer info */
}
recursivequeryrec;

/* Database containing all RR's recursively found */
static dbrootrec db;
static linkedlist queries = LINK_INITIALIZER (recursivequeryrec, ll);
static blocklist query_pool = BLOCK_INITIALIZER (recursivequeryrec, 4);

/*
 * mod_recursive_callback
 *
 *  Callback for the OSQ code. Will scan RR's, entering them into the
 * database if they pass all of the security checks. Will then run
 * the query through the chains again
 */
void
mod_recursive_callback (dnsosqueryrec * osq, packetrec * pkt,
			guchar * cp, dnspktrec * dnsrpkt)
{
  recursivequeryrec *rquery = (recursivequeryrec *) osq->data;
  dnspktrec dnsqpkt;
  int i, rc, totalrr;
  dnsrrrec rr;
  dnsdbrrrec *nrr;
  dbnoderec *node;
  dnsbuildrec *build;
  guchar ipv4_string[16];

  if (!at_inet4_ntop
      ((guchar *) & rquery->ip, ipv4_string, sizeof (ipv4_string))) {
    strcpy (ipv4_string, "<bad IPv4>");
  }

  DPRINTF (("mod_recursive_callback: original query from %s:%d id %d\n",
	    ipv4_string, rquery->port, rquery->id));

  /* Did our query result in an error? */
  switch (dnsrpkt->flags.rcode) {
  case DE_FORMERR:
    /* FIXME: We should resend the osq since the UDP packet may have
       been corrupted on the way there. */
    DPRINTF (("mod_recursive_callback: we got FORMERR!\n"));
    return;
  case DE_SERVFAIL:
    /* FIXME: Cycle onto the next server */
    return;
  case DE_NOERROR:
    /* This is the fall through case */
    break;
  default:
    /* We send back all other errors to the client */
    dns_osq_delete (osq);

    dns_buildpacket (pkt, &osq->query, NULL, rquery->id,
		     dnsrpkt->flags.rcode, rquery->flags, 1,
		     dnsrpkt->flags.aa);
    dns_sendpacket (rquery->sock, pkt, &rquery->ip, rquery->port);
    return;
  }

  /* Now we'll add all of the information into the cache */
  totalrr = dnsrpkt->ancount + dnsrpkt->nscount + dnsrpkt->arcount;
  for (i = 0; i < totalrr; i++) {
    while ((rc = dns_parserr (pkt, &cp, &rr)) < 0) {
      switch (rc) {
      case ERR_DNS_PARSE_LABEL:
      case ERR_DNS_PARSE_TOOSHORT:
	/* There's not much we can do here */
	/* FIXME: But do something */
	return;
      case ERR_DNS_PARSE_OVERFLOW:
	/* Allocate more buffer memory */
	packet_resizetbuf (pkt);
	break;
      default:
	err_warn ("mod_recursive_callback: unknown error parsing RR's");
	return;
      }
    }

    DPRINTF (("mod_recursive_callback: RR for %s, class %d, type %d\n",
	      rr.name, rr.class, rr.type));

    /* FIXME: We need to check the validity of the RR packet for cache */
    /* poisoning reasons */

    nrr = NULL;

    /* Check this information to see if we already have it in the cache */
    if ((node = dns_db_find (&db, rr.name)) != NULL) {
      for (nrr = node->data; nrr; nrr = nrr->next)
	if (dns_db_rr_compare (nrr, rr.class, rr.type, pkt->tempbuf,
			       rr.length)) break;
    }

    if (!nrr)
      /* The info's not there, so let's add it */
      dns_db_rr_add (&db, rr.name, rr.class, rr.type, rr.ttl,
		     pkt->tempbuf, rr.length);
  }

#ifdef DENTS_DEBUG
  db_display (&db);
#endif

  /* Ok, now rut it through the chains again */
  memcpy ((void *) &dnsqpkt.ip, (void *) &rquery->ip, sizeof (dnsqpkt.ip));
  dnsqpkt.port = rquery->port;
  dnsqpkt.id = rquery->id;
  dnsqpkt.socket = rquery->sock;
  memcpy (&dnsqpkt.query, &osq->query, sizeof (dnsqueryrec));
  memcpy (&dnsqpkt.flags, &rquery->flags, sizeof (dnsflagsrec));

  /* Delete the OSQ */
  dns_osq_delete (osq);

  DPRINTF (("mod_recursive_callback: running through chains again\n"));

  /* Allocate a build structure */
  build = dns_build_alloc ();
  if (!build) {
    err_warn ("mod_recursive_callback: Can't allocate a build record");
    return;
  }

  /* Run it through the chains again */
  rc = dns_ans_dochains (&dnsqpkt, build, pkt);

  if (rc == DE_WAITFORMOREINFO) {
    dns_build_delete (build);
    return;
  }

  /* Send a reply */
  dns_buildpacket (pkt, &dnsqpkt.query, build, dnsqpkt.id, rc,
		   dnsqpkt.flags, 1, (dnsqpkt.qru.query.zone->flags &
				      DNS_ZONE_AUTHORITATIVE));
  dns_sendpacket (rquery->sock, pkt, &rquery->ip, rquery->port);

  /* return the build structure back to the heap */
  dns_build_delete (build);

  /* FIXME: free rquery structure */
}

/*
 * mod_recursive_db
 *
 *  Check's the database to see if we have received this query before and
 * cached the response
 */
int
mod_recursive_db (dnspktrec * dnsqpkt, dnsbuildrec * build, packetrec * pkt)
{
  int rc;
  dbnoderec *node;

  /* Find the node */
  if ((node = dns_db_find (&db, dnsqpkt->query.name)) == NULL) {
    DPRINTF (("mod_recursive_db: dns_db_find return NULL on %s\n",
	      dnsqpkt->query.name));
    return (DE_NOINFO);
  }

  /* Find the answer */
  if ((rc = dns_build_findans_and_pack (&db, node, &dnsqpkt->query, build)) !=
      DE_NOERROR)
    /* Even though we have a name, doesn't mean we have that specific RR */
    return (DE_NOINFO);

  /* Find the authority information */
  dns_build_findauth_and_pack (&db, node, &dnsqpkt->query, build);

  /* Find any additional information */
  dns_build_findaddit_and_pack (&db, node, &dnsqpkt->query, build);

  return (DE_NOERROR);
}

/*
 * mod_recursive_send
 *
 *  Will take the query, find the best information we have and send a
 * query to the NS listed. If rd == 0, then we return the best information
 * and be done with it
 */
int
mod_recursive_send (dnspktrec * dnsqpkt, dnsbuildrec * build, packetrec * pkt)
{
  dbnoderec *node;
  dnsdbrrrec *referrer, *iprr;
  dnsosqueryrec *osq;
  struct in_addr ip;
  recursivequeryrec *rquery;
  guchar ipv4_string[16];

  /* If there's no pkt structure, then we can't send out queries. This */
  /* is only used if we don't want to do recursive queries as used in */
  /* control facility for instance */
  if (!pkt)
    return (DE_NOINFO);

  /* Try to find the best NS record */
  if ((node = dns_db_findbestns (&db, dnsqpkt->query.class,
				 dnsqpkt->query.name)) == NULL)
    /* Hmm, didn't find anything in the db, let's see if we can find it */
    /* in our hint cache */
    if ((node = dns_db_findbestns (&dns.hintdb, dnsqpkt->query.class,
				   dnsqpkt->query.name)) == NULL) {
      /* If we get here, then we couldn't even find the the NS' for the */
      /* root servers. That is bad */
      err_warn ("mod_recursive_send: couldn't find NS for root!");

      return (DE_SERVFAIL);
    }

  referrer = iprr = NULL;

  do {
    /* Then grab the NS record for it */
    if ((referrer = dns_db_rr_findmatch (node, referrer, dnsqpkt->query.class,
					 DT_NS)) == NULL)
      break;

    DPRINTF (("mod_recursive_send: found NS %s\n", referrer->data));

    /* Then find the node for the NS record */
    if ((node = dns_db_find (&db, referrer->data)) == NULL)
      /* Hmm, didn't find anything in the db, let's see if we can find it */
      /* in our hint cache */
      if ((node = dns_db_find (&dns.hintdb, referrer->data)) == NULL) {
	/* If we get here, then we couldn't find an A for the NS record */
	/* This can occur because of something that's misconfigured on */
	/* someone's end or because we don't have it cached yet. We */
	/* should try recursing here */
	/* FIXME: Recurse, until then, do nothing */

	DPRINTF (("mod_recursive_send: NS but no node for %s!\n",
		  referrer->data));
      }

    if (node) {
      if (
	  (iprr =
	   dns_db_rr_findmatch (node, NULL, dnsqpkt->query.class,
				DT_A)) == NULL) {
	/* We're here if we have no A records for a server listed as a NS */
	/* This could be because of a misconfiguration or we just don't have */
	/* it cached. Let's recurse */
	/* FIXME: Recurse to find the A, until then, do nothing */

	DPRINTF (("mod_recursive_send: NS but no A records for %s!\n",
		  referrer->data));
	continue;
      }
    }
  } while (!iprr);

  if (!referrer) {
    /* This should only happen if we have a node for atleast ., but no */
    /* NS records for it */
    err_warn
      ("mod_recursive_send: Couldn't find any referrer information for %s",
       dnsqpkt->query.name);
    return (DE_SERVFAIL);
  }

  DPRINTF (
	   ("mod_recursive_send: referrer = %p, iprr = %p\n", referrer,
	    iprr));

  /* Woo, we have usable information! */

  /* Grab the IP for the A */
  ip = *((struct in_addr *) iprr->data);

  /* Check to see if there's already a query sent on this triplet */
  if ((osq = dns_osq_find (&dnsqpkt->query)) != NULL) {
#ifdef DENTS_DEBUG
    int printed = 1;
#endif

    /* Ok, we got a duplicate query, now find out what kind of duplicate it */
    /* is (resent duplicate, parallel dup) */
    for (rquery = osq->data; rquery;
	 rquery = link_next (rquery, rquery->link)) {
      if (!memcmp (&rquery->ip, &dnsqpkt->ip, sizeof (struct in_addr))
	  && (rquery->port == dnsqpkt->port) && (rquery->id == dnsqpkt->id)) {
	/* Resent duplicate */
	DPRINTF (("mod_recursive_send: resent duplicate\n"));
	return (DE_WAITFORMOREINFO);
      } else {
#ifdef DENTS_DEBUG
	if (!printed) {
	  printed = 1;
	  DPRINTF (("mod_recursive_send: parallel duplicate\n"));
	}

	if (!at_inet4_ntop ((guchar *) & rquery->ip, ipv4_string,
			    sizeof (ipv4_string))) {
	  strcpy (ipv4_string, "<bad IPv4>");
	}

	DPRINTF (("mod_recursive_send:   from %s:%d id %d\n",
		  ipv4_string, rquery->port, rquery->id));
#endif
#ifndef DENTS_DEBUG
	break;
#endif
      }
    }
  }
#ifdef DENTS_DEBUG
  else
    DPRINTF (("mod_recursive_send: original query\n"));
#endif

  rquery = block_allocandaddtolink (&query_pool, &queries);

  rquery->id = dnsqpkt->id;
  memcpy ((void *) &rquery->ip, (void *) &dnsqpkt->ip, sizeof (dnsqpkt->ip));
  rquery->port = dnsqpkt->port;
  rquery->sock = dnsqpkt->socket;

  /* Grab an osq for it */
  if (!osq) {
    osq =
      dns_osq_create (ip, DNS_PORT, &dnsqpkt->query, mod_recursive_callback,
		      rquery);
    /* Then send it */
    dns_osq_send (pkt, osq);
  }
#if 0
  /* FIXME: This doesn't look quite right (perhaps a link function to do */
  /* this?) */
  osq->link.next = node->data;
  osq->link.prev = NULL;
#endif

  return (DE_WAITFORMOREINFO);
}

/*
 * Exported functions
 */

MODULE_FUNC (int, dns_zone_add, (dnszonerec * zone))
{
  printf ("mod_recursive_dns_add_zone: Called\n");

  return (0);
}

MODULE_FUNC (int, dns_zone_delete, (dnszonerec * zone))
{
  printf ("mod_recursive_dns_zone_delete: Called\n");

  return (DE_NOTIMP);
}

/*
 * module_init
 *
 *  Print banner and init the database
 */
MODULE_FUNC (int, init, (void))
{
  DPRINTF (("mod_recursive_init: version %d.%d.%d by Johannes Erdfelt\n",
	    MODULE_VER_MAJOR, MODULE_VER_MINOR, MODULE_VER_MICRO));

  /* Init the database */
  db_init (&db);

  return (0);
}

MODULE_FUNC (void, destroy, (void))
{
  printf ("mod_recursive_destroy: Called\n");
}

/* Define some symbols */
MODULE_DNS_ANSWER = {
mod_recursive_db, mod_recursive_send, NULL};

/* Exported symbols */
MODULE_SYMS (MODULE_SYMS_INIT MODULE_SYMS_DESTROY MODULE_SYMS_DNS)
