/* addr_tools.c - convert-functions for network addresses.

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

/* Author: Michel Onstein <michel@promera.nl>

   Provides functions to do conversion between human readable and
   internal stored network addresses. Includes error checking.  Providing
   our own functions makes sure that dents can still be compiled on a
   system where the specific (e.g. IPv6) protocol isn't available in
   libc.  See test cases in ../test/test_addr_tools.c */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

/* Add dents headers here */
#include "../include/common.h"
#include "addr_tools.h"

/* Return the packed IPv4 address in *ap as a human readable string.
   Returns *cp on success, NULL on error, which can only happen on a
   parameter problem. */
char *
at_inet4_ntop (const guint8 * ap, char *cp, size_t len)
{
  char wbuf[4 * 3 + 3 + 1];
  if (len < sizeof (wbuf))
    return NULL;		/* please pass a larger buffer */
  g_snprintf (wbuf, sizeof (wbuf), "%d.%d.%d.%d", ap[0], ap[1], ap[2], ap[3]);
  if (strlen (wbuf) >= len)
    return NULL;
  strncpy (cp, wbuf, len);
  return cp;
}

/* Return the packed IPv6 address in *ap as a human readable string.
   Returns *cp on success, NULL on error, which can only happen on a
   parameter problem. */
char *
at_inet6_ntop (const guint8 * ap, char *cp, size_t len)
{
  char *wbuf;
  int i;
  int lcount = 0;		/* What is largest set of rep. zeros */
  int lpos = -1;
  int ccount = 0;		/* Current amount of zeros */
  int cpos = -1;

  /* Sanity checking */
  if ((ap == NULL) || (cp == NULL))
    return NULL;

  if (!(wbuf = g_new0 (char, MAX ((4 * 8 + 7 + 1), len))))
      return NULL;

  for (i = 0; i < 8; i++) {
    guint16 val = ap[i * 2] * 256 + ap[(i * 2) + 1];

    if ((val != 0) || ((val == 0) && (i == 7))) {
      if ((val == 0) && (i == 7))
	ccount++;

      if (ccount >= lcount) {
	lpos = cpos;
	lcount = ccount;
      }

      ccount = 0;
      cpos = -1;
    } else {
      ccount++;
      if (cpos == -1)
	cpos = i;
    }
  }

  /* Main formatting loop */
  for (i = 0; i < 8; i++) {
    guint16 val = ap[i * 2] * 256 + ap[(i * 2) + 1];

    if (lpos == i) {
      if (lpos == 0)
	strcat (wbuf, ":\0");

      lpos++;
      if (--lcount == 0) {
	lpos--;
	strcat (wbuf, ":\0");
      }
    } else {
      sprintf (wbuf, "%s%X", wbuf, val);

      /* if we're not last, add : */
      if (i != 7)
	strcat (wbuf, ":");
    }

  }

  g_memmove (cp, wbuf, len);
  if (strlen (wbuf) >= len)
    cp = NULL;
  g_free (wbuf);
  return cp;
}

/*
 * at_hexword, return value for word size hexstring.
 *
 * at_hexword(char *cp, guint16 *val)
 *
 * Returns:     1 on success
 *              0 on failure
 */
static int
at_hexword (char *cp, guint16 * val)
{
  int i;

  *val = 0;

  if (strlen (cp) > 4)
    return 0;			/* Value will always be out of bound */

  /* Make uppercase and check for errors and add */
  for (i = 0; i < strlen (cp); i++) {
    if ((cp[i] >= 'a') && (cp[i] <= 'f'))
      cp[i] = cp[i] - 'a' + 'A';

    if (!(((cp[i] >= '0') && (cp[i] <= '9')) ||
	  ((cp[i] >= 'A') && (cp[i] <= 'F')))) return 0;

    *val *= 16;
    *val += ((cp[i] <= '9') ? (cp[i] - '0') : (cp[i] - 'A' + 10));
  }

  return 1;
}

/*
 * at_inet4_pton, return the human readable IPv4 string as a packed IPv4 address
 *
 * at_inet4_pton(char *cp, guint8 *ap)
 *
 * Returns *ap on success, NULL on error, in which case *cp is not a valid
 * formatted ipv4 address
 */
guint8 *
at_inet4_pton (char *cp, guint8 * ap)
{
  guint8 tbuf[4];
  int i;
  char localcopy[4 * 3 + 3 + 1];
  char *lp = (char *) &localcopy;

  /* Some sanity checking */
  if (strlen (cp) > (4 * 3 + 3))
    return NULL;		/* There is no IP address this big */

  strncpy (localcopy, cp, sizeof (localcopy));

#ifdef AT_OBSCURE_IPV4
  /* See if we have a dotted quad or maybe just an 32 bits value */
  if (strchr (lp, '.') == NULL) {
    char *p;
    long val;

    val = strtol (lp, &p, 0);

    if (*p != '\0')
      return NULL;

    if ((val < 0) || (val > 0xffffffff))
      return NULL;

    tbuf[0] = val >> 24;
    tbuf[1] = (val >> 16) & 0xff;
    tbuf[2] = (val >> 8) & 0xff;
    tbuf[3] = val & 0xff;

    memcpy (ap, tbuf, 4);

    return ap;
  }
#endif

  for (i = 0; i < 4; i++) {
    char *tp = strchr (lp, '.');
    int val;
    char cstr[4];

    if ((tp == NULL) && (i != 3))
      return NULL;		/* Not engouh dots */

    if (tp != NULL)
      *tp = '\0';

    val = atoi (lp);

    if ((val < 0) || (val > 255))	/* U wish ;-) */
      return NULL;

    g_snprintf (cstr, sizeof (cstr), "%d", val);

    if (strcmp (cstr, lp) != 0)
      return NULL;		/* Perhaps alpha chars? */

    /* We're sane, store */
    tbuf[i] = (guint8) val;

    lp = tp;			/* Proceed to next one */
    if (lp != NULL)
      lp++;

  }

  if (lp)
    return NULL;		/* trailing garbage */

  memcpy (ap, tbuf, 4);

  return ap;
}

/*
 * at_inet6_pton, return the human readable IPv6 string as a packed IPv6 address
 *
 * at_inet6_pton(char *cp, guint8 *ap)
 *
 * Returns *ap on success, NULL on error, in which case *cp is not a valid
 * formatted ipv6 address
 */
guint8 *
at_inet6_pton (char *cp, guint8 * ap)
{
  int i;
  int expanded = 0;		/* Have we found a :: */
  char localcopy[8 * 4 + 7 + 1];
  char *mp = localcopy;

  if (strlen (cp) + 1 > sizeof (localcopy)) {
    /* this address is too long */
    return NULL;
  }

  strcpy (mp, cp);
  bzero (ap, 16);

  /* Go and scan all words */
  for (i = 0; i < 8; i++) {
    char *p = strchr (mp, ':');
    guint16 val;

    /* Maybe this is the :: situation? */
    if (mp == p) {
      int j = 0;
      char *p2 = p;

      if (expanded) {
	return NULL;		/* More than 1 expansion
				   not allowed */
      }
      if (*(p + 1) == '\0')
	break;			/* We're at the end, finish */

      p2++;
      /* Find out how much there is following */
      while (p2 != NULL) {
	p2 = strchr (p2, ':');
	j++;
	if (p2)
	  p2++;
      }

      if ((i + j) >= 8) {
	return NULL;		/* To much */
      }

      i = 8 - j;		/* Advance us */
      p++;
      p = strchr (p, ':');	/* And continue */

      mp++;
      expanded = 1;
    }

    if ((p == NULL) && (i != 7)) {
      return NULL;
    }

    if (p != NULL)
      *p = '\0';

    if (!at_hexword (mp, &val)) {
      return NULL;
    }

    ap[i * 2] = val >> 8;
    ap[(i * 2) + 1] = val & 0xff;

    mp = p;
    mp++;
  }

  if ((mp - 1) != NULL)
    return NULL;		/* extra characters */
  return ap;
}

/* at_inet4_aton, convert human readable IPv4 address to packed in_addr
 * structure. It is just an interface to at_inet4_pton modeled after inet_ntoa
 *
 * int at_inet4_aton(const char *cp, struct in_addr *inp)
 *
 */
int at_inet4_aton(const char *cp, struct in_addr *inp)
{
	return (NULL != at_inet4_pton((char *) cp, (guint8 *) inp));
}

#ifdef HAVE_AF_INET6
/* at_inet6_aton, convert human readable IPv6 address to packed in6_addr
 * structure. It is just an interface to at_inet6_pton modeled after inet_ntoa
 * And is only provided if the host support IPv6 in it's libc
 * 
 * int at_inet6_aton(const char *cp, struct in6_addr *inp)
 *
 */
int at_inet6_aton(const char *cp, struct in6_addr *inp)
{
	return (NULL != at_inet6_pton((char *) cp, (guint8 *) inp));
}
#endif
