/* conf.h - All of the structures used for the config routines.

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */
#ifndef CONF_H
#define CONF_H

#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>

#include "../include/common.h"
#include "db_rb.h"
#include "link.h"

#define CONF_NAMEDBOOT	1
#define CONF_NAMEDCONF	2
#define CONF_DENTSCONF	3

/* Controls how the program deals with invalid domain names */
typedef enum
{
  CN_FAIL,			/* don't accept it */
  CN_WARN,			/* log warning */
  CN_IGNORE,			/* ignore it, don't do anything */
}
check_names_t;

typedef enum
{
  FWD_FIRST,
  FWD_ONLY,
  FWD_IF_NO_ANSWER,
  FWD_IF_NO_DOMAIN,
}
forward_policy_t;

typedef enum
{
  TRANS_ONE,
  TRANS_MANY,
}
transfer_format_t;

struct ipnm
{
  struct in_addr a;
  u_char l;
};

struct ip_name
{
  struct in_addr a;
  u_char l;

  guint8 flags;			/* not (!) */

  dbrbnode namenode;
  u_char name[33];
};

typedef enum
{
  IPL_TYPE_INDIRECT,
  IPL_TYPE_DIRECT,
}
iplist_type_t;

#define IPMATCH_NOT		0x0001

struct iplist
{
  struct iplist *next, *prev;

  linkedlist e;
};

struct iplist_entry
{
  struct iplist_entry *next, *prev;

  iplist_type_t type;
  guint8 flags;			/* not (!) */
  union
  {
    struct ipnm ipnm;
    struct iplist *iplist;
  }
  u;
};

struct listenon
{
  struct listenon *next, *prev;

  guint16 port;
  struct iplist *iplist;
};

/*
 * Common configuration information. Information specified from a
 *  config file or control facility
 *
 * All file options contain the conplete path. Must point to a file. They
 *  are all malloc'd to save space in the structure
 */
typedef struct
{
  u_char *directory;		/* default directory to look for */
  /* DNS related files */
  u_char *pid_file;		/* .pid file location */
  u_char *stats_file;		/* statistics file location */
  u_char *mem_stats_file;	/* mem statistics file location */
  u_char *dump_file;		/* database dump file */

  check_names_t cn_master;	/* check-names master */
  check_names_t cn_slave;	/* check-names slave */
  check_names_t cn_response;	/* check-names response */

  forward_policy_t forward;	/* forwarding policy */

  transfer_format_t transfer;	/* transfer format */

  gboolean host_statistics;	/* info on who gave us this answer */
  gboolean dealloc_on_exit;	/* deallocate all memory on exit? */
  gboolean recursion;		/* allow recursion? */
  gboolean notify;		/* send NOTIFY messages? */
  gboolean auth_nxdomain;	/* set AA on nxdomain? */
  gboolean multiple_cnames;	/* allow multiple CNAMEs? */

  guint16 transfers_in;		/* maximum zt in at one time */
  guint16 transfers_per_ns;	/* maximum zt per ns at one time */
  guint16 transfers_out;	/* maximum outbound zt */
  guint16 max_transfer_time_in;	/* number of minutes max for a zt */

  /* resource limits */
  int datasize;
  int stacksize;
  int coresize;
  int files;

  linkedlist listen;		/* ports/IP's to listen on */

  struct iplist *allow_query;	/* who can query from us? */
  struct iplist *allow_transfer;	/* who can zone transfer from us? */

  struct iplist *topology;	/* artificial network topology */

  struct sockaddr_in query_source;	/* where to send packets from? */
}
configrec;

#ifndef DECLARE_CONFIG
extern configrec config;
#else
configrec config = {
  /* Be careful setting strings here, if they get overriden in the config */
  /* file, we just free() the previous. free() doesn't work on string */
  /* constants */
  NULL,				/* directory */
  NULL,				/* pid_file */
  NULL,				/* stats_file */
  NULL,				/* mem_stats_file */
  NULL,				/* dump_file */

  CN_WARN,			/* cn_master */
  CN_WARN,			/* cn_slave */
  CN_IGNORE,			/* cn_response */

  FWD_FIRST,			/* forward */

  0,				/* host_statistics */
  0,				/* dealloc_on_exit */
  1,				/* recursion */
  0,				/* notify */
  1,				/* auth_nxdomain */
  0,				/* multiple_cnames */

  10,				/* transfers_in */
  2,				/* transfers_per_ns */
  10,				/* transfers_out */
  7200,				/* max_transfer_time_in */

  -1,				/* datasize */
  -1,				/* stacksize */
  -1,				/* coresize */
  -1,				/* files */
};
#endif

/* Prototypes */

/* conf.c */
struct iplist *iplist_create (void);
void iplist_add (struct iplist *ipl, struct iplist_entry *iple);
struct iplist_entry *iplist_entry_nm (struct in_addr ip, u_char length);
struct iplist_entry *iplist_entry_acl (struct iplist *ipl);
struct iplist_entry *iplist_name (u_char * name);
void conf_init (void);

#ifdef DENTS_DEBUG
void print_iplist (struct iplist *ipl);
#endif /* DENTS_DEBUG */

/* conf_nb.c */
int conf_nb_read (unsigned char *configfn);

/* conf_nc_parser.y */
int conf_nc_read (unsigned char *configfn);

#endif
