/* conf_nb.c - Reads a named.boot (BIND <= 4.9) configfile.

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>		/* inet_aton */
#include <ctype.h>		/* isspace */
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>		/* struct timeval (FreeBSD) */
#endif /* HAVE_SYS_TIME_H */
#include <sys/resource.h>	/* {get,set}rlimit */

/* Add dents headers here */
#include "../include/common.h"
#include "common.h"
#include "dns.h"
#include "string.h"
#include "module.h"

/*
 * The maximum length of a line allowed in the configuration file
 */
#define READLINEBUF 4096

/*
 * Maximum number of arguments allowed per line. Just be pessimistic and we
 *  won't have any problems (i hope)
 */
#define MAX_CONFIG_ARGS 255

typedef struct {
  guint16 linenum;
  guchar directory[PATH_MAX + 1];
} conf_nb_staterec;

/*
 * Each keyword get's it's own function to do with the arguments as it wishes
 */
static int conf_nb_primary(conf_nb_staterec * state, int argc,
			    guchar * argv[]);
static int conf_nb_secondary(conf_nb_staterec * state, int argc,
			      guchar * argv[]);
static int conf_nb_slave(conf_nb_staterec * state, int argc,
			  guchar * argv[]);
static int conf_nb_cache(conf_nb_staterec * state, int argc,
			  guchar * argv[]);
static int conf_nb_directory(conf_nb_staterec * state, int argc,
			      guchar * argv[]);
static int conf_nb_options(conf_nb_staterec * state, int argc,
			    guchar * argv[]);
static int conf_nb_xfernets(conf_nb_staterec * state, int argc,
			     guchar * argv[]);
static int conf_nb_sortlist(conf_nb_staterec * state, int argc,
			     guchar * argv[]);
static int conf_nb_include(conf_nb_staterec * state, int argc,
			    guchar * argv[]);
static int conf_nb_limit(conf_nb_staterec * state, int argc,
			  guchar * argv[]);
static int conf_nb_checknames(conf_nb_staterec * state, int argc,
			       guchar * argv[]);
static int conf_nb_forwarders(conf_nb_staterec * state, int argc,
			       guchar * argv[]);
static int conf_nb_noforward(conf_nb_staterec * state, int argc,
			      guchar * argv[]);
static int conf_nb_bogusns(conf_nb_staterec * state, int argc,
			    guchar * argv[]);
static int conf_nb_maxfetch(conf_nb_staterec * state, int argc,
			     guchar * argv[]);
static int conf_nb_mod_primary(conf_nb_staterec * state, int argc,
				guchar * argv[]);

struct conf_nb_keywordrec {
  char *keyword;
  int (*func) (conf_nb_staterec *, int, guchar *[]);
} conf_nb_keywords[] = {
  { "primary", conf_nb_primary },
  { "secondary", conf_nb_secondary },
  { "slave", conf_nb_slave },
  { "cache", conf_nb_cache },
  { "directory", conf_nb_directory },
  { "options", conf_nb_options },
  { "xfernets", conf_nb_xfernets },
  { "sortlist", conf_nb_sortlist },
  { "include", conf_nb_include },
  { "limit", conf_nb_limit },
  { "check-names", conf_nb_checknames },
  { "forwarders", conf_nb_forwarders },
  { "noforward", conf_nb_noforward },
  { "bogusns", conf_nb_bogusns },
  { "max-fetch", conf_nb_maxfetch },
  { "mod-primary", conf_nb_mod_primary }
};

#define NUM_CONF_NB_KEYWORDS (sizeof(conf_nb_keywords)/sizeof(struct conf_nb_keywordrec))

/*
 * Callbacks
 */

/*
 * conf_nb_primary
 *
 *  Adds a zone and sets self as master
 *
 * usage: primary <zone name> <filename>
 */
static int
conf_nb_primary (conf_nb_staterec * state, int argc, guchar * argv[])
{
  dnszonerec *zone;
  guchar zonename[DNS_MAXNAMELENGTH];
  dnsmodulerec *module;

  if (argc < 3)
    return (1);

  if (argc > 3)
    err_warn ("ignoring extra arguments for primary command on line %d",
	      state->linenum);

  dns_qualify_domain (zonename, argv[1], "");

  /* Make sure the module is loaded */
  module = dns_module_open ("mod_stddb");

  /* Add in the zone */
  zone = dns_zone_create (zonename);

  /* Make the full filename */
  dns_zone_setfilename (zone, "%s/%s", state->directory, argv[2]);

  /* Set the master IP */
  zone->flags |= DNS_ZONE_MASTER;
  zone->flags |= DNS_ZONE_AUTHORITATIVE;

  dns_zone_addmodule (zone, module);
  dns_zone_setloadmod (zone, module);

  return (0);
}

/*
 * conf_nb_secondary
 *
 *  Adds a zone and sets IP address as master, self as slave
 *
 * usage: secondary <zone name> <master IP> [<addit IP> ...] <filename>
 */
static int
conf_nb_secondary (conf_nb_staterec * state, int argc, guchar * argv[])
{
  struct in_addr master;
  dnszonerec *zone;
  guchar zonename[DNS_MAXNAMELENGTH];
  dnsmodulerec *module;

  if (argc < 4)
    return (1);

  /* FIXME: Handle multiple IP's */
  if (-1 == (master.s_addr = inet_addr (argv[2])))
    return 1;

  dns_qualify_domain (zonename, argv[1], "");

  /* Make sure the module is loaded */
  module = dns_module_open ("mod_stddb");

  /* Add in the zone */
  zone = dns_zone_create (zonename);

  /* Make the full filename */
  dns_zone_setfilename (zone, "%s/%s", state->directory, argv[3]);

  /* Set the master IP */
  zone->flags &= ~DNS_ZONE_MASTER;
  dns_zone_setslaveip (zone, master);

  zone->flags |= DNS_ZONE_AUTHORITATIVE;

  dns_zone_addmodule (zone, module);
  dns_zone_setloadmod (zone, module);

  return (0);
}

/*
 * conf_nb_slave
 *
 *  Same as "options forward-only"
 *
 * usage: slave
 */
static int
conf_nb_slave (conf_nb_staterec * state, int argc, guchar * argv[])
{
  if (argc > 1)
    err_warn ("ignoring extra arguments for slave command on line %d",
	      state->linenum);

  /* FIXME: Do something */

  return (0);
}

/*
 * conf_nb_cache
 *
 *  Adds a hint cache
 *
 * usage: cache <origin> <filename>
 */
static int
conf_nb_cache (conf_nb_staterec * state, int argc, guchar * argv[])
{
  dnshintrec *hint;

  /* I have no clue as to why BIND calls them a cache. The RFC's */
  /*  explicitily state this information should never be given back */
  /*  to a user. It should only be used as a "hint" for recursive */
  /*  queries --jerdfelt */

  if (argc < 3)
    return (1);

  if (argc > 3)
    err_warn ("ignoring extra arguments for cache command on line %d",
	      state->linenum);

  /* Allocate some memory */
  hint = block_allocandadd (&dns.hint);

  strncpyzt (hint->zonename, argv[1], DNS_MAXNAMELENGTH);
  g_snprintf (hint->filename, PATH_MAX, "%s/%s", state->directory, argv[2]);

  return (0);
}

/*
 * conf_nb_directory
 *
 *  Changes the default directory prepend
 *
 * usage: directory <directory>
 */
static int
conf_nb_directory (conf_nb_staterec * state, int argc, guchar * argv[])
{
  int len;

  if (argc < 2)
    return (1);

  if (argc > 2)
    err_warn ("ignoring extra arguments for directory command on line %d",
	      state->linenum);

  /* Copy the directory name */
  strncpyzt (state->directory, argv[1], PATH_MAX);

  /* If there was nothing there, then error out */
  if (!(len = strlen (state->directory))) {
    state->directory[0] = 0;
    return (2);
  }

  /* Chop off the trailing /, if it's there */
  if (state->directory[len - 1] == '/')
    state->directory[len - 1] = 0;

  /* FIXME: now we should check to make sure it's a directory */

  return (0);
}

/*
 * conf_nb_options
 *
 *  Specifies options to be turned on
 *
 * usage: options <option> [...]
 */
static int
conf_nb_options (conf_nb_staterec * state, int argc, guchar * argv[])
{
  int i, error = 0;

  /*
   * Possible options:
   *  no-recursion      Turns off recursion
   *  no-fetch-glue     Don't fetch glue information
   *  query-log         Turns on query logging
   *  forward-only      Send queries to only those in forwarders list
   *  fake-iquery       Send bogus response to Inverse Queries
   */

  if (argc < 2)
    return (1);

  for (i = 1; i < argc; i++) {
    if (stricmp (argv[i], "no-rcursion") == 0) {
      /* FIXME: Turn off recursion */
    } else if (stricmp (argv[i], "no-fetch-glue") == 0) {
      /* We don't do this anyway, so we'll ignore it */
    } else if (stricmp (argv[i], "query-log") == 0) {
      /* FIXME: Log queries */
    } else if (stricmp (argv[i], "forward-only") == 0) {
      /* FIXME: Add code here along with forwarders */
    } else if (stricmp (argv[i], "fake-iquery") == 0) {
      /* We'll prolly never support this since it's lame */
    } else {
      err_warn ("unknown option \"%s\" on line %d", argv[i], state->linenum);
      error++;
    }
  }

  return (error);
}

/*
 * conf_nb_xfernets
 *
 *  Limits who can do zone transfers
 *
 * usage: xfernets <mask> [...]
 */
static int
conf_nb_xfernets (conf_nb_staterec * state, int argc, guchar * argv[])
{
  if (argc < 2)
    return (1);

  /* FIXME: Do something */

  return (0);
}

/*
 * conf_nb_sortlist
 *
 *  List used for sorting DNS servers closest and thusly "best" server to
 * query
 *
 * usage: sortlist <mask> [...]
 */
static int
conf_nb_sortlist (conf_nb_staterec * state, int argc, guchar * argv[])
{
  if (argc < 2)
    return (1);

  /* FIXME: Do something */

  return (0);
}

/*
 * conf_nb_include
 *
 *  Includes another file containing more configuration
 *
 * usage: include <filename>
 */
static int
conf_nb_include (conf_nb_staterec * state, int argc, guchar * argv[])
{
  if (argc < 2)
    return (1);

  if (argc > 2)
    err_warn ("ignoring extra arguments for include command on line %d",
	      state->linenum);

  /* FIXME: Do something */

  return (0);
}

#if !defined(RLIMIT_NOFILE) && defined(RLIMIT_OFILE)
#define RLIMIT_NOFILE RLIMIT_OFILE
#endif

/*
 * conf_nb_limit
 *
 *  Set's up process limits on the daemon and other limits
 *
 * usage: limit <type> <size>
 */
static int
conf_nb_limit (conf_nb_staterec * state, int argc, guchar * argv[])
{
  struct rlimit rlim;
  unsigned long size;

  /*
   * Possible limit types:
   *  datasize          Process data size
   *  files             Maximum number of file descriptors
   *  transfers-in      Maximum number of pull zone transfers
   *  transfers-per-ns  Maximum number of pull zone transfers to one NS
   */

  if (argc < 3)
    return (1);

  if (argc > 3)
    err_warn ("ignoring extra arguments for limit command on line %d",
	      state->linenum);

  /* Parse out the size they want */
  if ((size = strtoul (argv[2], NULL, 10)) == ULONG_MAX) {
    err_warn ("invalid size \"%s\" on line %d", argv[2], state->linenum);
    return (1);
  }

  if (stricmp (argv[1], "datasize") == 0) {
    /* Get the limit */
    if (getrlimit (RLIMIT_DATA, &rlim) < 0) {
      err_warn ("getrlimit(RLIMIT_DATA) returned < 0: %m");
      return (0);
    }

    DPRINTF (("conf_nb_limit: setting RLIMIT_DATA to %lu, max is %ld\n",
	      size, rlim.rlim_max));

    /* Warn if the size they specified is > than the max */
    if (size > rlim.rlim_max)
      err_warn ("size %d > rlim_max on line %d, trying anyway", size,
		state->linenum);

    /* Try setting it */
    rlim.rlim_cur = size;
    if (setrlimit (RLIMIT_DATA, &rlim) < 0) {
      /* Woops, didn't work */
      err_warn ("setrlimit(RLIMIT_DATA, %d) returned < 0: %m", size);
      return (0);
    }
  } else if (stricmp (argv[1], "files") == 0) {
    /* Get the limit */
    if (getrlimit (RLIMIT_NOFILE, &rlim) < 0) {
      err_warn ("getrlimit(RLIMIT_NOFILE) returned < 0: %m");
      return (0);
    }

    DPRINTF (("conf_nb_limit: setting RLIMIT_NOFILE to %lu, max is %ld\n",
	      size, rlim.rlim_max));

    /* Warn if the size they specified is > than the max */
    if (size > rlim.rlim_max)
      err_warn ("size %d > rlim_max on line %d, trying anyway", size,
		state->linenum);

    /* Try setting it */
    rlim.rlim_cur = size;
    if (setrlimit (RLIMIT_NOFILE, &rlim) < 0) {
      /* Woops, didn't work */
      err_warn ("setrlimit(RLIMIT_NOFILE, %d) returned < 0: %m", size);
      return (0);
    }
  } else if (stricmp (argv[1], "transfers-in") == 0) {
    /* FIXME: Do something */
  } else if (stricmp (argv[1], "transfers-per-ns") == 0) {
    /* FIXME: Do something */
  } else {
    err_warn ("unknown limit type \"%s\" on line %d", argv[1],
	      state->linenum);
    return (1);
  }

  return (0);
}

/*
 * conf_nb_checknames
 *
 *  Specifies what to do with illegal domain names
 *
 * usage: check-names <type> <action>
 */
static int
conf_nb_checknames (conf_nb_staterec * state, int argc, guchar * argv[])
{
  /*
   * Possible types:
   *  primary           Primary zone
   *  secondary         Secondary zone
   *  response          Response
   *
   * Possible actions:
   *  fail              Name is invalid
   *  warn              Passes, but logs
   *  ignore            Ignores any errors
   */
  if (argc < 3)
    return (1);

  if (argc > 3)
    err_warn ("ignoring extra arguments for check-names command on line %d",
	      state->linenum);

  /* FIXME: Do something */

  return (0);
}

/*
 * conf_nb_forwarders
 *
 *  List of IP's to forward queries to
 *
 * usage: forwarders <IP> [...]
 */
static int
conf_nb_forwarders (conf_nb_staterec * state, int argc, guchar * argv[])
{
  if (argc < 2)
    return (1);

  /* FIXME: Do something */

  return (0);
}

/*
 * conf_nb_noforward
 *
 * this contains a list of zones *not* to forward queries for. It's
 * a patch to named, but in order to get around some bugs in bind 4.9,
 * a lot of sites use this (We use this on our internal nameserver to keep
 * from forwarding queries for  non-published internal domains)
 *
 * usage: noforward <domain> [...]
 */
static int
conf_nb_noforward (conf_nb_staterec * state, int argc, guchar * argv[])
{
  if (argc < 2)
    return (1);

  /* FIXME: Do something */

  return (0);
}

/*
 * conf_nb_bogusns
 *
 *  List of blacklisted Name Servers
 *
 * usage: bogusns <IP> [...]
 */
static int
conf_nb_bogusns (conf_nb_staterec * state, int argc, guchar * argv[])
{
  if (argc < 2)
    return (1);

  /* FIXME: Do something */
  return (0);
}

/*
 * conf_nb_maxfetch
 *
 *  Identical to "limit transfers-in"
 *
 * usage: max-fetch <num>
 */
static int
conf_nb_maxfetch (conf_nb_staterec * state, int argc, guchar * argv[])
{
  if (argc < 2)
    return (1);

  if (argc > 2)
    err_warn ("ignoring extra arguments for max-fetch command on line %d",
	      state->linenum);

  /* FIXME: Do something */
  return (0);
}

/*
 * conf_nb_mod_primary
 *
 *  Adds a zone controled by a module and sets self as master
 *
 * usage: mod_primary <zone name> <module_name> [ other args ...]
 */
static int
conf_nb_mod_primary (conf_nb_staterec * state, int argc, guchar * argv[])
{
#define ROUTINE "conf_nb_mod_primary"
  dnszonerec *zone;
  guchar zonename[DNS_MAXNAMELENGTH];
  dnsmodulerec *module;
  char *zone_name_in,		/* argument 2 on the config line */
   *module_name,		/* argument 3 on the config line */
    module_path[256];		/* Scratch buffer for formatting the shared lib path */

  DPRINTF (("%s: Starting. %d args\n", ROUTINE, argc));

  if (argc < 3) {
    err_warn
      ("Invalid syntax in config file.  Expected mod_primary <zone> <filename> <module_name>\n");
    return (1);
  }

  if (argc > 3)
    err_warn ("Passing extra arguments for mod_primary command on line %d",
	      state->linenum);

  zone_name_in = argv[1];
  module_name = argv[2];

  dns_qualify_domain (zonename, zone_name_in, "");

  /* SECURITY - buffer overrun possible (?) */
  sprintf (module_path, "../modules/%s/.libs", module_name);

  /* Make sure the module is loaded */
  module_dyn_path_add (module_path);
  module = dns_module_open (module_name);

  if (!module) {
    err_warn ("couldn't initialize module %s in %s\n", module_name, ROUTINE);
    return (1);
  }

  /* Add in the zone */
  zone = dns_zone_create (zonename);

  /* Set the master IP */
  zone->flags |= DNS_ZONE_MASTER;
  zone->flags |= DNS_ZONE_AUTHORITATIVE;

  dns_zone_addmodule (zone, module);
  dns_zone_setloadmod (zone, module);

  /* copy extra args into the zone record argc & argv members */
  if (argc > 3) {
    int numargs = argc - 3;	/* # of args to pass along */
    int i;			/* loop control variable */

    zone->argc = numargs;

    /* TODO: check for malloc() errors */
    zone->argv = (char **) malloc ((argc - 3) * sizeof (char *));

    for (i = 0; i < numargs; i++) {
      zone->argv[i] = strdup (argv[3 + i]);
    }
  }

  return (0);
}

/*
 * End callbacks
 */

/*
 * conf_nb_read
 *
 *  Reads in the named.boot style configuration file and stores everything
 * in memory for later reference
 */
/* FIXME: Should count and deal with the # of errors better */
int conf_nb_read(u_char *configfn)
{
  FILE *configf;
  guchar *p, *p1;
  guchar *argv[MAX_CONFIG_ARGS], linebuf[READLINEBUF + 1];
  guint16 numargs, i, parseerror = 0;
  conf_nb_staterec state;

  /* FIXME: HACK HACK HACK */
  /* Definately the wrong place for this, oh well */
  module_dyn_path_add("../modules/mod_frl/.libs");
  module_dyn_path_add("../modules/mod_recursive/.libs");
  module_dyn_path_add("../modules/mod_stddb/.libs");
  module_dyn_path_add(".");

  /* Reset the state */
  state.linenum = 0;		/* Line 0 */
  strncpyzt(state.directory, ".", PATH_MAX);	/* Current directory */

  /* Open up the configuration file */
  if ((configf = fopen(configfn, "rt")) == NULL) {
    err_warn("conf_nb_read: Couldn't open %s!", configfn);
    return -1;
  }

  /* Loop while we have data */
  while (!feof(configf)) {
    /* Read in the configuration file a line at a time */
    if (fgets(linebuf, READLINEBUF, configf) == NULL) {
      if (ferror(configf))
	err_fatal("conf_nb_read: Error reading configuration file %s!",
		   configfn);

      continue;
    }

    /* Check to see if the line we read in was longer than our buffer and */
    /*  give a warning */
    if (strlen(linebuf) >= (READLINEBUF - 1))
      err_warn("conf_nb_read: Warning: Line %d is >= %s characters long, truncating",
	 state.linenum, READLINEBUF);
    else
      state.linenum++;

    p = linebuf;

    /* Let's dump any beginning white space */
    while (isspace(*p))
      p++;

    /* Let's start ruling out blank lines and comments */
    if (!strlen(p))
      continue;

    /* Chop any trailing new line chars */
    p1 = p + strlen(p) - 1;
    while ((p1 >= p) && strchr("\r\n", *p1)) {
      *p1 = 0;
      p1--;
    }

    if (!strlen(p))
      continue;

    /* Comments are preceded by a ; or # character. Ignore any lines with */
    /*  those characters. Technically, the # character isn't supposed to */
    /*  be used as a comment, however they cannot appear in a domain name */
    /*  so we treat it as a comment to make users happy --jerdfelt */
    if ((*p == ';') || (*p == '#'))
      continue;

    /* Let's start parsing out the arguments on this line */
    numargs = 0;

    /* Next find the argument */
    while (p && *p && (numargs < MAX_CONFIG_ARGS)) {
      /* Dump any beginning white space */
      while (isspace(*p) && *p)
	p++;

      argv[numargs++] = p1 = p;

      while (!isspace(*p1) && *p1)
	p1++;

      if (*p1)
	p = p1 + 1;
      else
	p = NULL;

      *p1 = 0;
    }

    /*
       #ifdef DENTS_DEBUG
       for (i = 0; i < numargs; i++) 
       printf("[%s] ",argv[i]);
       printf("\n");
       #endif
     */

    /* Let's run through all of the different possibilites for the keyword */
    /* FIXME: Use the RB tree stuff for quick lookups (like it makes a */
    /*  difference) */
    for (i = 0; i < NUM_CONF_NB_KEYWORDS; i++)
      if (!stricmp(conf_nb_keywords[i].keyword, argv[0])) {
	parseerror += conf_nb_keywords[i].func(&state, numargs, argv);
	break;
      }

    if (i >= NUM_CONF_NB_KEYWORDS) {
      /* Dunno what they're trying to do */
      err_warn("conf_nb_read: Unknown keyword %s at line %d", argv[0],
		state.linenum);
      parseerror++;
    }
  }
  fclose(configf);

  return parseerror;
}

