/* conf_nc_lexer.c - Lexical analyzer for named.conf style config files.

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "../include/common.h"
#include "dents.h"
#include "../include/err.h"
#include "../include/db.h"
#include "db_rb.h"

#include "conf_nc_parser.h"
#include "conf_nc_lexer.h"
#include "addr_tools.h"		/* at_inet4_aton() */

struct keyword
{
  int token;			/* Token number (per conf_nc_parser.h) */
  dbrbnode labelnode;
  u_char *label;
};

static struct keyword *lexer_find_keyword (u_char * string);

struct keyword keywords[] = {
  LEX_KW ("acl", C_ACL),
  LEX_KW ("address", C_ADDRESS),
  LEX_KW ("algorithm", C_ALGORITHM),
  LEX_KW ("allow-query", C_ALLOW_QUERY),
  LEX_KW ("allow-transfer", C_ALLOW_TRANSFER),
  LEX_KW ("allow-update", C_ALLOW_UPDATE),
  LEX_KW ("also-notify", C_ALSO_NOTIFY),
  LEX_KW ("auth-nxdomain", C_AUTH_NXDOMAIN),
  LEX_KW ("bogus", C_BOGUS),
  LEX_KW ("category", C_CATEGORY),
  LEX_KW ("channel", C_CHANNEL),
  LEX_KW ("check-names", C_CHECK_NAMES),
  LEX_KW ("cleaning-interval", C_CLEANING_INTERVAL),
  LEX_KW ("coresize", C_CORESIZE),
  LEX_KW ("datasize", C_DATASIZE),
  LEX_KW ("deallocate-on-exit", C_DEALLOC_ON_EXIT),
  LEX_KW ("debug", C_DEBUG),
  LEX_KW ("default", C_DEFAULT),
  LEX_KW ("directory", C_DIRECTORY),
  LEX_KW ("dump-file", C_DUMP_FILE),
  LEX_KW ("dynamic", C_DYNAMIC),
  LEX_KW ("fail", C_FAIL),
  LEX_KW ("fake-iquery", C_FAKE_IQUERY),
  LEX_KW ("false", C_FALSE),
  LEX_KW ("fetch-glue", C_FETCH_GLUE),
  LEX_KW ("file", C_FILE),
  LEX_KW ("files", C_FILES),
  LEX_KW ("first", C_FIRST),
  LEX_KW ("forward", C_FORWARD),
  LEX_KW ("forwarders", C_FORWARDERS),
  LEX_KW ("hint", C_HINT),
  LEX_KW ("host-statistics", C_HOST_STATISTICS),
  LEX_KW ("if-no-answer", C_IF_NO_ANSWER),
  LEX_KW ("if-no-domain", C_IF_NO_DOMAIN),
  LEX_KW ("ignore", C_IGNORE),
  LEX_KW ("include", C_INCLUDE),
  LEX_KW ("interface-interval", C_INTERFACE_INTERVAL),
  LEX_KW ("key", C_KEY),
  LEX_KW ("keys", C_KEYS),
  LEX_KW ("listen-on", C_LISTEN_ON),
  LEX_KW ("logging", C_LOGGING),
  LEX_KW ("many-answers", C_MANY_ANSWERS),
  LEX_KW ("master", C_MASTER),
  LEX_KW ("masters", C_MASTERS),
  LEX_KW ("max-transfer-time-in", C_MAX_TRANSFER_TIME_IN),
  LEX_KW ("memstatistics-file", C_MEMSTATISTICS_FILE),
  LEX_KW ("multiple-cnames", C_MULTIPLE_CNAMES),
  LEX_KW ("named-xfer", C_NAMED_XFER),
  LEX_KW ("no", C_NO),
  LEX_KW ("notify", C_NOTIFY),
  LEX_KW ("null", C_NULL),
  LEX_KW ("one-answer", C_ONE_ANSWER),
  LEX_KW ("only", C_ONLY),
  LEX_KW ("options", C_OPTIONS),
  LEX_KW ("pid-file", C_PID_FILE),
  LEX_KW ("port", C_PORT),
  LEX_KW ("print-category", C_PRINT_CATEGORY),
  LEX_KW ("print-severity", C_PRINT_SEVERITY),
  LEX_KW ("print-time", C_PRINT_TIME),
  LEX_KW ("query-source", C_QUERY_SOURCE),
  LEX_KW ("recursion", C_RECURSION),
  LEX_KW ("response", C_RESPONSE),
  LEX_KW ("secret", C_SECRET),
  LEX_KW ("server", C_SERVER),
  LEX_KW ("severity", C_SEVERITY),
  LEX_KW ("size", C_SIZE),
  LEX_KW ("slave", C_SLAVE),
  LEX_KW ("stacksize", C_STACKSIZE),
  LEX_KW ("statistics-file", C_STATISTICS_FILE),
  LEX_KW ("statistics-interval", C_STATISTICS_INTERVAL),
  LEX_KW ("stub", C_STUB),
  LEX_KW ("syslog", C_SYSLOG),
  LEX_KW ("topology", C_TOPOLOGY),
  LEX_KW ("transfer-format", C_TRANSFER_FORMAT),
  LEX_KW ("transfer-source", C_TRANSFER_SOURCE),
  LEX_KW ("transfers", C_TRANSFERS),
  LEX_KW ("transfers-in", C_TRANSFERS_IN),
  LEX_KW ("transfers-out", C_TRANSFERS_OUT),
  LEX_KW ("transfers-per-ns", C_TRANSFERS_PER_NS),
  LEX_KW ("true", C_TRUE),
  LEX_KW ("type", C_TYPE),
  LEX_KW ("unlimited", C_UNLIMITED),
  LEX_KW ("versions", C_VERSIONS),
  LEX_KW ("warn", C_WARN),
  LEX_KW ("yes", C_YES),
  LEX_KW ("zone", C_ZONE),
};

#define NUM_KEYWORDS (sizeof(keywords) / sizeof(struct keyword))

typedef enum
{
  SCAN_SCANNING,		/* We're scanning for the next token */
  SCAN_QUOTED_STRING,		/* Quoted string */
  SCAN_LITERAL,			/* Normal string */
  SCAN_NUMBER,			/* Number */
  SCAN_IPV4,			/* IPv4 Address */
  SCAN_IPV6,			/* IPv6 Address (not working yet) */
}
lexer_scan_state_t;

#define LEX_EOF			0x01

struct lexer_state
{
  struct lexer_state *parent;	/* Original file that did the "include" */

  u_char *filename;		/* Filename */
  FILE *f;			/* File handle */

  int lineno;			/* Line number */
  lexer_scan_state_t state;	/* Current state */
  int flags;			/* Flags */

  int warnings;			/* Number of warnings */
  int errors;			/* Number of errors */
};

static struct lexer_state *current = NULL;

#define LEX_LAST_WAS_DOT	0x01
#define LEX_CONSECUTIVE_DOTS	0x02

#define LEX_MAXSTRINGLEN	1024

struct lexer_current_parse
{
  u_char string[LEX_MAXSTRINGLEN + 1];
  int len;
  int numdots;
  int flags;
};

#define whitespace(c) ((c) == ' ' || (c) == '\t' || (c) == '\n')
#define domain_char(c) (isalnum((c)) || (c) == '.' || (c) == '-')
#define special_char(c) (c == ';' || c == '{' || c == '}' || c == '!' || \
		c == '/' || c == '\"' || c == '*')
#define literal_char(c) (!whitespace(c) && !special_char(c))

static int last_token;
static YYSTYPE last_yylval;

u_char *
lexer_token_to_text (int token, YYSTYPE * yylval)
{
  static u_char buffer[LEX_MAXSTRINGLEN + 1];

  if (!token)
    strncpyzt (buffer, "<EOF>", sizeof (buffer) - 1);
  else if (token < 255)
    g_snprintf (buffer, sizeof (buffer) - 1, "'%c'", token);
  else {
    switch (token) {
    case EOS:
      strncpyzt (buffer, ";", sizeof (buffer) - 1);
      break;
    case STRING:
      g_snprintf (buffer, sizeof (buffer) - 1, "\'%s\'", yylval->string);
      break;
    case QSTRING:
      g_snprintf (buffer, sizeof (buffer) - 1, "\"%s\"", yylval->string);
      break;
    case IPADDR:
      strncpyzt (buffer, inet_ntoa (yylval->ipaddr), sizeof (buffer) - 1);
      break;
    case NUMBER:
      g_snprintf (buffer, sizeof (buffer) - 1, "%ld", yylval->num);
      break;
    case END_INCLUDE:
      strncpyzt (buffer, "<EOI>", sizeof (buffer) - 1);
      break;
    }
  }

  return buffer;
}

/*
 * Wrapper functions for errors
 */
void
lexer_warning (int show_context, u_char * format, ...)
{
  va_list va;
  u_char buf[LEX_MAXSTRINGLEN + 1];
  struct lexer_state *ns;

  va_start (va, format);
  g_vsnprintf (buf, sizeof (buf) - 1, format, va);
  va_end (va);

  /* If there's no file open, skip the fancy stuff below */
  if (!current) {
    err_warn (buf);
    return;
  }

  if (show_context)
    err_warn ("%s in %s, line %d near %s", buf, current->filename,
	      current->lineno, lexer_token_to_text (last_token,
						    &last_yylval));
  else
    err_warn ("%s in %s, line %d", buf, current->filename, current->lineno);

  /* Now show the chain of included files */
  for (ns = current->parent; ns; ns = ns->parent)
    err_warn ("  included from %s, line %d", current->filename,
	      current->lineno);

  current->warnings++;
}

void
lexer_error (int show_context, u_char * format, ...)
{
  va_list va;
  u_char buf[LEX_MAXSTRINGLEN + 1];
  struct lexer_state *ns;

  va_start (va, format);
  g_vsnprintf (buf, sizeof (buf) - 1, format, va);
  va_end (va);

  /* If there's no file open, skip the fancy stuff below */
  if (!current) {
    err_warn (buf);
    return;
  }

  if (show_context)
    err_warn ("error: %s in %s, line %d near %s", buf, current->filename,
	      current->lineno, lexer_token_to_text (last_token,
						    &last_yylval));
  else
    err_warn ("error: %s in %s, line %d", buf, current->filename,
	      current->lineno);

  /* Now show the chain of included files */
  for (ns = current->parent; ns; ns = ns->parent)
    err_warn ("  included from %s, line %d", current->filename,
	      current->lineno);

  current->errors++;
}

void
yyerror (char *format, ...)
{
  va_list va;
  u_char buf[LEX_MAXSTRINGLEN + 1];

  va_start (va, format);
  g_vsnprintf (buf, sizeof (buf) - 1, format, va);
  va_end (va);

  lexer_error (1, buf);

#ifdef DENTS_DEBUG
  abort ();
#endif
}

/*
 * Character Input
 */

void
lexer_scan_comment (int scan_to_end)
{
  int c, prev_star = 0;

  /* Scan till the end of the commnet */
  while (1) {
    c = getc (current->f);

    switch (c) {
    case EOF:
      /* EOF is fine in a scan_to_end comment */
      if (!scan_to_end)
	lexer_error (0, "EOF in comment");

      current->flags |= LEX_EOF;
      return;
    case '*':
      /* Perhaps this is the end of a C style comment */
      prev_star = 1;
      break;
    case '/':
      if (prev_star && !scan_to_end)
	return;

      prev_star = 0;
      break;
    case '\n':
      if (scan_to_end) {
	/* Don't slurp the \n, other portions need it */
	ungetc (c, current->f);
	return;
      } else
	current->lineno++;

      prev_star = 0;
      break;
    default:
      prev_star = 0;
    }
  }
}

int
lexer_get_char (int comment_ok)
{
  int c, nc;

  if (current->flags & LEX_EOF)
    return (EOF);

  c = getc (current->f);

  if (comment_ok) {
    while (c == '/' || c == '#') {
      if (c == '#') {
	lexer_scan_comment (1);
	if (current->flags & LEX_EOF)
	  return (EOF);
	c = getc (current->f);
      } else {
	nc = getc (current->f);
	switch (nc) {
	case EOF:
	  current->flags |= LEX_EOF;
	  return ('/');
	case '*':
	case '/':
	  lexer_scan_comment ((nc == '/'));
	  if (current->flags & LEX_EOF)
	    return (EOF);
	  c = getc (current->f);
	  break;
	default:
	  ungetc ((nc), current->f);
	  return ('/');
	}
      }
    }
  }

  if (c == EOF)
    current->flags |= LEX_EOF;
  else if (c == '\n')
    current->lineno++;

  return (c);
}

void
lexer_unget_char (int c)
{
  if (c == EOF)
    current->flags |= LEX_EOF;
  else {
    ungetc (c, current->f);
    if (c == '\n')
      current->lineno--;
  }
}

void
lexer_clear_lcp (struct lexer_current_parse *lcp)
{
  lcp->len = 0;
  lcp->numdots = 0;
  lcp->flags = 0;
}

char *
lexer_duplicate_lcp (struct lexer_current_parse *lcp)
{
  return strdup (lcp->string);
}

void
lexer_finish_lcp (struct lexer_current_parse *lcp)
{
  lcp->string[lcp->len] = 0;
}

void
lexer_add_to_lcp (struct lexer_current_parse *lcp, int c)
{
  lcp->string[lcp->len] = c;
  lcp->len++;

  if (lcp->len >= LEX_MAXSTRINGLEN) {
    lexer_error (1, "lcp string too long");
    current->state = SCAN_SCANNING;

    lexer_clear_lcp (lcp);
  } else {
    /* Used for determining IP addresses */
    if (c == '.') {
      if (lcp->flags & LEX_LAST_WAS_DOT)
	lcp->flags |= LEX_CONSECUTIVE_DOTS;

      lcp->flags |= LEX_LAST_WAS_DOT;
      lcp->numdots++;
    } else
      lcp->flags &= ~LEX_LAST_WAS_DOT;
  }
}

/*
 * yylex
 *
 *  Breaks naming convention because of YACC. Returns the next token
 */
int
yylex (void)
{
  int c;
  int comment_ok = 1;
  int token = -1;
  struct lexer_current_parse lcp;
  struct keyword *value;

  lexer_clear_lcp (&lcp);

  while (token < 0) {
    c = lexer_get_char (comment_ok);

    switch (current->state) {
    case SCAN_SCANNING:
      if (c == EOF) {
	if (!current->parent)
	  token = 0;
	else {
	  lexer_close ();
	  token = END_INCLUDE;
	}
	break;
      }

      if (whitespace (c))
	break;

      if (literal_char (c)) {
	if (isdigit (c))
	  current->state = SCAN_NUMBER;
	else
	  current->state = SCAN_LITERAL;

	lexer_clear_lcp (&lcp);
	lexer_add_to_lcp (&lcp, c);
      } else {
	if (special_char (c)) {
	  if (c == ';') {
	    token = EOS;
	    break;
	  }

	  if (c == '\"') {
	    lexer_clear_lcp (&lcp);
	    current->state = SCAN_QUOTED_STRING;
	    comment_ok = 0;
	    break;
	  }

	  token = c;
	} else
	  lexer_error (1, "invalid character \'%c\'", c);
      }
      break;

    case SCAN_NUMBER:
      if (literal_char (c)) {
	if (!isdigit (c))
	  current->state = (c == '.') ? SCAN_IPV4 : SCAN_LITERAL;

	lexer_add_to_lcp (&lcp, c);
      } else {
	lexer_unget_char (c);

	current->state = SCAN_SCANNING;
	lexer_finish_lcp (&lcp);
	yylval.num = atoi (lcp.string);
	token = NUMBER;
      }
      break;

    case SCAN_LITERAL:
      if (literal_char (c)) {
	lexer_add_to_lcp (&lcp, c);
      } else {
	lexer_unget_char (c);

	current->state = SCAN_SCANNING;
	lexer_finish_lcp (&lcp);

	/* is it a keyword? */
	if ((value = lexer_find_keyword (lcp.string)) != NULL) {
	  yylval.string = lcp.string;
	  token = value->token;
	} else {
	  yylval.string = lexer_duplicate_lcp (&lcp);
	  token = STRING;
	}
      }
      break;

    case SCAN_IPV4:
      if (literal_char (c)) {
	if (!isdigit (c)) {
	  if (c != '.' || (lcp.flags & LEX_CONSECUTIVE_DOTS))
	    current->state = SCAN_LITERAL;
	}

	lexer_add_to_lcp (&lcp, c);
      } else {
	lexer_unget_char (c);

	if (lcp.numdots > 3 || (lcp.flags & LEX_LAST_WAS_DOT))
	  current->state = SCAN_LITERAL;
	else {
	  if (lcp.numdots == 1) {
	    lexer_add_to_lcp (&lcp, '.');
	    lexer_add_to_lcp (&lcp, '0');
	    lexer_add_to_lcp (&lcp, '.');
	    lexer_add_to_lcp (&lcp, '0');
	  } else if (lcp.numdots == 2) {
	    lexer_add_to_lcp (&lcp, '.');
	    lexer_add_to_lcp (&lcp, '0');
	  }

	  current->state = SCAN_SCANNING;
	  lexer_finish_lcp (&lcp);
	  token = IPADDR;

	  if (at_inet4_aton (lcp.string, &yylval.ipaddr) == 0) {
	    yylval.string = lexer_duplicate_lcp (&lcp);
	    token = STRING;
	  }
	}
      }
      break;

    case SCAN_QUOTED_STRING:
      if (c == EOF) {
	lexer_error (0, "EOF in quoted string");
	return 0;
      } else {
	if (c == '"') {
	  comment_ok = 1;
	  current->state = SCAN_SCANNING;

	  lexer_finish_lcp (&lcp);

	  yylval.string = lexer_duplicate_lcp (&lcp);
	  token = QSTRING;
	} else {
	  /* XXX add backslash escapes here */
	  lexer_add_to_lcp (&lcp, c);
	}
      }
      break;

    default:
      err_fatal ("unknown state");
    }
  }

  last_token = token;
  memcpy ((void *) &last_yylval, (void *) &yylval, sizeof (last_yylval));
  return token;
}

static dbrb keyworddb;

static void
lexer_init_keywords (void)
{
  int i;

  db_rb_create (&keyworddb, db_rb_cmp_pstri);

  for (i = 0; i < NUM_KEYWORDS; i++)
    db_rb_insert (&keyworddb, &keywords[i].labelnode, &keywords[i]);
}

static struct keyword *
lexer_find_keyword (u_char * string)
{
  return db_rb_find (&keyworddb, &string);
}

int
lexer_open (u_char * filename)
{
  struct lexer_state *ns;
  FILE *f;

  if (!(f = fopen (filename, "r"))) {
    lexer_warning (0, "Couldn't open %sfile \"%s\"",
		   (current) ? "included " : "", filename);
    return -1;
  }

  if ((ns = malloc (sizeof (*ns))) == NULL)
    err_fatal ("Couldn't allocate memory for new lexer_state");

  ns->f = f;
  ns->filename = filename;
  ns->lineno = 1;
  ns->state = SCAN_SCANNING;
  ns->flags = 0;
  ns->warnings = ns->errors = 0;
  ns->parent = current;

  current = ns;

  return 0;
}

void
lexer_close (void)
{
  struct lexer_state *ns;

  ns = current;
  current = ns->parent;

  fclose (ns->f);

  free (ns);
}

void
lexer_init (void)
{
  lexer_init_keywords ();
}
