/* db_rb.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/*
 * db_rb.c - A Red Black tree implementation.
 *
 *
 * This code implements a red black tree (a type of self-balancing binary tree)
 * Implements insert, delete, find, minimum and maximum, and pre-order 
 * traversal.
 * 
 * A description from:  
 * http://yallara.cs.rmit.edu.au/~tgunning/CS544/RBdesc.html
 *
 * (begin quote)
 * Red-Black trees first appeared in 1978.(1) However, they were based
 * on a special case of the B-tree developed by Rudolf Bayer in 1972 (2),
 * the so-called 2-3-4 tree.
 *
 * An n node red/black tree has the property that its height is
 * O(lg(n)). Another important property is that a node can be added to a
 * red-black tree and, in O(lg(n)) time, the tree can be readjusted to
 * become a larger red/black tree. Insertion does not require two phases,
 * unlike the AVL tree, another self-balancing binary search tree, so it
 * is much more efficient. Similarly, a node can be deleted from a
 * red/black tree and, in O(lg(n)) time, the tree can be readjusted to
 * become a smaller red/black tree. Due to these properties, red/black
 * trees are useful for data storage and are used, for example, in
 * implementations of the STL in C++ to build fast containers.
 *
 * The structure of the RB tree insures that the tree does not degenerate
 * into a linked list, so lookups always run in time 0(log(n)).  There 
 * is a slight penalty for insert and delete as compared to a simple
 * binary tree, but the time for these operations also run in time 
 * O(log(n)).
 *
 * (end quote)
 *
 * See test driver in ../test/test_db_rb.c
 *
 */
#include "common.h"
#include "db_rb.h"

dbrbnode db_rb_nil = 
  { &db_rb_nil, &db_rb_nil, &db_rb_nil, DB_RB_BLACK, NULL };
	/*  the above global vairable defines an empty dbrbnode */

static void
walksub (void *x, void *y)
{
  dbrbnode *z1, *z2;

  z1 = (dbrbnode *) x;
  z2 = (dbrbnode *) y;

  if (z1 == z2) {
    DPRINTF (
	     ("walksub: we found a node that we either are about to insert that is already in the tree\nor a node that we just deleted that is still in the tree\n"));
  }
}

static void
db_rb_rotateleft (dbrb * db, dbrbnode * n)
{
  dbrbnode *n1;

  n1 = n->right;
  n->right = n1->left;
  if (n1->left != DB_RB_NIL)
    n1->left->parent = n;
  n1->parent = n->parent;
  if (n->parent == DB_RB_NIL) {
    db->root = n1;
  } else {
    if (n == n->parent->left) {
      n->parent->left = n1;
    } else {
      n->parent->right = n1;
    }
  }
  n1->left = n;
  n->parent = n1;
}

static void
db_rb_rotateright (dbrb * db, dbrbnode * n)
{
  dbrbnode *n1;

  n1 = n->left;
  n->left = n1->right;
  if (n1->right != DB_RB_NIL)
    n1->right->parent = n;
  n1->parent = n->parent;
  if (n->parent == DB_RB_NIL) {
    db->root = n1;
  } else {
    if (n == n->parent->left) {
      n->parent->left = n1;
    } else {
      n->parent->right = n1;
    }
  }
  n1->right = n;
  n->parent = n1;
}

static void
db_rb_insertfixup (dbrb * db, dbrbnode * n)
{
  dbrbnode *n1;

  while (n != db->root && n->parent->color == DB_RB_RED) {
    /* We have a violation */
    if (n->parent == n->parent->parent->left) {
      n1 = n->parent->parent->right;
      if (n1->color == DB_RB_RED) {
	n->parent->color = DB_RB_BLACK;
	n1->color = DB_RB_BLACK;
	n->parent->parent->color = DB_RB_RED;
	n = n->parent->parent;
      } else {
	if (n == n->parent->right) {
	  n = n->parent;
	  db_rb_rotateleft (db, n);
	}

	/* Recolor and rotate */
	n->parent->color = DB_RB_BLACK;
	n->parent->parent->color = DB_RB_RED;
	db_rb_rotateright (db, n->parent->parent);
      }
    } else {
      n1 = n->parent->parent->left;
      if (n1->color == DB_RB_RED) {
	n->parent->color = DB_RB_BLACK;
	n1->color = DB_RB_BLACK;
	n->parent->parent->color = DB_RB_RED;
	n = n->parent->parent;
      } else {
	if (n == n->parent->left) {
	  n = n->parent;
	  db_rb_rotateright (db, n);
	}

	/* Recolor and rotate */
	n->parent->color = DB_RB_BLACK;
	n->parent->parent->color = DB_RB_RED;
	db_rb_rotateleft (db, n->parent->parent);
      }
    }
  }

  /* Root is ALWAYS black */
  db->root->color = DB_RB_BLACK;
}				/* end db_rb_insertfixup () */

/*
 * db_rb_insert - insert a new node into an initialized RB tree
 *
 * Insert a new node in the tree.  Note that the key value must
 * be unique.  
 *
 * Inputs:      db - Database handle initialized by db_rb_create()
 *              new - pointer to the dbrbnode field in the buffer
 *                   the caller wants to add to the tree.
 *              ptr - pointer to the beginning of the buffer the
 *                   caller wants to store in the tree.
 *
 * Outputs:     None.
 * Returns:     -1 on failure (e.g. duplicate key value already exists)
 *              0 on success.
 */
gint16
db_rb_insert (dbrb * db, dbrbnode * new, void *ptr)
{
  dbrbnode *n, *parent;
  gint32 ret = 0;

  parent = DB_RB_NIL;

  db_rb_walk (db->root, walksub, new);

  /* Start at the root */
  n = db->root;

  /* Find the correct place to insert the node */
  while (n != DB_RB_NIL) {

    if (!n) {
      DPRINTF (("db_rb_insert: This can not happen, n is null\n"));
      return (-1);
    }

    /* If we hit the exact number, then return this node */
    ret = db->compare ((void *) (new + 1), (void *) (n + 1));

    if (!ret) {
      DPRINTF (("db_rb_insert: Duplicate key already exists\n"));
      return (-1);
    }

    parent = n;
    n = (ret < 0) ? n->left : n->right;
  }

  /* Fill in all of the pertinent information */
  new->ptr = ptr;
  new->parent = parent;
  new->left = DB_RB_NIL;
  new->right = DB_RB_NIL;
  /* It starts off as red */
  new->color = DB_RB_RED;

  if (parent != DB_RB_NIL) {
    /* Insert him into the correct position */
    if (ret < 0)
      parent->left = new;
    else
      parent->right = new;
  } else
    db->root = new;

  /* Then fix it up */
  db_rb_insertfixup (db, new);
  db->numnodes++;

  return (0);

}				/* end db_rb_insert() */

static void
db_rb_deletefixup (dbrb * db, dbrbnode * n)
{
  dbrbnode *n1;

  while (n != db->root && n->color == DB_RB_BLACK) {
    if (n == n->parent->left) {
      n1 = n->parent->right;

      if (n1->color == DB_RB_RED) {
	n1->color = DB_RB_BLACK;
	n->parent->color = DB_RB_RED;
	db_rb_rotateleft (db, n->parent);
	n1 = n->parent->right;
      }

      if (n1->left->color == DB_RB_BLACK && n1->right->color == DB_RB_BLACK) {
	n1->color = DB_RB_RED;
	n = n->parent;
      } else {
	if (n1->right->color == DB_RB_BLACK) {
	  n1->left->color = DB_RB_BLACK;
	  n1->color = DB_RB_RED;
	  db_rb_rotateright (db, n1);
	  n1 = n->parent->right;
	}

	n1->color = n->parent->color;
	n->parent->color = DB_RB_BLACK;
	n1->right->color = DB_RB_BLACK;
	db_rb_rotateleft (db, n->parent);

	n = db->root;
      }
    } else {
      n1 = n->parent->left;

      if (n1->color == DB_RB_RED) {
	n1->color = DB_RB_BLACK;
	n->parent->color = DB_RB_RED;
	db_rb_rotateright (db, n->parent);
	n1 = n->parent->left;
      }

      if (n1->right->color == DB_RB_BLACK && n1->left->color == DB_RB_BLACK) {
	n1->color = DB_RB_RED;
	n = n->parent;
      } else {
	if (n1->left->color == DB_RB_BLACK) {
	  n1->right->color = DB_RB_BLACK;
	  n1->color = DB_RB_RED;
	  db_rb_rotateleft (db, n1);
	  n1 = n->parent->left;
	}

	n1->color = n->parent->color;
	n->parent->color = DB_RB_BLACK;
	n1->left->color = DB_RB_BLACK;
	db_rb_rotateright (db, n->parent);
	n = db->root;
      }
    }
  }

  /* Root is ALWAYS black */
  n->color = DB_RB_BLACK;
}

/*
 * db_rb_delete - remove a node from the RB tree.
 *
 * Inputs:      db - handle to a RB database created with db_rb_create()
 *              n - address of dbrbnode field in buffer to be removed 
 *                   from the tree
 * Outputs:     None.
 * 
 */
void
db_rb_delete (dbrb * db, dbrbnode * n)
{
  dbrbnode *n1, *n2;

  if (n->left == DB_RB_NIL || n->right == DB_RB_NIL) {
    n2 = n;
  } else {
    for (n2 = n->right; n2->left != DB_RB_NIL; n2 = n2->left);
  }

  if (n2->left != DB_RB_NIL)
    n1 = n2->left;
  else
    n1 = n2->right;

  if (n1 != DB_RB_NIL) {
    n1->parent = n2->parent;
  }
  if (n2->parent == DB_RB_NIL) {
    db->root = n1;
  } else {
    if (n2 == n2->parent->left)
      n2->parent->left = n1;
    else
      n2->parent->right = n1;
  }
  if (n2 != n) {
    n->ptr = n2->ptr;
    if (n->left != DB_RB_NIL)
      n->left->parent = n2;
    if (n->right != DB_RB_NIL)
      n->right->parent = n2;
    memcpy (n2, n, sizeof (dbrbnode));
    if (n == db->root)
      db->root = n2;
    if (n2->parent != DB_RB_NIL) {
      if (n2->parent->left == n)
	n2->parent->left = n2;
      else
	n2->parent->right = n2;
    }
  }
  if (n2->color == DB_RB_BLACK)
    db_rb_deletefixup (db, n1);

  memcpy (n, DB_RB_NIL, sizeof (dbrbnode));

  db->numnodes--;
  db_rb_walk (db->root, walksub, n);
}

/*
 * db_rb_find - find a key value in the RB tree
 *
 * Inputs:      db - RB handle initialized with db_rb_create
 *              data - ptr to a key value cast to a void. E.g. if the key is
 *                 a guint32  type:
 *              
 *                   guint32 search_val = 100;
 *              ptr = db_rb_find(&mydb, (void *)&search_val);
 *                  
 * Outputs:     None.
 * Returns:     the 'data' value passed to db_rb_insert ()
 *
 */
void *
db_rb_find (dbrb * db, void *data)
{
  dbrbnode *n;
  int ret;

  /* Start at the root */
  n = db->root;

  /* Loop while there are still more nodes */
  while (n != DB_RB_NIL) {
    if (!(ret = db->compare (data, (void *) (n + 1))))
      /* Bingo */
      return (n->ptr);

    n = (ret < 0) ? n->left : n->right;
  }

  /* Woops, didn't find an exact match */
  return (NULL);
}

/*
 * db_rb_minimum - find the node in the RB tree with the minimum key value 
 *
 * Inputs:      db - RB handle initialized with db_rb_create
 * Outputs:     None.
 * Returns:     NULL on failure
 *      pointer to the node in the RB tree with the minimum key value 
 *      on success
 */
void *
db_rb_minimum (dbrb * db)
{
  dbrbnode *n;

  /* Start at the root */
  n = db->root;

  /* Walk our way down the left */
  if (n != DB_RB_NIL) {
    while (n->left != DB_RB_NIL) {
      if (!n) {
	DPRINTF (("db_rb_minimum: n is null\n"));
	return (NULL);
      }

      n = n->left;
    }
  }

  return (n->ptr);
}

/*
 * db_rb_maximum - find the node in the RB tree with the maximum key value 
 *
 * Inputs:      db - RB handle initialized with db_rb_create
 * Outputs:     None.
 * Returns:     NULL on failure
 *      pointer to the node in the RB tree with the maximum key value 
 *      on success
 */
void *
db_rb_maximum (dbrb * db)
{
  dbrbnode *n;

  /* Start at the root */
  n = db->root;

  /* Walk our way down the right */
  if (n != DB_RB_NIL) {
    while (n->right != DB_RB_NIL) {
      n = n->right;
    }
  }

  return (n->ptr);
}

/*
 * db_rb_create - Initialize a new RB tree.
 * 
 * Inputs:  db - address of a dbrb buffer
 *          compare - a comparison function that takes 2 (void *)
 *            arguments and returns:
 *            0  if the two arguments are equal
 *            a value >0 if the first argument is greater than 
 *                  the second argument
 *            a value <0 if the first argument is less than 
 *                  the second argument
 * Returns: None
 *
 */
void
db_rb_create (dbrb * db, int (*compare) ())
{
  db->root = DB_RB_NIL;
  db->numnodes = 0;

  db->compare = compare;
}

/*
 * db_rb_walk - traverse an RB tree in pre-order traversal
 * 
 * Inputs:      rbnode - pointer to a dbrbnode value 
 *                 e.g. the  field declared in a user defined buffer, 
 *                      or db->root field in the dbrb structure.
 *              callback - a function that is called for each node
 *                 in the tree.  The callback function takes two 
 *                 arguments: 
 *
 *                 void callback (drbnode *rbnode, void *ptr)
 *                    rbnode - the current node pointer
 *                    ptr -    the argument passed to db_rb_walk()
 *
 *                 NOTE: you'd be smart not to add/remove nodes in the
 *                 tree while you are walking it!
 *              ptr - arbitrary value passed to the callback function
 *                 as each node is traversed.
 *
 * Outputs:     None.
 *
 */
void
db_rb_walk (dbrbnode * rbnode, void (*callback) (), void *ptr)
{
  if (rbnode != DB_RB_NIL) {
    if (!rbnode) {
      DPRINTF (("db_rb_walk: Uh-oh....  NULL rbnode\n"));
      return;
    }
    if (rbnode == rbnode->left || rbnode == rbnode->right
	|| rbnode == rbnode->parent) {
      DPRINTF (("*** db_rb_walk:  rbnode has a recursive loop\n"));
      return;
    }
  }

  if (rbnode->ptr) {
    callback (rbnode, ptr);
  }

  if (rbnode->left != DB_RB_NIL)
    db_rb_walk (rbnode->left, callback, ptr);

  if (rbnode->right != DB_RB_NIL)
    db_rb_walk (rbnode->right, callback, ptr);
}

/*
 * db_rb_print - print the nodes in an RB tree (sideways) (in-order traversal)
 * 
 * Inputs:      rbnode - pointer to a dbrbnode value 
 *                 e.g. the  field declared in a user defined buffer, 
 *                      or db->root field in the dbrb structure.
 *              callback - a function that is called for each node
 *                 in the tree to print the contents of the node to stdout.
 *                 The intention is to print out the key value.
 *                 The callback function takes two arguments: 
 *
 *                 void callback (drbnode *rbnode, void *data)
 *                    rbnode - the current node pointer
 *                    data -   the data ptr passed to db_rb_insert()
 *                 level - the level of this node.  Call with '0' from
 *                    a user function.
 * Outputs:     None.
 *
 */
void
db_rb_print (dbrbnode * rbnode, void (*callback) (), int level)
{
  if (rbnode != DB_RB_NIL) {
    if (!rbnode) {
      DPRINTF (("db_rb_print: Uh-oh....  NULL rbnode\n"));
      return;
    }
    if (rbnode == rbnode->left || rbnode == rbnode->right
	|| rbnode == rbnode->parent) {
      DPRINTF (("*** db_rb_print:  rbnode has a recursive loop\n"));
      return;
    }
  }

  if (rbnode->left != DB_RB_NIL)
    db_rb_print (rbnode->left, callback, level + 1);

  /* print this node */
  if (rbnode->ptr) {
    int i;

    /* print 2 spaces for each level */
    for (i = 0; i < level; i++) {
      printf ("  ");
    }

    printf ("%c", (rbnode->color == DB_RB_RED) ? '-' : ':');

    callback (rbnode, rbnode->ptr);
    printf ("\n");
  }

  if (rbnode->right != DB_RB_NIL)
    db_rb_print (rbnode->right, callback, level + 1);
}

/*
 * Built-in comparision functions for common types
 * 
 * db_rb_cmp_uint32 - compare two guint32 values
 * db_rb_cmp_int32  - compare two gint32 values
 * db_rb_cmp_uint16 - compare two guint16 values
 * db_rb_cmp_int16  - compare two gint16 values
 * db_rb_cmp_stri   - compare two null-terminated strings (case insensitive)
 * db_rb_cmp_pstri  - compare two null-terminated string ptrs(case insensitive)
 *
 */
int
db_rb_cmp_uint32 (guint32 * i, guint32 * i1)
{
  guint32 ti, ti1;

  memcpy ((void *) &ti, (void *) i, sizeof (guint32));
  memcpy ((void *) &ti1, (void *) i1, sizeof (guint32));

  return (ti > ti1 ? 1 : (ti < ti1 ? -1 : 0));
}

int
db_rb_cmp_int32 (gint32 * i, gint32 * i1)
{
  gint32 ti, ti1;

  memcpy ((void *) &ti, (void *) i, sizeof (gint32));
  memcpy ((void *) &ti1, (void *) i1, sizeof (gint32));

  return (ti > ti1 ? 1 : (ti < ti1 ? -1 : 0));
}

int
db_rb_cmp_uint16 (guint16 * i, guint16 * i1)
{
  guint16 ti, ti1;

  memcpy ((void *) &ti, (void *) i, sizeof (guint16));
  memcpy ((void *) &ti1, (void *) i1, sizeof (guint16));

  return (ti > ti1 ? 1 : (ti < ti1 ? -1 : 0));
}

int
db_rb_cmp_int16 (gint16 * i, gint16 * i1)
{
  gint16 ti, ti1;

  memcpy ((void *) &ti, (void *) i, sizeof (gint16));
  memcpy ((void *) &ti1, (void *) i1, sizeof (gint16));

  return (ti > ti1 ? 1 : (ti < ti1 ? -1 : 0));
}

int
db_rb_cmp_stri (guchar * s, guchar * s1)
{
  return (stricmp (s, s1));
}

int
db_rb_cmp_pstri (guchar ** s, guchar ** s1)
{
  return (stricmp (*s, *s1));
}
