/* dents.c - contains initialization and a mainloop.

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

/* This file parses out the command line.  Reads in the config file.
Does a precheck of the system to make sure it's ready. Starts reading
the zone files and then starts up all of the worker threads. */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* This tells conf.h to define the variable "config". */
#define DECLARE_CONFIG

/* Add dents headers here */
#include "../include/common.h"

#include <signal.h>

#include "dents.h"
#include "dns.h"
#include "module.h"
#include "signal_handlers.h"
#include "conf_nc_lexer.h"

#ifdef WITH_CTL
#include "ctlfac.h"
#include "ctlfac_stats.h"
#endif /* WITH_CTL */

/* NULL means default name */
guchar *configfile = NULL;

#ifdef WITH_CTL
time_t time_of_start;
time_t time_of_reset;
struct ctlfac_stats server_ctlfac_stats;

Dents_ctlfac_RecursionType server_recursion_type;
Dents_ctlfac_ResetState server_reset_state;

void start_ctlfac_threads (void);
void ctlfac_handler_thread (void);
#endif /* WITH_CTL */

#ifdef WITH_PTHREAD

/* List of receiver threads */
static blocklist receive_threads = BLOCK_INITIALIZER (receiverrec, 1);

/* Main thread which receives queries and then passes it on to
   dns_handle_packet for further processing. */
static void
udp_receiver_thread (receiverrec * receiver)
{
  int maxfd = 0, numfd;
  fd_set readfds, treadfds;
  socketrec *sock, *nsock;
  packetrec *pkt;

  /* Allocate a packet for those functions which need it */
  pkt = packet_alloc ();

  /* Zero out the fd's */
  FD_ZERO (&readfds);

  /* Make the fd_set */
  for (sock = link_first (&udpsockets); sock;
       sock = link_next (sock, sock->lt)) {
    FD_SET (sock->fd, &readfds);

    if (sock->fd > maxfd)
      maxfd = sock->fd;
  }

  /* Start up our infinite loop accepting packets */
  while (1) {

    /* Copy over the "master" list of readfds to a temporary readfd set */
    memcpy ((void *) &treadfds, (void *) &readfds, sizeof (fd_set));
    numfd = select (maxfd + 1, &treadfds, NULL, NULL, NULL);
    DPRINTF (("udp_receiver_thread: thread #%d out of select\n",
	      receiver->threadnum));
    if (numfd < 0)
      err_fatal ("udp_receiver_thread: select: %m");

    /* Find out which socket is waiting and read it. */
    for (nsock = link_first (&udpsockets); (sock = nsock);
	 nsock = link_next (nsock, nsock->lt))
      if (FD_ISSET (sock->fd, &treadfds))
	sock->readaction (sock, pkt);
  }
  /* This function does not return. */
}

/* Handles multiplexing of all of the TCP streams. */
static void
tcp_receiver_thread (receiverrec * receiver)
{
  int maxfd;
  fd_set readfds, writefds;
  socketrec *nsock, *sock;

  while (1) {
    /* Zero out the fd's */
    FD_ZERO (&readfds);
    FD_ZERO (&writefds);

    maxfd = 0;

    /* Make the fd_set */
    for (sock = link_first (&tcpsockets); sock;
	 sock = link_next (sock, sock->lt)) {
      /* We always try to read */
      FD_SET (sock->fd, &readfds);

      /* Do we need to write? */
      if ((sock->typepriv.tcp).buffersize)
	FD_SET (sock->fd, &writefds);

      if (sock->fd > maxfd)
	maxfd = sock->fd;
    }

    /* Select and wait for a packet to come in. */
    if (select (maxfd + 1, &readfds, &writefds, NULL, NULL) < 0)
      err_fatal ("tcp_receiver_thread: select: %m");

    DPRINTF (("tcp_receiver_thread: thread #%d out of select\n",
	      receiver->threadnum));

    /* Find out which socket is waiting */
    for (nsock = link_first (&tcpsockets); (sock = nsock);
	nsock = link_next (nsock, nsock->lt)) {
      /* Can we read? */
      if (FD_ISSET (sock->fd, &readfds)) {
	sock->readaction (sock, NULL);
	FD_CLR (sock->fd, &readfds);
      }
      /* Can we write? */
      if (FD_ISSET (sock->fd, &writefds)) {
	sock->writeaction (sock);
	FD_CLR (sock->fd, &writefds);
      }
    }
  }
}

/* Continually runs events in main thread. */
static void
mainloop (void)
{
#ifdef WITH_CTL
  server_reset_state = Dents_ctlfac_Running;
#endif

  /* Idle loop for now. We'll do some house keeping later. */
  /* FIXME - Grab stats from the worker threads. */
  /* FIXME - Every so often see if new interfaces have been added, etc. */
  g_main_run (g_main_new(FALSE));
}

/* Creates all necessary receiver threads, allocates memory, etc. */
static void
start_receiver_threads (void)
{
  guint16 i;
  receiverrec *r;

  /* Create the receiver threads */

  /* FIXME: Startup a configuration file number of */
  /* threads to start with */
  for (i = 0; i < NUM_PRESTART_RECV; i++) {
    /* Create the thread */
    r = block_allocandadd (&receive_threads);
    r->threadnum = i;

    pthread_create (&r->thread, NULL, (void *) udp_receiver_thread,
		    (void *) r);
  }

  r = block_allocandadd (&receive_threads);
  r->threadnum = i;

  pthread_create (&r->thread, NULL, (void *) tcp_receiver_thread, (void *) r);
}

#else /* not WITH_PTHREAD */

/* Main loop which polls all of the sockets to see if there's anything
   waiting to be read or if we can write to them (TCP). Passes it on
   to dns_handle_packet for further processing */
static void
mainloop (void)
{
#define ROUTINE "mainloop ()"

  int maxfd = 0, numfd;
  fd_set readfds, writefds;
  socketrec *nsock, *sock;
  struct timeval timeout;
  packetrec *pkt;

  /* Allocate a packet for those functions which need it. */
  pkt = packet_alloc ();

  /* Start up our infinite loop accepting packets */
  while (1) {
    /* Zero out the fd's */
    FD_ZERO (&readfds);
    FD_ZERO (&writefds);

    /* Temporary solution until we use GPollFD. */
    timeout.tv_sec = 0;
    timeout.tv_usec = 250000;

    maxfd = 0;

    /* Make the fd_set */
    for (sock = link_first (&udpsockets); sock;
	 sock = link_next (sock, sock->lt)) {
      /* We always try to read */
      FD_SET (sock->fd, &readfds);

      if (sock->fd > maxfd)
	maxfd = sock->fd;
    }

    for (sock = link_first (&tcpsockets); sock;
	 sock = link_next (sock, sock->lt)) {
      /* We always try to read */
      FD_SET (sock->fd, &readfds);

      /* Do we need to write? */
      if (sock->tp_tcp.buffersize)
	FD_SET (sock->fd, &writefds);

      if (sock->fd > maxfd)
	maxfd = sock->fd;
    }

    numfd = select (maxfd + 1, &readfds, &writefds, NULL, &timeout);
    if (numfd)
      DPRINTF (("%s: out of select\n", ROUTINE));

    if (numfd < 0 && errno != EINTR)
      err_fatal ("%s: select: %m", ROUTINE);

    /* Find out which socket is waiting */
    for (nsock = link_first (&udpsockets); (sock = nsock);
	 nsock = link_next (nsock, nsock->lt)) {
      /* Anything to read? */
      if (FD_ISSET (sock->fd, &readfds)) {
	sock->readaction (sock, pkt);
	FD_CLR (sock->fd, &readfds);
      }
    }

    /* Find out which socket is waiting */
    for (nsock = link_first (&tcpsockets); (sock = nsock);
	 nsock = link_next (nsock, nsock->lt)) {
      /* Anything to read? */
      if (FD_ISSET (sock->fd, &readfds)) {
	sock->readaction (sock, pkt);
	FD_CLR (sock->fd, &readfds);
      }
      /* Anything to write? */
      if (FD_ISSET (sock->fd, &writefds)) {
	sock->writeaction (sock);
	FD_CLR (sock->fd, &writefds);
      }
    }
    
    /* Passing FALSE tells GLib not to wait for an event. */
    while (g_main_iteration(FALSE))
      ; /* More events are waiting if true was returned. */
  }
  /* This function does not return. */
#undef ROUTINE
}

#endif /* not WITH_PTHREAD */

/*
 * dents_is_threaded
 *
 *  Returns if the server is compiled with a threading package. This is
 * useful for dynamic modules to call at load time to make sure someone
 * isn't doing something really stupid
 */
int
dents_is_threaded (void)
{
#ifdef WITH_PTHREAD
  return 1;
#else
  return 0;
#endif
}

/* Check the option argument to see if it specfies a valid syslog
   facility. */
static int
validate_syslog_fac (char *fac_string, int *ret_fac_ptr)
{
  if (!strcasecmp (fac_string, "USER")) {
    *ret_fac_ptr = LOG_USER;
  } else if (!strcasecmp (fac_string, "DAEMON")) {
    *ret_fac_ptr = LOG_DAEMON;
  } else if (!strcasecmp (fac_string, "LOCAL0")) {
    *ret_fac_ptr = LOG_LOCAL0;
  } else if (!strcasecmp (fac_string, "LOCAL1")) {
    *ret_fac_ptr = LOG_LOCAL1;
  } else if (!strcasecmp (fac_string, "LOCAL2")) {
    *ret_fac_ptr = LOG_LOCAL2;
  } else if (!strcasecmp (fac_string, "LOCAL3")) {
    *ret_fac_ptr = LOG_LOCAL3;
  } else if (!strcasecmp (fac_string, "LOCAL4")) {
    *ret_fac_ptr = LOG_LOCAL4;
  } else if (!strcasecmp (fac_string, "LOCAL5")) {
    *ret_fac_ptr = LOG_LOCAL5;
  } else if (!strcasecmp (fac_string, "LOCAL6")) {
    *ret_fac_ptr = LOG_LOCAL6;
  } else if (!strcasecmp (fac_string, "LOCAL7")) {
    *ret_fac_ptr = LOG_LOCAL7;
  } else {
    return -1;
  }

  /* return success */
  return 0;
}				/* end validate_syslog_fac() */

/* Prints out the command line usage. Usually called when an invalid
   argument was specified or they specifically asked for it.  Exit's
   after printing. */
static void
usage (void)
{
  fprintf (stderr,
	   "Usage: dents <-f> <-n> <-c location> <-C format> <-p port> <-d piddir>\n"
	   "             <-m moduledir> <-s syslogfac>\n"
	   "[" BUILDINFO "]\n\n"
	   "       -f tells the server to enable its control facility\n"
	   "       -n tells the server not to fork\n"
	   "       -c tells the server explicitly the location of the config file\n"
	   "       -C tells the server which config file format to use\n"
	   "              \"nb\" means named.boot format\n"
	   "              \"nc\" means named.conf format\n"
	   "              \"dc\" means dents.conf format\n"
	   "       -p <port> tells dents to listen on port\n"
	   "       -d <piddir> tells dents which directory to use to store its pidfile\n"
	   "       -m <moduledir> tells dents to add the specified dir to the module path\n"
	   "       -s <syslogfac> tells dents to log to the specified facility\n"
	   "          one of USER, DAEMON, LOCAL0, ... LOCAL7. The default is USER\n"
	   "\n");

  exit (1);
}

#ifdef DENTS_DEBUG
/* If the user presses ctrl-c in debug mode we flush buffers before
   exiting. */
void
sigint_handler (int signum)
{
#define ROUTINE "sigint_handler ()"
  fprintf (stderr, "%s: Exiting on signal %d flushing stdout, stderr\n",
	   ROUTINE, signum);
  fflush (stderr);
  fflush (stdout);
  exit (0);
#undef ROUTINE
}
#endif

/* Parses out command line. Then runs into the mainloop. */
int
main (int argc, char *argv[])
{
#ifdef DENTS_DEBUG
  guint8 nofork = 1;
  guint16 dnsport = DNS_PORT + 5000;
#else /* !DENTS_DEBUG */
  guint8 nofork = 0;
  guint16 dnsport = DNS_PORT;
#endif /* !DENTS_DEBUG */
#ifdef WITH_CTL
  guint8 do_ctlfac = 0;
#endif /* WITH_CTL */
  int c, ret;
  guint16 configformat = CONF_NAMEDBOOT;	/* Default format */
  char *endptr;
  FILE *pidfd;
  char *piddir = PIDDIR;
  char pidfile[PATH_MAX + 1];
  int logopts = LOG_PID;
  int syslog_fac = LOG_USER;

#ifdef WITH_CTL
  time_of_start = time_of_reset = time (NULL);
  server_recursion_type = Dents_ctlfac_RecursionAvailable;
  server_reset_state = Dents_ctlfac_Initializing;
#endif /* WITH_CTL */

  /* Cheezy. We shouldn't need this when we generate id's correctly */
  /* FIXME: Generate id's correctly so we can ditch this */
  srand (time (NULL));

  /* Lets strip off the path from argv[0] */
  if (strrchr (argv[0], '/'))
    argv[0] = strrchr (argv[0], '/') + 1;

  /* First we'll parse out the command line. */
  while (1) {
    if ((c = getopt (argc, argv, "nfC:c:d:p:m:s:")) == -1)
      break;

    switch (c) {
    case 'n':			/* Don't daemonize (ie don't fork) */
#ifdef LOG_PERROR
      logopts |= LOG_PERROR;
#else
      logopts |= LOG_CONS;
#endif
      nofork = 1;
      break;
    case 'f':			/* _Do_ use the control facility */
#ifdef WITH_CTL
      do_ctlfac = 1;
#else
      fprintf (stderr,
	       "This server was not compiled with support for the control facility.  '-f' flag ignored.\n");
#endif /* WITH_CTL */
      break;
    case 'd':			/* alternate piddir */
      piddir = (guchar *) strdup (optarg);
      break;
    case 'c':			/* Alternate configfile */
      if (configfile)
	free (configfile);

      if (!optarg)
	usage ();

      configfile = (guchar *) strdup (optarg);
      break;
    case 'C':			/* Config file format */
      if (!stricmp (optarg, "nb") || !stricmp (optarg, "named.boot") ||
	  !stricmp (optarg, "namedboot") || !stricmp (optarg, "bind4"))
	configformat = CONF_NAMEDBOOT;
      else if (!stricmp (optarg, "nc") || !stricmp (optarg, "named.conf") ||
	       !stricmp (optarg, "namedconf") || !stricmp (optarg, "bind8"))
	configformat = CONF_NAMEDCONF;
      else if (!stricmp (optarg, "dc") || !stricmp (optarg, "dents.conf") ||
	       !stricmp (optarg, "dentsconf") || !stricmp (optarg, "dents"))
	configformat = CONF_DENTSCONF;
      else {
	fprintf (stderr, "unknown config file format \"%s\"\n", optarg);
	usage ();
      }
      break;

      /* FIXME: These are kludges, we need to get the real config file stuff */
      /* done, so we can get rid of this */
    case 'p':
      dnsport = (guint16) strtol (optarg, &endptr, 0);
      if (endptr[0] != '\0')
	usage ();
      break;
    case 'm':
      /* configure a given module path */
      module_dyn_path_add (optarg);
      break;
    case 's':
      if (validate_syslog_fac (optarg, &syslog_fac))
	usage ();
      break;
    case '?':			/* Unknown, or usage */
      usage ();
      break;

    default:			/* Oops, something went wrong */
      fprintf (stderr, "getopt returned unknown character code 0x%02X\n", c);
      exit (1);
    }
  }

  /* This doesn't really do anything, except insure that the routines in
     dns_debug get linked into the server code.  */
  dns_debug_init ();

#ifdef DENTS_DEBUG
  /* handler for user pressing ctrl-C */
  signal (SIGINT, sigint_handler);
#endif

#if 0
  /* Still buggy; need machine with working gdb & pthreads. */
  install_chld_handler ();
#endif

  /* FIXME: This should definitly go somewhere else */
  /* Setup dir for module loading! */
  module_dyn_path_add (PKGLIBDIR);

  /* Open up syslog! */
  openlog (argv[0], logopts, syslog_fac);

#ifndef DENTS_DEBUG
  /* Banner */
  syslog (LOG_INFO, "dents v" DENTS_VERSION " starting");
#else
  syslog (LOG_INFO, "dents v" DENTS_VERSION " (debug) starting");
#endif

  /* Initialize the conf stuff and the lexer */
  lexer_init ();
  conf_init ();

  /* Scan the interfaces */
  interface_scan ();

  /* Glib has to be initialized to be threadsafe. */
  /* FIXME: Put the check into configure.in. */
#ifndef G_THREADS_IMPL_POSIX
#error Your version of GLib was not compiled for libpthread.
#endif /* not G_THREADS_IMPL_POSIX */
  g_thread_init (NULL);

  /* Setup the static modules */
  module_st_init ();

  /* Setup DNS. Reset all of the tree's, linked lists, etc */
  /* handler for child processes */
  install_chld_handler ();
  dns_setup (dnsport);

  /* If no config file, use the default. */
  ret = -1;			/* No warnings. */
  switch (configformat) {
  case CONF_NAMEDBOOT:
    if (!configfile)
      configfile = (guchar *) strdup (NAMED_BOOT_COMPAT_PATH);

    /* Read in the config file */
    ret = conf_nb_read (configfile);
    break;
  case CONF_NAMEDCONF:
  case CONF_DENTSCONF:
    if (!configfile)
      configfile = (guchar *) strdup (DENTS_CONF_COMPAT_PATH);

    /* Read in the config file */
    ret = conf_nc_read (configfile);
    break;
  default:
    err_fatal ("main: unknown configformat of %d\n", configformat);
  }

  if (ret < 0)
    exit (-1);

  /* We'll try to fork now */
  if (!nofork) {
    /* Standard fork and background code. */
    switch (fork ()) {
    case -1:			/* Oh shit, something went wrong. */
      perror ("fork");
      exit (-1);
    case 0:			/* Child: close off stdout, stdin and stderr. 
				 */
      close (0);
      close (1);
      close (2);
      break;
    default:			/* Parent: Just exit */
      exit (0);
    }
  }

  /* Write out the .pid file */
  g_snprintf (pidfile, sizeof (pidfile) - 1, "%s/dents.pid", piddir);
  if ((pidfd = fopen (pidfile, "w")) == NULL)
    err_fatal ("Could not open %s", pidfile);

  fprintf (pidfd, "%d", (int) getpid ());
  fclose (pidfd);

  /* Now we'll start up the zone loading threads and then the worker threads */
  /* at the same time so some queries can be answered when the information */
  /* is loaded */
  /* CHECKME - This may be a bad idea since it will thrash a bunch. Perhaps */
  /* we should load the zones, THEN bind? That way we start off immediately */
  /* with the full number of queries a second -jerdfelt */
  /* Initialize the control facility */

  /* Initialize the DNS related portions (read in the zone files, etc) */
  dns_init ();

#ifdef WITH_PTHREAD

#ifdef WITH_CTL
  if (do_ctlfac)
    start_ctlfac_threads ();
#endif /* WITH_CTL */

  /* Start up the receiver threads */
  start_receiver_threads ();
#endif /* WITH_PTHREAD */

  /* Run the main loop, this should never return */
  mainloop ();

  return 0;
}
