/* dns.c - Interface to the DNS module and internal DNS functions.

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

#include "common.h"
#include "dents.h"
#include "dns.h"
#include "module.h"		/* kludge down below */

dnsprotocolrec dns;
linkedlist dnssockets = LINK_INITIALIZER (socketrec, lp);

/* Send's a prebuilt packet. */
void
dns_sendpacket (socketrec * sock, packetrec * pkt, struct in_addr *ip,
		guint16 port)
{
  switch (sock->type) {
  case SOCKET_UDP:
    dns_sock_udp_send (sock, pkt->packetbuf, pkt->packetsize, ip, port);
    break;
  case SOCKET_TCP:
    dns_sock_tcp_send (sock, pkt->packetbuf, pkt->packetsize, ip, port);
    break;
  }
}

/*
 * dns_qualify_domain
 *
 *  Attempts to deduce if the domain name is fully qualified or not. If not,
 * the default domain name will be appended. In either case, the root (.) is
 * always explicitily stated
 */
void
dns_qualify_domain (guchar * dest, guchar * str, guchar * defaultdomain)
{
  /* Nothing? Put in the default domain */
  if (!strlen (str)) {
    strncpyzt (dest, defaultdomain, DNS_MAXNAMELENGTH);
    return;
  }

  /* Is there a trailing dot? If not, add the default domain */
  if (str[strlen(str) - 1] != '.') {
    g_snprintf(dest, DNS_MAXNAMELENGTH, "%s.%s", str, defaultdomain);
    return;
  }

  /* Copy the name */
  strncpyzt (dest, str, DNS_MAXNAMELENGTH);
}

/*
 * dns_setup
 *
 *  Set's up the initial DNS structures
 */
void
dns_setup (int port)
{
  /* Zero it all out */
  memset ((void *) &dns, 0, sizeof (dnsprotocolrec));

  /* Set up the port number */
  dns.port = port;

  /* Setup the zone and hint lists */
  block_setup(&dns.zone, sizeof(dnszonerec), 1);
  block_setup(&dns.hint, sizeof(dnshintrec), 1);

  /* Initialize the DB's */
  db_init(&dns.zonedb);		/* Zones (authoritative and non auth) */
  db_init(&dns.hintdb);		/* Hints used to start */

  dns_osq_setup ();
}

/*
 * dns_init
 *
 *  Set's up any socket's we need. Load's in any databases, and set's up
 * any zones
 */
void
dns_init (void)
{
  dnszonerec *zone;
  dbnoderec *node;
  dnsmodulerec *module;

  /* Initialize all of the type trees */
  dns_types_init ();

  /* bind() the sockets */
  dns_sock_udp_bind ();
  dns_sock_tcp_bind ();

  /* Load all of the zone's */
  dns_zone_loadall ();

  /*
   * We'll add it if and only if these conditions are met:
   *  1) The server is configured to be recursive
   *  2) We don't have a . zone
   */

  if (((node = dns_db_find (&dns.zonedb, ".")) == NULL) || (!node->data)) {
    /* FIXME: We only need this if we're recursive */
    module = dns_module_open ("mod_recursive");

    zone = dns_zone_create (".");
    zone->flags |= DNS_ZONE_CACHE;
    zone->flags &= ~DNS_ZONE_AUTHORITATIVE;

    dns_zone_addmodule (zone, module);
    dns_zone_setloadmod (zone, module);
  }
#ifdef DENTS_DEBUG
#if 0
  /* FIXME: HACK HACK HACK */
  /* Definately the wrong place for this, oh well */
  module_dyn_path_add ("/usr/local/src/dents/modules/mod_frl");
  module_dyn_path_add ("/usr/local/src/dents/modules/mod_recursive");
  module_dyn_path_add ("/usr/local/src/dents/modules/mod_stddb");
  module_dyn_path_add (".");
#endif

  module = dns_module_open ("mod_frl");
  zone = dns_zone_create ("frl.test.");

  dns_zone_setfilename (zone, "");

  zone->flags |= DNS_ZONE_MASTER;
  zone->flags |= DNS_ZONE_AUTHORITATIVE;

  dns_zone_addmodule (zone, module);
  dns_zone_setloadmod (zone, module);

  zone = dns_zone_create ("159.69.207.in-addr.arpa.");

  dns_zone_setfilename (zone, "");

  zone->flags |= DNS_ZONE_MASTER;
  zone->flags |= DNS_ZONE_AUTHORITATIVE;

  dns_zone_addmodule (zone, module);
  dns_zone_setloadmod (zone, module);
#endif

#ifdef DENTS_DEBUG
  printf("dns_init: zone database\n");
  db_display(&dns.zonedb);
  printf("dns_init: hint database\n");
  db_display(&dns.hintdb);
#endif /* DENTS_DEBUG */
}

