/* dns.h - structures that are used in the daemon.

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

/* DOES NOT contain configuration #define's, there should be no need
   to have to touch this file to compile the source.  This file
   contains only DNS specific stuff.  */

#ifndef DNS_H
#define DNS_H

#include <glib.h>		/* G_BYTE_ORDER */

/* We need some other stuff */
#include "packet.h"

/* Many of the definitions of the DNS types are kept in a separate file */
/*  for readability */
#include "dns_types.h"

/* DNS specific socket stuff */
#include "dns_sock.h"

/* These are some RFC defined limits */
#define DNS_MAXPACKETSIZE	64 * 1024
#define DNS_MAXNAMELENGTH	255
#define DNS_MAXLABELLENGTH	63
#define DNS_MAXRRDATALENGTH	64 * 1024

/* Maximum number of RR's per packet we send */
#define DNS_MAXRRPACKET		64

/* This is the IANA assigned port for DNS */
#define DNS_PORT		53

/* Sanity checking zones */
#define DNS_RDB_SANITYCHECK	1
#define DNS_RDB_NOSANITYCHECK	2

#define ERR_DNS_RDB_NOERROR	0
#define ERR_DNS_RDB_NOSUCHFILE	-1
#define ERR_DNS_RDB_PARSEERROR	-2

#define ERR_DNS_PARSE_LABEL	-1
#define ERR_DNS_PARSE_TOOSHORT	-2
#define ERR_DNS_PARSE_OVERFLOW	-3

/* See dns_comp.c/dns_compressrr for explanation */
#define DNS_MAXTXTSEGMENTLENGTH	255

#include "dns_zone.h"		/* DNS zone structures and #define's */
#include "dns_module.h"		/* DNS module stuff */

/* These are various DNS related structures */
typedef struct
{
#if (G_BYTE_ORDER == G_BIG_ENDIAN)
  guint8 qr:1;			/* response flag */
  guint8 opcode:4;		/* purpose of message */
  guint8 aa:1;			/* authoritive answer */
  guint8 tc:1;			/* truncated message */
  guint8 rd:1;			/* recursion desired */

  guint8 ra:1;			/* recursion available */
  guint8 pr:1;			/* primary server required (non standard) */
  guint8 unused:2;		/* unused bits (should always be zero) */
  guint8 rcode:4;		/* response code */
#elif (G_BYTE_ORDER == G_LITTLE_ENDIAN)
  guint8 rd:1;			/* recursion desired */
  guint8 tc:1;			/* truncated message */
  guint8 aa:1;			/* authoritive answer */
  guint8 opcode:4;		/* purpose of message */
  guint8 qr:1;			/* response flag */

  guint8 rcode:4;		/* response code */
  guint8 unused:2;		/* unused bits (should always be zero) */
  guint8 pr:1;			/* primary server required (non standard) */
  guint8 ra:1;			/* recursion available */
#elif (G_BYTE_ORDER == G_PDP_ENDIAN)
#  warning "PDP byteorder not supported"
#else
#  warning "Byteorder-code is buggy"
#endif
}
dnsflagsrec;

/* We parse out queries into this structure */
typedef struct
{
  guchar name[DNS_MAXNAMELENGTH + 1];	/* Domain Name for this query */

  guint16 type;			/* Type of query */
  guint16 class;		/* Class of query */
}
dnsqueryrec;

/* Internal structure used to store RR's */
typedef struct
{
  void *next;			/* Pointer to next RR */

  guchar name[DNS_MAXNAMELENGTH + 1];	/* Domain Name for this RR */

  guint16 type;			/* Type of RR */
  guint16 class;		/* Class of RR */
  gint32 ttl;			/* TTL of RR */

  guint16 length;		/* Length of the RDATA */
}
dnsrrrec;

/* Module chains */
typedef struct
{
  llentry ll;			/* Linked List (blocklist) */

  int flags;			/* Special characteristics about this module */

  dnsmodulerec *module;		/* Module */
}
dnschainrec;

/* Structures used for storing RR's in the db tree */
typedef struct
{
  void *next;			/* Next RR */
  void *parent;			/* Pointer to parent node */

  guint16 type;			/* Type of RR */
  guint16 class;		/* Class of RR */
  gint32 ttl;			/* TTL of RR */

  int length;			/* Length of the RDATA */
  guchar data[1];		/* The RDATA (allocated to size) */
}
dnsdbrrrec;

typedef struct
{
  /* Information off the socket */
  struct in_addr ip;
  guint16 port;
  socketrec *socket;

  /* Information from parsing */
  guint16 id;
  dnsqueryrec query;
  dnsflagsrec flags;
  guint16 qdcount;		/* number of queries in the query section */
  guint16 ancount;		/* number of answers in the answer section */
  guint16 nscount;		/* number of ns' in the ns section */
  guint16 arcount;		/* number of addit rr's in the ar section */

  union
  {
    struct
    {
      dnszonerec *zone;
    }
    query;
    struct
    {
      /* Nothing */
      int foo;			/* No bitching from the compiler */
    }
    reply;
  }
  qru;
}
dnspktrec;

#define DNS_BUILD_ANSWER 	0
#define DNS_BUILD_NS		1
#define DNS_BUILD_ADDITIONAL	2

/* Temporary structure used to store information when building packets. */
/*  We pass this off to all of the functions for the zone. They'll add */
/*  in RR's as needed */
typedef struct
{
  int sectype;

  guchar name[DNS_MAXNAMELENGTH + 1];

  guint16 class;
  guint16 type;
  guint32 ttl;

  void *data;
}
dnsbuildrrrec;

/*
 * dnsbuildrec - a structure to hold a list of RR's that are later 
 *              assembled into a DNS packet.
 *
 * Note, after declaring this structure, you should clear it with 
 * dns_build_reset(). When you are finished using this structure,
 * you should free the internal data it uses with 'dns_build_delete()'
 * 
 */
typedef struct dnsbuildrec_str dnsbuildrec;

/* For outstanding queries. We save the information for when the answer */
/*  comes back. This is used for outstanding sent (we generated it) and */
/*  received queries (from a remote client) */
typedef struct dnsosqueryrec
{
  llentry ll;			/* Linked List (blocklist) */
  /* Info on the query we sent */
  dbrbnode idnode;		/* Tree for the id's to find them quickly */
  guint16 toid;			/* id packet was sent with */
  struct in_addr toip;		/* IP packet was sent to */
  guint16 toport;		/* port packet was sent to */

  dnsqueryrec query;		/* Actual query made */

  guint8 resends;		/* Number of resends made so far */

  /* FIXME: typecast this correctly */
  void (*callback) ();		/* Callback function when we've received the reply */
  void *data;
} dnsosqueryrec;

/* The scratch structure used in for the protocols */
typedef struct
{
  guint16 port;			/* Port number we're using */

  socketrec sendsock;		/* Socket to send data out */

  blocklist zone;		/* Block list of zones */
  blocklist hint;		/* Block list of hints */
  blocklist osq;		/* Block list of outstanding queries */

  dbrootrec zonedb;		/* Tree for all of the zones we're responsible for */
  dbrootrec hintdb;		/* Tree for the hints used in recursion */
} dnsprotocolrec;

extern dnsprotocolrec dns;
extern linkedlist dnssockets;

#include "dns_comp.h"		/* DNS packet compression */

/*
 * Prototypes
 */

/* dns.c */
void dns_sendpacket (socketrec * sock, packetrec * pkt, struct in_addr *ip,
		     guint16 port);
void dns_qualify_domain (guchar * dest, guchar * str, guchar * defaultdomain);
void dns_setup (int port);
void dns_init (void);

/* dns_answer.c */
void dns_ans_zone_transfer (dnspktrec * dnsqpkt, dnsbuildrec * build,
			    packetrec * pkt);
int dns_ans_dochains (dnspktrec * dnsqpkt, dnsbuildrec * build,
		      packetrec * pkt);
int dns_ans_callback (dnspktrec * dnsqpkt, dnsbuildrec * build,
		      packetrec * pkt);

/* dns_build.c */
int dns_buildpacket (packetrec * pkt, dnsqueryrec * query,
		     dnsbuildrec * build, guint16 id, guint16 rcode,
		     dnsflagsrec flags, int reply, int aa);
void dns_build_add (dnsbuildrec * build, int sectype, guchar * name,
		    guint32 ttl, int class, int type, void *data);
void dns_build_addrr (dnsbuildrec * build, int sectype, dnsdbrrrec * rr);
dnsbuildrec *dns_build_alloc (void);
void dns_build_delete (dnsbuildrec * build);
guchar *dns_build_alloc_data (dnsbuildrec * build, int data_len);
void dns_build_reset (dnsbuildrec * build);
int dns_build_findans_and_pack (dbrootrec * db, dbnoderec * node,
				dnsqueryrec * query, dnsbuildrec * build);
void dns_build_findauth_and_pack (dbrootrec * db, dbnoderec * node,
				  dnsqueryrec * query, dnsbuildrec * build);
void dns_build_findaddit_and_pack (dbrootrec * db, dbnoderec * node,
				   dnsqueryrec * query, dnsbuildrec * build);

/* dns_db.c */
int dns_db_rr_compare (dnsdbrrrec * rr, guint16 class, guint16 type,
		       void *buf, int buflen);
dnsdbrrrec *dns_db_rr_findmatch (dbnoderec * node, dnsdbrrrec * rr,
				 guint16 class, guint16 type);
dbnoderec *dns_db_find (dbrootrec * db, guchar * domainname);
dbnoderec *dns_db_findbest (dbrootrec * db, guchar * domainname);
dbnoderec *dns_db_findbestns (dbrootrec * db, guint16 class,
			      guchar * domainname);
dbnoderec *dns_db_add (dbrootrec * db, guchar * domainname);
void dns_db_delete (dbrootrec * db, dbnoderec * node);
void dns_db_rr_add (dbrootrec * db, guchar * domainname, guint16 class,
		    guint16 type, guint32 ttl, guchar * rdata, int rlen);
void dns_db_getname (dbnoderec * node, guchar * name, int namelen);

/* dns_parse.c */
guchar *dns_unpackname (packetrec * p, guchar * cp, guchar * name,
			int maxlen);
int dns_parsequery (packetrec * p, guchar ** pcp, dnsqueryrec * query);
int dns_parserr (packetrec * p, guchar ** pcp, dnsrrrec * rr);

/* dns_read_file.c */
int dns_rdb_read (dbrootrec * db, int check, char *deforigin, char *filename);

/* dns_qr.c */
void dns_handle_packet (packetrec * pkt);

/* dns_qr_query.c */
void dns_query_handlepkt (dnspktrec * dnsqpkt, packetrec * pkt);

/* dns_qr_resp.c */
void dns_resp_handlepkt (dnspktrec * dnsrpkt, packetrec * pkt, guchar * cp);

/* dns_osq.c */
dnsosqueryrec *dns_osq_findid (guint16 id);
dnsosqueryrec *dns_osq_find (dnsqueryrec * query);
dnsosqueryrec *dns_osq_create (struct in_addr ip, guint16 port,
			       dnsqueryrec * query, void (*callback) (),
			       void *data);
void dns_osq_send (packetrec * p, dnsosqueryrec * osq);
void dns_osq_senderror (packetrec * p, dnsosqueryrec * osq, int rcode);
void dns_osq_delete (dnsosqueryrec * osq);
gboolean dns_osq_resend (gpointer);
void dns_osq_setup (void);

/* dns_types.c */
int dns_types_unpackrr (dnsdbrrrec * rr, guchar * buf, int buflen);
int dns_types_pack (dnsrrtyperec * rrtype, guchar * argv[], int numargs,
		    guchar * defaultdomain, guchar * rrbuf);
guint8 dns_types_findclass (guchar * label);
dnsrrtyperec *dns_types_findrrtype_by_label (guchar * label);
dnsrrtyperec *dns_types_findrrtype_by_type (guint8 type);
void dns_types_init (void);
int dns_types_pack_A (dnsbuildrec * buildrec, guchar ** rdata_ptr,
		      guchar * ipv4_addr);
int dns_types_pack_AAAA (dnsbuildrec * buildrec, guchar ** rdata_ptr,
			 guchar * ipv6_addr);
int dns_types_pack_1_domain_name (dnsbuildrec * buildrec, guchar ** rdata_ptr,
				  guchar * default_domain,
				  guchar * domain_name);
#define dns_types_pack_NS    dns_types_pack_1_domain_name
#define dns_types_pack_MD    dns_types_pack_1_domain_name
#define dns_types_pack_MF    dns_types_pack_1_domain_name
#define dns_types_pack_CNAME dns_types_pack_1_domain_name
#define dns_types_pack_MB    dns_types_pack_1_domain_name
#define dns_types_pack_MG    dns_types_pack_1_domain_name
#define dns_types_pack_MR    dns_types_pack_1_domain_name
#define dns_types_pack_PTR   dns_types_pack_1_domain_name
	/* The RR types above all use the same RDATA type 
	   (a single domain-name element) 
	 */

int dns_types_pack_2_domain_names (dnsbuildrec * buildrec,
				   guchar ** rdata_ptr,
				   guchar * default_domain,
				   guchar * domain_name1,
				   guchar * domain_name2);
#define dns_types_pack_RP    dns_types_pack_2_domain_names
#define dns_types_pack_MINFO dns_types_pack_2_domain_names
	/* The RR types above all use the same RDATA type 
	   (2 domain-name elements) 
	 */
int dns_types_pack_1_character_string (dnsbuildrec * buildrec,
				       guchar ** rdata_ptr,
				       guchar * character_string);
#define dns_types_pack_TXT   dns_types_pack_1_character_string
#define dns_types_pack_NSAP  dns_types_pack_1_character_string
	/* The RR's above all use the same RDATA type 
	   (a single character-string element) 
	 */
int dns_types_pack_2_character_strings (dnsbuildrec * buildrec,
					guchar ** rdata_ptr,
					guchar * character_string1,
					guchar * character_string2);
#define dns_types_pack_HINFO dns_types_pack_2_character_strings
#define dns_types_pack_ISDN  dns_types_pack_2_character_strings
int dns_types_pack_uint16_domain (dnsbuildrec * buildrec, guchar ** rdata_ptr,
				  guint16 int_arg,
				  guchar * default_domain,
				  guchar * domain_name);
#define dns_types_pack_MX    dns_types_pack_uint16_domain
#define dns_types_pack_AFSDB dns_types_pack_uint16_domain
int dns_types_pack_SOA (dnsbuildrec * buildrec, guchar ** rdata_ptr,
			guchar * default_domain, guchar * mname,
			guchar * rname, guint32 serial,
			gint32 refresh, gint32 retry, gint32 expire,
			guint32 minimum);

/* dns_debug.c */
char *dns_de_to_string (int error);
char *dns_dc_to_string (guint16 class);
char *dns_dt_to_string (guint16 type);
void dns_debug_print_pkt (guchar * pkt, int pktsize);
int dns_debug_init (void);

#endif
