/* dns_comp.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/*
 * dns_comp.c
 *
 *  Various functions to compress data into RFC 1035 style formats, etc
 */

/******************************************************************************
 *			Include Files
 *****************************************************************************/
#include "common.h"
#include "dns.h"

#include <ctype.h>		/* tolower */

/******************************************************************************
 *			Functions
 *****************************************************************************/

/*
 * dns_compressfind
 *
 * Given a list of pointers to labels, will try to find a pointer to compress
 * domain name with 
 *
 * Inputs:      ldname - the domain name we are looking for.
 *              buf - beginning of the packet that has been encoded 
 *                    thus far.
 *              cmp - a list of the offsets of the domain names 
 *                    that have already been encoded thus far.
 *
 */
static guchar *dns_compressfind
  (guchar * ldname, guchar * buf, dnscompressrec * cmp)
{
  int i, i1;
  guchar *p, *lp;

  /* Don't try compressing if we hit the root */
  if (!*ldname)
    return (NULL);

  /* Loop through all of the ptrs */
  for (i = 0; i < cmp->numcp; i++) {
    /* Setup some easy reference pointers */
    p = cmp->compptrs[i];
    lp = ldname;

    /* Then we'll go through and check if we match the rest of the name */
    while (*lp) {
      i1 = *(p++);

      /* Check if all of the characters match */
      for (; (i1) && (*lp != '.'); i1--, lp++, p++) {
	if ((*lp == '\\') && (*(lp + 1)))
	  lp++;

	/* Do a case insensitive match */
	if (tolower (*lp) != tolower (*p))
	  break;
      }

      if ((i1) || ((*lp != '.') && (!*lp)))
	break;

      /* Top 2 bits set means it's a pointer */
      if (*p >= 0xC0) {
	guint16 ti;

	memcpy ((void *) &ti, (void *) p, sizeof (guint16));
	p = buf + (ntohs (ti) & ~0xC000);
      }

      /* Are we at a period? The skip over it */
      if (*lp == '.')
	lp++;
    }

    /* If we hit the end, then the test passed */
    if (!*lp)
      return (cmp->compptrs[i]);
  }

  /* Otherwise, it failed */
  return (NULL);
}

/*
 * dns_compressname - compresses a <domain-name> and writes it into the packet
 *
 * Inputs:      cp - pointer to current position in packet to encode name
 *              name - the <domain-name> to encode
 *              buf - beginning of the packet buffer (used for pointers)
 *              cmp - compression record - a list of domain-names
 *                    already encoded.  This is used to create pointer
 *                    entries.
 * Outputs:     None.
 * Returns:     The new current position in the packet advanced past the
 *              data this function just encoded.
 *  
 */
guchar *
dns_compressname (guchar * cp, guchar * name, guchar * buf,
		  dnscompressrec * cmp)
{
  guchar *lp, *p;

  /* Now we'll go through trying to find a best match for compression */
  lp = name;

  /* If it's the root, short circuit here */
  if (*lp == '.') {
    *(cp++) = 0;
    return (cp);
  }

  /* Loop through the entire domain name */
  while (*lp) {
    /* Check to see if we can compress the rest of the domain with a */
    /*  pointer */
    p = dns_compressfind (lp, buf, cmp);

    if (!p) {
      int count = 0;

      /* Nope, let's copy over the rest of the label over */

      /* Save a pointer to this label for future reference */
      cmp->compptrs[cmp->numcp] = cp;

      /* Skip over the length byte, we'll fill it in later */
      cp++;

      /* Loop through to we hit a label seperator (.) or NUL */
      /* Exit pre-maturely if the label exceeds 64 bytes in length */
      count = 0;
      while ((*lp != '.') && (*lp != 0)) {

	count++;

	if ((*lp == '\\') && (*(lp + 1)))
	  lp++;

	*(cp++) = *(lp++);

	/* Don't allow the label length to exceed 64 bytes, or else
	   it will look like a pointer.
	 */
	if (count >= 63) {
	  err_warn ("Label longer than 63 bytes truncated in %s\n", name);
	  /* skip to the end of this label */
	  while ((*lp != '.') && (*lp != 0)) {
	    lp++;
	  }

	  break;
	}

      }				/* end while no label separator encountered */

      /* Save the length of the label */
      *(cmp->compptrs[cmp->numcp]) = (cp - cmp->compptrs[cmp->numcp]) - 1;

      cmp->numcp++;

      lp++;
    } else {
      /* Whee, we can use a pointer, let's figure out the math and add it */
      PUTINT16 (cp, (p - buf) | 0xC000);

      /* We don't null terminate after pointers. The pointer takes care of */
      /*  that. Plus we'll exit now */
      return (cp);
    }
  }

  /* Null terminate at the end of the label */
  *(cp++) = 0;

  return (cp);
}

/*
 * dns_compressquery
 *
 *  Add's a query style structure to the buffer first compressing the
 * domain name
 */
guchar *
dns_compressquery (packetrec * p, guchar * cp, guchar * dname,
		   dnsqueryrec * query, dnscompressrec * cmp)
{
  cp = dns_compressname (cp, dname, p->packetbuf, cmp);

  /* FIXME: Make sure we don't overflow anything */
  PUTINT16 (cp, query->type);
  PUTINT16 (cp, query->class);

  return (cp);
}

/*
 * dns_comp_character_string - place a <character-string> value in rdata
 *
 * Inputs:      output_rdata - pointer to position in packet to write data
 *              input_string - null terminated string to copy into packet
 * Outputs:     ret_len - # of bytes used in encoding.
 * Returns:     pointer to new position in packet to write more data
 *
 *
 * Found in contrib/host/host.c (from bind-4.9.5+P1):
 *
 * --Begin Quote--
 *
 * The treatment of TXT records has changed from 4.8.2 to 4.8.3. Formerly,
 * the data consisted simply of the text string. Now, the text string is
 * preceded by the character count with a maximum of 255, and multiple
 * strings are embedded if the total character count exceeds 255.
 * We handle only the new situation in this program, assuming that nobody
 * uses TXT records before 4.8.3 (unfortunately this is not always true:
 * current vendor supplied software may sometimes be even pre-BIND 4.8.2).
 *
 * --End Quote--
 * 
 * When RFC 1035 mentions a <character-string>, it is referring to
 * the definition in section 3.3:
 *
 * "<domain-name> is a domain name represented as a series of labels, 
 * and terminated by a label with zero length.  <character-string> is a 
 * single length octet followed by that number of characters.  
 * <character-string> is treated as binary information, and can be up 
 * to 256 characters in length (including the length octet)."
 *
 */
static guchar *dns_comp_character_string
  (guchar * output_rdata, guchar * input_string, int *ret_len)
{
  guint8 segmentlen;
  guint16 size;

  *ret_len = 0;

  /* The size of the string */
  size = strlen (input_string);

  while (size) {

    /* Is the TXT string too long to put in by itself? */
    if (size > DNS_MAXTXTSEGMENTLENGTH)
      segmentlen = DNS_MAXTXTSEGMENTLENGTH;
    else
      segmentlen = size;

    /* Write the length of this segment */
    *output_rdata = segmentlen;
    output_rdata++;

    /* Write the segment */
    memcpy (output_rdata, input_string, segmentlen);

    /* Update counts and pointers */
    size -= segmentlen;
    input_string += segmentlen;
    output_rdata += segmentlen;

    (*ret_len) += segmentlen + 1;

  }				/* end while () */

  /* return pointer past the end of the rdata that we have encoded */
  return output_rdata;

}				/* end dns_comp_character_string() */

/*
 * dns_comp_hex_string - encode a 'hex string' in a RR as binary data.
 *
 * Inputs:      output_rdata - pointer to position in packet to write data
 *              input_string - null terminated string to copy into packet
 *                    contains the ascii characters 0-9 and a-f
 * Outputs:     ret_len - # of bytes used in encoding.
 * Returns:     pointer to new position in packet to write more data
 *
 * rfc 1637 introduces something called a 'hex string'
 *
 * --Begin Quote--
 *
 * RDATA Format:
 *
 * RDATA: a variable length string of octets containing the NSAP.
 * The value is the binary encoding of the NSAP as it would appear in
 * the CLNP source or destination address field. A typical example of
 * such an NSAP (in hex) is shown below. For this NSAP, RDLENGTH is
 * 20 (decimal); "."s have been omitted to emphasize that they don't
 * appear in the DNS packets.
 *
 *      39840f80005a0000000001e13708002010726e00
 *
 * Master File Format:
 *
 * The NSAP RR introduces a new hex string format for the RDATA field.
 * The format is "0x" (i.e., a zero followed by an 'x' character)
 * followed by a variable length string of hex characters (0 to 9, a to
 * f). The hex string is case-insensitive. "."s (i.e., periods) may be
 * inserted in the hex string anywhere after the "0x" for readability.
 * The "."s have no significance other than for readability and are not
 *  propagated in the protocol (e.g., queries or zone transfers).
 *
 * 
 */
static guchar *dns_comp_hex_string
  (guchar * output_rdata, guchar * input_string, int *ret_len)
{
  int i,			/* loop control variable */
    hexstring_len;		/* length of the string to encode */
  guchar high_nybble = 0,	/* set to 1 if we are encoding the high nibble */
    curr_byte = 0;		/* accumulate current byte we are encoding */

  /* skip the initial 0x if it exists */
  if ((strcmp (input_string, "0x") == 0)
      || (strcmp (input_string, "0X") == 0))
    input_string += 2;

  hexstring_len = strlen (input_string);
  *ret_len = 0;

  curr_byte = 0;

  /* loop through the input string and encode each nybble */
  for (i = 0; i < hexstring_len; i++) {
    /* ignore periods and non-hex digits */
    if (isxdigit (input_string[i])) {
      guchar char_val = tolower (input_string[i]);
      guchar int_val = 0;

      if (char_val >= '0' && char_val <= '9') {
	int_val = char_val - '0';
      } else {
	int_val = 10 + char_val - 'a';
      }

      if (high_nybble) {
	*output_rdata = (curr_byte << 4) + int_val;
	high_nybble = 0;
	output_rdata++;
	(*ret_len)++;
	curr_byte = 0;
      } else {
	curr_byte = int_val;
	high_nybble = 1;
      }
    }
  }				/* end for each character in the hex string */

  /* output the last character if there was an odd # of characters */
  if (high_nybble) {
    *output_rdata = curr_byte << 4;
    output_rdata++;
    (*ret_len)++;
  }

  return output_rdata;

}				/* end dns_comp_hex_string_string() */

/*
 * dns_compressrr
 *
 * Add's a RR style structure to the buffer first compressing the domain
 * name and compressing any rdata as needed
 *
 * Inputs:      p - packet that is being output. 
 *              cp - pointer to current position in compressed packet data
 *                   (points to a position inside p->packetbuf)
 *              dname - domain name field in RR.
 *              rr - Build record created by the module (Includes uncompressed
 *                  information)
 *              cmp - structure containing info on labels that have
 *                  already been compressed for use in building pointers.
 * Outputs:     None.
 * 
 * Side Effects:The data in the 'rr' argument is appended to the 
 *              p->packetbuf buffer as a valid DNS RR ready to send
 *              out as a packet.  
 *
 * Returns:     updated pointer to current position (passed in as 'cp')
 * 
 */
guchar *
dns_compressrr (packetrec * p, guchar * cp,	/* current position */
		guchar * dname, dnsbuildrrrec * rr, dnscompressrec * cmp)
{
  guchar *sp;			/* 'start position' - used for remembering where to put RLENGTH */
  unsigned long foo;
  u_int16_t bar[2];


  /* Compress the name */
  cp = dns_compressname (cp, dname, p->packetbuf, cmp);

  /* FIXME: Make sure we don't overflow anything */
  PUTINT16 (cp, rr->type);
  PUTINT16 (cp, rr->class);
  PUTINT32 (cp, rr->ttl);

  /* Save the location of the rlength */
  sp = cp;
  cp += sizeof (guint16);

  /* Based on the type of record, convert the rdata passed in from the module
     into the outgoing packet format.  This applies any compression or
     addition of meta-data bits (e.g. a length field)
   */

  switch (rr->type) {

    /* RR's that consist of a single 32 bit address */
  case DT_A:
    memcpy (cp, rr->data, sizeof (struct in_addr));
    cp += sizeof (struct in_addr);
    break;

    /* RR's that contain a 16 bit integer followed by a domain name */
  case DT_MX:
  case DT_AFSDB:
  case DT_RT:
    PUTINTP16 (cp, rr->data);
       foo = (ulong) & bar[1];
       foo = foo - (ulong) & bar[0];
       foo = foo + (ulong) rr->data;

    cp =
      dns_compressname (cp, foo, p->packetbuf, cmp);
    break;

    /* RR's that contain a single domain name */
  case DT_CNAME:
  case DT_NS:
  case DT_PTR:
  case DT_MB:
    cp = dns_compressname (cp, rr->data, p->packetbuf, cmp);
    break;

    /* the SOA RR has a format all its own! */
  case DT_SOA:
    {
      guchar *tp;		/* Position in the source rdata buffer from 'rr->rdata' */

      tp = rr->data;

      /* Host */
      cp = dns_compressname (cp, tp, p->packetbuf, cmp);
      tp += strlen (tp) + 1;

      /* Email address */
      cp = dns_compressname (cp, tp, p->packetbuf, cmp);
      tp += strlen (tp) + 1;

      PUTINTP32 (cp, tp);	/* Serial number */
      tp += sizeof (guint32);
      PUTINTP32 (cp, tp);	/* Refresh time */
      tp += sizeof (guint32);
      PUTINTP32 (cp, tp);	/* Retry time */
      tp += sizeof (guint32);
      PUTINTP32 (cp, tp);	/* Expire time */
      tp += sizeof (guint32);
      PUTINTP32 (cp, tp);	/* Minimum time */
      tp += sizeof (guint32);
    }
    break;

    /* RR's that consist of a single character string */
  case DT_TXT:
  case DT_X25:
    {
      int rdata_len = 0;	/* We don't need this value, since cp is updated */
      cp = dns_comp_character_string (cp, rr->data, &rdata_len);
    }
    break;

    /* RR's that consist of two character strings */
  case DT_HINFO:
  case DT_ISDN:
    {
      int rdata_len = 0;

      /* the RDATA is 2 <character-string> values */

      /* the CPU data */
      cp = dns_comp_character_string (cp, rr->data, &rdata_len);
       foo = (ulong) rr->data;
       foo = foo + (ulong) rdata_len;

      /* the OS data */
      cp = dns_comp_character_string (cp, foo, &rdata_len);
    }
    break;

    /* RR's that consist of two domain names */
  case DT_MINFO:
  case DT_RP:
    {
      int first_domain_len = strlen (rr->data);
      /* length of the first string in the RR */
       foo = (ulong) rr->data + (ulong)first_domain_len + 1;;

      /* the RDATA is 2 <domain-name> values */
      cp = dns_compressname (cp, rr->data, p->packetbuf, cmp);
      cp = dns_compressname (cp, foo,
			     p->packetbuf, cmp);
    }
    break;

    /* The WKS RR has a format all its own */
  case DT_WKS:
    /* This one is tough to implement.  in particular, how does the 
       module pass the variable length bitmap ??? I think it should send
       an ASCII string of zeroes and ones.  Fortunately, this is not an
       oft used RR field.
     */
    DPRINTF (("dns_compressrr: DT_WKS not implemented\n"));
    break;

    /* RR's that use a single 'hex string' (rfc1637) */
  case DT_NSAP:
    {
      int rdata_len = 0;
      cp = dns_comp_hex_string (cp, rr->data, &rdata_len);
    }
    break;

  case DT_NSAP_PTR:
    DPRINTF (("dns_compressrr: DT_NSAP_PTR not implemented\n"));
    break;

  case DT_SIG:
    {
      /* excerpted from from rfc2064

         --Begin Quote--

         The RDATA portion of a SIG RR is as shown below.  The integrity of
         the RDATA information is protected by the signature field.

         1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 3 3
         0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
         +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
         |        type covered           |  algorithm    |     labels    |
         +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
         |                         original TTL                          |
         +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
         |                      signature expiration                     |
         +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
         |                         time signed                           |
         +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
         |         key footprint         |                               /
         +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+         signer's name         /
         /                                                               /
         +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
         |                                                               /
         +-                          signature                           /
         /                                                               /
         +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

         holy cow, there is a lot more to it - see the RFC for more info.

       */

      DPRINTF (("dns_compressrr: DT_SIG not implemented \n"));
    }
    break;

  case DT_AAAA:
    /* what RFC does this come from??? */
    /* Well the defs are in 1886 but i don't know about the compression */
    memcpy (cp, rr->data, sizeof (guint8) * 16);
    cp += sizeof (guint8) * 16;
    break;

  case DT_LOC:
    /*  from rfc1876
       MSB                                           LSB
       +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
       0|        VERSION        |         SIZE          |
       +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
       2|       HORIZ PRE       |       VERT PRE        |
       +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
       4|                   LATITUDE                    |
       +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
       6|                   LATITUDE                    |
       +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
       8|                   LONGITUDE                   |
       +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
       10|                   LONGITUDE                   |
       +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
       12|                   ALTITUDE                    |
       +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
       14|                   ALTITUDE                    |
       +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
       (octet)      
     */
    DPRINTF (("dns_compressrr: DT_LOC not implemented \n"));
    break;

  case DT_NXT:

    /* from rfc 2065:

       NXT RDATA Format

       The RDATA for an NXT RR consists simply of a domain name followed by
       a bit map.

       The type number for the NXT RR is 30.

       1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 3 3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |         next domain name                                      /
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                    type bit map                               /
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

       The NXT RR type bit map is one bit per RR type present for the owner
       name similar to the WKS socket bit map.  The first bit represents RR
       type zero (an illegal type which should not be present.) A one bit
       indicates that at least one RR of that type is present for the owner
       name.  A zero indicates that no such RR is present.  All bits not
       specified because they are beyond the end of the bit map are assumed
       to be zero.  Note that bit 30, for NXT, will always be on so the
       minimum bit map length is actually four octets.  The NXT bit map
       should be printed as a list of RR type mnemonics or decimal numbers
       similar to the WKS RR.

       The domain name may be compressed with standard DNS name compression
       when being transmitted over the network.  The size of the bit map can
       be inferred from the RDLENGTH and the length of the next domain name.
     */
    DPRINTF (("dns_compressrr: DT_NXT not implemented \n"));
    break;

  default:
    DPRINTF (("dns_compressrr: unknown type: %d\n", rr->type));
    break;
  }

  /* Now fill in the rlength */
  PUTINT16 (sp, (cp - sp) - sizeof (guint16));

  return (cp);

}				/* end dns_compressrr() */
