/* dns_db.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/*
 * dns_db.c
 *
 *  Controls all of the access to the dbtree from the DNS functions
 */
#include "common.h"
#include "db.h"
#include "dns.h"

/*
 * dns_db_rr_compare
 *
 *  Compares an RR in the database to one just parsed to see if they're
 * identical
 *
 * 1 == same, else 0
 */
int
dns_db_rr_compare (dnsdbrrrec * rr, guint16 class, guint16 type,
		   void *buf, int buflen)
{
  unsigned long foo;
  guint16 bar[2];

  switch (type) {
  case DT_A:
    if (buflen != sizeof (struct in_addr))
      return (1);

    return (!memcmp (rr->data, buf, sizeof (struct in_addr)));
  case DT_MX:
    if (rr->length != buflen)
      return (0);

    /* We compare the priority first */
    if (memcmp (rr->data, buf, sizeof (guint16)))
      return (0);

    foo=(ulong)&bar[1]-(ulong)&bar[0];
    foo=foo+(ulong)buf;

    /* Then check the domain name */
    return (!stricmp (rr->data + sizeof (guint16), foo));
  case DT_SOA:
    /* We can only have one SOA and since they found one, we say this */
    /*  is identical */
    return (1);
  case DT_CNAME:
  case DT_NS:
  case DT_PTR:
    if (rr->length != buflen)
      return (0);

    /* Check the domain name */
    return (!stricmp (rr->data, buf));
  case DT_TXT:
    if (rr->length != buflen)
      return (0);

    /* FIXME: Should different case be the same? */
    return (!stricmp (rr->data, buf));
  }

  return (0);
}

/*
 * dns_db_rr_findmatch
 *
 *  Given a node, will return back the first matching RR of that type. If
 * rr is NULL, will start at the first RR, or if !NULL, will start relative
 * to that RR
 */
dnsdbrrrec *
dns_db_rr_findmatch (dbnoderec * node, dnsdbrrrec * rr,
		     guint16 class, guint16 type)
{
  dnsdbrrrec *nrr;

  /* Where do we start? */
  if (rr)
    /* Relative to a previous RR? */
    nrr = rr->next;
  else
    /* Beginning of the RR's? */
    nrr = node->data;

  /* Now try to find anything that matches it */
  for (; nrr; nrr = nrr->next) {
    /* If the class matches exactly or the query is DC_ANY and */
    /*  the type matches xactly or the query is DT_ANY */
    if (((class == DC_ANY) || (nrr->class == class)) &&
	((type == DT_ANY) || (nrr->type == type)))
      /*
         ((type == DT_ANY) || (nrr->type == type) || (nrr->type == DT_CNAME)))
       */
      return (nrr);
  }

  return (NULL);
}

/*
 * dns_db_find
 *
 *  Finds a specific domainname in the db tree
 */
dbnoderec *
dns_db_find (dbrootrec * db, guchar * domainname)
{
  guchar *p, dname[DNS_MAXNAMELENGTH + 1];
  dbnoderec *node;

  if (!*domainname)
    return (NULL);

  /* Copy over the domain name temporarily */
  strncpyzt (dname, domainname, DNS_MAXNAMELENGTH);

  /* The root gets displayed differently, so let's correct for that */
  if (dname[0] == '.' && (!*(dname + 1)))
    dname[0] = 0;

  /* Start at the top of this db */
  node = &db->node;

  /* We'll step down each level in the tree finding if this label is already */
  /*  there. If not, we'll add it */
  do {
    if ((p = (guchar *) strrchr (dname, '.')) != NULL) {
      node = db_findnode (node, p + 1);
      *p = 0;
    } else
      node = db_findnode (node, dname);

    if (!node)
      return (NULL);
  } while (p);

  return (node);
}

/*
 * dns_db_findbest
 *
 *  Tries to find the best information we can on a domain
 */
dbnoderec *
dns_db_findbest (dbrootrec * db, guchar * domainname)
{
  guchar *p, dname[DNS_MAXNAMELENGTH + 1];
  dbnoderec *node, *bestnode = NULL;

  if (!*domainname)
    return (NULL);

  /* Copy over the domain name temporarily */
  strncpyzt (dname, domainname, DNS_MAXNAMELENGTH);

  /* The root gets displayed differently, so let's correct for that */
  if (dname[0] == '.' && (!*(dname + 1)))
    dname[0] = 0;

  /* Start at the top of this db */
  node = &db->node;

  /* We'll step down each level in the tree finding if this label is already */
  /*  there. If not, we'll add it */
  do {
    /* Find the next seperator */
    if ((p = (guchar *) strrchr (dname, '.')) != NULL) {
      node = db_findnode (node, p + 1);
      *p = 0;
    } else
      node = db_findnode (node, dname);

    if (!node)
      return (bestnode);

    if (node->data)
      bestnode = node;
  } while (p);

  return (bestnode);
}

/*
 * dns_db_findbestns
 *
 *  Tries to find the best information we can on a domain
 */
dbnoderec *
dns_db_findbestns (dbrootrec * db, guint16 class, guchar * domainname)
{
  dbnoderec *node;

  /* Find the best node */
  node = dns_db_findbest (db, domainname);

  /* Keep going down checking if we've found an NS */
  for (; node && !(dns_db_rr_findmatch (node, NULL, class, DT_NS));
       node = node->parent);

  return (node);
}

/*
 * dns_db_add
 *
 *  Create's a new node in a dbtree, creating all intermediary nodes on the
 * way. If it already exists, it returns a pointer to existing node, else
 * returns a pointer to the new node
 */
dbnoderec *
dns_db_add (dbrootrec * db, guchar * domainname)
{
  guchar dname[DNS_MAXNAMELENGTH + 1], *p, *p1;
  dbnoderec *node, *n;

  if (!*domainname)
    return (NULL);

  /* Does it already exist? */
  if ((node = dns_db_find (db, domainname)) != NULL)
    return (node);

  /* Now that we know it's in this name space, we'll chop off the end */
  strncpyzt (dname, domainname, DNS_MAXNAMELENGTH);

  /* The root gets displayed differently, so let's correct for that */
  if (dname[0] == '.' && (!*(dname + 1)))
    dname[0] = 0;

  /* Start at the top of this db */
  node = &db->node;

  /* We'll step down each level in the tree finding if this label is already */
  /*  there. If not, we'll add it */
  do {
    if ((p = (guchar *) strrchr (dname, '.')) != NULL) {
      p1 = p + 1;
      *p = 0;
    } else
      p1 = dname;

    if ((n = db_findnode (node, p1)) == NULL)
      n = db_addnode (node, p1);

    node = n;
  } while (p);

  return (node);
}

/*
 * dns_db_delete
 *
 *  Delete's a node from a DBtree
 */
void
dns_db_delete (dbrootrec * db, dbnoderec * node)
{
  dbnoderec *parent;

  do {
    parent = node->parent;
    db_deletenode (node);
    node = parent;
  } while (parent && (parent->children.numnodes == 0));
}

/*
 * dns_db_rr_add
 *
 *  Add's an RR to an DBtree
 */
void
dns_db_rr_add (dbrootrec * db, guchar * domainname, guint16 class,
	       guint16 type, guint32 ttl, guchar * rdata, int rlen)
{
  dnsdbrrrec *rr;
  dbnoderec *node;

  /* Allocate some memory for this entry */
  if ((rr = malloc (sizeof (dnsdbrrrec) + rlen - sizeof (char))) == NULL)
    err_fatal ("dns_db_rr_add: couldn't allocate rr");

  /* Save all of the info related to this RR */
  rr->class = class;
  rr->type = type;
  rr->ttl = ttl;
  memcpy (rr->data, rdata, rlen);
  rr->length = rlen;

  if ((node = dns_db_add (db, domainname)) == NULL)
    return;

  /* Setup the parent pointer */
  rr->parent = node;

  /* Add it into the linked list */
  rr->next = node->data;
  node->data = rr;
}

/*
 * dns_db_getname
 *
 *  Builds the Fully Qualified name for a given node
 */
void
dns_db_getname (dbnoderec * node, guchar * name, int namelen)
{
  name[0] = 0;

  do {
    strncatzt (name, node->label, namelen);
    strncatzt (name, ".", namelen);

    node = node->parent;
  } while (node && node->label[0]);
}
