/* dns_module.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/*
 * dns_module.c
 *
 *  DNS specific module code
 */
#include "common.h"
#include "dns.h"
#include "module.h"

linkedlist dnsmodules = LINK_INITIALIZER (dnsmodulerec, ll);
blocklist dnsmodule_pool = BLOCK_INITIALIZER (dnsmodulerec, 1);

/*
 * Null functions for thos modules which don't implement them
 */

/*
 * dns_module_null_zone_load
 *
 *  A null zone load function
 */
static int
dns_module_null_zone_load (dnszonerec * zone)
{
  return (0);
}

/*
 * dns_module_null_zone_reload
 *
 *  A null zone reload function
 */
static int
dns_module_null_zone_reload (dnszonerec * zone)
{
  return (DE_NOTIMP);
}

/*
 * dns_module_null_zone_dump
 *
 *  A null zone dump function
 */
static int
dns_module_null_zone_dump (void)
{
  return (DE_NOTIMP);
}

/*
 * dns_module_null_rr_add
 *
 *  A null RR add function
 */
static int
dns_module_null_rr_add (void)
{
  return (DE_NOTIMP);
}

/*
 * dns_module_null_rr_delete
 *
 *  A null RR delete function
 */
static int
dns_module_null_rr_delete (void)
{
  return (DE_NOTIMP);
}

/*
 * Normal functions
 */

/*
 * dns_module_populate
 *
 *  Populates a module structure with syms
 */
static int
dns_module_populate (dnsmodulerec * module)
{
  /* Mandatory */
  if ((module->zone_add = module_sym (module->mod,
				      "module_dns_zone_add")) == NULL) {
    err_warn
      ("dns_module_populate: unable to find symbol module_dns_zone_add");
    return (-1);
  }

  if ((module->zone_delete = module_sym (module->mod,
					 "module_dns_zone_delete")) == NULL) {
    err_warn
      ("dns_module_populate: unable to find symbol module_dns_zone_delete");
    return (-1);
  }

  if ((module->answer = module_sym (module->mod,
				    "module_dns_answer")) == NULL) {
    err_warn ("dns_module_populate: unable to find symbol module_dns_answer");
    return (-1);
  }

  /* Optional */
  if ((module->zone_load = module_sym (module->mod,
				       "module_dns_zone_load")) == NULL)
    module->zone_load = dns_module_null_zone_load;

  if ((module->zone_reload = module_sym (module->mod,
					 "module_dns_zone_reload")) == NULL)
    module->zone_reload = dns_module_null_zone_reload;

  if ((module->zone_dump = module_sym (module->mod,
				       "module_dns_zone_dump")) == NULL)
    module->zone_dump = dns_module_null_zone_dump;

  if ((module->rr_add = module_sym (module->mod,
				    "module_dns_rr_add")) == NULL)
      module->rr_add = dns_module_null_rr_add;

  if ((module->rr_delete = module_sym (module->mod,
				       "module_dns_rr_delete")) == NULL)
    module->rr_delete = dns_module_null_rr_delete;

  return (0);
}

/*
 * dns_module_alloc
 *
 *  Allocate's a dns module structure
 */
dnsmodulerec *
dns_module_alloc (void)
{
  return (block_allocandaddtolink (&dnsmodule_pool, &dnsmodules));
}

/*
 * dns_module_free
 *
 *  Free's a previously akkicated dns module structure
 */
void
dns_module_free (dnsmodulerec * module)
{
  block_delete (&dnsmodule_pool, &module->ll);
}

/*
 * dns_module_open
 *
 *  Open's a module for use in DNS
 */
dnsmodulerec *
dns_module_open (guchar * modname)
{
  modulerec *module;
  dnsmodulerec *dnsmodule;

  /* Allocate a structure */
  dnsmodule = dns_module_alloc ();

  /* Open up the module */
  if ((module = module_open (modname)) == NULL) {
    dns_module_free (dnsmodule);
    return (NULL);
  }

  /* Save the pointer to the generic module */
  dnsmodule->mod = module;

  /* Populate the module */
  dns_module_populate (dnsmodule);

  return (dnsmodule);
}

/*
 * dns_module_close
 *
 *  Closes'a module used for DNS
 */
void
dns_module_close (dnsmodulerec * module)
{
  module_close (module->mod);

  dns_module_free (module);
}
