/* dns_osq.c - Handles dealing with outstanding queries

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>		/* inet_ntoa */

/* Add dents headers here */
#include "../include/common.h"
#include "dns.h"

/* Database sorted by id */
dbrb osq_ids;

/* Database sorted by name */
dbrootrec osqdb;
blocklist osq_list = BLOCK_INITIALIZER (dnsosqueryrec, 5);

/* Attempts to find an OSQ by id. */
dnsosqueryrec *
dns_osq_findid (guint16 id)
{
  return (db_rb_find (&osq_ids, &id));
}

/* Attempts to find an OSQ by name/type/class pair. */
dnsosqueryrec *
dns_osq_find (dnsqueryrec * query)
{
  dbnoderec *node;
  dnsosqueryrec *osq;

  if ((node = dns_db_find (&osqdb, query->name)) != NULL) {
    osq = node->data;

    if ((osq->query.type == query->type)
	&& (osq->query.class == query->class))
      return (osq);
  }
  return (NULL);
}

/* Allocates an osq and then fills in all of the pertinent information. */
dnsosqueryrec *
dns_osq_create (struct in_addr ip, guint16 port,
		dnsqueryrec * query, void (*callback) (), void *data)
{
  dnsosqueryrec *osq;
  dbnoderec *node;

  /* Allocate a some memory for this new query */
  osq = block_allocandadd (&osq_list);

  /* Reset some variables */
  osq->resends = 0;

  /* Fill in the information we need to keep track of to match reponses */
  memcpy ((void *) &osq->toip, (void *) &ip, sizeof (osq->toip));
  osq->toport = port;

  /* Copy over the query */
  memcpy (&osq->query, query, sizeof (dnsqueryrec));

  /* Function to call when we receive a reply */
  osq->callback = callback;
  osq->data = data;

  /* FIXME: This is REALLY lame. It's not particularly secure */
  do {
    osq->toid = rand ();
    DPRINTF (("dns_osq_create: trying id %d\n", osq->toid));
    /* Add it into the id db */
  } while (db_rb_insert (&osq_ids, &osq->idnode, osq) < 0);

  /* Schedule the resend of this every 5 seconds */
  /* FIXME: Get a "resend_delay"-variable from our configsystem. */
  g_timeout_add (5000, dns_osq_resend, osq);

  /* Add into the dbtree for osq */
  node = dns_db_add (&osqdb, osq->query.name);
  node->data = NULL;
  node->data = osq;

  return (osq);
}

/* Send a previously created osq. */
void
dns_osq_send (packetrec * pkt, dnsosqueryrec * osq)
{
  dnsflagsrec flags;

  if (!osq)
    return;

  DPRINTF (("dns_osq_send: sending query to %s:%d for %s\n",
	    inet_ntoa (osq->toip), osq->toport, osq->query.name));

  flags.rd = flags.tc = flags.unused = 0;
  /* FIXME: Kludged until we can make recursion user configurable. */
  flags.ra = 0;
  flags.opcode = DO_STANDARD;

  dns_buildpacket (pkt, &osq->query, NULL, osq->toid, DE_NOERROR,
		   flags, 0, 0);
  dns_sendpacket (&dns.sendsock, pkt, &osq->toip, osq->toport);
}

/* Send a previously created osq. */
void
dns_osq_senderror (packetrec * pkt, dnsosqueryrec * osq, int rcode)
{
#if 0
  /* FIXME: We should send the error back to the other dups as well */
  /* It should eventually end deleting the node from the tree */
  DPRINTF (("dns_osq_send: sending error to %s:%d for %s\n",
	    inet_ntoa (osq->fromip), osq->fromport, osq->query.name));

  dns_buildpacket (pkt, &osq->query, NULL, osq->fromid, rcode, osq->flags, 1);
  dns_sendpacket (&dns.sendsock, pkt, osq->fromip, osq->fromport);
#endif
}

/* Delete's an OSQ. */
void
dns_osq_delete (dnsosqueryrec * osq)
{
  dbnoderec *node;
  if (g_source_remove_by_user_data (osq)) {
    /* We have an event so we're in the idnode tree. Delete it. */
    db_rb_delete (&osq_ids, &osq->idnode);
  }
  /* Find the name in the osq database. */
  if ((node = dns_db_find (&osqdb, osq->query.name)) != NULL)
    dns_db_delete (&osqdb, node);
  block_delete (&osq_list, &osq->ll);
}

/* Resend an already built osq.  When we've resent it too many times,
   send an error back and delete it. */
gboolean
dns_osq_resend (gpointer ptr_osq)
{
  dnsosqueryrec *osq = ptr_osq;
  struct packetrec *pkt = packet_alloc();
  /* FIXME: Get a "resend_retries"-variable from our configsystem. */
  if (osq->resends > 5) {
    /* FIXME: Should we send back SERVFAIL on a recursive query? */
    DPRINTF (("dns_osq_resend: no response, giving up\n"));
    dns_osq_senderror (pkt, osq, DE_SERVFAIL);
    dns_osq_delete (osq);
    packet_free(pkt);
    return FALSE;		/* Don't try again. */
  }
  /* Resend the packet and increment the number of times we've sent it. */
  dns_osq_send (pkt, osq);
  osq->resends++;
  packet_free(pkt);
  return TRUE;			/* Call me again. */
}

/* Initializes the OSQ code. */
void
dns_osq_setup (void)
{
  /* Initialize the name sorted OSQ list */
  db_init (&osqdb);

  /* Initialize the id sorted OSQ list */
  db_rb_create (&osq_ids, db_rb_cmp_uint16);
}
