/* dns_parse.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/*
 * dns_parse.c
 *
 *  Parses out the different parts of a DNS packet into easy to use structures
 */
#include "common.h"
#include "dns.h"

#define MAX_UNPACK_NESTING 5

/* FIXME: Handle \.'s correctly in domain names */

/*
 * _dns_unpackname
 *
 *  Unpacks a name from a raw DNS packet into a string. This function is only
 * called by dns_unpackname itself
 */
static guchar *
_dns_unpackname (packetrec * p, guchar * cp, guchar * name,
		 int maxlen, int level)
{
  int len;
  int offset;
  guchar *lp = name;
  guint16 ti;

  /* This is to prevent packing bombs which would otherwise bring down the */
  /*  program */
  if (level > MAX_UNPACK_NESTING) {
    DPRINTF (("_dns_unpackname: maximum nesting reached\n"));
    return (NULL);
  }

  /* Make sure the pointer that was passed isn't out of range */
  if (cp > (p->packetbuf + p->packetsize)) {
    DPRINTF (("_dns_unpacklebel: cp is past buffer end\n"));
    return (NULL);
  }

  /* Check if it's just the root, and if so, short circuit here */
  if (!(len = *cp)) {
    *(lp++) = '.';
    *(lp++) = 0;
    return (cp + 1);
  }

  /* Loop through now */
  do {
    /* Check the bit mask to see what kinda compression is being done */
    switch (len & 0xC0) {
    case 0xC0:
      /* Go recursively */
      memcpy ((void *) &ti, (void *) cp, sizeof (guint16));
      cp += sizeof (guint16);

      offset = (ntohs (ti) & ~0xC000);

      /* Call the routine */
      _dns_unpackname (p, p->packetbuf + offset, lp, maxlen, level + 1);

      return (cp);
    case 0x80:
    case 0x40:
      DPRINTF (("_dns_unpackname: unsupported bit pattern 0x%02X\n",
		len & 0xC0));
      return (NULL);
    }

    /* Skip past the length byte */
    cp++;

    /* Check if there's enough room in the buffer */
    if ((len + 1) > maxlen) {
      DPRINTF (
	       ("_dns_unpackname: not enough space in name to store label (%d, %d)\n",
		len + 1, maxlen));
      return (NULL);
    }

    /* Make sure we're not copying past the buffer as well */
    if ((cp + (len + 1)) > (p->packetbuf + p->packetsize)) {
      DPRINTF (("_dns_unpackname: cp + (len + 1) is past buffer end\n"));
      return (NULL);
    }

    /* Copy over everything */
    memcpy (lp, cp, len);

    /* Update the pointers */
    lp += len;
    cp += len;

    /* Add the label seperator in */
    *(lp++) = '.';

    /* Update the maximum length */
    maxlen -= (len + 1);

    len = *cp;
  } while (len);

  /* Null terminate it */
  *lp = 0;

  return (cp + 1);
}

/*
 * dns_unpackname
 *
 *  Unpacks a name from a DNS packet
 */
guchar *
dns_unpackname (packetrec * p, guchar * cp, guchar * name, int maxlen)
{
  return (_dns_unpackname (p, cp, name, maxlen, 0));
}

/*
 * dns_parsequery
 *
 *  Parses out information pertaining to the question part of the dns packet
 */
int
dns_parsequery (packetrec * p, guchar ** pcp, dnsqueryrec * query)
{
  guchar *cp;

  /* Grab the pointer */
  cp = *pcp;

  /* Unpack the name */
  if ((cp = dns_unpackname (p, cp, query->name, DNS_MAXNAMELENGTH)) == NULL)
    return (ERR_DNS_PARSE_OVERFLOW);

  /* Make sure there's enough data left for the type and class */
  if ((cp + sizeof (guint16) + sizeof (guint16)) >
      (p->packetbuf + p->packetsize)) return (ERR_DNS_PARSE_TOOSHORT);

  /* Grab the type and class */
  GRABINT16 (cp, query->type);
  GRABINT16 (cp, query->class);

  /* Stuff it back in */
  *pcp = cp;

  return (0);
}

/*
 * dns_parserr
 *
 *  Parses out information pertaining to the resource records used in different
 * parts of the dns packet
 */
int
dns_parserr (packetrec * p, guchar ** pcp, dnsrrrec * rr)
{
  int rlength;
  guchar *cp, *dp, *dpend;
  guint16 tempu16;

  /* Grab the pointer */
  cp = *pcp;

  /* Grab the name of the RR */
  if ((cp = dns_unpackname (p, cp, rr->name, DNS_MAXNAMELENGTH)) == NULL)
    return (ERR_DNS_PARSE_LABEL);

  /* Make sure there's enough data left for the type, class, ttl and length */
  if ((cp + sizeof (guint16) + sizeof (guint16) + sizeof (guint32)) >
      (p->packetbuf + p->packetsize))
    return (ERR_DNS_PARSE_TOOSHORT);

  /* Grab the type, class and ttl */
  GRABINT16 (cp, rr->type);
  GRABINT16 (cp, rr->class);
  GRABINT32 (cp, rr->ttl);

  /* Grab the RDATA length */
  GRABINT16 (cp, rlength);

  /* Make sure there's enough rdata */
  if ((cp + rlength) > (p->packetbuf + p->packetsize))
    return (ERR_DNS_PARSE_TOOSHORT);

  dp = p->tempbuf;
  dpend = p->tempbuf + p->tbufsize;

  /* Now let's parse out the RR specific stuff */
  switch (rr->type) {
  case DT_A:
    if ((dp + sizeof (guint16)) > dpend)
      return (ERR_DNS_PARSE_OVERFLOW);

    memcpy (dp, cp, sizeof (struct in_addr));
    dp += sizeof (struct in_addr);
    cp += sizeof (struct in_addr);
    break;
  case DT_MX:
    if ((dp + sizeof (guint16)) > dpend)
      return (ERR_DNS_PARSE_OVERFLOW);

    GRABINT16 (cp, tempu16);
    memcpy (dp, (void *) &tempu16, sizeof (guint16));
    dp += sizeof (guint16);
    /* Fallthrough to grab the name */
  case DT_CNAME:
  case DT_NS:
  case DT_PTR:
    if ((cp = dns_unpackname (p, cp, dp, dpend - dp)) == NULL)
      return (ERR_DNS_PARSE_OVERFLOW);

    dp += strlen (dp) + 1;	/* Don't forget termination nul */
    break;
  case DT_TXT:{
      /* See dns_comp.c/dns_compressrr for explanation */
      guint8 segmentlen;
      guint16 size;

      size = rlength;

      while (size) {
	/* Grab the segment length */
	memcpy ((void *) &segmentlen, cp, sizeof (segmentlen));
	cp += sizeof (segmentlen);

	if ((dp + segmentlen) > dpend)
	  return (ERR_DNS_PARSE_OVERFLOW);

	/* Grab the data */
	memcpy (dp, cp, segmentlen);
	cp += segmentlen;

	/* Update counters and pointers */
	size -= segmentlen;
	dp += segmentlen;
      }

      /* Null terminate it */
      *(dp++) = 0;
      break;
    }
  default:
    DPRINTF (("dns_parserr: unknown type: %d\n", rr->type));
    break;
  }

  /* Calculate the length */
  rr->length = dp - p->tempbuf;

  /* Stuff it back in */
  *pcp = cp;

  return (0);
}
