/* dns_qr.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/*
 * dns_qr.c
 *
 *  Handles queries and responses for DNS.
 */
#include "common.h"
#include "dns.h"

#include <arpa/inet.h>		/* inet_ntoa */

/*
 * dns_handle_packet
 *
 *  Handle's an already read packet
 */
void
dns_handle_packet (packetrec * pkt)
{
  dnspktrec dnspkt;
  guchar *cp;
#ifdef DENTS_DEBUG
  int i;

  DPRINTF (("dns_handle_packet: received from %s:%d %hu bytes\n  ",
	    inet_ntoa (pkt->ip), pkt->port, pkt->packetsize));

  /* Print out the raw packet */
  for (i = 0; i < pkt->packetsize; i++) {
    printf ("%02X ", pkt->packetbuf[i]);

    if ((i & 15) == 15)
      printf ("\n  ");
  }
  printf ("\n\n");

  /* NOTE: this may not always be safe!!! If we ever receive non-DNS messages
     through this routine, debug_print_pkt will probably make the server
     toss its cookies.
   */
  dns_debug_print_pkt (pkt->packetbuf, pkt->packetsize);
#endif

  /* Grab the IP addresses from the packet structure for internal */
  /*  abstraction purposes */
  memcpy ((void *) &dnspkt.ip, (void *) &pkt->ip, sizeof (struct in_addr));
  dnspkt.port = pkt->port;
  dnspkt.socket = pkt->sock;

  /* Is the packet we read big enough for atleast the headers? (id + flags + */
  /*  {qd, an, ns, ar}count) */
  if (pkt->pbufsize < (sizeof (guint16) + sizeof (dnsflagsrec) +
		       (sizeof (guint16) * 4))) {
    /* We send back FORMERR here */
    err_warn ("dns_handle_packet: malformed packet, too short for header");
    dns_buildpacket (pkt, NULL, NULL, dnspkt.id, DE_FORMERR, dnspkt.flags, 1,
		     0);
    dns_sendpacket (pkt->sock, pkt, &pkt->ip, pkt->port);
    return;
  }

  /* Grab the header */
  cp = pkt->packetbuf;
  GRABINT16 (cp, dnspkt.id);
  memcpy (&dnspkt.flags, cp, sizeof (dnsflagsrec));
  cp += sizeof (dnsflagsrec);
  GRABINT16 (cp, dnspkt.qdcount);
  GRABINT16 (cp, dnspkt.ancount);
  GRABINT16 (cp, dnspkt.nscount);
  GRABINT16 (cp, dnspkt.arcount);

  /* We can only handle one query at a time since error codes are returned */
  /*  on a per packet basis, not per name queried */

  /* Check if the packet has no queries whatsoever */
  if (!dnspkt.qdcount) {
    err_warn ("dns_handle_packet: malformed packet. query count == 0!");
    dns_buildpacket (pkt, NULL, NULL, dnspkt.id, DE_FORMERR, dnspkt.flags, 1,
		     0);
    dns_sendpacket (pkt->sock, pkt, &pkt->ip, pkt->port);
    return;
  }

  /* Parse out the first query */
  if (dns_parsequery (pkt, &cp, &dnspkt.query)) {
    err_warn ("dns_handle_packet: formerr, dns_parsequery returned !0");
    dns_buildpacket (pkt, NULL, NULL, dnspkt.id, DE_FORMERR, dnspkt.flags, 1,
		     0);
    dns_sendpacket (pkt->sock, pkt, &pkt->ip, pkt->port);
    return;
  }

  /*
   *  The DNS packet has a count field which indicates there can be more than
   * one query. But there's a packet wide error code. I dunno how to mark
   * each query as being successful, etc. Until this is resolved, we'll
   * follow the lead BIND has set and ignore any additional queries
   */
  if (dnspkt.qdcount > 1) {
    dnsqueryrec tquery;
    int i;

    err_warn
      ("dns_handle_packet: packet has more than 1 query, ignoring extras");

    /* Skip over the rest of the packets */
    for (i = 1; i < dnspkt.qdcount; i++) {
      /* Parse out the query */
      if (dns_parsequery (pkt, &cp, &tquery)) {
	/* Woops */
	err_warn ("dns_handle_packet: formerr, dns_parsequery returned !0");
	dns_buildpacket (pkt, NULL, NULL, dnspkt.id, DE_FORMERR, dnspkt.flags,
			 1, 0);
	dns_sendpacket (pkt->sock, pkt, &pkt->ip, pkt->port);
	return;
      }

      /* Print them out, maybe this is an attack? If not, we can spot the */
      /*  broken resolver libraries this way */
      err_info
	("dns_handle_packet: extra query. name: %s, class: %d, type: %d",
	 tquery.name, tquery.class, tquery.type);
    }
  }

  /* Let's check to see if this is a query or not */
  if (dnspkt.flags.qr)
    dns_resp_handlepkt (&dnspkt, pkt, cp);
  else
    dns_query_handlepkt (&dnspkt, pkt);
}
