/* dns_qr_query.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/*
 * dns_qr_query.c
 *
 *  Code/Data path for DNS packets we receive which are queries
 */
#include "common.h"
#include "dns.h"

#include <arpa/inet.h>		/* inet_ntoa */

/*
 * dns_query_handlepkt
 *
 *  Called when we get a DNS packet which is a query. Will parse out all of
 * the queries in the packet and send off the task of getting the information
 * for the particular queries to the correct threads.
 */
void
dns_query_handlepkt (dnspktrec * dnsqpkt, packetrec * pkt)
{
#define ROUTINE "dns_query_handlepkt"
  dnsbuildrec *build;
  int rc;

  DPRINTF (("%s: name: %s, class: %d (%s), type: %d (%s)\n",
	    ROUTINE,
	    dnsqpkt->query.name,
	    dnsqpkt->query.class, dns_dc_to_string (dnsqpkt->query.class),
	    dnsqpkt->query.type, dns_dt_to_string (dnsqpkt->query.type)));

  /* Get a new build data structure */
  build = dns_build_alloc ();
  assert (build);

  /* Check the opcode to see what they want us to do */
  switch (dnsqpkt->flags.opcode) {

  case DO_STANDARD:

    /* Handle zone transfers specially */
    if (dnsqpkt->query.type == DT_AXFR) {
      /* Only allow zone transfers using TCP to happen */
      if (pkt->sock->type == SOCKET_UDP) {
	err_warn
	  ("dns_query_handlepkt: Rejecting UDP Zone Transfer from %s:%d",
	   inet_ntoa (pkt->ip), ntohs (pkt->port));

	/* FORMERR because UDP is technically wrong, plus it's what */
	/*  BIND does */
	dns_buildpacket (pkt, &dnsqpkt->query, NULL, dnsqpkt->id, DE_FORMERR,
			 dnsqpkt->flags, 1, 0);
	dns_sendpacket (pkt->sock, pkt, &pkt->ip, pkt->port);

	/* don't return here - fall through the switch() statement so
	   we can clean up memory at the end of this function.
	 */
	break;
      }

      /* FIXME: Access control */

      /* Do it */
      dns_ans_zone_transfer (dnsqpkt, build, pkt);

      /* don't return here - fall through the switch() statement so
         we can clean up memory at the end of this function.
       */
      break;
    }

    /* Roll through all of the functions */
    rc = dns_ans_dochains (dnsqpkt, build, pkt);

    if (rc == DE_WAITFORMOREINFO)
      break;

    dns_buildpacket (pkt, &dnsqpkt->query, build, dnsqpkt->id, rc,
		     dnsqpkt->flags, 1, (dnsqpkt->qru.query.zone->flags &
					 DNS_ZONE_AUTHORITATIVE));
    dns_sendpacket (pkt->sock, pkt, &pkt->ip, pkt->port);
    break;

  case DO_STATUS:
#ifdef TAKE_THE_GURU_APPROVED_ACTION
    /*
       Subject:      shutdown()
       From:         alan@cymru.net (Alan Cox)
       Date:         1997/04/09
       Message-Id:   <199704090925.KAA07858@snowcrash.cymru.net>
       Newsgroups:   comp.security.ssh

       The shutdown() operation is defined solely for
       sockets. Nothing in POSIX 1003.1g draft 6.4 requires
       shutdown has any meaning on a pipe. Indeed since normal
       (read non BSD4) pipes are unidirectional it isn't needed
       as close has the full semantics.

       To make matters more complex FreeBSD pipes are no longer
       sockets as far as I can tell from their code, but do
       seem to implement shutdown().

       As far as I can tell from 1003.1g calling shutdown()
       on a non socket can do anything you like (eg run nettrek
       or start emacs) but its a fairly safe bet that you will
       get an error such as -ENOTSOCK so you can use shutdown
       and ignore the return code if you wish.

       Alan
     */

    execl ("/usr/bin/xemacs", "xemacs");
    execl ("/usr/X11R6/bin/nettrek", "nettrek");
#else
    err_warn ("Received DO_STATUS from %s:%d", inet_ntoa (dnsqpkt->ip),
	      dnsqpkt->port);
#endif
    break;
  case DO_INVERSE:
    /* We don't support this (will we ever?) */
    dns_buildpacket (pkt, &dnsqpkt->query, NULL, dnsqpkt->id,
		     DE_NOTIMP, dnsqpkt->flags, 1, 0);
    dns_sendpacket (pkt->sock, pkt, &pkt->ip, pkt->port);
    break;

  default:
    err_warn ("Unknown opcode %d", dnsqpkt->flags.opcode);
    break;
  }

  /* free the build data structure */
  dns_build_delete (build);

#undef ROUTINE
}				/* end dns_query_handlepkt() */
