/* dns_read_file.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/*
 * dns_read_file.c
 *
 *  Reads in RFC 883 style zone files into memory.
 */
#include "common.h"
#include "dns.h"

#include <ctype.h>		/* isspace */

enum
{
  DNS_RDB_SPACE,
  DNS_RDB_NOTSPACE
};

/*
 * Amount of memory that is allocated to read in each line of the config file.
 *  The line can't be longer than this
 */
#define DNS_RDB_MAXLINELENGTH 4096

/*
 * Maximum number of arguments that can be on a line in the DB file
 */
#define DNS_RDB_MAXARGS 255

/*
 * Maximum number of parsing errors before it exit's out
 */
#define DNS_RDB_MAXPARSEERRORS 5

/*
 * Maximum levels $include is allowed to nest. Basically to stop going into
 *  an infite loop
 */
#define DNS_RDB_MAXINCLUDENESTING 50

typedef struct
{
  guchar filename[PATH_MAX + 1];
  guchar origin[DNS_MAXNAMELENGTH + 1];

  int linenum;
}
dnsrdbstaterec;

/*
 * dns_rdb_readline
 *
 *  Reads in one line, following parentheses, etc
 */
static int
dns_rdb_readline (FILE * f, dnsrdbstaterec * state, guchar * linebuf)
{
  guint8 openparen = 0;
  guchar *p;

  p = linebuf;
  do {
    /* Read in the zone file a line at a time */
    /* Woo, pointer arithmetic! */
    if (fgets (p, DNS_RDB_MAXLINELENGTH - (p - linebuf), f) == NULL) {
      /* FIXME: fatal is a little wrong here */
      if (ferror (f))
	err_fatal ("Error reading database file!\n");

      if (openparen)
	err_fatal ("Unexpected end of file!");

      return (-1);
    }

    state->linenum++;

    while (*p) {
      /* Check for quoted charcters */
      switch (*p) {
      case '\\':		/* Escape the next character */
	p++;
	break;
      case '(':		/* This line wraps (ie ignore end of lines) */
	openparen = 1;
	memmove (p, p + 1, strlen (p));
	break;
      case ')':		/* Unignore end of lines */
	openparen = 0;
	memmove (p, p + 1, strlen (p));
	break;
      case ';':		/* Rest of the line is a comment */
	*p = 0;
	break;
      }

      if (*p)
	p++;
    }

    /* Check to see if the line we read in was longer than our buffer and */
    /*  give a warning */
    if (strlen (linebuf) >= (DNS_RDB_MAXLINELENGTH - 1)) {
      err_warn ("Warning: Line is >= %d characters long",
		DNS_RDB_MAXLINELENGTH);
      break;
    }
  } while (openparen);

  /* Check to see if there's some trailing CR-LF's */
  if (linebuf[0])
    for (p = &linebuf[strlen (linebuf) - 1];
	 (p >= linebuf) && (((*p == '\r') || (*p == '\n'))); p--)
      *p = 0;

  return (0);
}

/*
 * dns_rdb_scanline
 *
 *  Scan's the line for a specific character (or set of characters). Skips
 * over \ which means to quote the next character per the RFC
 */
guchar *
dns_rdb_scanline (guchar * p, guint16 ch)
{
  int match;

  while (*p) {
    /* Is it the quote character? */
    if ((*p == '\\') && (*(p + 1) != 0)) {
      memmove (p, p + 1, strlen (p));
      p++;
    }

    switch (ch) {
    case DNS_RDB_NOTSPACE:
      match = !isspace (*p);
      break;
    case DNS_RDB_SPACE:
      match = isspace (*p);
      break;
    default:
      match = (*p == (guchar) ch);
      break;
    }

    /* Did it match? */
    if (match)
      return (p);

    p++;
  }

  return (p);
}

/*
 * dns_rdb_parseline
 *
 *  Parse's a line read from a zone file
 */
int
dns_rdb_parseline (guchar * linebuf, guchar * argv[], dnsrdbstaterec * state)
{
  guchar *p, *p1;
  int numargs = 0, i;

  p = linebuf;

  /* Let's start ruling out blank lines and comments */
  if (!strlen (p))
    return (0);

  if (isspace (*p)) {
    /* If the beginning is a space, we'll set argv[0][0] to 0 as a special */
    /*  case so later on we know to use the previous domain name */
    argv[numargs++] = (guchar *) strchr (p, 0);
    while (isspace (*p))
      p++;
  }

  /* Next, split the line up by spaces */
  while (p && *p && (numargs < DNS_RDB_MAXARGS)) {
    /* Skip all of the whitespace */
    p = dns_rdb_scanline (p, DNS_RDB_NOTSPACE);

    /* Quote character? */
    if (*p == '\"' || *p == '\'') {
      /* The argument starts at the next character */
      argv[numargs++] = p + 1;

      /* Find the next occurance of the quote character */
      p1 = dns_rdb_scanline (p + 1, *p);

      /* If it's NULL, then there's not close quote, bitch and continue */
      if (!*p1) {
	err_warn ("dns_rdb_readlevel: No close %c on line %d", *p,
		  state->linenum);
	return (-1);
      }
    } else {
      /* Save the position */
      argv[numargs++] = p1 = p;

      /* Skip all of the regular text */
      p1 = dns_rdb_scanline (p1, DNS_RDB_SPACE);
    }

    /* Null terminate it if necessary */
    if (*p1)
      p = p1 + 1;
    else
      p = NULL;

    *p1 = 0;

    /* Was this last one blank? */
    if (!*argv[numargs - 1])
      numargs--;
  }

#ifdef DENTS_DEBUG
  for (i = 0; i < numargs; i++)
    printf ("[%s] ", argv[i]);
  printf ("\n");
#endif

  return (numargs);
}

/*
 * dns_rdb_read_level
 *
 *  Recursive function to read in all of the lines of a zone file and
 * parse them out
 * FIXME: Make sure the SOA is at top and there's only one SOA
 */
static int
dns_rdb_read_level (dbrootrec * db, int level, int check,
		    dnsrdbstaterec * state, guchar * linebuf, guchar * rrbuf)
{
  FILE *f;
  guchar *p, *argv[DNS_RDB_MAXARGS];
  guchar domainname[DNS_MAXNAMELENGTH + 1];
  guchar prevdomainname[DNS_MAXNAMELENGTH + 1];
  guchar defaultdomain[DNS_MAXNAMELENGTH + 1];
  int parseerror = 0, buflen, numargs, curarg;
  guint16 class, prevclass = DC_IN;
  guint32 ttl, defttl = 0xFFFF;
  dnsrrtyperec *rrtype;

  /* Make sure we don't go recursively into $include's forever */
  if (level > DNS_RDB_MAXINCLUDENESTING) {
    err_warn ("Too many nested $include's in %s", state->filename);
    return (ERR_DNS_RDB_PARSEERROR);
  }

  /* Setup the default domain */
  strncpyzt (defaultdomain, state->origin, DNS_MAXNAMELENGTH);

  /* Open up the zone file for reading */
  if ((f = fopen (state->filename, "rt")) == NULL)
    return (ERR_DNS_RDB_NOSUCHFILE);

  /* Loop while we have no more data to read in */
  while (!feof (f)) {
    /* FIXME: Check to make sure we haven't had too many parse errors */

    if (dns_rdb_readline (f, state, linebuf) < 0)
      break;

    /* If we're out of data, continue */
    if (feof (f))
      continue;

    if ((numargs = dns_rdb_parseline (linebuf, argv, state)) < 0) {
      /* If we get a negative number then something went wrong */
      parseerror++;
      continue;
    }

    /* If there's 0 arguments, go to the next line */
    if (!numargs)
      continue;

    /* Now let's check if it's a command directive (ie starting with a $ */
    /*  sign, those require slightly different processing) */
    if (argv[0][0] == '$') {
      if (stricmp (argv[0], "$origin") == 0) {
	if (numargs < 2) {
	  err_warn ("Not enough arguments for $ORIGIN command on line %d",
		    state->linenum);
	  parseerror++;
	} else
	  strncpyzt (state->origin, argv[1], DNS_MAXNAMELENGTH);
      }

      if (stricmp (argv[0], "$include") == 0) {
	dnsrdbstaterec newstate;

	strncpyzt (newstate.filename, argv[1], PATH_MAX);
	if (numargs < 3)
	  strncpyzt (newstate.origin, state->origin, DNS_MAXNAMELENGTH);
	else
	  strncpyzt (newstate.origin, argv[2], DNS_MAXNAMELENGTH);

	newstate.linenum = 0;

	dns_rdb_read_level (db, level + 1, check, &newstate, linebuf, rrbuf);
      }

      continue;
    }

    /*
     * If the lines doesn't start at the beginning (ie space or tab before
     *  first character) then we use the last domain. The @ character by
     *  itself means the same thing
     */
    if ((strlen (argv[0]) == 1) && (*argv[0] == '@')) {
      strncpyzt (domainname, defaultdomain, DNS_MAXNAMELENGTH);

      strncpyzt (prevdomainname, domainname, DNS_MAXNAMELENGTH);
    } else if (!*argv[0]) {
      strncpyzt (domainname, prevdomainname, DNS_MAXNAMELENGTH);
    } else {
      dns_qualify_domain (domainname, argv[0], defaultdomain);

      strncpyzt (prevdomainname, domainname, DNS_MAXNAMELENGTH);
    }

    /* Make sure we have enough arguments */
    if (numargs <= 1) {
      parseerror++;
      err_warn ("Not enough arguments! (numargs <= 1)");
      continue;
    }

    curarg = 1;

    /* Let's check if this is a TTL or not */
    ttl = strtoul (argv[curarg], (char **) &p, 10);

    if (!*p) {
      curarg++;

      if ((class = dns_types_findclass (argv[curarg])) > 0) {
	prevclass = class;
	curarg++;
      } else
	class = prevclass;
    } else {
      if ((class = dns_types_findclass (argv[curarg])) > 0) {
	prevclass = class;
	curarg++;

	ttl = strtoul (argv[curarg], (char **) &p, 10);

	if (*p)
	  ttl = defttl;
      } else {
	ttl = defttl;
	class = prevclass;
      }
    }

    /* Make sure we have enough arguments for this line */
    if (curarg >= numargs) {
      parseerror++;
      err_warn ("Not enough arguments! (curarg >= numargs)");
      continue;
    }

    /* Find the type by label */
    if (!(rrtype = dns_types_findrrtype_by_label (argv[curarg++]))) {
      parseerror++;
      err_warn ("Invalid RR type");
      continue;
    }

    /* Is this an obsoleted or unsupported type? */
    if (rrtype->numargs < 0) {
      err_warn ("Obsoleted RR type %s", rrtype->label);
      continue;
    } else if (rrtype->numargs == 0) {
      err_warn ("Unsupported RR type %s", rrtype->label);
      continue;
    }

    /* Check to see if we have enough arguments for this type */
    if ((curarg + rrtype->numargs) > numargs) {
      parseerror++;
      err_warn ("Not enough arguments for %s", rrtype->label);
      continue;
    }

    buflen = dns_types_pack (rrtype, &argv[curarg], numargs - curarg,
			     defaultdomain, rrbuf);

    if (buflen <= 0) {
      parseerror++;
      err_warn ("Unknown/Invalid RR type %s", rrtype->label);
      continue;
    }

    /* Grab the default TTL */
    if (rrtype->qtype == DT_SOA) {
      defttl = strtoul (argv[9], NULL, 10);

      /* The default TTL applies to the SOA as well */
      if (ttl == 0xFFFF)
	ttl = defttl;
    }

    /* Now we'll actually store this information in our DB */
    dns_db_rr_add (db, domainname, class, rrtype->qtype, ttl, rrbuf, buflen);
  }

  fclose (f);

  /* Any parse errors? */
  if (parseerror)
    return (ERR_DNS_RDB_PARSEERROR);

  return (ERR_DNS_RDB_NOERROR);
}

/*
 * dns_rdb_read
 *
 *  Attempts to read in the zone file specified by filename. This file
 * should be in the database format specified by RFC 1035.
 */
int
dns_rdb_read (dbrootrec * db, int check, char *deforigin, char *filename)
{
  dnsrdbstaterec state;
  guchar *linebuf, *rrbuf;
  int rc;

  /* First we'll allocate some memory for the buffer to read in lines */
  if ((linebuf = (guchar *) malloc (DNS_RDB_MAXLINELENGTH)) == NULL)
    err_fatal ("Couldn't allocate %d bytes for line buffer!",
	       DNS_RDB_MAXLINELENGTH);

  if ((rrbuf = (guchar *) malloc (DNS_MAXRRDATALENGTH)) == NULL)
    err_fatal ("Couldn't allocate %d bytes for RR buffer!",
	       DNS_MAXRRDATALENGTH);

  /* Reset some of the state information */
  strncpyzt (state.filename, filename, PATH_MAX);
  strncpyzt (state.origin, deforigin, DNS_MAXNAMELENGTH);
  state.linenum = 0;

  rc = dns_rdb_read_level (db, 0, check, &state, linebuf, rrbuf);

  /* Free those allocated variables */
  free (linebuf);
  free (rrbuf);

  return (rc);
}
