/* dns_sock_tcp.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/*
 * dns_sock_tcp.c
 *
 *  Handles reading/writing packets over TCP both synchronously
 * and asynchronously
 */
#include "common.h"
#include "dns.h"

#include <arpa/inet.h>		/* inet_ntoa */
#include <sys/socket.h>		/* bind */

/* Private data for DNS TCP sockets */
blocklist dnstcpsockets = BLOCK_INITIALIZER (dnstcprec, 1);

/*
 * dns_sock_tcp_accept
 *
 *  accept()'s a connection on a listening TCP socket
 */
void
dns_sock_tcp_accept (socketrec * asock, packetrec * pkt)
{
  int fd;
  struct sockaddr_in sa;
  int salen;
  socketrec *sock;
  dnstcprec *priv;

  /* Allocate the socket */
  sock = dns_sock_tcp_alloc ();
  priv = (dnstcprec *) sock->priv;

  /* Zero it out */
  memset ((void *) &sa, 0, sizeof (sa));

  /* accept() the socket */
  salen = sizeof (sa);
  if ((fd = accept (asock->fd, (struct sockaddr *) &sa, &salen)) < 0) {
    err_warn ("dns_sock_tcp_accept: accept: %m");
    return;
  }

  DPRINTF (("dns_sock_tcp_accept: connection from %s:%d on fd %d\n",
	    inet_ntoa (sa.sin_addr), htons (sa.sin_port), fd));

  /* Set up the socket structure */
  sock->fd = fd;
  sock->readaction = dns_sock_tcp_readsize;
  sock->writeaction = dns_sock_tcp_sendbuf;

  /* Set up the private data */
  memcpy ((void *) &priv->ip, (void *) &sa.sin_addr, sizeof (priv->ip));
  priv->port = htons (sa.sin_port);
  memcpy ((void *) &priv->packet.ip, (void *) &sa.sin_addr,
	  sizeof (priv->packet.ip));
  priv->packet.port = ntohs (sa.sin_port);
  priv->packet.sock = sock;
  priv->packetread = priv->packetsize = 0;

  /* Allocate the packet buffer */
  priv->packet.pbufsize = 4096;
  priv->packet.tbufsize = 4096;

  if ((priv->packet.packetbuf = malloc (priv->packet.pbufsize)) == NULL)
    err_fatal ("dns_sock_tcp_accept: malloc: %m");

  if ((priv->packet.tempbuf = malloc (priv->packet.tbufsize)) == NULL)
    err_fatal ("dns_sock_tcp_accept: malloc: %m");

  /* Make sure it's non blocking */
  sock_setnonblock (sock->fd);
}

/*
 * dns_sock_tcp_readsize
 *
 *  Simple function which reads the size of the packet from the TCP stream
 * per the RFC
 */
void
dns_sock_tcp_readsize (socketrec * sock, packetrec * unused)
{
  int numread;
  dnstcprec *priv = (dnstcprec *) sock->priv;

  /* Read in a maximum of guint16 bytes for the length of the DNS packet */
  if ((numread = read (sock->fd, priv->packet.tempbuf + priv->packetread,
		       sizeof (guint16) - priv->packetread)) < 0) {
    err_warn ("dns_sock_tcp_readsize: read: %m");
    close (sock->fd);
    dns_sock_tcp_free (sock);
    return;
  }

  /* Did they close the connection on us */
  if (!numread) {
    close (sock->fd);
    dns_sock_tcp_free (sock);
    return;
  }

  /* Update the count */
  priv->packetread += numread;

  /* Have we read enough? */
  if (priv->packetread < sizeof (guint16))
    return;

  /* Ok, now we know enough */
  priv->packetread = 0;
  priv->packetsize = ntohs (*((guint16 *) priv->packet.tempbuf));

  /* Change the action */
  sock->readaction = dns_sock_tcp_readpacket;
}

/*
 * dns_sock_tcp_readpacket
 *
 *  Read's in the entire DNS packet and when done, processes the packet
 */
void
dns_sock_tcp_readpacket (socketrec * sock, packetrec * unused)
{
  int numread;
  dnstcprec *priv = (dnstcprec *) sock->priv;

  /* Read in the maximum size of the packet length given to us */
  if ((numread = read (sock->fd, priv->packet.packetbuf + priv->packetread,
		       priv->packetsize - priv->packetread)) < 0) {
    err_warn ("dns_sock_tcp_readpacket: read: %m");
    close (sock->fd);
    dns_sock_tcp_free (sock);
    return;
  }

  /* Did they close the connection */
  if (numread == 0) {
    close (sock->fd);
    dns_sock_tcp_free (sock);
    return;
  }

  /* Update the count */
  priv->packetread += numread;

  /* Have we read enough? */
  if (priv->packetread < priv->packetsize)
    /* Not enough, return until we do */
    return;

  priv->packet.packetsize = priv->packetsize;

  /* Ok, we have enough now */
  dns_handle_packet (&priv->packet);

  /* Flip back */
  sock->readaction = dns_sock_tcp_readsize;
  priv->packet.packetsize = 0;

  priv->packetsize = priv->packetread = 0;
}

/*
 * dns_sock_tcp_send
 *
 *  Send's a packet back across a TCP connection asynchronously
 */
void
dns_sock_tcp_send (socketrec * sock, guchar * buf, int length,
		   struct in_addr *ip, guint16 port)
{
  guint16 packetsize;
  dnstcprec *priv = (dnstcprec *) sock->priv;
#ifdef DENTS_DEBUG
  int i;

  DPRINTF (("dns_sock_tcp_send: sending packet\n "));

  for (i = 0; i < priv->packet.packetsize; i++) {
    printf ("%02X ", (unsigned char) priv->packet.packetbuf[i]);
    if ((i & 15) == 15)
      printf ("\n ");
  }
  printf ("\n\n");
#endif

  DPRINTF (("dns_sock_tcp_send: sending packet via TCP\n"));

  /* Write the size of the packet */
  packetsize = htons (length);
  if (sock_tcp_write (sock, (guchar *) & packetsize, sizeof (packetsize)) < 0) {
    err_warn ("dns_sock_tcp_send: error writing: %m");
    dns_sock_tcp_free (sock);
    return;
  }

  /* Then write the actual packet */
  if (sock_tcp_write (sock, buf, length) < 0) {
    err_warn ("dns_sock_tcp_send: error writing: %m");
    dns_sock_tcp_free (sock);
    return;
  }
}

/*
 * dns_sock_tcp_sendbuf
 *
 *  Wrapper for TCP's dns_sock_tcp_sendbuf to close appropriate socket on
 * error
 */
void
dns_sock_tcp_sendbuf (socketrec * sock)
{
  /* Simple */
  if (sock_tcp_sendbuf (sock) < 0) {
    err_warn ("dns_sock_tcp_sendbuf: write: %m");
    dns_sock_tcp_free (sock);
  }
}

/*
 * dns_sock_tcp_pullzone
 *
 *  Called when we want to pull down a zone from a zone transfer
 */
void
dns_sock_tcp_pullzone (socketrec * sock, packetrec * pkt)
{
  DPRINTF (("dns_sock_tcp_pullzone: called\n"));
}

/*
 * dns_sock_tcp_sendaxfr
 *
 *  Sends the AXFR request
 */
void
dns_sock_tcp_sendaxfr (socketrec * sock)
{
  DPRINTF (("dns_sock_tcp_sendaxfr: called\n"));
}

/*
 * Administrative functions
 */

/*
 * dns_sock_tcp_bind
 *
 *  bind()'s all TCP sockets
 */
void
dns_sock_tcp_bind (void)
{
  int fd;
  socketrec *sock;
  struct in_addr addr;

  addr.s_addr = INADDR_ANY;

  /* FIXME: we should let the user choose the port to bind to */
  fd = sock_open (SOCKET_TCP, addr, dns.port, 1);

  if (listen (fd, MAX_LISTEN) < 0)
    err_fatal ("dns_sock_tcp_bind: listen: %m");

  /* Allocate a socket */
  sock = dns_sock_tcp_alloc ();

  sock->fd = fd;
  sock->readaction = dns_sock_tcp_accept;
  sock->writeaction = NULL;

  sock->interface = NULL;
}

/*
 * dns_sock_tcp_free
 *
 *  Free's a DNS TCP socket
 */
void
dns_sock_tcp_free (socketrec * sock)
{
  sock_free (sock);
}

/*
 * dns_sock_tcp_alloc
 *
 *  Allocate's a DNS TCP socket
 */
socketrec *
dns_sock_tcp_alloc (void)
{
  socketrec *sock;
  dnstcprec *priv;

  /* Allocate the socket */
  sock = sock_tcp_alloc ();
  priv = (dnstcprec *) (sock->priv = block_allocandadd (&dnstcpsockets));

  priv->socket = sock;

  /* Add to our chain of sockets */
  link_add (&dnssockets, &sock->lp);

  return (sock);
}
