/* dns_sock_udp.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/*
 * dns_sock_udp.c
 *
 *  Handles reading/writing packets over UDP
 */
#include "common.h"
#include "dns.h"

#include <arpa/inet.h>		/* inet_ntoa */

/* Private data for DNS UDP sockets */
blocklist dnsudpsockets = BLOCK_INITIALIZER (dnsudprec, 1);

/*
 * dns_sock_udp_read
 *
 *  Called whenever select() returns that something is happening with a UDP
 * packet
 */
void
dns_sock_udp_read (socketrec * sock, packetrec * pkt)
{
#define ROUTINE "dns_sock_udp_read ()"

  int numread;
  struct sockaddr_in sa;
  int salen = sizeof (sa);

  DPRINTF (("%s: reading up to %hu bytes\n", ROUTINE, pkt->pbufsize));

  /* Read in the actual packet */
  errno = 0;
  if ((numread = recvfrom (sock->fd, pkt->packetbuf, pkt->pbufsize, 0,
			   (struct sockaddr *) &sa, &salen)) < 0) {
    err_warn ("dns_sock_udp_read: recvfrom: %m");
    return;
  }

  /* The socket we received this packet from */
  pkt->sock = sock;

  /* Then record where the packet came from */
  memcpy ((void *) &pkt->ip, (void *) &sa.sin_addr, sizeof (pkt->ip));
  pkt->port = ntohs (sa.sin_port);

  /* Fill in the information */
  pkt->packetsize = numread;

  DPRINTF (("%s: received from %s:%d %d bytes. errno=%d\n",
	    ROUTINE, inet_ntoa (pkt->ip), pkt->port, numread, errno));

  /* Then handle the packet */
  dns_handle_packet (pkt);

#undef ROUTINE
}

/*
 * dns_sock_udp_send
 *
 *  Send's a UDP DNS packet back across to the correct destination
 */
void
dns_sock_udp_send (socketrec * sock, guchar * buf, int length,
		   struct in_addr *ip, guint16 port)
{
#ifdef DENTS_DEBUG
  int i;

  DPRINTF (
	   ("dns_sock_udp_send: sending to %s:%d\n  ", inet_ntoa (*ip),
	    port));

  /* Print out the raw packet */
  for (i = 0; i < length; i++) {
    printf ("%02X ", (unsigned char) buf[i]);

    if ((i & 15) == 15)
      printf ("\n  ");
  }
  printf ("\n\n");

  /* NOTE: this may not always be safe!!! If we ever send non-DNS messages
     through this routine, debug_print_pkt will probably make the server
     toss its cookies.
   */
  dns_debug_print_pkt (buf, length);

#endif

  /* Send it */
  if (sock_udp_write (sock, buf, length, ip, port) < 0) {
    err_warn ("dns_sock_udp_send: error writing: %m");
    dns_sock_udp_free (sock);
  }
}

/*
 * Administrative functions
 */

/*
 * dns_sock_udp_alloc
 *
 *  Allocates a regular socket as well as any private data
 */
socketrec *
dns_sock_udp_alloc (void)
{
  socketrec *sock;
  dnsudprec *priv;

  /* Allocate the socket */
  sock = sock_udp_alloc ();
  priv = (dnsudprec *) (sock->priv = block_allocandadd (&dnsudpsockets));

  priv->socket = sock;

  /* Add to our chain of sockets */
  link_add (&dnssockets, &sock->lp);

  return (sock);
}

/*
 * dns_sock_udp_free
 *
 *  Free's a DNS UDP socket
 */
void
dns_sock_udp_free (socketrec * sock)
{
  sock_udp_free (sock);
}

/*
 * dns_sock_udp_bind
 *
 *  bind()'s all of the UDP sockets needed for DNS
 */
void
dns_sock_udp_bind (void)
{
  int fd;
  socketrec *sock;
  struct sockaddr_in sa;
  interfacerec *iface;

  /* Clear, clear, clear */
  memset ((void *) &sa, 0, sizeof (sa));

  /* Loop through all of the interfaces */
  for (iface = link_first (&interfaces); iface;
       iface = link_next (iface, iface->ll)) {
    /* FIXME: we should check this address against a "dont bind" list from */
    /*  a config file */

    /* Open up a socket */
    /* FIXME: we should let the user choose the port to bind to */
    fd = sock_open (SOCKET_UDP, iface->address, dns.port, 1);

    sock = dns_sock_udp_alloc ();

    sock->fd = fd;
    sock->readaction = dns_sock_udp_read;
    sock->writeaction = NULL;

    sock->interface = iface;
    /* FIXME: protocol ops */
  }

  /* FIXME: Don't do this if we don't do recursive queries */
  {
    sa.sin_addr.s_addr = INADDR_ANY;
    dns.sendsock.fd = sock_open (SOCKET_UDP, sa.sin_addr, dns.port, 1);

    dns.sendsock.readaction = NULL;
    dns.sendsock.writeaction = NULL;
    /* FIXME: protocol ops */
  }
}
