/* dns_zone.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/*
 * dns_zone.c
 *
 *  Handles adding/deleting/loading of all zones
 *
 *  Prototypes in dns_zone.h
 *
 */

/***************************************************************************
 *			Include Files
 ***************************************************************************/
#include "common.h"
#include "dns.h"

#include <stdarg.h>		/* va_list */

/***************************************************************************
 *			Macro Definitions
 ***************************************************************************/

/***************************************************************************
 *			Internal Function Prototypes
 ***************************************************************************/

/***************************************************************************
 *			Global Data
 ***************************************************************************/

blocklist chain_pool = BLOCK_INITIALIZER (dnschainrec, 1);

/***************************************************************************
 ***************************************************************************
 *			Function Definitions
 ***************************************************************************
 ***************************************************************************/

/*
 * dns_zone_addmodule
 *
 *  Add's a function to a zone's function table
 */
void
dns_zone_addmodule (dnszonerec * zone, dnsmodulerec * module)
{
  dnschainrec *chain;

  chain = block_allocandaddtolink (&chain_pool, &zone->lookup);

  chain->module = module;
}

/*
 * dns_zone_create
 *
 *  Add's a zone to the tree. Doesn't read in the zone file yet, just records
 * the name for future use.
 */
dnszonerec *
dns_zone_create (guchar * zonename)
{
  dnszonerec *zone;

  /* Allocate a zone */
  zone = block_allocandadd (&dns.zone);

  strncpyzt (zone->zonename, zonename, DNS_MAXNAMELENGTH);
  zone->status = DNS_ZONE_NOTLOADED;

  /* Add this zone into the DB */
  dns_db_addzone (&dns.zonedb, zonename, zone);

  return (zone);
}

/*
 * dns_zone_setloadmod
 *
 *  Set's the module which will (re)load the zone
 */
void
dns_zone_setloadmod (dnszonerec * zone, dnsmodulerec * module)
{
  zone->loadmodule = module;
}

/*
 * dns_zone_setslaveip
 *
 *  Set's a zone up as a slave and fills in master's IP address
 */
void
dns_zone_setslaveip (dnszonerec * zone, struct in_addr ip)
{
  zone->flags &= ~DNS_ZONE_MASTER;
  memcpy ((void *) &zone->master, (void *) &ip, sizeof (ip));
}

/* Set's the filename for a zone's data. */
void
dns_zone_setfilename (dnszonerec * zone, guchar * fnfmt, ...)
{
  va_list va;
  va_start (va, fnfmt);
  g_vsnprintf (zone->filename, PATH_MAX, fnfmt, va);
  va_end (va);
}

/* Load's all of the zones listed in the config structure. */
void
dns_zone_loadall (void)
{
  dnszonerec *zone;
  dnszonerec *next_zone;
  dnshintrec *hint;
  int rc;

  /* Loop through all of the zones. */
  for (zone = link_first (&dns.zone.used); zone; zone = next_zone) {

    /* Go ahead and get the next zone in the list in case we have to
       invalidate this zone entry.  */
    next_zone = link_next (zone, zone->ll);

    if (!zone->loadmodule) {
      err_warn ("dns_zone_loadall: No loadmodule symbol. ignoring zone %s",
		zone->zonename);
      dns_zone_cleanup (zone);
      continue;
    }

    rc = zone->loadmodule->zone_load (zone);

    switch (rc) {
    case ERR_DNS_RDB_PARSEERROR:
    case ERR_DNS_RDB_NOSUCHFILE:
      if (zone->flags & DNS_ZONE_MASTER) {
	/* Uhh, it's a primary zone and the file is missing. We'll
	   play dumb for now, after bitching a bit. */
	err_warn ("dns_zone_loadall: ignoring zone %s", zone->zonename);
	dns_zone_cleanup (zone);
	continue;

      } else {
	/* It's a seconday zone, so let's setup an event to fetch it. */
	dns_zone_scheduletransfer (zone);
      }
      break;
    }
  }

  /* Now load in all of the hint information. */
  for (hint = link_first (&dns.hint.used); hint;
       hint = link_next (hint, hint->ll)) {
    /* These are always going to be standard zone files */
    rc = dns_rdb_read (&dns.hintdb, DNS_RDB_NOSANITYCHECK, hint->zonename,
		       hint->filename);
    switch (rc) {
    case ERR_DNS_RDB_NOSUCHFILE:
      err_fatal ("dns_zone_loadall: hints file @ %s not found!",
		 hint->filename);
      break;
    case ERR_DNS_RDB_PARSEERROR:
      err_fatal ("dns_zone_loadall: hints file @ %s had parsing problems!",
		 hint->filename);
      break;
    }
  }
}

/*
 * dns_zone_scheduletransfer
 *
 *  Schedules a zone transfer to take place
 */
void
dns_zone_scheduletransfer (dnszonerec * zone)
{
  return;
  /* Broken code, #if'd out for now */
#if 0
  int fd;
  struct sockaddr_in sa;
  dnstcpsocketrec *ts;
  struct servent *portinfo;

  memset ((void *) &sa, 0, sizeof (sa));

  /* Let's open up the listening TCP sockets */
  if ((fd = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
    err_fatal ("dns_zone_scheduletransfer: socket: %m");

  /* Set it to be non blocking */
  sock_setnonblock (fd);

  /* You should use getservbyname() -kwy */
  portinfo = getservbyname ("domain", "tcp");
  if (portinfo)
    sa.sin_port = portinfo->s_port;
  else
    sa.sin_port = htons (DNS_PORT);

  memcpy ((void *) &sa.sin_addr, (void *) &zone->master,
	  sizeof (sa.sin_addr));
  sa.sin_family = AF_INET;

  DPRINTF (("dns_zone_scheduletransfer: connect()'ing to %s:%d\n",
	    inet_ntoa (sa.sin_addr), ntohs (sa.sin_port)));

  if (connect (fd, (struct sockaddr *) &sa, sizeof (sa)) < 0)
    if (errno != EINPROGRESS)
      err_fatal ("dns_zone_schedule_transfer: connect: %m");

  if ((ts = malloc (sizeof (dnstcpsocketrec))) == NULL)
    err_fatal ("dns_zone_scheduletransfer: malloc: %m");

  /* Add it to the linked list */
  link_add (&config.sockets, ts);

  ts->fd = fd;
  ts->action = dns_sock_tcp_pullzone;
  ts->send = dns_sock_tcp_sendaxfr;
  ts->protocol = PROT_DNS;
  ts->type = SOCKET_TCP;
#endif
}

/*
 * dns_db_addzone
 *
 *  Add a zone to a database
 */
void
dns_db_addzone (dbrootrec * db, guchar * domainname, dnszonerec * zone)
{
  guchar dname[DNS_MAXNAMELENGTH + 1], *p, *p1;
  dbnoderec *node, *n;

  /* We butcher the string, so copy it first */
  strncpyzt (dname, domainname, DNS_MAXNAMELENGTH);

  /* The root gets displayed differently, so let's correct for that */
  if (dname[0] == '.' && (!*(dname + 1)))
    dname[0] = 0;

  /* Start at the top of this db */
  node = &db->node;

  /* We'll step down each level in the tree finding if this label is already */
  /*  there. If not, we'll add it */
  do {
    if ((p = (guchar *) strrchr (dname, '.')) != NULL) {
      p1 = p + 1;
      *p = 0;
    } else
      p1 = dname;

    if ((n = db_findnode (node, p1)) == NULL)
      n = db_addnode (node, p1);

    node = n;
  } while (p);

  node->data = zone;
  zone->dbnode = node;
}

/*
 * dns_zone_cleanup ()
 *
 * This zone record is no good anymore. Remove it from the list
 * and cleanup memory associated with it.
 */
void
dns_zone_cleanup (dnszonerec * zone)
{

  /* remove the zone from the rb tree */
  db_deletenode (zone->dbnode);

  /* remove the zone from the linked list tree */
  block_delete (&dns.zone, &zone->ll);

}				/* end dns_zone_cleanup () */
