/* interface.c - Deals with the interfaces on the system.

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

#include "common.h"
#include "interface.h"

#include <sys/socket.h>		/* getsock* */
#include <net/if.h>		/* structures for ioctl(,SIOCGIFCONF,) */
#define BSD_COMP		/* for solaris' SIOCGIFCONF */
#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>		/* SIOCGIFCONF, ioctl */
#endif /* HAVE_SYS_IOCTL_H */
#include <arpa/inet.h>		/* inet_ntoa */

linkedlist interfaces;		/* Interfaces on the system */

/* Scan's the system for all of the interfaces on it. */
void
interface_scan (void)
{
  struct ifconf ifc;
  struct ifreq *ifr;
  u_char *buf;
  guint16 i, numifaces;
  int ifsock;
  interfacerec *iface;

  /* Open a socket to communicate to the kernel */
  if ((ifsock = socket (AF_INET, SOCK_DGRAM, 0)) < 0)
    err_fatal ("interface_scan: socket(AF_INET, SOCK_DGRAM, 0): %m");

  /* Arbitrary size, it will dynamically grow */
  numifaces = 32;

  /* Allocate some memory for the interfaces */
  if ((buf = malloc (sizeof (struct ifreq) * numifaces)) == NULL)
    err_fatal ("interface_scan: malloc: %m");

  while (1) {
    /* Setup the stuff for the ioctl and call it (hey i'm getting lazy with */
    /* the comments, ok?) */
    ifc.ifc_len = sizeof (struct ifreq) * numifaces;
    ifc.ifc_buf = buf;

    /* FIXME: We should probably not fatal here */
    if (ioctl (ifsock, SIOCGIFCONF, &ifc) < 0)
      err_fatal ("interface_scan: ioctl(SIOCGIFCONF): %m");

    /* If we read in the same amount we allocated, assume there's more */
    /* There is the rare chance that we actuall hit it exactly, but it's */
    /* a small price to pay */
    if ((ifc.ifc_len / sizeof (struct ifreq)) == numifaces) {
      /* Double the size */
      numifaces *= 2;

      /* Reallocate the buffer */
      if ((buf = realloc (buf, sizeof (struct ifreq) * numifaces)) == NULL)
	err_fatal ("interface_scan: realloc: %m");
    } else
      break;
  }

  /* Loop through all of the interfaces, getting the IP address for each */
  for (i = 0; i < ifc.ifc_len;) {
    ifr = (struct ifreq *) ((u_char *) ifc.ifc_req + i);

    /* For some reason BSD has to be difficult */
#ifdef HAVE_SA_LEN
    i += sizeof (ifr->ifr_name) +
      (ifr->ifr_addr.sa_len > sizeof (struct sockaddr)
       ? ifr->ifr_addr.sa_len : sizeof (struct sockaddr));
#else
    i += sizeof (struct ifreq);
#endif

    /* Make sure it's an AF_INET address. This is mostly to cover the BSD's */
    /* which send AF_LINK interface's back as well. */
    if (ifr->ifr_addr.sa_family != AF_INET)
      continue;

    /* Grab the flags of the interface */
    if (ioctl (ifsock, SIOCGIFFLAGS, ifr) < 0) {
      perror ("interface_scan: getting flags");
      err_fatal ("interface_scan: ioctl(SIOCGIFFLAGS): %m");
    }

    /* Make sure it's marked up */
    if (!(ifr->ifr_flags & IFF_UP)) {
      DPRINTF (("interface_scan: interface %s not up, skipping\n",
		ifr->ifr_name));
      continue;
    }

    /* Grab the IP address of the interface */
    if (ioctl (ifsock, SIOCGIFADDR, ifr) < 0)
      err_fatal ("interface_scan: ioctl(SIOCGIFADDR): %m");

    /* Create the socket structure */
    if ((iface = malloc (sizeof (interfacerec))) == NULL)
      err_fatal ("inteface_scan: malloc: %m");

    /* Copy over the IP address */
    memcpy (&iface->address,
	    &((struct sockaddr_in *) &ifr->ifr_addr)->sin_addr,
	    sizeof (struct in_addr));
    DPRINTF (("interface_scan: found interface %s with IP %s\n",
	      ifr->ifr_name, inet_ntoa (iface->address)));

    strncpyzt (iface->name, ifr->ifr_name, IF_MAXNAMELENGTH);

    /* Add it to the linked list */
    link_add (&interfaces, &iface->ll);
  }

  /* Close any sockets we've opened */
  close (ifsock);

  /* Free the memory we allocated */
  free (buf);
}
