/* module.c - Generic interface to some common module code.

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

#include "common.h"
#include "module.h"

linkedlist modules = LINK_INITIALIZER (modulerec, ll);
blocklist module_pool = BLOCK_INITIALIZER (modulerec, 1);

/* Populate's a module's structure with all of the symbols. */
int
module_populate (modulerec * module)
{
  void *ptr;

  /* Defaults */
  module->version = NULL;
  module->init = (func_i_v) NULL;
  module->destroy = (func_v_v) NULL;

  /* 
   * Mandatory symbols
   */
  /* Check if it's a pthreads module */
  ptr = module_sym (module, "module_compiled_pthreads");
#ifdef WITH_PTHREAD
  if (ptr == NULL) {
    err_warn
      ("module_populate: dents compiled with pthread support, but module was not");
    return (-1);
  }
#else
  if (ptr != NULL) {
    err_warn
      ("module_populate: dents compiled without pthread support, but module was");
    return (-1);
  }
#endif

  if ((module->modname = module_sym (module, "module_name")) == NULL) {
    err_warn ("module_populate: unable to find symbol module_name");
    return (-1);
  }

  if ((ptr = module_sym (module, "module_compiled_version")) == NULL) {
    err_warn
      ("module_populate: unable to find symbol module_compiled_version");
    return (-1);
  }
  module->compiled_version = *(char **) ptr;

  /* Optional symbols */
  if ((ptr = module_sym (module, "module_init")) != NULL)
    module->init = ptr;

  if ((ptr = module_sym (module, "module_destroy")) != NULL)
    module->destroy = ptr;

  if ((ptr = module_sym (module, "module_version")) != NULL)
    module->version = *(char **) ptr;

  if (module->version)
    DPRINTF (("module_populate: module %s v%s loaded. "
	      "linked against %s\n", module->modname, module->version,
	      module->compiled_version));
  else
    DPRINTF (("module_populate: module %s loaded. linked against %s\n",
	      module->modname, module->compiled_version));

  return (0);
}

/* Allocate's a module structure. */
modulerec *
module_alloc (void)
{
  return (block_allocandaddtolink (&module_pool, &modules));
}

/* Free's a module structure. */
void
module_free (modulerec * module)
{
  block_delete (&module_pool, &module->ll);
}

/* Open's a module name, handle both static and dynamic modules.
 Assumes all statically linked modules are already in the list and if
 it doesn't find anything in the list, it will try to load it as a
 dynamic module. */
modulerec *
module_open (guchar * modname)
{
  modulerec *module;

  /* First, find out if it's already loaded */
  for (module = link_first (&modules); module;
       module = link_next (module, module->ll)) {
    if (strncmp (modname, module->name, MODULE_MAXNAMELENGTH) == 0)
      break;
  }

  /* Did we find a module? */
  /* All statically linked modules should have module structs already, so */
  /* we assume they're trying to load a dynamic module */
  if ((!module) && ((module = module_dyn_open (modname)) == NULL))
    return (NULL);

  /* IF this is the first module that's trying to use this module */
  /* and the refcount is 0 then we should try to init it */
  if (!module->refcount && (module->init != NULL)) {
    /* Return code of < 0 is bad, so we mark it inactive */
    if (module->init () < 0) {
      module->state = INACTIVE;
      return (NULL);
    } else
      module->state = ACTIVE;
  } else
    /* If there's no init function, assume it's active */
    module->state = ACTIVE;

  module->refcount++;

  return (module);
}

/* Close's a module. When the last reference is closed, it will be
   really closed. */
int
module_close (modulerec * module)
{
  /* Decrement the reference count */
  module->refcount--;

  /* Did we hit 0? */
  if (!module->refcount) {
    /* Free it then */
    DPRINTF (("module_close: refcount == 0, destroying\n"));

    /* Call the destroy function */
    if (module->destroy)
      module->destroy ();

    if (module->type == DYNAMIC)
      module_dyn_close (module);

  }

  return (0);
}

/* Returns a pointer to the symbol. */
void *
module_sym (modulerec * module, guchar * name)
{
  switch (module->type) {
  case STATIC:
    return (module_st_sym (module, name));
  case DYNAMIC:
    return (module_dyn_sym (module, name));
  }

  /* Huh? */
  err_warn ("module_sym: unknown module type %d", module->type);

#ifdef DENTS_DEBUG
  abort ();
#endif

  return (NULL);
}
