/* module_dynamic.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/*
 * module_dynamic.c
 *
 *  Handles dynamically loaded modules
 */
#include <sys/stat.h>
#include <string.h>

#include "common.h"
#include "module.h"

guchar **module_dyn_paths = NULL;

#ifdef WITH_PTHREAD
pthread_mutex_t mod_dyn_lock = PTHREAD_MUTEX_INITIALIZER;
#endif

/*
 * module_dyn_isfile
 *
 *  Tells us if a file is a regular file or not
 */
static int
module_dyn_isfile (const guchar * filename)
{
  struct stat s;

  if (stat (filename, &s) != 0)
    return (0);

  return (S_ISREG (s.st_mode));
}

/*
 * module_dyn_path_add
 *
 *  Add's a directory to the module search path
 */
int
module_dyn_path_add (guchar * pathname)
{
  guchar **path, *p;
  int numpath = 0;

  LOCK ("module_dyn_path_add", mod_dyn_lock);

  /* See if it's in the path first */
  for (path = module_dyn_paths; path && *path; path++, numpath++) {
    if (strncmp (pathname, *path, PATH_MAX) == 0) {
      /* Woops, don't add it twice */
      UNLOCK ("module_dyn_path_add", mod_dyn_lock);
      return (0);
    }
  }

  /* Allocate some memory for it */
  if ((p = strdup (pathname)) == NULL) {
    err_warn ("module_dyn_path_add: strdup returned NULL");
    UNLOCK ("module_dyn_path_add", mod_dyn_lock);
    return (-1);
  }

  /* Resize the array of pointers (or allocate it in the first place) */
  if (module_dyn_paths == NULL)
    module_dyn_paths = malloc (2 * sizeof (guchar *));
  else
    module_dyn_paths = realloc (module_dyn_paths,
				(numpath + 2) * sizeof (guchar *));

  /* Problems? */
  if (module_dyn_paths == NULL)
    err_fatal ("module_dyn_path_add: realloc returned NULL");

  /* Add it in */
  module_dyn_paths[numpath] = p;
  module_dyn_paths[numpath + 1] = NULL;

  UNLOCK ("module_dyn_path_add", mod_dyn_lock);

  return (0);
}

/*
 * module_dyn_path_del
 *
 *  Delete's a directory from the module search path
 */
int
module_dyn_path_del (guchar * pathname)
{
  guchar **path, **npath;
  int numpath = 0, pathpos;

  LOCK ("module_dyn_path_del", mod_dyn_lock);

  /* Linearly search for the path in the list */
  for (path = module_dyn_paths; path && *path; path++, numpath++) {
    if (strncmp (pathname, *path, PATH_MAX) == 0)
      break;
  }

  /* Did we hit the end of the list? If so, we didn't find it */
  if (*path == NULL) {
    UNLOCK ("module_dyn_path_del", mod_dyn_lock);
    return (0);
  }

  /* Record the current position in the array for later */
  pathpos = numpath;

  /* Find out how many total there are */
  for (npath = path; *npath; npath++, numpath++);

  /* Move the paths down, overwriting the deleted path */
  memmove (path, path + 1, (numpath - pathpos - 1) * sizeof (guchar *));

  /* Reallocate the memory */
  numpath--;
  module_dyn_paths = realloc (module_dyn_paths, numpath * sizeof (guchar *));

  /* Why reallocating to less memory can fail is beyond me, but we check */
  /*  anyway since it's the "right thing to do" */
  if (!module_dyn_paths)
    err_fatal ("module_dyn_path_del: relloc returned NULL!");

  /* Free the string for the path */
  free (path);

  UNLOCK ("module_dyn_path_del", mod_dyn_lock);

  return (0);
}

/*
 * module_dyn_find
 *
 *  Find's a module in the search path
 */
static int
module_dyn_find (modulerec * module, guchar * modname)
{
  guchar **path, pathbuf[PATH_MAX + 1];

  LOCK ("module_dyn_find", mod_dyn_lock);

  /* Try it find it */
  for (path = module_dyn_paths; path && *path; path++) {
    /* Try it raw first */
    g_snprintf (pathbuf, PATH_MAX, "%s/%s", *path, modname);
    if (module_dyn_isfile (pathbuf))
      goto mdf_found;

    /* Try it with .so */
    g_snprintf (pathbuf, PATH_MAX, "%s/%s.so", *path, modname);
    if (module_dyn_isfile (pathbuf))
      goto mdf_found;

    /* Try it with .la */
    g_snprintf (pathbuf, PATH_MAX, "%s/%s.la", *path, modname);
    if (module_dyn_isfile (pathbuf))
      goto mdf_found;

    /* Try it with /lib/ */
    g_snprintf (pathbuf, PATH_MAX, "%s/lib/%s", *path, modname);
    if (module_dyn_isfile (pathbuf))
      goto mdf_found;

    /* Try it with /lib/ and .so */
    g_snprintf (pathbuf, PATH_MAX, "%s/lib/%s.so", *path, modname);
    if (module_dyn_isfile (pathbuf))
      goto mdf_found;

    /* Try it with /lib/ and .la */
    g_snprintf (pathbuf, PATH_MAX, "%s/lib/%s.la", *path, modname);
    if (module_dyn_isfile (pathbuf))
      goto mdf_found;
  }

  /* Couldn't find it */
  module->pathname[0] = 0;

  UNLOCK ("module_dyn_find", mod_dyn_lock);

  return (-1);

mdf_found:
  /* Found it */
  strncpyzt (module->pathname, pathbuf, PATH_MAX);
  UNLOCK ("module_dyn_find", mod_dyn_lock);

  return (0);
}

/*
 * module_dyn_sym
 *
 *  Returns a pointer to the symbol
 */
void *
module_dyn_sym (modulerec * module, guchar * symname)
{
  return (module_i_sym (module->handle, symname));
}

/*
 * module_dyn_open
 *
 *  Open's a dynamically loaded module
 */
modulerec *
module_dyn_open (guchar * modname)
{
  modulerec *module;

  if ((module = module_alloc ()) == NULL) {
    err_warn ("module_dyn_open: couldn't allocate memory for module");
    return (NULL);
  }
#if 0
  module->refcount = 1;		/* Start off at one */
#else
  module->refcount = 0;		/* Start off at zero */
#endif

  module->type = DYNAMIC;

  /* Save the module name passed via the config file */
  strncpyzt (module->name, modname, MODULE_MAXNAMELENGTH);

  /* Find the module in the path */
  if (module_dyn_find (module, modname) < 0) {
    err_warn
      ("module_dyn_open: module %s not in config path, hoping module_i_open finds it",
       modname);
    strncpyzt (module->pathname, modname, MODULE_MAXNAMELENGTH);
  }

  /* Open up the module */
  if ((module->handle = module_i_open (module->pathname)) == NULL) {
    err_warn ("module_dyn_open: module %s @ %s could not be opened",
	      modname, module->pathname);
    module_free (module);
    return (NULL);
  }

  DPRINTF (
	   ("module_dyn_open: opened module %s @ %s\n", modname,
	    module->pathname));

  /* Try populating the symbols */
  if (module_populate (module) < 0) {
    err_warn ("module_dyn_open: module %s @ %s was missing symbols",
	      modname, module->pathname);
    module_free (module);
    return (NULL);
  }
#if 0
  /* Call the init function */
  if (module->init)
    module->init ();
#endif

  return (module);
}

/*
 * module_dyn_close
 *
 *  Called when the reference count gets to 0
 */
int
module_dyn_close (modulerec * module)
{
  void *handle;

  /* Save the handle for closing it */
  handle = module->handle;

  /* Free the module structure */
  module_free (module);

  return (module_i_close (handle));
}
