/* module_syms.h - Macros for easy programming of modules.

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

/* Internally, we use A LOT of indirection to get things done like we
 want them to. Read the comments near the macros to find out what they
 do since it's not easy gathering it from the code */

#ifndef MODULE_SYMS_H
#define MODULE_SYMS_H

#include "common.h"
#include "module.h"

/* This should only be included if we're compiling a module. */
#if !defined(STATIC_MODULE) && !defined(DYNAMIC_MODULE)
#error This header file should only be used for compiling modules
#error STATIC_MODULE and DYNAMIC_MODULE not defined
#endif

/* Check if they included MODULE_NAME. */
#ifndef MODULE_NAME
#error Must define MODULE_NAME before including header file
#endif

/* Check if they defined a module version. */
#if defined(MODULE_VER_MAJOR) && defined(MODULE_VER_MINOR) &&	\
	defined(MODULE_VER_MICRO)
#define _MOD_VERSIONS
#endif

/* External macros. */

/* The MODULE_FUNC(r, n, p) macro defines a function that will be
exported via the normal interfaces.  The r field is the return
portion, the n field is the name, and the p field is the parameters.

Example:
 
 MODULE_FUNC(int, dns_zone_add, (dnszonerec *zone))

will generate:
 
 int mod_name_dns_zone_add(dnszonerec *zone)  */

#define MODULE_FUNC(r, n, p)	r _MOD_UNIQUE(n) p

/* MODULE_SYMS will define the 4 mandatory symbols and add any extra
symbols to the list. Example:
 
  MODULE_SYMS(
   MODULE_SYMS_INIT
   MODULE_SYMS_DESTROY
  )
*/

/* Internal header macros. */

/* The _MOD_UNIQUE(x) macro will generate a unique name, depending on
the style of the module. As a static module, _MOD_UNIQUE(x) will take
the variable name and prepend MODULE_NAME##_. As a dynamic module,
_MOD_UNIQUE(x) will take the variable name and prepend it with
module_.

Example:
 
  _MOD_UNIQUE(version)

will generate:

  mod_name_version    (for a static module-driver)
  module_version      (for a shared module-driver)  */

/* The _MOD_NAME_S macro will stringify the MODULE_NAME */
#define _MOD_NAME_S__(x)	#x
#define _MOD_NAME_S_(x)		_MOD_NAME_S__(x)
#define _MOD_NAME_S		_MOD_NAME_S_(MODULE_NAME)

/* Static modules */
#ifdef STATIC_MODULE

#define _MOD_UNIQUE__(x, y)	x ## _ ## y
#define _MOD_UNIQUE_(x, y)	_MOD_UNIQUE__(x, y)
#define _MOD_UNIQUE(x)		_MOD_UNIQUE_(MODULE_NAME, x)

/* Did they set versions? Make sure it gets added to the symbols. */
#if defined(_MOD_VERSIONS)
# define _MOD_VER_SYMS { "module_version", &_MOD_UNIQUE(version) },
#else /* !_MOD_VERSIONS */
# define _MOD_VER_SYMS
#endif /* !_MOD_VERSIONS */

#ifdef WITH_PTHREAD
int _MOD_UNIQUE (compiled_pthreads) = 1;
# define _MOD_PTHREAD \
  { "module_compiled_pthreads", &_MOD_UNIQUE(compiled_pthreads) },
#else /* !WITH_PTHREAD */
# define _MOD_PTHREAD
#endif /* !WITH_PTHREAD */

#define MODULE_SYMS(x)					\
  modulesymbolrec _MOD_UNIQUE(syms)[] = {		\
    { "module_name",			_MOD_NAME_S },	\
    _MOD_PTHREAD					\
    { "module_compiled_version",	&_MOD_UNIQUE(compiled_version) }, \
    _MOD_VER_SYMS					\
    x							\
    { NULL, NULL }					\
  };



#define MODULE_SYMS_INIT		\
  { "module_init",			_MOD_UNIQUE(init) },

#define MODULE_SYMS_DESTROY		\
  { "module_destroy",			_MOD_UNIQUE(destroy) },

/* Dynamic modules */
#elif defined(DYNAMIC_MODULE)

#define _MOD_UNIQUE(x)		module_##x

#ifdef WITH_PTHREAD
int _MOD_UNIQUE (compiled_pthreads) = 1;
#endif /* WITH_PTHREAD */

/* We don't need to create a table of the symbols since the low level
   library code for the OS will take care of that for us.  So most of
   these macros are NOOPs. */

#define MODULE_SYMS(x) x
#define MODULE_SYMS_INIT
#define MODULE_SYMS_DESTROY

/* Module name.  Be careful, we want the string to be in the correct
   data segment.  Hence the compile time allocated string instead of a
   pointer to a string. */
guchar _MOD_UNIQUE (name)[] = _MOD_NAME_S;

#endif /* STATIC_MODULE | DYNAMIC_MODULE */

/* Predefined symbols */

/* Did they set versions? */

#ifndef MODULE_VERSION
# define MODULE_VERSION CONCAT_VER(MODULE_VER_MAJOR, MODULE_VER_MINOR, MODULE_VER_MICRO)
# define CONCAT_VER(a,b,c) CONCAT_VER_(a,b,c)
# define CONCAT_VER_(a,b,c) #a ## "." ## #b ## "." ## #c
#endif /* MODULE_VERSION */

#if defined(_MOD_VERSIONS)
  char *_MOD_UNIQUE (version) = MODULE_VERSION;
#endif
/* Version of dents we compiled against */
  char *_MOD_UNIQUE (compiled_version) = DENTS_VERSION;
#endif /* !MODULE_SYMS_H */
