/* redblack.c - implements a red/black tree structure.

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/* Implement the red/black tree structure. It is designed to emulate
   ** the standard tsearch() stuff. i.e. the calling conventions are
   ** exactly the same
 */
#include <stddef.h>
#include <stdlib.h>
#include "redblack.h"

#ifdef assert
#undef assert
#endif /* assert */
#define assert(expr)

enum nodecolour
{ BLACK, RED };

struct rbnode
{
  struct rbnode *left;		/* Left down */
  struct rbnode *right;		/* Right down */
  struct rbnode *up;		/* Up */
  enum nodecolour colour;	/* Node colour */
  void *key;			/* Pointer to user's key (and data) */
};

/* Dummy (sentinel) node, so that we can make X->left->up = X
   ** We then use this instead of NULL to mean the top or bottom
   ** end of the rb tree. It is a black node.
 */
struct rbnode rb_null = { &rb_null, &rb_null, &rb_null, BLACK, NULL };

#define RBNULL (&rb_null)

static struct rbnode *rb_traverse __P ((int, void *, struct rbtree *));
static struct rbnode *rb_lookup __P ((int, void *, struct rbtree *));
static void rb_left_rotate __P ((struct rbnode **, struct rbnode *));
static void rb_right_rotate __P ((struct rbnode **, struct rbnode *));
static void rb_delete __P ((struct rbnode **, struct rbnode *));
static struct rbnode *rb_successor __P ((struct rbnode *));
static void rb_delete_fix __P ((struct rbnode **, struct rbnode *));
static struct rbnode *rb_alloc __P (());
static void rb_free __P ((struct rbnode *));
static void rb_walk
__P ((struct rbnode *, void (*)(void *, VISIT, int, void *), void *, int));
static void rb_walk_pre
__P ((struct rbnode *, void (*)(void *, VISIT, int, void *), void *, int));
static void rb_walk_post
__P ((struct rbnode *, void (*)(void *, VISIT, int, void *), void *, int));
static void rb_walk_end
__P ((struct rbnode *, void (*)(void *, VISIT, int, void *), void *, int));
static RBLIST *rb_openlist __P ((struct rbnode *));
static void *rb_readlist __P ((RBLIST *));
static void rb_closelist __P ((RBLIST *));
static struct rbnode *rb_successor __P ((struct rbnode *));
static struct rbnode *rb_preccessor __P ((struct rbnode *));

/*
   ** OK here we go, the balanced tree stuff. The algorithm is the
   ** fairly standard red/black taken from "Introduction to Algorithms"
   ** by Cormen, Leiserson & Rivest. Maybe one of these days I will
   ** fully understand all this stuff.
   **
   ** Basically a red/black balanced tree has the following properties:-
   ** 1) Every node is either red or black (colour is RED or BLACK)
   ** 2) A leaf (RBNULL pointer) is considered black
   ** 3) If a node is red then its children are black
   ** 4) Every path from a node to a leaf contains the same no
   **    of black nodes
   **
   ** 3) & 4) above guarantee that the longest path (alternating
   ** red and black nodes) is only twice as long as the shortest
   ** path (all black nodes). Thus the tree remains fairly balanced.
 */

/*
   ** Initialise RB tree - needs to be called for every seperate tree
   ** we have.
   ** Args: comparison routine
   **     : Any config data for cmp routine (usually NULL)
   **
   ** Returns: pointer to rbtree structure which is the handle for the
   ** entire tree
   **
   ** The cmp routine will be called with:-
   ** args: first key
   **       second key
   **       data (as registered in rbinit)
 */

struct rbtree *
rbinit (int (*cmp) (void *, void *, void *), void *data)
{
  struct rbtree *retval;

  if ((retval = (struct rbtree *) malloc (sizeof (struct rbtree))) == NULL)
    return (NULL);

  retval->rb_cmp = cmp;
  retval->rb_sdata = data;
  retval->rb_root = RBNULL;

  return (retval);
}

/* Look for and if not found add a key into the RB tree
   ** Args: pointer to key
   **       rbtree handle
   ** 
   ** Returns: pointer to key found (NULL on error)
   **
 */
void *
rbsearch (void *key, struct rbtree *rbinfo)
{
  struct rbnode *x;

  if (rbinfo == NULL)
    return (NULL);

#if 0
  /* If we have a NULL root (empty tree) then convert it to RBNULL */
  if (rbinfo->rb_root == NULL)
    rbinfo->rb_root = (void *) RBNULL;
#endif

  x = rb_traverse (1, key, rbinfo);

#if 0
  /* If root is RBNULL, then convert it back to NULL */
  if (rbinfo->rb_root == (void *) RBNULL)
    rbinfo->rb_root = NULL;
#endif

  return ((x == RBNULL) ? NULL : x->key);
}

/* Look for key into the RB tree
   ** Args: pointer to key
   **       rbtree handle
   ** 
   ** Returns: pointer to key found or NULL if not found
 */
void *
rbfind (void *key, struct rbtree *rbinfo)
{
  struct rbnode *x;

  if (rbinfo == NULL)
    return (NULL);

  /* If we have a NULL root (empty tree) then just return */
  if (rbinfo->rb_root == RBNULL) {
    return (NULL);
  }

  x = rb_traverse (0, key, rbinfo);

  return ((x == RBNULL) ? NULL : x->key);
}

/* Search for and delete key in the RB tree
   ** Args: pointer to key
   **       rbtree handle
   ** 
   ** Returns: pointer to key deleted or NULL if not found
 */
void *
rbdelete (void *key, struct rbtree *rbinfo)
{
  struct rbnode *x;
  void *y;

  if (rbinfo == NULL)
    return (NULL);

#if 0
  /* If we have a NULL root (empty tree) then convert it to RBNULL */
  if (rbinfo->rb_root == NULL)
    rbinfo->rb_root = (void *) RBNULL;
#endif

  x = rb_traverse (0, key, rbinfo);

  if (x == RBNULL) {
#if 0
    /* If root is RBNULL, then convert it back to NULL */
    if (rbinfo->rb_root == (void *) RBNULL)
      rbinfo->rb_root = NULL;
#endif

    return (NULL);
  } else {
    y = x->key;
    rb_delete (&rbinfo->rb_root, x);

#if 0
    /* If root is RBNULL, then convert it back to NULL */
    if (rbinfo->rb_root == (void *) RBNULL)
      rbinfo->rb_root = NULL;
#endif

    return (y);
  }
}

/* Walk down a RB tree
   **
   ** rbwalk - behaves like the 'twalk()' function from the Standard C library
   ** rbwalkpre - just calls the callback for pre-order traversal - also
   **             visits all leaf nodes
   ** rbwalkpost - just calls the callback for post-order traversal - also
   **             visits all leaf nodes
   ** rbwalkend - just calls the callback for end-order traversal - also
   **             visits all leaf nodes
   ** 
   ** Args: rbtree handle
   **       action function to call at each node
   **       arg to be pased untouched to action function
   ** 
   ** Return: None
   **
   ** Action function is called with the following
   ** args: node key
   **       visit (preorder, postorder, endorder leaf) see twalk
   **       How many nodes down the tree we are
   **       arg as passed to rbwalk()
   **
 */
void
rbwalk (struct rbtree *root, void (*action) (void *, VISIT, int, void *),
	void *arg)
{
  if (root == NULL)
    return;
  rb_walk (root->rb_root, action, arg, 0);
}

void
rbwalkpost (struct rbtree *root, void (*action) (void *, VISIT, int, void *),
	    void *arg)
{
  if (root == NULL)
    return;
  rb_walk_post (root->rb_root, action, arg, 0);
}

void
rbwalkpre (struct rbtree *root, void (*action) (void *, VISIT, int, void *),
	   void *arg)
{
  if (root == NULL)
    return;
  rb_walk_pre (root->rb_root, action, arg, 0);
}

void
rbwalkend (struct rbtree *root, void (*action) (void *, VISIT, int, void *),
	   void *arg)
{
  if (root == NULL)
    return;
  rb_walk_end (root->rb_root, action, arg, 0);
}

/* Function to extract a list of nodes from a tree
   ** Do not use! This hasn't been converted to new calling conventions
   ** Args: rbtree handle
   **
   ** Returns: pointer to RBLIST structure.
 */
RBLIST *
rbopenlist (void *root)
{
  if (root == NULL)
    return (NULL);

  return (rb_openlist ((struct rbnode *) root));
}

/* Read next node from tree
   ** Do not use! This hasn't been converted to new calling conventions
   ** Args: RBLIST handle
   **
   ** Returns next key found, or NULL if reached the end
 */
void *
rbreadlist (RBLIST * rblistp)
{
  if (rblistp == NULL)
    return (NULL);

  return (rb_readlist (rblistp));
}

/* Close the RBLIST handle and free associated memory
   ** Do not use! This hasn't been converted to new calling conventions
   ** Args: Handle to be closed
   ** Return: none
 */
void
rbcloselist (RBLIST * rblistp)
{
  if (rblistp == NULL)
    return;

  rb_closelist (rblistp);
}

/* Lookup key in tree according to mode
   ** Args: mode which is one of:-
   **            RB_LUEQUAL  - Only exact match
   **            RB_LUGTEQ   - Exact match or greater
   **            RB_LULTEQ   - Exact match or less
   **            RB_LULESS   - Less than key (not equal to)
   **            RB_LUGREAT  - Greater than key (not equal to)
   **            RB_LUNEXT   - Next key after current
   **            RB_LUPREV   - Prev key before current
   **            RB_LUFIRST  - First key in tree
   **            RB_LULAST   - Last key in tree
   **       pointer to key
   **       rbtree handle
   **
   ** Returns key found or NULL if there is no match
 */
void *
rblookup (int mode, void *key, struct rbtree *rbinfo)
{
  struct rbnode *x;

  /* If we have a NULL root (empty tree) then just return NULL */
  if (rbinfo->rb_root == NULL) {
    return (NULL);
  }

  x = rb_lookup (mode, key, rbinfo);

  return ((x == RBNULL) ? NULL : x->key);
}

void *
rbmin (struct rbtree *rbinfo)
{
  struct rbnode *n;

  n = rbinfo->rb_root;

  if (n != RBNULL) {
    while (n->left != RBNULL) {
      n = n->left;
    }
  }

  if (n != RBNULL) {
    return n->key;
  } else {
    return NULL;
  }
}

void *
rbmax (struct rbtree *rbinfo)
{
  struct rbnode *n;

  n = rbinfo->rb_root;

  if (n != RBNULL) {
    while (n->right != RBNULL) {
      n = n->right;
    }
  }

  if (n != RBNULL) {
    return n->key;
  } else {
    return NULL;
  }
}

void *
rbdestroy (struct rbtree *rbinfo)
{
  /* FIXME: shouldn't this alos delete all nodes in the tree !?!?  -EZA */
  free (rbinfo);

  /* FIXME: what is the intended return value from this function?   */
  return NULL;
}

void *
rbinsert (void *key, struct rbtree *rbinfo)
{
  struct rbnode *x;

  if (rbinfo == NULL)
    return (NULL);

#if 0
  /* If we have a NULL root (empty tree) then convert it to RBNULL */
  if (rbinfo->rb_root == NULL)
    rbinfo->rb_root = (void *) RBNULL;
#endif

  x = rb_traverse (0, key, rbinfo);

  if (x != RBNULL)
    return NULL;

  x = rb_traverse (1, key, rbinfo);

#if 0
  /* If root is RBNULL, then convert it back to NULL */
  if (rbinfo->rb_root == (void *) RBNULL)
    rbinfo->rb_root = NULL;
#endif

  return ((x == RBNULL) ? NULL : x->key);
}

/* --------------------------------------------------------------------- */
/* Everything below here is static and is not accesible outside */

/* Search for and if not found and insert is true, will add a new
   ** node in. Returns a pointer to the new node, or the node found
 */
static struct rbnode *
rb_traverse (int insert, void *key, struct rbtree *rbinfo)
{
  struct rbnode *x, *y, *z;
  int cmp;
  int found = 0;
  int cmpmods ();

  y = RBNULL;			/* points to the parent of x */
  x = rbinfo->rb_root;

  /* walk x down the tree */
  while (x != RBNULL && found == 0) {
    y = x;
    /* printf("key=%s, x->key=%s\n", key, x->key); */
    cmp = (*rbinfo->rb_cmp) (key, x->key, rbinfo->rb_sdata);
    if (cmp < 0)
      x = x->left;
    else if (cmp > 0)
      x = x->right;
    else
      found = 1;
  }

  if (found || !insert)
    return (x);

  if ((z = rb_alloc ()) == NULL) {
    /* Whoops, no memory */
    return (RBNULL);
  }

  z->key = key;
  z->up = y;
  if (y == RBNULL) {
    rbinfo->rb_root = z;
  } else {
    cmp = (*rbinfo->rb_cmp) (z->key, y->key, rbinfo->rb_sdata);
    if (cmp < 0)
      y->left = z;
    else
      y->right = z;
  }

  z->left = RBNULL;
  z->right = RBNULL;

  /* colour this new node red */
  z->colour = RED;

  /* Having added a red node, we must now walk back up the tree balancing
     ** it, by a series of rotations and changing of colours
   */
  x = z;

  /* While we are not at the top and our parent node is red
     ** N.B. Since the root node is garanteed black, then we
     ** are also going to stop if we are the child of the root
   */

  while (x != rbinfo->rb_root && (x->up->colour == RED)) {
    /* if our parent is on the left side of our grandparent */
    if (x->up == x->up->up->left) {
      /* get the right side of our grandparent (uncle?) */
      y = x->up->up->right;
      if (y->colour == RED) {
	/* make our parent black */
	x->up->colour = BLACK;
	/* make our uncle black */
	y->colour = BLACK;
	/* make our grandparent red */
	x->up->up->colour = RED;

	/* now consider our grandparent */
	x = x->up->up;
      } else {
	/* if we are on the right side of our parent */
	if (x == x->up->right) {
	  /* Move up to our parent */
	  x = x->up;
	  rb_left_rotate (&rbinfo->rb_root, x);
	}

	/* make our parent black */
	x->up->colour = BLACK;
	/* make our grandparent red */
	x->up->up->colour = RED;
	/* right rotate our grandparent */
	rb_right_rotate (&rbinfo->rb_root, x->up->up);
      }
    } else {
      /* everything here is the same as above, but
         ** exchanging left for right
       */

      y = x->up->up->left;
      if (y->colour == RED) {
	x->up->colour = BLACK;
	y->colour = BLACK;
	x->up->up->colour = RED;

	x = x->up->up;
      } else {
	if (x == x->up->left) {
	  x = x->up;
	  rb_right_rotate (&rbinfo->rb_root, x);
	}

	x->up->colour = BLACK;
	x->up->up->colour = RED;
	rb_left_rotate (&rbinfo->rb_root, x->up->up);
      }
    }
  }

  /* Set the root node black */
  (rbinfo->rb_root)->colour = BLACK;

  return (z);
}

/* Search for a key according to mode (see redblack.h)
 */
static struct rbnode *
rb_lookup (int mode, void *key, struct rbtree *rbinfo)
{
  struct rbnode *x, *y;
  int cmp = 0;
  int found = 0;

  y = RBNULL;			/* points to the parent of x */
  x = rbinfo->rb_root;

  if (mode == RB_LUFIRST) {
    /* Keep going left until we hit a NULL */
    while (x != RBNULL) {
      y = x;
      x = x->left;
    }

    return (y);
  } else if (mode == RB_LULAST) {
    /* Keep going right until we hit a NULL */
    while (x != RBNULL) {
      y = x;
      x = x->right;
    }

    return (y);
  }

  /* walk x down the tree */
  while (x != RBNULL && found == 0) {
    y = x;
    /* printf("key=%s, x->key=%s\n", key, x->key); */
    cmp = (*rbinfo->rb_cmp) (key, x->key, rbinfo->rb_sdata);
    if (cmp < 0)
      x = x->left;
    else if (cmp > 0)
      x = x->right;
    else
      found = 1;
  }

  if (found && (mode == RB_LUEQUAL || mode == RB_LUGTEQ || mode == RB_LULTEQ))
    return (x);

  if (!found
      && (mode == RB_LUEQUAL || mode == RB_LUNEXT
	  || mode == RB_LUPREV)) return (RBNULL);

  if (mode == RB_LUGTEQ || (!found && mode == RB_LUGREAT)) {
    if (cmp > 0)
      return (rb_successor (y));
    else
      return (y);
  }

  if (mode == RB_LULTEQ || (!found && mode == RB_LULESS)) {
    if (cmp < 0)
      return (rb_preccessor (y));
    else
      return (y);
  }

  if (mode == RB_LUNEXT || (found && mode == RB_LUGREAT))
    return (rb_successor (x));

  if (mode == RB_LUPREV || (found && mode == RB_LULESS))
    return (rb_preccessor (x));

  /* Shouldn't get here */
  return (RBNULL);
}

/*
   ** Rotate our tree thus:-
   **
   **             X        rb_left_rotate(X)--->            Y
   **           /   \                                     /   \
   **          A     Y     <---rb_right_rotate(Y)        X     C
   **              /   \                               /   \
   **             B     C                             A     B
   **
   ** N.B. This does not change the ordering.
   **
   ** We assume that neither X or Y is NULL
 */

static void
rb_left_rotate (struct rbnode **rootp, struct rbnode *x)
{
  struct rbnode *y;

  assert (x != RBNULL);
  assert (x->right != RBNULL);

  y = x->right;			/* set Y */

  /* Turn Y's left subtree into X's right subtree (move B) */
  x->right = y->left;

  /* If B is not null, set it's parent to be X */
  if (y->left != RBNULL)
    y->left->up = x;

  /* Set Y's parent to be what X's parent was */
  y->up = x->up;

  /* if X was the root */
  if (x->up == RBNULL) {
    *rootp = y;
  } else {
    /* Set X's parent's left or right pointer to be Y */
    if (x == x->up->left) {
      x->up->left = y;
    } else {
      x->up->right = y;
    }
  }

  /* Put X on Y's left */
  y->left = x;

  /* Set X's parent to be Y */
  x->up = y;
}

static void
rb_right_rotate (struct rbnode **rootp, struct rbnode *y)
{
  struct rbnode *x;

  assert (y != RBNULL);
  assert (y->left != RBNULL);

  x = y->left;			/* set X */

  /* Turn X's right subtree into Y's left subtree (move B) */
  y->left = x->right;

  /* If B is not null, set it's parent to be Y */
  if (x->right != RBNULL)
    x->right->up = y;

  /* Set X's parent to be what Y's parent was */
  x->up = y->up;

  /* if Y was the root */
  if (y->up == RBNULL) {
    *rootp = x;
  } else {
    /* Set Y's parent's left or right pointer to be X */
    if (y == y->up->left) {
      y->up->left = x;
    } else {
      y->up->right = x;
    }
  }

  /* Put Y on X's right */
  x->right = y;

  /* Set Y's parent to be X */
  y->up = x;
}

/* Return a pointer to the smallest key greater than x
 */
static struct rbnode *
rb_successor (struct rbnode *x)
{
  struct rbnode *y;

  if (x->right != RBNULL) {
    /* If right is not NULL then go right one and
       ** then keep going left until we find a node with
       ** no left pointer.
     */
    for (y = x->right; y->left != RBNULL; y = y->left);
  } else {
    /* Go up the tree until we get to a node that is on the
       ** left of its parent (or the root) and then return the
       ** parent.
     */
    y = x->up;
    while (y != RBNULL && x == y->right) {
      x = y;
      y = y->up;
    }
  }
  return (y);
}

/* Return a pointer to the largest key smaller than x
 */
static struct rbnode *
rb_preccessor (struct rbnode *x)
{
  struct rbnode *y;

  if (x->left != RBNULL) {
    /* If left is not NULL then go left one and
       ** then keep going right until we find a node with
       ** no right pointer.
     */
    for (y = x->left; y->right != RBNULL; y = y->right);
  } else {
    /* Go up the tree until we get to a node that is on the
       ** right of its parent (or the root) and then return the
       ** parent.
     */
    y = x->up;
    while (y != RBNULL && x == y->left) {
      x = y;
      y = y->up;
    }
  }
  return (y);
}

/* Delete the node z, and free up the space
 */
static void
rb_delete (struct rbnode **rootp, struct rbnode *z)
{
  struct rbnode *x, *y;

  if (z->left == RBNULL || z->right == RBNULL)
    y = z;
  else
    y = rb_successor (z);

  if (y->left != RBNULL)
    x = y->left;
  else
    x = y->right;

  x->up = y->up;

  if (y->up == RBNULL) {
    *rootp = x;
  } else {
    if (y == y->up->left)
      y->up->left = x;
    else
      y->up->right = x;
  }

  if (y != z) {
    z->key = y->key;
  }

  if (y->colour == BLACK)
    rb_delete_fix (rootp, x);

  rb_free (y);
}

/* Restore the reb-black properties after a delete */
static void
rb_delete_fix (struct rbnode **rootp, struct rbnode *x)
{
  struct rbnode *w;

  while (x != *rootp && x->colour == BLACK) {
    if (x == x->up->left) {
      w = x->up->right;
      if (w->colour == RED) {
	w->colour = BLACK;
	x->up->colour = RED;
	rb_left_rotate (rootp, x->up);
	w = x->up->right;
      }

      if (w->left->colour == BLACK && w->right->colour == BLACK) {
	w->colour = RED;
	x = x->up;
      } else {
	if (w->right->colour == BLACK) {
	  w->left->colour = BLACK;
	  w->colour = RED;
	  rb_right_rotate (rootp, w);
	  w = x->up->right;
	}

	w->colour = x->up->colour;
	x->up->colour = BLACK;
	w->right->colour = BLACK;
	rb_left_rotate (rootp, x->up);
	x = *rootp;
      }
    } else {
      w = x->up->left;
      if (w->colour == RED) {
	w->colour = BLACK;
	x->up->colour = RED;
	rb_right_rotate (rootp, x->up);
	w = x->up->left;
      }

      if (w->right->colour == BLACK && w->left->colour == BLACK) {
	w->colour = RED;
	x = x->up;
      } else {
	if (w->left->colour == BLACK) {
	  w->right->colour = BLACK;
	  w->colour = RED;
	  rb_left_rotate (rootp, w);
	  w = x->up->left;
	}

	w->colour = x->up->colour;
	x->up->colour = BLACK;
	w->left->colour = BLACK;
	rb_right_rotate (rootp, x->up);
	x = *rootp;
      }
    }
  }

  x->colour = BLACK;
}

static void
rb_walk (struct rbnode *x, void (*action) (void *, VISIT, int, void *),
	 void *arg, int level)
{
  if (x == RBNULL)
    return;

  if (x->left == RBNULL && x->right == RBNULL) {
    /* leaf */
    (*action) (x->key, leaf, level, arg);
  } else {
    (*action) (x->key, preorder, level, arg);

    rb_walk (x->left, action, arg, level + 1);

    (*action) (x->key, postorder, level, arg);

    rb_walk (x->right, action, arg, level + 1);

    (*action) (x->key, endorder, level, arg);
  }
}

static void
rb_walk_pre (struct rbnode *x, void (*action) (void *, VISIT, int, void *),
	     void *arg, int level)
{
  if (x == RBNULL)
    return;

  (*action) (x->key, preorder, level, arg);

  if (x->left)
    rb_walk_pre (x->left, action, arg, level + 1);

  if (x->right)
    rb_walk_pre (x->right, action, arg, level + 1);
}

static void
rb_walk_post (struct rbnode *x, void (*action) (void *, VISIT, int, void *),
	      void *arg, int level)
{
  if (x == RBNULL)
    return;

  if (x->left)
    rb_walk_post (x->left, action, arg, level + 1);

  (*action) (x->key, postorder, level, arg);

  if (x->right)
    rb_walk_post (x->right, action, arg, level + 1);

}

static void
rb_walk_end (struct rbnode *x, void (*action) (void *, VISIT, int, void *),
	     void *arg, int level)
{
  if (x == RBNULL)
    return;

  if (x->left)
    rb_walk_end (x->left, action, arg, level + 1);

  if (x->right)
    rb_walk_end (x->right, action, arg, level + 1);

  (*action) (x->key, endorder, level, arg);

}

static RBLIST *
rb_openlist (struct rbnode *rootp)
{
  RBLIST *rblistp;

  rblistp = (RBLIST *) malloc (sizeof (RBLIST));

  rblistp->rootp = rootp;
  rblistp->nextp = rootp;

  if (rootp != RBNULL) {
    while (rblistp->nextp->left != RBNULL) {
      rblistp->nextp = rblistp->nextp->left;
    }
  }

  return (rblistp);
}

static void *
rb_readlist (RBLIST * rblistp)
{
  void *key = NULL;

  if (rblistp->nextp != RBNULL) {
    key = rblistp->nextp->key;
    rblistp->nextp = rb_successor (rblistp->nextp);
  }

  return (key);
}

static void
rb_closelist (RBLIST * rblistp)
{
  free (rblistp);
}

/*
 * Allocate space for our nodes, allowing us to get space from
 * malloc() in larger chunks.
 */
static struct rbnode *rbfreep = NULL;

#define RBNODEALLOC_CHUNK_SIZE 1000
static struct rbnode *
rb_alloc ()
{
  struct rbnode *x;
  int i;

  if (rbfreep == NULL) {
    /* must grab some more space */
    rbfreep =
      (struct rbnode *) malloc (sizeof (struct rbnode) *
			      RBNODEALLOC_CHUNK_SIZE);

    if (rbfreep == (struct rbnode *) -1) {
      return (NULL);
    }

    /* tie them together in a linked list (use the up pointer) */
    for (i = 0, x = rbfreep; i < RBNODEALLOC_CHUNK_SIZE - 1; i++, x++) {
      x->up = (x + 1);
    }
    x->up = NULL;
  }

  x = rbfreep;
  rbfreep = rbfreep->up;
  return (x);
}

/* free (dealloc) an rbnode structure - add it onto the front of the list
   ** N.B. rbnode need not have been allocated through rb_alloc()
 */
static void
rb_free (struct rbnode *x)
{
  x->up = rbfreep;
  rbfreep = x;
}

#if 0
int
rb_check (struct rbnode *rootp)
{
  if (rootp == NULL || rootp == RBNULL)
    return (0);

  if (rootp->up != RBNULL) {
    SddbReport (0, "Root up pointer not RBNULL");
    dumptree (rootp, 0);
    return (1);
  }

  if (rb_check1 (rootp)) {
    dumptree (rootp, 0);
    return (1);
  }

  if (count_black (rootp) == -1) {
    dumptree (rootp, 0);
    return (-1);
  }

  return (0);
}

int
rb_check1 (struct rbnode *x)
{
  if (x->left == NULL || x->right == NULL) {
    SddbReport (0, "Left or right is NULL");
    return (1);
  }

  if (x->colour == RED) {
    if (x->left->colour != BLACK && x->right->colour != BLACK) {
      SddbReport (0, "Children of red node not both black, x=%ld", x);
      return (1);
    }
  }

  if (x->left != RBNULL) {
    if (x->left->up != x) {
      SddbReport (0, "x->left->up != x, x=%ld", x);
      return (1);
    }

    if (rb_check1 (x->left))
      return (1);
  }

  if (x->right != RBNULL) {
    if (x->right->up != x) {
      SddbReport (0, "x->right->up != x, x=%ld", x);
      return (1);
    }

    if (rb_check1 (x->right))
      return (1);
  }
  return (0);
}

count_black (struct rbnode * x)
{
  int nleft, nright;

  if (x == RBNULL)
    return (1);

  nleft = count_black (x->left);
  nright = count_black (x->right);

  if (nleft == -1 || nright == -1)
    return (-1);

  if (nleft != nright) {
    SddbReport (0, "Black count not equal on left & right, x=%ld", x);
    return (-1);
  }

  if (x->colour == BLACK) {
    nleft++;
  }

  return (nleft);
}

dumptree (struct rbnode * x, int n)
{
  char *prkey ();

  if (x != NULL && x != RBNULL) {
    n++;
    SddbReport (15, "Tree: %*s %ld: left=%ld, right=%ld, colour=%s, key=%s",
		n,
		"",
		x,
		x->left,
		x->right,
		(x->colour == BLACK) ? "BLACK" : "RED", prkey (x->key));

    dumptree (x->left, n);
    dumptree (x->right, n);
  }
}
#endif

/*
 * $Log: redblack.c,v $
 * Revision 1.6  1999/08/08 16:23:04  eric
 * removed reference to sbrk()
 *
 * Revision 1.5  1999/07/30 20:58:51  illume
 * Several changes made while trying to compile dents on an old solaris
 * host.
 *
 * Revision 1.4  1999/07/25 10:46:19  eric
 *
 * commented out debugging from event.c
 * Added a few new function calls to the redblack tree (just for walking the tree)
 * Also got rid of some compiler warnings.
 *
 * Revision 1.3  1999/07/21 02:54:14  grumple
 * Okay, I give.  These work and pass the tests.  Screw it.
 *
 * Revision 2.5  1999/02/10 03:28:43  damian
 * Maror re-do of redblack.c so that it uses the rbtree structure. Now we
 * can pass arbitrary data to the compare routine.
 *
 * Revision 2.4  1998/11/29 11:18:51  damian
 * Re-orged the header files and renamed them. Fixed that stupid bug
 * in sddbc as well.
 *
 * Revision 2.3  1998/11/11 02:55:14  damian
 * More stuff to fix the makedepend (remove assert.h)
 *
 * Revision 2.2  1997/05/27 19:24:52  damian
 * Changed the name of the report() funtion to SddbReport() to avoid the name polution
 *
 * Revision 2.1  1997/01/22 03:24:50  damian
 * Jump to new new revision
 *
 * Revision 1.1  1996/06/21 19:04:09  root
 * Initial revision
 *
 */
