/* signal_handlers.c - supports registering a function for a signal.
 * 
 * Dents - a threaded modular dns-server.
 * Copyright (C) 1999  Dents Team  (http://www.dents.org)
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version. This program is distributed in the
 * hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

#include <sys/types.h>		/* waitpid */
#include <sys/wait.h>		/* waitpid */
#include <signal.h>		/* SIGCHLD, sigaction, etc. */
#include <stdio.h>		/* NULL, which should be a bultin */
#include <stdlib.h>		/* free() */

#include "signal_handlers.h"

/* Data types */
typedef struct registered_child {
    pid_t   pid;
    void    (*handler) (pid_t, int);
    struct registered_child *prev;
    struct registered_child *next;
} registered_child;

/* Globals */
static registered_child *child_list = NULL;
G_LOCK_DEFINE_STATIC(chldlock);

/* Prototypes */
static RETSIGTYPE catch_chld(int);
static registered_child *find_child_by_pid(pid_t pid);

/* Implementations */
RETSIGTYPE
catch_chld(int this_should_be_chld)
{
    int     status;
    pid_t   cpid;
    registered_child *rc;
    if (this_should_be_chld != SIGCHLD)
	return;
    fprintf(stderr, "catch_chld() called on %i.\n", this_should_be_chld);

    G_LOCK(chldlock);
    while (1) {
	cpid = waitpid(-1, &status, WNOHANG);
	if (cpid < 1) {
	    break;
	}
	else {
	    fprintf(stderr, "catch_chld() waitpid() got %i.\n", cpid);
	    if ((rc = find_child_by_pid(cpid))) {
		fprintf(stderr, "Calling handler for %i at %p.\n", cpid, rc->handler);
		rc->handler(cpid, status);
		if (rc->prev) {
		    rc->prev->next = rc->next;
		}
		if (rc->next) {
		    rc->next->prev = rc->prev;
		}
		if (child_list == rc) {
		    child_list = rc->next;
		}
		free(rc);
	    }
	    else {
		fprintf(stderr, "No handler found for %i.\n", cpid);
	    }
	}
    }
    G_UNLOCK(chldlock);
}

/* N.b., you should own the chldlock before calling this.  Returns
 * pointer to child structure on success, NULL on failure. */
static registered_child *
find_child_by_pid(pid_t pid)
{
    registered_child *rval;

    for (rval = child_list; rval; rval = rval->next) {
	if (rval->pid == pid) {
	    return (rval);
	}
    }
    return (NULL);
}

/* Returns 0 on success, other on failure. */
int
reg_chld_handler(void (*handler) (pid_t, int), pid_t pid)
{
    registered_child *rc;

    if (!(rc = (registered_child *) malloc(sizeof(registered_child)))) {
	return (-1);
    }

    rc->pid = pid;
    rc->handler = handler;

    G_LOCK(chldlock);

    if (child_list) {
	child_list->prev = rc;
    }
    rc->next = child_list;
    child_list = rc;

    G_UNLOCK(chldlock);
    return 0;
}

/* Returns the previous value of the signal handler, or SIG_ERR on
 * error. */
void    (*install_chld_handler(void)) (int) {
    return signal(SIGCHLD, catch_chld);
}
