/* sock.c - Handles socket routines

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

#include "common.h"
#include "sock.h"


#ifdef HAVE_UNISTD_H
#include <unistd.h>             /* fcntl */
#endif /* HAVE_UNISTD_H */
#ifdef HAVE_FCNTL_H
#include <fcntl.h>		/* fcntl */
#endif /* HAVE_FCNTL_H */

#include <arpa/inet.h>		/* inet_ntoa */

/* Global list of all sockets */
linkedlist sockets = LINK_INITIALIZER (socketrec, ll);

/* Pool of sockets */
blocklist socket_pool = BLOCK_INITIALIZER (socketrec, 5);

/*
 * sock_open
 *
 *  Open's up a generic socket (builtin) and bind()'s it to a addr/port
 */
int
sock_open (int proto, struct in_addr ip, guint16 port, int nonblock)
{
  int fd;
  struct sockaddr_in sa;

  /* Clear it out */
  memset ((void *) &sa, 0, sizeof (sa));

  /* Grab a socket */
  switch (proto) {
  case SOCKET_UDP:
    fd = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    break;
  case SOCKET_TCP:
    fd = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP);
    break;
  default:
    err_fatal ("sock_open: unknown proto %d", proto);
    return (-1);
  }

  /* Error */
  if (fd < 0)
    err_fatal ("sock_open: socket: %m");

  /* Set the socket to non blocking */
  if (nonblock)
    sock_setnonblock (fd);

  sa.sin_family = AF_INET;
  memcpy ((void *) &sa.sin_addr, (void *) &ip, sizeof (struct in_addr));
  sa.sin_port = htons (port);

  DPRINTF (("sock_open: bind(%s)'ing to %s:%d\n",
	    (proto == SOCKET_UDP) ? "IPPROTO_UDP" : "IPPROTO_TCP",
	    inet_ntoa (sa.sin_addr), port));

  /* bind() the socket to the interface */
  if (bind (fd, (struct sockaddr *) &sa, sizeof (struct sockaddr)) < 0)
    err_fatal ("sock_open: bind()'ing to %s:%d, %m", inet_ntoa (ip), port);

  return (fd);
}

/*
 * sock_setnonblock
 *
 *  Set's a socket to a non blocking state
 */
void
sock_setnonblock (int fd)
{
  int one = 1, flags;

  /* Reuse the address */
  if (setsockopt (fd, SOL_SOCKET, SO_REUSEADDR, (char *) &one, sizeof (one))
      != 0)
    err_warn ("sock_setnonblock: setsockopt(fd, SO_REUSEADDR): %m");

  /* Get the current flags for the socket */
  if ((flags = fcntl (fd, F_GETFL, 0)) < 0)
    err_fatal ("sock_setnonblock: fcntl(F_GETFL): %m");

  /* Set the flags + Non Blocking */
  if (fcntl (fd, F_SETFL, flags | O_NONBLOCK) != 0)
    err_fatal ("sock_setnonblock: fcntl(F_SETFL): %m");
}

/*
 * sock_alloc
 *
 *  Allocate's a socket
 */
socketrec *
sock_alloc (void)
{
  return (block_allocandaddtolink (&socket_pool, &sockets));
}

/*
 * sock_free
 *
 *  Free's a socket
 */
void
sock_free (socketrec * sock)
{
  /* Type specific free functions */
  switch (sock->type) {
  case SOCKET_TCP:
    sock_tcp_free (sock);
    break;
  case SOCKET_UDP:
    sock_udp_free (sock);
    break;
  }

  /* Put it back into the pool */
  block_delete (&socket_pool, &sock->ll);
}
