/* arg_list_api.c - library routines for passing argument lists.

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team    (http://www.dents.org)
   Copyright (C) 1999  Eric Z. Ayers (Eric.Ayers@mindspring.com)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

/* Add dents headers here */
#include "../include/common.h"
#include "../include/arg_list_api.h"	/* Exported Defs for this source */
#include "al_parse_defs.h"

/******************************************************************************
 *				Macro Definitions
 *****************************************************************************/
#define ASSERT g_assert

#define ARG_HND_MAGIC		0x33EA56EE
#define ARG_LIST_MAGIC		0x32EA56DD
#define _AL_ENTRY_MAGIC		0x31EA56CC
#define _AL_NAMED_ARG_MAGIC	0x30EA56AA
/* Randomly chosen identifiers to detect memory probs */

#define CHECK_ARG_HND(ptr) \
{ ASSERT(ptr); ASSERT((ptr)->alh_magic == ARG_HND_MAGIC); }
#define CHECK_ARG_LIST(ptr) \
{ ASSERT(ptr); ASSERT((ptr)->all_magic == ARG_LIST_MAGIC); }
#define CHECK_AL_ENTRY(ptr) \
{ ASSERT(ptr); ASSERT((ptr)->ale_magic == _AL_ENTRY_MAGIC); }
#define CHECK_AL_NAMED_ARG(ptr) \
{ ASSERT(ptr); ASSERT((ptr)->aln_magic == _AL_NAMED_ARG_MAGIC); }

/******************************************************************************
 *				Type Definitions
 *****************************************************************************/

/*
 * ARG_HND 
 *
 * A handle to a data structure that can contain a 
 *      collection of named arguments.
 *
 */
struct arg_handle_str
{
  unsigned int alh_magic;
  /* A variable to check for consistency when passed into */
  /* this module. */

  GHashTable *alh_named_arg_hash;
  /* Hash table of named arguments _AL_NAMED_ARG structs   */

  ARG_LIST alh_named_arg_list;
  /* A list containing the names of the arguments */
};

/*  
 * ARG_LIST
 *
 * A collection of ARG_ENTRY structures 
 */
struct arg_list_str
{
  unsigned int all_magic;
  /* A variable to check for consistency when passed into */
  /* this module. */

  GSList *all_list;
  /* glib singly linked list for holding data structures */
};

/*
 * _AL_ENTRY 
 *  One element of an ARG_LIST 
 */
typedef struct al_entry_str
{
  unsigned int ale_magic;
  /* A variable to check for consistency when passed into */
  /* this module. */

  int ale_type;
  /* One of the AL_TYPE_definitions. determines which of the */
  /* union members below is valid.                       */

  union
  {
    char *string_value;
    const void *void_value;
    int int_value;
    ARG_LIST list_value;
    ARG_HND hnd_value;
  }
  ale_value;			/* The value of this argument                          */

}
_AL_ENTRY;

/*
 * AL_NAMED_ARG
 * A single argument/value pair
 */
typedef struct al_named_arg
{
  unsigned int aln_magic;
  /* A variable to check for consistency when passed into */
  /* this module. */
  char *aln_key;
  /* The 'name' of this argument                          */
  int aln_type;
  /* One of the AL_TYPE_definitions. determines which of the */
  /* union members below is valid.                       */
  union
  {
    char *string_value;
    const void *void_value;
    int int_value;
    ARG_LIST list_value;
    ARG_HND hnd_value;
  }
  aln_value;			/* The value of this argument                          */

}
_AL_NAMED_ARG;

/* _ALH_WRITE_CTX: used by al_hnd_write() */
typedef struct
{
  int level;
  FILE *out_file;
  int return_val;		/* how else will we return an error from  fprintf()? */
}
_ALH_WRITE_CTX;

/******************************************************************************
 *				Function Prototypes
 *****************************************************************************/

static int _al_named_arg_delete (gpointer key,	/* char * name of this argument structure */
				 gpointer value,	/* _AL_NAMED_ARG * */
				 gpointer user_data	/* not used */
  );

static gint _al_entry_str_cmp (gconstpointer p1,	/* _AL_ENTRY #1 */
			       gconstpointer p2	/* _AL_ENTRY #2 */
  );

static void _al_entry_delete (gpointer value,	/* _AL_ENTRY * */
			      gpointer user_data	/* not used */
  );

static void _al_hnd_write_list (ARG_LIST al, _ALH_WRITE_CTX * ctx);

static void _al_hnd_write_arg (ARG_HND hnd, _ALH_WRITE_CTX * ctx);

/******************************************************************************
 ******************************************************************************
 **				Internal Functions
 ******************************************************************************
 *****************************************************************************/

/*
 * _al_named_arg_delete() - erase a _AL_NAMED_ARG structure from the hash table
 * 
 * This function is meant to be passed to g_hash_foreach () or 
 * g_hash_foreach_remove().  All the memory associated with this structure
 * is freed.
 *
 * Inputs:
 *      key        char * name of this argument structure 
 *      value      _AL_NAMED_ARG *  in disguise
 *      user_data  not used 
 *
 * Outputs: None.
 *      
 * Returns: Always returns 1 to indicate that this node should be removed
 *          from the hash table.
 */
static int
_al_named_arg_delete (gpointer key,	/* char * name of this argument structure */
		      gpointer value,	/* _AL_NAMED_ARG * */
		      gpointer user_data	/* not used */
  )
{
#define ROUTINE "_al_named_arg_delete ()"
  _AL_NAMED_ARG *na = (_AL_NAMED_ARG *) value;

  CHECK_AL_NAMED_ARG (na);

  switch (na->aln_type) {
  case AL_TYPE_UNDEF:
  case AL_TYPE_INT:
  case AL_TYPE_VOID:
    /* there is nothing we can do in these cases */
    break;

  case AL_TYPE_STRING:
    g_free (na->aln_value.string_value);
    na->aln_value.string_value = NULL;
    break;

  case AL_TYPE_LIST:
    al_list_delete (&na->aln_value.list_value);
    break;

  case AL_TYPE_HND:
    al_hnd_delete (&na->aln_value.hnd_value);
    break;

  default:
    fprintf (stderr, "%s: Unknown argument type %d\n", ROUTINE, na->aln_type);
    ASSERT (NULL);
  }

  /* free the memory associated with the key value */
  g_free (na->aln_key);

  /* free the memory associated with this node */
  g_free (na);

  return 1;
#undef ROUTINE
}				/* end _al_named_arg_delete() */

/*
 * _al_entry_str_cmp - compare function for two string values of _AL_ENTRY
 *
 * This is a standard function of type GCompareFunc
 */
static gint
_al_entry_str_cmp (gconstpointer p1, gconstpointer p2)
{
  _AL_ENTRY *e1 = (_AL_ENTRY *) p1;
  _AL_ENTRY *e2 = (_AL_ENTRY *) p2;

  ASSERT (e1->ale_type == AL_TYPE_STRING && e2->ale_type == AL_TYPE_STRING);

  if (g_str_equal (e1->ale_value.string_value, e2->ale_value.string_value))
    return 0;
  return 1;
}

/*
 * _al_entry_delete() - erase a _AL_ENTRY structure from the list
 * 
 * This function is meant to be passed to g_slist_foreach () or 
 *  All the memory associated with this structure is freed.
 *
 * Inputs:
 *      value      _AL_ENTRY *  in disguise
 *      user_data  not used 
 *
 * Outputs: None.
 *      
 * Returns: None
 *          
 */
static void
_al_entry_delete (gpointer value,	/* _AL_ENTRY * */
		  gpointer user_data	/* not used */
  )
{
#define ROUTINE "al_entry_delete ()"
  _AL_ENTRY *e = (_AL_ENTRY *) value;

  CHECK_AL_ENTRY (e);

  switch (e->ale_type) {
  case AL_TYPE_UNDEF:
  case AL_TYPE_INT:
  case AL_TYPE_VOID:
    /* there is nothing we can do in these cases */
    break;

  case AL_TYPE_STRING:
    g_free (e->ale_value.string_value);
    e->ale_value.string_value = NULL;
    break;

  case AL_TYPE_LIST:
    al_list_delete (&e->ale_value.list_value);
    break;

  case AL_TYPE_HND:
    al_hnd_delete (&e->ale_value.hnd_value);
    break;

  default:
    fprintf (stderr, "%s: Unknown argument type %d\n", ROUTINE, e->ale_type);
    ASSERT (NULL);
  }

  /* free the memory associated with this node */
  g_free (e);
#undef ROUTINE
}				/* end _al_entry_delete() */

/******************************************************************************
 ******************************************************************************
 **				Exported Functions
 ******************************************************************************
 *****************************************************************************/

/*
 * al_hnd_new - allocate a new ARG_HND handle
 *
 * Inputs:
 *      None.
 * Outputs:
 *      None.
 * Returns: NULL pointer value on failure.
 *          non-zero pointer value on success
 *
 */
ARG_HND
al_hnd_new (void)
{
  struct arg_handle_str *hnd;

  hnd = (ARG_HND) g_malloc0 (sizeof (struct arg_handle_str));

  if (!hnd) {
    return NULL;
  }

  hnd->alh_magic = ARG_HND_MAGIC;

  /* The aln_key element in the structure is a char buffer, 
     so we should just use the provided string compare functions here.
   */
  hnd->alh_named_arg_hash = g_hash_table_new (g_str_hash, g_str_equal);
  hnd->alh_named_arg_list = al_list_new ();

  return hnd;

}				/* end al_hnd_new */

/*
 * al_hnd_delete - free an ARG_HND and all of its resources 
 *
 * Inputs:      hdn_ptr - valid addr of an ARG_HND structure allocated
 *              with arg_hnd_new
 * Outputs:     hdn_ptr - is invalidated
 *
 * Returns:     None.
 */
void
al_hnd_delete (ARG_HND * hnd_ptr)
{
  CHECK_ARG_HND (*hnd_ptr);

  /* free everything in the hash table */
  g_hash_table_foreach_remove ((*hnd_ptr)->alh_named_arg_hash,
			       _al_named_arg_delete, NULL);

  g_hash_table_destroy ((*hnd_ptr)->alh_named_arg_hash);
  (*hnd_ptr)->alh_named_arg_hash = NULL;

  /* free the list of argument names */
  al_list_delete (&(*hnd_ptr)->alh_named_arg_list);

  /* free memory associated with the handle */
  g_free (*hnd_ptr);
  *hnd_ptr = 0;

}				/* end al_hnd_delete () */

/* ------------------------------------------------------------------------- */
/*
 * al_hnd_add_string - add a string argument to the handle
 * al_hnd_add_int    - add a int argument to the handle
 * al_hnd_add_void   - add a void argument to the handle
 * al_hnd_add_list   - add a ARG_LIST argument to the handle
 * al_hnd_add_hnd    - add a ARG_HND argument to the handle
 * 
 * Inputs:      hnd - value ARG_HND value returned from al_hnd_new ()
 *              name - name of the argument to add
 *              value - value of the argument to add.  
 *      
 *                 In the case of a string, the routine does a  strdup() 
 *                 of the contents, so you don't have to worry about saving 
 *                 the memory from the argument you added. In the case of a 
 *                 void, ARG_LIST, or ARG_HND, we just add a pointer to the 
 *                 list, so DON'T free it!
 *      
 * Ouptuts:
 *              None.
 * Returns:
 *      AL_E_OK on success (0) 
 *      AL_E_DUP if this has already been added.
 */
int
al_hnd_add_string (ARG_HND hnd, const char *name, const char *value)
{
#define ROUTINE "al_hnd_add_string ()"
  _AL_NAMED_ARG *na;

  CHECK_ARG_HND (hnd);
  ASSERT (name);
  ASSERT (value);

  /* Check for an exisiting key.  If it already exists, return DUP */
  if (g_hash_table_lookup (hnd->alh_named_arg_hash, name)) {
    return AL_E_DUP;
  }

  na = (_AL_NAMED_ARG *) g_malloc0 (sizeof (_AL_NAMED_ARG));
  na->aln_magic = _AL_NAMED_ARG_MAGIC;
  na->aln_key = g_strdup (name);
  na->aln_type = AL_TYPE_STRING;
  na->aln_value.string_value = g_strdup (value);

  g_hash_table_insert (hnd->alh_named_arg_hash, (gpointer) na->aln_key,
		       (gpointer) na);
  al_list_add_string (hnd->alh_named_arg_list, name);

  return AL_E_OK;

#undef ROUTINE
}				/* end al_hnd_add_string () */

int
al_hnd_add_int (ARG_HND hnd, char *name, int value)
{
  _AL_NAMED_ARG *na;

  CHECK_ARG_HND (hnd);
  ASSERT (name);
  ASSERT (value);

  /* Check for an exisiting key.  If it already exists, return DUP */
  if (g_hash_table_lookup (hnd->alh_named_arg_hash, name)) {
    return AL_E_DUP;
  }

  na = (_AL_NAMED_ARG *) g_malloc0 (sizeof (_AL_NAMED_ARG));
  na->aln_magic = _AL_NAMED_ARG_MAGIC;
  na->aln_key = g_strdup (name);
  na->aln_type = AL_TYPE_INT;
  na->aln_value.int_value = value;

  g_hash_table_insert (hnd->alh_named_arg_hash, (gpointer) na->aln_key,
		       (gpointer) na);
  al_list_add_string (hnd->alh_named_arg_list, name);

  return AL_E_OK;

#undef ROUTINE
}				/* end al_hnd_add_int () */

int
al_hnd_add_void (ARG_HND hnd, char *name, const void *value)
{
  _AL_NAMED_ARG *na;

  CHECK_ARG_HND (hnd);
  ASSERT (name);
  ASSERT (value);

  /* Check for an exisiting key.  If it already exists, return DUP */
  if (g_hash_table_lookup (hnd->alh_named_arg_hash, name)) {
    return AL_E_DUP;
  }

  na = (_AL_NAMED_ARG *) g_malloc0 (sizeof (_AL_NAMED_ARG));
  na->aln_magic = _AL_NAMED_ARG_MAGIC;
  na->aln_key = g_strdup (name);
  na->aln_type = AL_TYPE_VOID;

  na->aln_value.void_value = value;

  g_hash_table_insert (hnd->alh_named_arg_hash, (gpointer) na->aln_key,
		       (gpointer) na);
  al_list_add_string (hnd->alh_named_arg_list, name);

  return AL_E_OK;

}				/* end al_hnd_add_void () */

int
al_hnd_add_list (ARG_HND hnd, char *name, ARG_LIST value)
{
  _AL_NAMED_ARG *na;

  CHECK_ARG_HND (hnd);
  ASSERT (name);
  ASSERT (value);

  /* Check for an exisiting key.  If it already exists, return DUP */
  if (g_hash_table_lookup (hnd->alh_named_arg_hash, name)) {
    return AL_E_DUP;
  }

  na = (_AL_NAMED_ARG *) g_malloc0 (sizeof (_AL_NAMED_ARG));
  na->aln_magic = _AL_NAMED_ARG_MAGIC;
  na->aln_key = g_strdup (name);
  na->aln_type = AL_TYPE_LIST;
  na->aln_value.list_value = value;

  g_hash_table_insert (hnd->alh_named_arg_hash,
		       (gpointer) na->aln_key, (gpointer) na);

  al_list_add_string (hnd->alh_named_arg_list, name);

  return AL_E_OK;

}				/* end al_hnd_add_list() */

int
al_hnd_add_hnd (ARG_HND hnd, char *name, ARG_HND value)
{
  _AL_NAMED_ARG *na;

  CHECK_ARG_HND (hnd);
  ASSERT (name);
  CHECK_ARG_HND (value);

  /* Check for an exisiting key.  If it already exists, return DUP */
  if (g_hash_table_lookup (hnd->alh_named_arg_hash, name)) {
    return AL_E_DUP;
  }

  na = (_AL_NAMED_ARG *) g_malloc0 (sizeof (_AL_NAMED_ARG));
  na->aln_magic = _AL_NAMED_ARG_MAGIC;
  na->aln_key = g_strdup (name);
  na->aln_type = AL_TYPE_HND;
  na->aln_value.hnd_value = value;

  g_hash_table_insert (hnd->alh_named_arg_hash,
		       (gpointer) na->aln_key, (gpointer) na);

  al_list_add_string (hnd->alh_named_arg_list, name);

  return AL_E_OK;

}				/* end al_hnd_add_hnd () */

/* ========================================================================= */

/*
 * al_hnd_find - Find an argument & return its type + value
 *
 * Inputs:      name - name of the argument to retrieve
 * Outputs:     type_ptr - returns one of the AL_TYPE_XXX values
 *              value_ptr - returns the value. You'll need to cast this
 *                          back to the type that is indicated by type_ptr
 *
 * Returns:
 *      AL_E_OK (0) on success
 *      AL_E_NOT_FOUND if the argument 'name' isn't defined
 */

int
al_hnd_find (ARG_HND hnd, char *name, int *type_ptr, void **value_ptr)
{
#define ROUTINE "al_hnd_find ()"
  _AL_NAMED_ARG *na;

  CHECK_ARG_HND (hnd);
  ASSERT (name);
  ASSERT (type_ptr);
  ASSERT (value_ptr);

  na = (_AL_NAMED_ARG *) g_hash_table_lookup (hnd->alh_named_arg_hash, name);

  if (!na)
    return AL_E_NOT_FOUND;

  *type_ptr = na->aln_type;

  /* this is kinda dumb, i mean we could just cast to a void * or memcpy 
     the size of a void*, but this is a litte easier to follow.
   */
  switch (na->aln_type) {
  case AL_TYPE_INT:
    /* ignore compiler whining on this line if you are on a 64 bit machine */
    *value_ptr = (void *) na->aln_value.int_value;
    break;

  case AL_TYPE_STRING:
    *value_ptr = (void *) na->aln_value.string_value;
    break;

  case AL_TYPE_VOID:
    *value_ptr = (void *) na->aln_value.void_value;
    break;

  case AL_TYPE_LIST:
    *value_ptr = (void *) na->aln_value.list_value;
    break;

  case AL_TYPE_HND:
    *value_ptr = (void *) na->aln_value.hnd_value;
    break;

  case AL_TYPE_UNDEF:
  default:
    ASSERT (0);
  }

  return AL_E_OK;
#undef ROUTINE
}				/* end al_hnd_find() */

/*
 * al_hnd_find_string - retrieve a string value parameter from the handle
 *
 * Inputs:      hnd - valid ARG_HND 
 *              name - name of the string argument to retrieve
 * Outputs:     None.
 * Returns:     a valid string on success. DO NOT free this string - it's
 *              just a pointer to data allocated in the handle structure.
 */
char *
al_hnd_find_string (ARG_HND hnd, char *name)
{
#define ROUTINE "al_hnd_find_string ()"
  int ret_type = AL_TYPE_UNDEF;
  char *ret_value = NULL;

  if (al_hnd_find (hnd, name, &ret_type, (void **) &ret_value) != AL_E_OK) {
    return NULL;
  }

  if (ret_type != AL_TYPE_STRING)
    return NULL;

  return ret_value;
#undef ROUTINE
}				/* end al_hnd_find_string() */

/*
 * al_hnd_find_int - retrieve an integer argument
 *
 * Inputs:      hnd - valid ARG_HND 
 *              name - name of the string argument to retrieve
 * Outputs:     None.
 * Returns:     The integer value.  It returns -1 if the value was not found.
 *              If you really want to know if the argument was there 
 *              and -1 is a valid value, use al_hnd_find() instead.
 */
int
al_hnd_find_int (ARG_HND hnd, char *name)
{
#define ROUTINE "al_hnd_find_int ()"
  int ret_type = AL_TYPE_UNDEF;
  int ret_value = -1;

  if (al_hnd_find (hnd, name, &ret_type, (void **) &ret_value) != AL_E_OK) {
    return -1;
  }

  if (ret_type != AL_TYPE_INT)
    return -1;

  return ret_value;
#undef ROUTINE
}				/* end al_hnd_find_int() */

/*
 * al_hnd_find_void - retrieve a void argument
 *
 * Inputs:      hnd - valid ARG_HND 
 *              name - name of the string argument to retrieve
 * Outputs:     None.
 * Returns:     NULL on failure
 *              A non-null void pointer on success. If you want to test
 *              to see if the named argument is really there and
 *              and NULL is a valid value, use al_hnd_find() instead.
 */
void *
al_hnd_find_void (ARG_HND hnd, char *name)
{
#define ROUTINE "al_hnd_find_void ()"
  int ret_type = AL_TYPE_UNDEF;
  void *ret_value = NULL;

  if (al_hnd_find (hnd, name, &ret_type, (void **) &ret_value) != AL_E_OK) {
    return NULL;
  }

  if (ret_type != AL_TYPE_VOID)
    return NULL;

  return ret_value;
#undef ROUTINE
}				/* end al_hnd_find_void() */

/*
 * al_hnd_find_list - retrieve a list argument
 *
 * Inputs:      hnd - valid ARG_HND 
 *              name - name of the string argument to retrieve
 * Outputs:     None.
 * Returns:     NULL on failure
 *              A non-null ARG_LIST on success. 
 */
ARG_LIST
al_hnd_find_list (ARG_HND hnd, char *name)
{
#define ROUTINE "al_hnd_find_list ()"
  int ret_type = AL_TYPE_UNDEF;
  ARG_LIST ret_value = NULL;

  if (al_hnd_find (hnd, name, &ret_type, (void **) &ret_value) != AL_E_OK) {
    return NULL;
  }

  if (ret_type != AL_TYPE_LIST)
    return NULL;

  return ret_value;

#undef ROUTINE
}				/* end al_hnd_find_list() */

/*
 * al_hnd_find_hnd - retrieve an ARG_HND argument
 *
 * Inputs:      hnd - valid ARG_HND 
 *              name - name of the string argument to retrieve
 * Outputs:     None.
 * Returns:     NULL on failure
 *              A non-null ARG_HND on success. 
 */
ARG_HND
al_hnd_find_hnd (ARG_HND hnd, char *name)
{
#define ROUTINE "al_hnd_find_hnd ()"
  int ret_type = AL_TYPE_UNDEF;
  ARG_HND ret_value = NULL;

  if (al_hnd_find (hnd, name, &ret_type, (void **) &ret_value) != AL_E_OK) {
    return NULL;
  }

  if (ret_type != AL_TYPE_HND)
    return NULL;

  return ret_value;
#undef ROUTINE
}				/* end al_hnd_find_hnd() */

/* ========================================================================= */

/*
 * al_hnd_get_names ()
 *
 * Inputs:      hnd - argument handle
 * Outputs:     none.
 *
 * Returns:
 *      a non-zero ARG_LIST pointer on success
 *      NULL on failure, or if no names have been configured in this hnd
 *      WARNING: do not free this list!  It's kept inside the hnd structure.
 *
 */
ARG_LIST
al_hnd_get_names (ARG_HND hnd)
{
  CHECK_ARG_HND (hnd);
  return hnd->alh_named_arg_list;
}

/*
 * al_hnd_declare_type -  declare what type an argument name/value has
 * 
 * The idea here is to allow someone to declare the type ahead of time
 * w/o actually assigning a value. This would help so that someone could
 * define what argument names and types are allowed before parsing a config
 * file, then the parser could return an error if the argument contained
 * the wrong type.
 *
 */
int
al_hnd_declare_type (ARG_HND hnd, char *name)
{
  /* not implemented */
  return AL_E_NOT_IMPL;
}				/* end al_hnd_declare_type () */

/*
 * al_hnd_read - populate a handle from a text file
 *
 * Inputs:      in_file - open file to read
 *              namelen - length of buffer passed in name
 * Outputs:     name - name of hash to return
 * 
 * Returns:     A valid ARG_HND value on success
 *              NULL on failure.
 */
ARG_HND
al_hnd_read (FILE * in_file, char *name, int namelen)
{
  ASSERT (in_file);
  ASSERT (name);

  yy_al_lineno = 0;

  /* reset the parser to the start state */
  yy_al_restart (in_file);

  if (!yy_al_parse ()) {
    strncpy (name, global_al_parsed_name, namelen);
    free (global_al_parsed_name);
    return global_al_parsed_hnd;
  } else {
    return NULL;
  }

}				/* end al_hnd_read () */

/*
 * al_hnd_read_string - populate a handle from a string
 *
 * Inputs:      in_string - null terminated string to read from
 *              namelen - length of buffer passed in name
 * Outputs:
 *              name - name of the hash to return.
 * Returns:     A valid ARG_HND value on success.
 *              NULL on failure.
 *
 */
ARG_HND
al_hnd_read_string (char *in_string, char *name, int namelen)
{
  ASSERT (in_string);
  ASSERT (name);

  yy_al_lineno = 0;

  /* reset the parser to the start state - point the scanner to the string */
  yy_al_scanner (in_string);

  if (!yy_al_parse ()) {
    strncpy (name, global_al_parsed_name, namelen);
    free (global_al_parsed_name);
    return global_al_parsed_hnd;
  } else {
    return NULL;
  }

}				/* end al_hnd_read_string () */

/*
 * _al_hnd_write_list() - helper function for al_hnd_write()
 * This iterates through the list
 */
static void
_al_hnd_write_list (ARG_LIST al, _ALH_WRITE_CTX * ctx)
{
  GSList *curr_entry;

  for (curr_entry = al->all_list;
       curr_entry; curr_entry = g_slist_next (curr_entry)) {

    _AL_ENTRY *ale_ptr = (_AL_ENTRY *) curr_entry->data;

    CHECK_AL_ENTRY (ale_ptr);

    switch (ale_ptr->ale_type) {
    case AL_TYPE_INT:
      fprintf (ctx->out_file, "%d%c", ale_ptr->ale_value.int_value,
	       g_slist_next (curr_entry) ? ',' : ' ');
      break;

    case AL_TYPE_STRING:
      fprintf (ctx->out_file, "'%s'%c", ale_ptr->ale_value.string_value,
	       g_slist_next (curr_entry) ? ',' : ' ');
      break;

    case AL_TYPE_VOID:
      /* ha! we can't print this! */
      fprintf (ctx->out_file, "%p%c\n", ale_ptr->ale_value.void_value,
	       g_slist_next (curr_entry) ? ',' : ' ');
      break;

    case AL_TYPE_LIST:
      /* hmm. we can just use this routine all over again! */
      fprintf (ctx->out_file, "[ ");
      _al_hnd_write_list (ale_ptr->ale_value.list_value, ctx);
      fprintf (ctx->out_file, "]%c\n", g_slist_next (curr_entry) ? ',' : ' ');
      break;

    case AL_TYPE_HND:
      ctx->level++;
      fprintf (ctx->out_file, "{\n");
      _al_hnd_write_arg (ale_ptr->ale_value.hnd_value, ctx);
      fprintf (ctx->out_file, "%*c}%c # end hnd (list entry)\n",
	       ctx->level * 2, ' ', g_slist_next (curr_entry) ? ',' : ' ');

      ctx->level--;
    }				/* end switch() */
  }				/* end for each element in the list  */

}				/* end al_hnd_write_list () */

/*
 * _al_hnd_write_arg() - helper function for al_hnd_write()
 * called as a callback from g_hash_table_foreach() 
 */
static void
_al_hnd_write_arg (ARG_HND hnd, _ALH_WRITE_CTX * ctx)
{
  GSList *curr_entry;

  CHECK_ARG_HND (hnd);

  /* Go through the list of strings that represent keys in the hash
     table.

     Why did I do it this way?  Why didn't I just do a g_hash_table_foreach?

     Well, I did at first, but I changed it in order to print commas.
     The list allows this function to have a look-ahead to see if it needs
     to print another comma.  I thought about maintaining a stack of 
     'ctx' variables, but that seemed like a lot of work.  I already 
     had a list there to support the function al_hnd_get_names().
   */
  for (curr_entry = hnd->alh_named_arg_list->all_list;
       curr_entry; curr_entry = g_slist_next (curr_entry)) {
    _AL_ENTRY *ale_ptr = (_AL_ENTRY *) curr_entry->data;
    _AL_NAMED_ARG *aln_ptr =
      (_AL_NAMED_ARG *) g_hash_table_lookup (hnd->alh_named_arg_hash,
					     ale_ptr->ale_value.string_value);

    CHECK_AL_NAMED_ARG (aln_ptr);

    /* %*c , num_chars, char_to_print */
    fprintf (ctx->out_file, "%*c", ctx->level * 2, ' ');
    switch (aln_ptr->aln_type) {
    case AL_TYPE_INT:
      fprintf (ctx->out_file, "'%s' : %d\n",
	       aln_ptr->aln_key, aln_ptr->aln_value.int_value);
      break;

    case AL_TYPE_STRING:
      fprintf (ctx->out_file, "'%s' : '%s'%c\n", aln_ptr->aln_key,
	       aln_ptr->aln_value.string_value,
	       g_slist_next (curr_entry) ? ',' : ' ');

      break;

    case AL_TYPE_VOID:
      /* ha! we can't print this! */
      fprintf (ctx->out_file, "# '%s' : %p%c\n", aln_ptr->aln_key,
	       aln_ptr->aln_value.void_value,
	       g_slist_next (curr_entry) ? ',' : ' ');
      break;

    case AL_TYPE_LIST:
      fprintf (ctx->out_file, "'%s' : [", aln_ptr->aln_key);
      _al_hnd_write_list (aln_ptr->aln_value.list_value, ctx);
      fprintf (ctx->out_file, "]%c\n", g_slist_next (curr_entry) ? ',' : ' ');
      break;

    case AL_TYPE_HND:
      /* hmm. we can just use this routine all over again! */
      ctx->level++;
      fprintf (ctx->out_file, "'%s' : {\n", aln_ptr->aln_key);

      _al_hnd_write_arg (aln_ptr->aln_value.hnd_value, ctx);

      /* %*c , num_chars, char_to_print */
      fprintf (ctx->out_file, "%*c}%c# end hnd %s\n", ctx->level * 2, ' ',
	       g_slist_next (curr_entry) ? ',' : ' ', aln_ptr->aln_key);

      ctx->level--;

    }				/* end switch */
  }				/* end for each element in the list */

}				/* end al_hnd_write_arg () */

/*
 * al_hnd_write - dump a handle to a text file
 *
 * Write something out to a file that is vaguely like named.boot
 * (it won't be, of course)
 *
 * FIXME: This function doesn't check for errors from fprintf()
 *
 * Inputs:  hnd - valid ARG_HND value
 *          out_file - a stdio FILE * open for writing.
 *               Note, it is the responsibility of the caller to
 *               close this function. 
 *              
 * Outputs: None.
 * Returns AL_E_OK on success
 *
 */
int
al_hnd_write (ARG_HND hnd, FILE * out_file, char *name)
{
  _ALH_WRITE_CTX ctx;

  CHECK_ARG_HND (hnd);
  ASSERT (out_file);
  ASSERT (name);

  ctx.level = 1;
  ctx.out_file = out_file;
  ctx.return_val = AL_E_OK;

  printf ("%s = {\n", name);
  _al_hnd_write_arg (hnd, &ctx);
  printf ("} # end configuration %s\n", name);
  return ctx.return_val;

}				/* end al_hnd_write() */

ARG_LIST
al_list_new (void)
{
  ARG_LIST al;

  al = (ARG_LIST) g_malloc0 (sizeof (struct arg_list_str));
  al->all_magic = ARG_LIST_MAGIC;
  al->all_list = NULL;

  return al;

}				/* end al_list_new () */

void
al_list_delete (ARG_LIST * list_ptr)
{
#define ROUTINE "al_list_delete ()"

  CHECK_ARG_LIST (*list_ptr);

  /* Free each entry in the list */
  g_slist_foreach ((*list_ptr)->all_list, _al_entry_delete, NULL);

  g_slist_free ((*list_ptr)->all_list);

  g_free (*list_ptr);
  *list_ptr = NULL;

#undef ROUTINE
}				/* end al_list_delete() */

/*
 * al_list_add_string - add a string value to an ARG_LIST
 *
 * Inputs:  list_ptr - list allocated with al_list_new()
 *          value - the string to add to the list. This library
 *                  makes a copy of the string.
 *
 * Outputs: none.
 *
 * Returns: AL_E_OK on success.
 */
int
al_list_add_string (ARG_LIST list_ptr, const char *value)
{
  _AL_ENTRY *ae;

  CHECK_ARG_LIST (list_ptr);
  ASSERT (value);

  ae = (_AL_ENTRY *) g_malloc (sizeof (_AL_ENTRY));
  ae->ale_magic = _AL_ENTRY_MAGIC;
  ae->ale_type = AL_TYPE_STRING;
  ae->ale_value.string_value = g_strdup (value);

  list_ptr->all_list = g_slist_append (list_ptr->all_list, (gpointer) ae);

  return AL_E_OK;
}				/* end al_list_add_string */

/*
 * al_list_length () - number of ARG_ENTRY values in this list
 *
 * Inputs:
 *      list_ptr - valid ARG_LIST handle.
 * Outputs:
 *      None.
 * Returns:
 *      The number of entries in the list.
 *
 */
int
al_list_length (ARG_LIST list_ptr)
{
  CHECK_ARG_LIST (list_ptr);
  return g_slist_length (list_ptr->all_list);
}				/* end al_list_num_items() */

/*
 * al_list_first - retrieve the first ARG_ENTRY in the list
 *
 */
ARG_ENTRY
al_list_first (ARG_LIST list_ptr)
{
  CHECK_ARG_LIST (list_ptr);

  return (ARG_ENTRY) list_ptr->all_list;
}

/*
 * al_list_next - retrieve the subsequent ARG_ENTRY in the list
 *
 */
ARG_ENTRY
al_list_next (ARG_ENTRY entry)
{
  GSList *sl = (GSList *) entry;
  return g_slist_next (sl);
}

/*
 * al_entry_get_string - Pull the string value out of an ARG_ENTRY
 *
 * Inputs:
 *      entry - return from al_list_first() or al_list_next()
 * Outputs:
 *      None.
 * Returns:
 *      NULL if the type of the ARG_ENTRY is not a string
 *      non-nul string pointer on success.   DO NOT free this string.
 *      this is just a pointer to a value allocated in the list.
 */
char *
al_entry_get_string (ARG_ENTRY entry)
{
  GSList *sl = (GSList *) entry;

  _AL_ENTRY *e = (_AL_ENTRY *) sl->data;
  CHECK_AL_ENTRY (e);

  if (e->ale_type != AL_TYPE_STRING)
    return NULL;

  return e->ale_value.string_value;
}

/*
 * al_list_string_exists - Find a string value in the list
 *
 * Inputs:      list - a valid ARG_LIST
 *              value - a string to find in the list.
 * Outpus:      None.
 * Returns:
 *      Returns 0 if the value cannot be found.
 *      Returns 1 if the value is found in the list.
 *
 */
int
al_list_string_exists (ARG_LIST list, char *value)
{
  GSList *found;
  _AL_ENTRY temp;

  temp.ale_magic = _AL_ENTRY_MAGIC;
  temp.ale_type = AL_TYPE_STRING;
  temp.ale_value.string_value = value;

  found = g_slist_find_custom (list->all_list, (gpointer) & temp,
			       _al_entry_str_cmp);
  if (found)
    return 1;

  return 0;

}				/* end al_list_string_exists() */

/******************************************************************************
 ******************************************************************************
 **				Test Driver
 ******************************************************************************
 *****************************************************************************/

#ifdef TEST_DRIVER
int
main (int argc, char *argv)
{
  ARG_HND test_hnd = NULL, test_hnd2 = NULL;
  ARG_LIST test_list = NULL;
  ARG_ENTRY test_entry = NULL;
  float my_float_val = 3.14;

  test_hnd = al_hnd_new ();

  if (al_hnd_add_string (test_hnd, "test-string", "test-string-value") !=
      AL_E_OK) {
    fprintf (stderr, "Error inserting string.\n");
    exit (1);
  }

  if (al_hnd_add_int (test_hnd, "test-int", 666) != AL_E_OK) {
    fprintf (stderr, "Error inserting string.\n");
    exit (1);
  }

  if (al_hnd_add_void (test_hnd, "test-void", (void *) &my_float_val) !=
      AL_E_OK) {
    fprintf (stderr, "Error inserting string.\n");
    exit (1);
  }

  test_list = al_list_new ();
  al_list_add_string (test_list, "apples");
  al_list_add_string (test_list, "oranges");
  al_list_add_string (test_list, "bananas");

  if (al_hnd_add_list (test_hnd, "test-list", test_list) != AL_E_OK) {
    fprintf (stderr, "Error inserting  ARG_LIST.\n");
    exit (1);
  }

  /* allocate a new handle structure */
  test_hnd2 = al_hnd_new ();
  al_hnd_add_string (test_hnd2, "test2", "test2-value");

  if (al_hnd_add_hnd (test_hnd, "test-hnd", test_hnd2) != AL_E_OK) {
    fprintf (stderr, "Error inserting ARG_HND.\n");
    exit (1);
  }

  /* get the list of names */
  if (NULL == (test_list = al_hnd_get_names (test_hnd))) {
    fprintf (stderr, "Retrieving list of names returned NULL\n");
    exit (1);
  }

  for (test_entry = al_list_first (test_list);
       test_entry; test_entry = al_list_next (test_entry)) {
    char *name = al_entry_get_string (test_entry);
    printf ("%s ", name);
  }
  printf ("\n");

  /* test the al_list_string_exists() function */
  if (al_list_string_exists (test_list, "test-int") != 1) {
    fprintf (stderr, "should have found 'test-int' in the list\n");
    exit (1);
  }

  if (al_list_string_exists (test_list, "bogus") == 1) {
    fprintf (stderr, "should NOT have found 'bogus' in the list\n");
    exit (1);
  }

  /* get the integer arg */
  {
    int int_val = al_hnd_find_int (test_hnd, "test-int");
    if (int_val != 666) {
      fprintf (stderr, "Error fetching int.\n");
      exit (1);
    }
  }

  /* get the string arg */
  {
    char *ret_val = al_hnd_find_string (test_hnd, "test-string");
    if (!ret_val || (!g_str_equal (ret_val, "test-string-value"))) {
      fprintf (stderr, "Error fetching string.\n");
      exit (1);
    }
  }

  /* get the void * arg */
  {
    void *ret_val = al_hnd_find_void (test_hnd, "test-void");
    if (!ret_val) {
      fprintf (stderr, "Error fetching void.\n");
      exit (1);
    }
    printf ("fetched ptr to float %f\n", *(float *) ret_val);
  }

  /* get the ARG_LIST arg */
  {
    ARG_LIST ret_val = al_hnd_find_list (test_hnd, "test-list");
    ARG_ENTRY entry;
    if (!ret_val) {
      fprintf (stderr, "Error fetching ARG_LIST.\n");
      exit (1);
    }
    for (entry = al_list_first (ret_val); entry; entry = al_list_next (entry)) {
      printf ("%s ", al_entry_get_string (entry));
    }
    printf ("\n");
  }

  /* get the ARG_HND arg */
  {
    ARG_HND ret_val = al_hnd_find_hnd (test_hnd, "test-hnd");
    char *test2_ret_val;

    if (!ret_val) {
      fprintf (stderr, "Error fetching ARG_HND.\n");
      exit (1);
    }
    if (ret_val != test_hnd2) {
      fprintf (stderr,
	       "Error fetching ARG_HND - I should have found test2_hnd!.\n");
      exit (1);
    }

    if ((test2_ret_val = al_hnd_find_string (ret_val, "test2")) == NULL) {
      fprintf (stderr, "Error fetching from returned ARG_HND.\n");
      exit (1);
    }

    if (!g_str_equal (test2_ret_val, "test2-value")) {
      fprintf (stderr, "Wrong string returned from returned ARG_HND.\n");
      exit (1);
    }

    printf ("returned from test2_hnd = %s\n", test2_ret_val);

  }

  /* get the int arg with the string arg function (should fail) */
  {
    char *ret_val = al_hnd_find_string (test_hnd, "test-int");
    if (ret_val != NULL) {
      fprintf (stderr,
	       "Tried to fetch an int value as a string and it worked??!?\n");
      exit (1);
    }
  }

  /* get the int arg with the void arg function (should fail) */
  {
    void *ret_val = al_hnd_find_void (test_hnd, "test-int");
    if (ret_val != NULL) {
      fprintf (stderr,
	       "Tried to fetch an int value as a void * and it worked??!?\n");
      exit (1);
    }
  }

  /* get the string arg with the int arg function (should fail) */
  {
    int ret_val = al_hnd_find_int (test_hnd, "test-string");
    if (ret_val != -1) {
      fprintf (stderr,
	       "Tried to fetch a string value as a int and it worked??!?\n");
      exit (1);
    }
  }

  /* get the int arg with the list arg function (should fail) */
  {
    ARG_LIST ret_val = al_hnd_find_list (test_hnd, "test-int");
    if (ret_val != NULL) {
      fprintf (stderr,
	       "Tried to fetch an int value as a list and it worked??!?\n");
      exit (1);
    }
  }

  /* get the int arg with the ARG_HND function (should fail) */
  {
    ARG_HND ret_val = al_hnd_find_hnd (test_hnd, "test-int");
    if (ret_val != NULL) {
      fprintf (stderr,
	       "Tried to fetch an int value as a hnd and it worked??!?\n");
      exit (1);
    }
  }

  /* dump this hash to stdout */
  al_hnd_write (test_hnd, stdout, "TEST_DRIVER");

  al_hnd_delete (&test_hnd);

  printf ("All tests passed\n");

}
#endif
