/* db.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/*
 * db.c
 *
 *  All of the functions needed for the databases management. Adding, removing,
 * modifying, etc.
 *
 *  The database is now a tree of Red-Black trees. More specifically, each
 * node, has a RB tree for any children of it
 */
#include "common.h"
#include "db.h"

/*
 * db_deletenode
 * 
 *  Delete's a node
 */
void
db_deletenode (dbnoderec * node)
{
  /* Don't delete the root */
  if (!node->parent)
    return;

  db_rb_delete (&((dbnoderec *) node->parent)->children, &node->rb);

  free (node);
}

/*
 * db_addnode
 *
 *  Add's a new node off a current node
 */
dbnoderec *
db_addnode (dbnoderec * node, guchar * label)
{
  dbnoderec *newnode;

  /* Allocate some memory for this hash entry */
  if ((newnode = malloc (sizeof (dbnoderec))) == NULL)
    err_fatal ("couldn't allocate memory for node");

  /* Save the node label */
  strncpyzt (newnode->label, label, DB_MAXLABELLENGTH);

  /* Setup the pointers */
  newnode->parent = node;

  /* Create a RB tree for any children */
  db_rb_create (&newnode->children, db_rb_cmp_stri);

  /* Insert this into the parent */
  db_rb_insert (&node->children, &newnode->rb, newnode);

  /* Reset the data field to NULL */
  newnode->data = NULL;

  return (newnode);
}

/*
 * db_findnode
 *
 *  Find's a node in a hash table
 */
dbnoderec *
db_findnode (dbnoderec * node, guchar * label)
{
  return (db_rb_find (&node->children, label));
}

#ifdef DENTS_DEBUG
#include "dns.h"		/* For DNS_MAXNAMELENGTH */

#if 0
void
db_tree_display (dbnoderec * node, guchar * rootname)
{
  guchar name[DNS_MAXNAMELENGTH + 1];

  /* If we're at the root, then name is blank */
  if (node->label[0])
    g_snprintf (name, DNS_MAXNAMELENGTH, "%s.%s", node->label, rootname);
  else
    name[0] = 0;

  if (node->data) {
    /* However, the root prefers to be displayed as just . */
    if (!name[0])
      printf ("%s.\n", name);
    else
      printf ("%s\n", name);
  }

  /* Walk down to electric avenue */
  db_rb_walk (node->children.root, db_tree_display, name);
}
#endif

#define PUSH(x) stack[stackptr++] = x
#define POP() stack[stackptr--]

void
db_tree_display (dbnoderec * node, guchar * rootname)
{
  guchar name[DNS_MAXNAMELENGTH + 1];
#if 0
  dbnoderec *stack[(DNS_MAXNAMELENGTH / 2) + 1];
#endif
  int stackptr = 0;

  do {
    /* If we're at the root, then name is blank */
    if (node->label[0])
      g_snprintf (name, DNS_MAXNAMELENGTH, "%s.%s", node->label, rootname);
    else
      name[0] = 0;

    if (node->data) {
      /* However, the root prefers to be displayed as just . */
      if (!name[0])
	printf ("%s.\n", name);
      else
	printf ("%s\n", name);
    }

  } while (stackptr);
}

void
db_display (dbrootrec * db)
{
  db_tree_display (&db->node, "");
}
#endif

/*
 * db_init
 *
 *  Initializes a db
 */
void
db_init (dbrootrec * db)
{
  /* We start out with 0 entries */
  db->numentries = 0;

  /* Zero out the label */
  db->node.label[0] = 0;

  /* This node has no data */
  db->node.data = NULL;

  /* Create the RB tree for it */
  db_rb_create (&db->node.children, db_rb_cmp_stri);

  /* We have no parent */
  db->node.parent = NULL;
}

#if 0
/* This function, albeit not used at all, is somewhat useful and we may need */
/*  it again at some time so we'll keep it */
/*
 * db_hashvalue
 *
 *  Routine that actually calculates the hash given a certain number of bits
 */
static guint32
db_hashvalue (guchar * key, guint8 bits)
{
  guchar lower;
  guint16 hash, mask;

  /* Go through each character in the key and convert it to lowercase, then */
  /*  figure it into the hash value */
  for (; *key; key++) {
    lower = tolower (*key);

/*    hash += (hash << 3) + lower; *//* This one seems to work well too */
    hash += (hash << 1) ^ lower;
  }

  /* Figure out the mask */
  mask = ((1 << bits) - 1);

  return (hash & mask);
}
#endif
