/* dns_build.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/*
 * dns_build.c
 *
 * Library routines that help build's DNS packets into a dnsbuildrec *
 * structure.
 */

/*****************************************************************************
 *			Include Files
 *****************************************************************************/

#include "dns.h"
#include "link.h"

#include <arpa/inet.h>		/* inet_ntoa */

/*****************************************************************************
 *			Macro Definitions
 *****************************************************************************/

/* Macro to append Resource Records into a buffer                       */
#define DNS_BP_PACKRR(build,type)  \
	if (_build_packrr(build, type, pkt, &cp, &cmp) != 0)  \
		return 1;

/*****************************************************************************
 *			Type Definitions
 *****************************************************************************/

struct dnsbuildrec_str
{

  llentry ll;			/* for including in a linked list */

  /* This section is used to build a packet to be sent out */
  dnsbuildrrrec rr[DNS_MAXRRPACKET];	/* RR's to be packed */

  int lastrr;			/* Last RR, because I'm lazy */

  int count[3];			/* Total of three sections */

  /* The following fields allow a module writer to get some data
     to write RDATA into for building a packet that will hang around
     until the packet is built with dns_buildpacket().  This way,
     the module writer doesn't have to use a global static buffer
     and the module can be thread safe.
   */

  guchar data_buf[DNS_MAXRRDATALENGTH * 2];
  /* a persistent buffer for building RDATA */
  int bytes_avail;		/* total # of bytes available in data_buf */
  int bytes_alloced;		/* total # of bytes alloced so far        */

};

/*****************************************************************************
 *			Global Data
 *****************************************************************************/

static blocklist gbl_buildrec_chain_pool = BLOCK_INITIALIZER (dnsbuildrec, 1);

/*****************************************************************************
 *****************************************************************************
 *			Internal Function Definitions
 *****************************************************************************
 *****************************************************************************/

/*
 * _bp_packrr - pack all RR's for a response section into the packet
 *
 * Inputs:      build - contains the RR's to add to the packet
 *              sectype - a DNS_BUILD_XXX definition (ANSWER, AUTHORITY, 
 *                      or ADDITIONAL)
 *              pkt - record containing packet to send
 *              cp - pointer to the current end of the packet were new
 *                   rr's should be written.
 *              cmp - compress record - list of domain names alredy compressed
 * Outputs:     cp - modified to point to the new end of the packet
 * Returns:     0 on success, 1 on failure
 *
 */
static int _build_packrr
  (dnsbuildrec * build,
   int sectype, packetrec * pkt, guchar ** cp_ptr, dnscompressrec * cmp_ptr)
{
#define ROUTINE "_build_packrr ()"
  int i;

  DPRINTF (("%s: Starting. lastrr=%d\n", ROUTINE, build->lastrr));

  for (i = 0; i < build->lastrr; i++) {

    if (build->rr[i].sectype != sectype) {
      DPRINTF (("%s: %d doesn't match pack section type %d\n",
		ROUTINE, build->rr[i].sectype, sectype));
      continue;
    }

    /* Compress it */
    if ((*cp_ptr = dns_compressrr (pkt,
				   *cp_ptr, build->rr[i].name, &build->rr[i],
				   cmp_ptr)) == NULL) {
      err_warn ("%s: dns_compressrr returned NULL", ROUTINE);
      return (1);
    }

  }				/* end for () */

  return 0;

#undef ROUTINE
}				/* end _build_packrr () */

/*****************************************************************************
 *****************************************************************************
 *			External Function Definitions
 *****************************************************************************
 *****************************************************************************/

/*
 * dns_buildpacket
 *
 * Build's a DNS packet, packing all information necessary into given
 * buffer
 *
 * Inputs:
 *    query - contains a valid DNS Query
 *    buildrec - filled in with data to return in packet.
 *    id -     Packet ID number (to uniquely identify this packet for replies)
 *    rcode -  Response code to add to packet (DE_XXX error code)
 *    flags -  Flag structure from the original query
 *    reply -  1 == build a response packet 0 == build a query packet
 *    aa -     1 == authoritive answer 0 == non-authoritive
 *
 * Outputs:
 *    pkt   - Fields filled in with the built packet.
 *    flags - Flag structure in response.
 *    query -  
 *
 * Return Values:
 *    0 == success
 *    1 == failure
 *    
 */
int
dns_buildpacket (packetrec * pkt, dnsqueryrec * query, dnsbuildrec * build,
		 guint16 id, guint16 rcode, dnsflagsrec flags, int reply,
		 int aa)

{
  dnscompressrec cmp = { 0 };
  guchar *cp;

#if 0
  DPRINTF (("dns_buildpacket: Starting. rcode=%s (%d)\n",
	    dns_de_to_string (rcode), rcode));
#endif

  /* Setup the flags first */

  /* Unused is per RFC, 0. We'll assume it's not truncated for now */
  flags.tc = flags.unused = 0;
  /* FIXME: Kludged until we can make recursion user configurable */
  flags.ra = 0;

  /* Query? Reply? */
  flags.qr = reply;
  /* Set the return code */
  flags.rcode = rcode;

  /* Check if we're authoritative over this zone */
  flags.aa = (aa) ? 1 : 0;

#if 0
  if (zone)
    flags.aa = (zone->flags & DNS_ZONE_AUTHORITATIVE) ? 1 : 0;
  else
    flags.aa = 0;
#endif

  /* Grab the RD, PR and opcode from the original query */
  flags.rd = flags.rd;
  flags.pr = flags.pr;
  flags.opcode = flags.opcode;

  /* Ok, let's start actually building the packet now */
  /* Pointers, mmm */
  cp = pkt->packetbuf;

  /* Let's put in the id */
  PUTINT16 (cp, id);

  /* Then the flags */
  memcpy (cp, (void *) &flags, sizeof (dnsflagsrec));
  cp += sizeof (dnsflagsrec);

  /* If we have a query pointer, then we're packing it */
  if (query != NULL)
    PUTINT16 (cp, 1);		/* Query count */
  else
    PUTINT16 (cp, 0);		/* Query count */

  /* If we have a build pointer, then we have RR's to pack */
  if (build != NULL) {
    PUTINT16 (cp, build->count[DNS_BUILD_ANSWER]);	/* Answer count */
    PUTINT16 (cp, build->count[DNS_BUILD_NS]);	/* NS count */
    PUTINT16 (cp, build->count[DNS_BUILD_ADDITIONAL]);	/* Additional count */
  } else {
    PUTINT16 (cp, 0);		/* Answer count */
    PUTINT16 (cp, 0);		/* NS count */
    PUTINT16 (cp, 0);		/* Additional count */
  }

  /* Compress the query back into the packet */
  if (query != NULL) {
    if ((cp = dns_compressquery (pkt, cp, query->name, query, &cmp)) == NULL) {
      err_warn ("dns_buildpacket: dns_compressquery returned NULL");
      return (1);
    }
  }

  /* Compress all of the RR's into the packet now */
  if (build != NULL) {
    /* Add all of the RR's */
    DPRINTF (("dns_buildpacket: Adding RR\n"));

    /* Do the Answer RR's first */
    DNS_BP_PACKRR (build, DNS_BUILD_ANSWER);

    /* And then the NS */
    DNS_BP_PACKRR (build, DNS_BUILD_NS);

    /* And then lastly the additional */
    DNS_BP_PACKRR (build, DNS_BUILD_ADDITIONAL);
  }

  /* Whee, using pointer arithmetic, figure out the size of the packet now */
  pkt->packetsize = cp - pkt->packetbuf;

  return (0);

}				/* end dns_buildpacket() */

/*
 * dns_build_add -  Adds an RR for building later
 *
 * This adds an RR entry to the dnsbuildrec for the specified section.
 * The RR's will later be built into a DNS packet using dns_buildpacket()
 *
 * The format of the 'data' section follows the RFC definition for each
 * RDATA type as specified by the 'type' parameter in the following manner:
 *
 *    address quantities are encoded as struct sin_addr values which 
 *       have already been converted into network byte order.
 *    Other integer quantities are encoded in binary in HOST byte order.  
 *    <domain-name> quantities are encoded as null terminated strings.
 *    <character-string> quantities are encoded as null terminated strings.
 *    <hex-string> quantities are encoded as ASCII strings e.g. "0x0123abcdef"
 *    
 * The conversion of fields into network byte order and compression of 
 * domain names is done in dns_buildpacket().
 *
 * e.g.  The 'data' field of a  DT_MX type record is defined in the RFC as:
 *
 *
 *   +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *   |                  PREFERENCE                   |
 *   +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *   /                   EXCHANGE                    /
 *   /                                               /
 *   +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *
 * PREFERENCE      A 16 bit integer which specifies the preference given to
 *                 this RR among others at the same owner.  Lower values
 *                 are preferred.
 *
 * EXCHANGE        A <domain-name> which specifies a host willing to act as
 *                 a mail exchange for the owner name.
 *
 * The first 2 bytes of the 'data' field would contain a 16 bit integer
 * in the byte-order of the host machine.  The 3rd byte would begin a
 * string containing the domain name to return. The last byte in the buffer
 * would be a NULL character to indicate the end of the string.
 *
 * Inputs:      build - a build record (passed into the module)
 *              sectype - section this RR should be added to.
 *                   A DNS_BUILD_XXX definition (ANSWER, AUTHORITY, 
 *                      or ADDITIONAL) 
 *              name - the domain-name portion of the RR to add
 *              ttl - time to live field of the RR to add
 *              class - the query class (DC_XXX definition)
 *              type - the RR type (DT_XXX definition)
 *              data - data to add to the RDATA field .  Note that 
 *                     this must be persistent data.  If you aren't storing
 *                     your RR's in memory, use dns_build_alloc_data() to
 *                     retrieve some buffer space.
 * Outputs:     None.
 * Side Effects: The RR is appended to the buildrecord struture
 * Returns:     None.
 *
 */
void
dns_build_add (dnsbuildrec * build, int sectype, guchar * name,
	       guint32 ttl, int class, int type, void *data)
{
  dnsbuildrrrec *buildrr;

  buildrr = &build->rr[build->lastrr++];

  build->count[sectype]++;

  buildrr->sectype = sectype;
  strncpyzt (buildrr->name, name, DNS_MAXNAMELENGTH);
  buildrr->ttl = ttl;
  buildrr->class = class;
  buildrr->type = type;
  buildrr->data = data;

}				/* end dns_build_add () */

/*
 * dns_build_addrr - Add a pre-built RR to the build record.
 *
 *  This function is a convenience/performance enhanced version of 
 *  dns_build_add().  If you are storing your RR data in binary format, 
 *  it is easy to use the dnsdbrrrec as your storage data structure 
 *  and just pass the data to this routine unmodified when it is queried.  
 * 
 *  This routine just copies the pointer to the rdata field, not the
 *  entire record. Later, all of the rr's will be compressed into a 
 *  packet in the correct order.   For this reason, the data of the 
 *  'rr' field  that you pass must be persistent data.
 *
 * Inputs:      build - a build record (passed into the module)
 *              sectype - section this RR should be added to.
 *                   A DNS_BUILD_XXX definition (ANSWER, AUTHORITY, 
 *                      or ADDITIONAL) 
 *              rr - a pre-build resource record to add.
 * Outputs:     None.
 * Side Effects: The RR is appended to the buildrecord struture
 * Returns:     None.
 * 
 */
void
dns_build_addrr (dnsbuildrec * build, int sectype, dnsdbrrrec * rr)
{
  dnsbuildrrrec *buildrr;

  buildrr = &build->rr[build->lastrr++];

  build->count[sectype]++;

  buildrr->sectype = sectype;
  dns_db_getname (rr->parent, buildrr->name, DNS_MAXNAMELENGTH);
  buildrr->ttl = rr->ttl;
  buildrr->class = rr->class;
  buildrr->type = rr->type;
  buildrr->data = rr->data;

}				/* end dns_build_addrr */

/*
 * dns_build_alloc - allocate a build record.
 * 
 * This allocates a build record from the list.  When you are
 * finished using this structure, call dns_build_delete()
 *
 * Inputs:   None
 * Outputs:  None
 * Returns: a valid buffer for a dnsbuildrec that has been reset to zero.
 * 
 */
dnsbuildrec *
dns_build_alloc (void)
{
  dnsbuildrec *build;

  build = (dnsbuildrec *) block_allocandadd (&gbl_buildrec_chain_pool);

  dns_build_reset (build);

  return build;

}				/* end dns_build_alloc () */

/*
 * dns_build_delete - dispose of a dnsbuildrec allocated with dns_build_alloc()
 *
 * Inputs:      build - a build record handle allocated with  dns_build_alloc()
 * Outputs:     None.
 * Returns:     None
 * Side Effects: The 'build' parameter is returned to the free heap
 *               of build records.
 */
void
dns_build_delete (dnsbuildrec * build)
{
  dns_build_reset (build);
  block_delete (&gbl_buildrec_chain_pool, &build->ll);

}				/* end dns_build_delete () */

/*
 * dns_build_alloc_data - retrieve a buffer to use to build RDATA  of an RR
 *
 * Inputs:       build - a build structure passed to the module as
 *                   a handle to hold query results.
 *               data_len - the # of bytes needed by the caller for
 *                   the rdata field.
 * Outputs:      None.  
 * Side Effects: This allocates more space for in the build record that
 *               will be persistent until dns_build_free() is called for
 *               this record.
 * Returns:      On success, a pointer to the beginning of a buffer 
 *               that will hold 'data_len' bytes worth of data.
 *               NULL on failure.
 */
guchar *
dns_build_alloc_data (dnsbuildrec * build, int data_len)
{
#define ROUTINE "dns_build_alloc_data ()"
  guchar *ret_val = &build->data_buf[build->bytes_alloced];

  /* Align the size that we are alloc'ing to the largest quantity
     the architecture supports to make sure the return values are
     properly aligned.
   */
  data_len += data_len % sizeof (void *);

  if (build->bytes_alloced + data_len > build->bytes_avail) {
    DPRINTF (
	     ("%s: not enough free space for %d bytes!\n", ROUTINE,
	      data_len));
    return NULL;
  }

  build->bytes_alloced += data_len;

  return ret_val;

#undef ROUTINE
}				/* end dns_build_alloc_data() */

/*
 * dns_build_reset
 *
 *  Clears a build structure. 
 *     - Removes the list of RRs added with dns_build_add ()
 *     - sets all counts to 0
 *     - inits data buffer fields used by dns_build_alloc_data ()
 *
 * Inputs:
 *     build - An unallocated dnsbuildrec structure.
 * Outputs:
 *     build - all values in the build record are zeroed out.
 *             (RR's in the buildrec are removed)
 * Returns:
 *     None.
 */
void
dns_build_reset (dnsbuildrec * build)
{
  build->lastrr = 0;
  build->count[DNS_BUILD_ANSWER] = build->count[DNS_BUILD_NS] =
    build->count[DNS_BUILD_ADDITIONAL] = 0;

  build->bytes_alloced = 0;
  build->bytes_avail = sizeof (build->data_buf);

}				/* end dns_build_reset () */

/*
 * dns_build_findans_and_pack
 *
 * Try's to find an answer for a name/class/type triplet in the database
 * specified and packs them into the build structure.
 *
 * Inputs:      db - root of the DNS database tree
 *              node - ???
 *              dnsqueryrec - query we are tyring to satisfy.
 * Outputs:     None
 * Returns:     DE_NOERROR on success
 * Side Effects: 'answer' RR's are added to the 'build' data structure.
 *
 */
int
dns_build_findans_and_pack (dbrootrec * db, dbnoderec * node,
			    dnsqueryrec * query, dnsbuildrec * build)
{
  dnsdbrrrec *rr;

  DPRINTF (("dns_build_findans_and_pack: finding info on %s\n", query->name));

  /* Loop through and find all of the RR's for this requested type */
  for (rr = NULL; (rr = dns_db_rr_findmatch (node, rr, query->class,
					     query->type)) != NULL;)
    /* Save the pointer for packing/compression later */
    dns_build_addrr (build, DNS_BUILD_ANSWER, rr);

  /* Did we find any info on that which was wanted? */
  if (build->count[DNS_BUILD_ANSWER])
    return (DE_NOERROR);

  /* Then search for CNAME's */
  for (rr = NULL;
       (rr = dns_db_rr_findmatch (node, rr, query->class, DT_CNAME)) != NULL;)
    /* Save the pointer for packing/compression later */
    dns_build_addrr (build, DNS_BUILD_ANSWER, rr);

  return (DE_NOERROR);
}

/*
 * dns_build_findauth_and_pack
 *
 * Find's authority information on a query and packs it into the build
 * structure.
 *
 * Inputs:      db - root of the DNS database tree
 *              node - ???
 *              dnsqueryrec - query we are tyring to satisfy.
 * Outputs:     None
 * Returns:     DE_NOERROR on success
 * Side Effects: 'answer' RR's are added to the 'build' data structure.
 * 
 */
void
dns_build_findauth_and_pack (dbrootrec * db, dbnoderec * node,
			     dnsqueryrec * query, dnsbuildrec * build)
{
  dnsdbrrrec *rr;

  DPRINTF (("dns_build_findauth_and_pack: finding auth info on %s\n",
	    query->name));

  /* If the original query was for NS records, then we skip this */
  if (query->type == DT_NS)
    return;

  /* Find some authority information on this domain name */
  /* FIXME: Don't duplicate NS' with what's in the answer section */
  for (;
       node && !(rr = dns_db_rr_findmatch (node, NULL, query->class, DT_NS));
       node = node->parent);

  /* Find anything? */
  if (!node)
    return;

  /* Add those in */
  for (rr = NULL; (rr = dns_db_rr_findmatch (node, rr, query->class, DT_NS))
       != NULL;)
    dns_build_addrr (build, DNS_BUILD_NS, rr);
}

/*
 * dns_build_findaddit_and_pack
 *
 *  Find's additional information and pack's the results in the build record.
 *
 * Inputs:      db - root of the DNS database tree
 *              node - ???
 *              dnsqueryrec - query we are tyring to satisfy.
 * Outputs:     None
 * Returns:     DE_NOERROR on success
 * Side Effects: 'answer' RR's are added to the 'build' data structure.
 * *
 */
void
dns_build_findaddit_and_pack (dbrootrec * db, dbnoderec * node,
			      dnsqueryrec * query, dnsbuildrec * build)
{
  int i, i1;
  dnsdbrrrec *rr;
  guchar dname[DNS_MAXNAMELENGTH + 1];
  unsigned long foo;
  u_int16_t bar[2];


  DPRINTF (
	   ("dns_build_findaddit_and_pack: finding additional information\n"));

  /* Loop through all of the records before this */
  for (i = 0; i < build->count[DNS_BUILD_ANSWER] + build->count[DNS_BUILD_NS];
       i++) {
    switch (build->rr[i].type) {
    case DT_MX:
      foo=(ulong)&bar[1]-(ulong)&bar[0];
      foo=foo+(ulong)build->rr[i].data;
      node = dns_db_find (db, foo);
      break;
    case DT_CNAME:
    case DT_NS:
      node = dns_db_find (db, build->rr[i].data);
      break;
    default:
      node = NULL;
      break;
    }

    /* If node is NULL, then either we didn't find any glue info or */
    /*  we shouldn't do any additional processing */
    if (!node)
      continue;

    dns_db_getname (node, dname, DNS_MAXNAMELENGTH);
    DPRINTF (("\ton %s", dname));

    /* Search through for A records on that node */
    rr = NULL;
    while ((rr = dns_db_rr_findmatch (node, rr, query->class, DT_A)) != NULL) {
      /* Make sure we don't duplicate data */
      for (i1 = 0; i1 < build->lastrr; i1++)
	if (!stricmp (build->rr[i1].name, dname) && (build->rr[i1].class ==
						     rr->class)
	    && (build->rr[i1].type == rr->type))
	  break;

      /* If we didn't find this one previously, then add it in */
      if (i1 >= build->lastrr) {
	DPRINTF ((" %s", inet_ntoa (*((struct in_addr *) rr->data))));
	dns_build_addrr (build, DNS_BUILD_ADDITIONAL, rr);
      }
    }

    DPRINTF (("\n"));
  }
}
