/* dns_debug.c - Helpful routines that help with debugging DENTS.

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team     (http://www.dents.org)
   Copyright (C) 1999  Eric Z. Ayers  (Eric.Ayers@mindspring.com)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"


/******************************************************************************
 *				Include Files
 *****************************************************************************/
/* #include <sys/socket.h>  requires sys/types.h? */
#include <netinet/in.h>
#include <arpa/inet.h>

#include "dns.h"
#include "dns_types.h"
#include "addr_tools.h"
/******************************************************************************
 *				Functions
 *****************************************************************************/

/*
 * dns_dt_to_string
 *
 * Description: Translates the RR type macro value into a string for 
 *                printing in a debugging statements
 *
 * Inputs:
 *      type - A RR type value (DT_XX)
 * Outputs:     
 *      None
 *
 * Returns:     
 *      A string representation of the value - e.g. "DT_XX"
 *
 */
char *
dns_dt_to_string (guint16 type)
{
  dnsrrtyperec *dnsrr;

  if (((dnsrr = dns_types_findrrtype_by_type (type)) != NULL)
      && (dnsrr->label != NULL))
    return (dnsrr->label);

  if (type == DT_ANY)
    return "DT_ANY";

  return ("?? Unknown Type ??");

}				/* end dns_dt_to_string () */

/*
 * dns_dc_to_string
 *
 * Description: Translates the Class macro value into a string for 
 *                printing in a debugging statements
 *
 * Inputs:
 *      type - A DNS Class value (DC_XX)
 * Outputs:     
 *      None
 *
 * Returns:     
 *      A string representation of the value - e.g. "DC_XX"
 *
 */
char *
dns_dc_to_string (guint16 class)
{
  switch (class) {
  case DC_IN:
    return "DC_IN";
  case DC_CS:
    return "DC_CS";
  case DC_CH:
    return "DC_CH";
  case DC_HS:
    return "DC_HS";
  case DC_ANY:
    return "DC_ANY";
  }

  return "DC_???";

}				/* end dns_dc_to_string () */

/*
 * dns_de_to_string
 *
 * Description: Translates the Error macroo value into a string for 
 *                printing in a debugging statements
 *
 * Inputs:
 *      error - a DE DNS error code
 * Outputs:     
 *      None
 *
 * Returns:     
 *      A string representation of the value - e.g. "DE_XXX"
 *
 */
char *
dns_de_to_string (int error)
{

  switch (error) {
  case DE_WAITFORMOREINFO:
    return "DE_WAITFORMOREINFO";
  case DE_NOINFO:
    return "DE_NOINFO";
  case DE_NOERROR:
    return "DE_NOERROR";
  case DE_FORMERR:
    return "DE_FORMERR";
  case DE_SERVFAIL:
    return "DE_SERVFAIL";
  case DE_NXDOMAIN:
    return "DE_NXDOMAIN";
  case DE_NOTIMP:
    return "DE_NOTIMP";
  case DE_REFUSED:
    return "DE_REFUSED";
  }

  return "DE_???";
}				/* end dns_de_to_string() */

/* 
 *   _debug_print_pkt_header 
 *
 * Work routine for  dns_debug_print_pkt () 
 *  prints just the packet header to stdout.
 *
 *  Inputs:     pkt - ptr to the start of the header section
 *
 *  Outputs:    qdcount - # of entries in the question section
 *              ancount - # of entries in the answer section
 *              nscount - # of entries in the authority section
 *              arcount - # of entries in the additional section
 */
static void _debug_print_pkt_header
  (guchar * pkt, int *qdcount, int *ancount, int *nscount, int *arcount)
{
  guint16 tmp_uint16;
  int opcode;
  int reserved_zero;
  int rcode;
  char *tmp_string;

  /* packet header layout from rfc 1035
     1  1  1  1  1  1
     0  1  2  3  4  5  6  7  8  9  0  1  2  3  4  5
     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
     |                      ID                       |
     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
     |QR|   Opcode  |AA|TC|RD|RA|   Z    |   RCODE   |
     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
     |                    QDCOUNT                    |
     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
     |                    ANCOUNT                    |
     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
     |                    NSCOUNT                    |
     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
     |                    ARCOUNT                    |
     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
   */

  /* the 16 bit ID for this query */
  memcpy (&tmp_uint16, pkt, 2);
  printf ("  ID:  %010d                    ", ntohs (tmp_uint16));

  /* a set of bit flags */
  memcpy (&tmp_uint16, pkt + 2, 2);
  tmp_uint16 = ntohs (tmp_uint16);

  printf ("  QR: %-20s\n", (tmp_uint16 & 0x0001) ? "Query" : "Response");

  opcode = (tmp_uint16 << 1) & 0x000F;

  switch (opcode) {
  case 0:
    tmp_string = "QUERY";
    break;
  case 1:
    tmp_string = "IQUERY";
    break;
  case 2:
    tmp_string = "STATUS";
    break;
  default:
    tmp_string = "Reserved";
    break;
  }

  printf ("  Opcode: %02d %-20s  ", opcode, tmp_string);

  printf ("  AA:     %01d  %-20s\n",
	  tmp_uint16 & 0x0010,
	  (tmp_uint16 & 0x0010) ? "Authoritive" : "Not Authorative");

  printf ("  TC:     %01d  %-20s  ",
	  tmp_uint16 & 0x0020,
	  (tmp_uint16 & 0x0020) ? "TRUNCATED" : "Not Truncated");

  printf ("  RD:     %01d  %-20s\n",
	  tmp_uint16 & 0x0040,
	  (tmp_uint16 & 0x0040) ? "Recursion Desired" : "No Recursion Des.");

  printf ("  RA:     %01d  %-20s  ",
	  tmp_uint16 & 0x0080,
	  (tmp_uint16 & 0x0080) ? "Recursion Available" :
	  "No Recursion Avail.");

  reserved_zero = (tmp_uint16 << 9) & 0x0007;

  printf ("  Z:      %d %-20s\n",
	  reserved_zero, reserved_zero ? "* expected zero *" : "(Reserved)");

  rcode = (tmp_uint16 << 12) & 0x000F;

  printf ("  RCODE:  %02d %s\n", rcode, dns_de_to_string (rcode));

  /* The 4 remaining fields in the header */
  memcpy (&tmp_uint16, pkt + 4, 2);
  *qdcount = ntohs (tmp_uint16);

  memcpy (&tmp_uint16, pkt + 6, 2);
  *ancount = ntohs (tmp_uint16);

  memcpy (&tmp_uint16, pkt + 8, 2);
  *nscount = ntohs (tmp_uint16);

  memcpy (&tmp_uint16, pkt + 10, 2);
  *arcount = ntohs (tmp_uint16);

  printf
    ("  QDCNT:  %02d entries in question     ANCNT:  %02d entries in answer\n",
     *qdcount, *ancount);

  printf
    ("  NSCNT:  %02d entries in authority    ARCNT:  %02d entries in additional\n",
     *nscount, *arcount);

}				/* end _debug_print_pkt_header () */

/*
 * _debug_print_pkt_domain_name - print out a text label.  
 *
 * Inputs:      pkt_start - ptr to the start of the packet (needed to decode pointers
 *              pkt - ptr to the start of the label
 * Outputs:     label_size - # of bytes in the label
 */
static void
_debug_print_pkt_domain_name (guchar * pkt_start, guchar * pkt,
			      int *label_size, int level)
{
#define MAX_UNPACK_NESTING 5

  guchar *p = pkt;

  *label_size = 0;

  if (level > MAX_UNPACK_NESTING) {
    printf (" * maximum nesting %d reached *\n", MAX_UNPACK_NESTING);
    return;
  }
  level++;

  while (*p != 0) {
    int i, len;

    /* the first octet is a length */
    len = *p;

    /* If the top 2 bytes are 1, then this is a 2 byte long pointer */
    /* 
       The pointer takes the form of a two octet sequence:

       +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
       | 1  1|                OFFSET                   |
       +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+

       The first two bits are ones.  This allows a pointer to be distinguished
       from a label, since the label must begin with two zero bits because
       labels are restricted to 63 octets or less.  (The 10 and 01 combinations
       are reserved for future use.)  The OFFSET field specifies an offset from
       the start of the message (i.e., the first octet of the ID field in the
       domain header).  A zero offset specifies the first byte of the ID field,
       etc.
     */

    if (len & 0xC0) {
      guint16 ptr_size;
      int dummy = 0;

      /* pull out the first 2 bits */
      memcpy (&ptr_size, p, 2);
      ptr_size = ntohs (ptr_size) & 0x3FFF;

      printf ("->(0x%X)", ptr_size);
      (*label_size) += 2;
      p += 2;
      _debug_print_pkt_domain_name (pkt_start, pkt_start + ptr_size,
				    &dummy, level);
      return;
    }

    (*label_size)++;
    p++;

    for (i = 0; i < len; i++) {
      putchar (*p);
      p++;
    }
    putchar ('.');

    (*label_size) += len;
  }
  p++;
  (*label_size) += 1;

#undef MAX_UNPACK_NESTING
}				/* end _debug_print_pkt_domain_name() */

/*
 * _debug_print_pkt_character_string
 *
 * Print a <character-string> type of data in the rdata section
 * a <character-string> is a single byte to tell the length followed the
 * string value.
 *
 * NOTE: jerdfelt wrote that for TXT records, there was some kind of 
 *       extension to this.  Multiple strings are embedded if the 
 *       total character count exceeds 255. (does this apply just to 
 *       TXT records??  what RFC??)
 *
 * Inputs:      pkt - start of character string
 * Outputs:     length_ptr - # of bytes in this character-string
 *
 */
static void
_debug_print_pkt_character_string (guchar * pkt, int *length_ptr)
{
  int len;
  int i;

  char *p = pkt;
  len = *p;
  p++;

  /* Note, this string is NOT null terminated!  we could copy it
     first and then print it out with printf() instead, but this was 
     too easy!
   */
  for (i = 0; i < len; i++, p++)
    putchar (*p);

  /* one byte for the length byte and the actual length of the string. */
  *length_ptr = len + 1;
}

/* 
 *  _debug_print_pkt_question - print the question section of the query
 *  
 * Inputs:      pkt - start of the question section
 *              qd_count - # of entries in the question section
 * Outputs:     question_size - # of bytes in the question section.
 *
 */
static void _debug_print_pkt_question
  (guchar * pkt, int qd_count, int *question_size)
{
  int label_len = 0;
  guchar *p;
  guint16 tmp_uint16;
  int i;

  *question_size = 0;

  if (0 == qd_count)
    printf ("  (Empty)\n");

  for (i = 0; i < qd_count; i++) {
    printf ("  QNAME: ");
    _debug_print_pkt_domain_name (pkt, pkt, &label_len, 0);
    printf ("\n");

    p = pkt + label_len;

    memcpy (&tmp_uint16, p, 2);
    tmp_uint16 = ntohs (tmp_uint16);

    printf ("  QTYPE: %05d %-20s", tmp_uint16, dns_dt_to_string (tmp_uint16));

    p += 2;
    memcpy (&tmp_uint16, p, 2);
    tmp_uint16 = ntohs (tmp_uint16);

    printf ("  QCLASS: %05d %-20s\n",
	    tmp_uint16, dns_dc_to_string (tmp_uint16));

    (*question_size) += label_len + 4;
  }

}				/* end _debug_print_pkt_question () */

/*
 * _debug_print_pkt_rr - print a single resource record
 *
 * Inputs:      pkt_start - ptr to the start of the packet
 *              pkt - ptr to the start of this RR in the packet
 * Outputs:     rr_size -returned size of this RR         
 *
 */
static void
_debug_print_pkt_rr (guchar * pkt_start, guchar * pkt, int *rr_size)
{
  int label_size = 0;
  guint16 type, class, rdlen;
  guint32 ttl;

  *rr_size = 0;

  printf ("  NAME:  ");
  _debug_print_pkt_domain_name (pkt_start, pkt, &label_size, 0);
  printf ("\n");
  *rr_size = label_size;

  memcpy (&type, pkt + *rr_size, 2);
  type = ntohs (type);
  printf ("  TYPE:  %02d %-20s   ", type, dns_dt_to_string (type));
  (*rr_size) += 2;

  memcpy (&class, pkt + *rr_size, 2);
  class = ntohs (class);
  printf ("  CLASS: %02d %-20s\n", class, dns_dc_to_string (class));
  (*rr_size) += 2;

  memcpy (&ttl, pkt + *rr_size, 4);
  ttl = ntohl (ttl);
  printf ("  TTL:   %09d                 ", ttl);
  (*rr_size) += 4;

  memcpy (&rdlen, pkt + *rr_size, 2);
  rdlen = ntohs (rdlen);
  printf ("  RDLEN: %05d\n", rdlen);
  (*rr_size) += 2;

  printf ("  RDATA: ");

  if (class != DC_IN) {
    printf ("  (Can't interpret RDATA of non IN class data)\n");
  } else {

    switch (type) {

      /* RDATA that contains a single 32 bit integer */
    case DT_A:
      {
	char ipv4addr[16];
	if (at_inet4_ntop (pkt + *rr_size, ipv4addr, sizeof (ipv4addr))) {
	  printf ("%s\n", ipv4addr);
	} else {
	  printf ("<problem printing addr>\n");
	}
      }
      break;
    case DT_AAAA:
      {
	char ipv6addr[40];
	if (at_inet6_ntop (pkt + *rr_size, ipv6addr, sizeof (ipv6addr))) {
	  printf ("%s\n", ipv6addr);
	} else {
	  printf ("<problem printing addr>\n");
	}

      }
      break;

      /* RDATA that contains a single <domain-name> */
    case DT_CNAME:
    case DT_NS:
    case DT_MD:
    case DT_MF:
    case DT_MB:
    case DT_MG:
    case DT_MR:
    case DT_PTR:
      _debug_print_pkt_domain_name (pkt_start, pkt + *rr_size, &label_size,
				    0);
      printf ("\n");
      break;

    case DT_NULL:
      printf ("  (see hex dump above)\n");
      break;

      /* RDATA that consists of 2 <character-strings> */
    case DT_HINFO:
    case DT_ISDN:
      _debug_print_pkt_character_string (pkt + *rr_size, &label_size);
      printf ("    ");
      _debug_print_pkt_character_string (pkt + *rr_size + label_size,
					 &label_size);
      printf ("\n");
      break;

      /* RDATA that consists of 2 <domain-names> */
    case DT_MINFO:
    case DT_RP:
      _debug_print_pkt_domain_name (pkt_start, pkt + *rr_size, &label_size,
				    0);
      _debug_print_pkt_domain_name (pkt_start, pkt + *rr_size + label_size,
				    &label_size, 0);
      break;

      /* RDATA that consists of a 16 bit int followed by a domain name */
    case DT_MX:
    case DT_AFSDB:
    case DT_RT:
      printf ("%d ", ntohs (*(guint16 *) (pkt + *rr_size)));
      _debug_print_pkt_domain_name (pkt_start, pkt + *rr_size + 2,
				    &label_size, 0);
      printf ("\n");
      break;

      /* RDATA that contains a single character string */
    case DT_TXT:
      _debug_print_pkt_character_string (pkt + *rr_size, &label_size);
      printf ("\n");
      break;

    case DT_SOA:
      {
	int offset = 0;
	guint32 serial;
	gint32 refresh, retry, expire;

	_debug_print_pkt_domain_name (pkt_start, pkt + *rr_size, &label_size,
				      0);
	printf ("   ");
	offset += label_size;

	_debug_print_pkt_domain_name (pkt_start, pkt + *rr_size + offset,
				      &label_size, 0);
	offset += label_size;
	printf ("\n");

	serial = ntohl (*(guint32 *) (pkt + *rr_size + offset));
	offset += 4;
	printf ("      SERIAL: %-10u ", serial);

	refresh = ntohl (*(gint32 *) (pkt + *rr_size + offset));
	offset += 4;
	printf ("  REFRESH: %-10d\n", refresh);

	retry = ntohl (*(gint32 *) (pkt + *rr_size + offset));
	offset += 4;
	printf ("      RETRY:  %-10d ", retry);

	expire = ntohl (*(gint32 *) (pkt + *rr_size + offset));
	offset += 4;
	printf ("  EXPIRE:  %-10d\n", expire);
	break;
      }

      /* RDATA that contains a single hex string */
    case DT_NSAP:
      {
	int i;
	for (i = 0; i < rdlen; i++) {

	  printf ("%02X", *(pkt + *rr_size + i));
	}
	printf ("\n");
      }
      break;
    case DT_WKS:
    default:
      printf (" (Printing RDATA of this type not implemented\n");
    }
  }

  (*rr_size) += rdlen;

  return;

}				/* end _debug_print_pkt_rr() */

/*
 * _debug_print_pkt_rrs - print all the RRs in one section
 *
 * Input:       pkt_start - ptr to the first byte of the pkt
 *              pkt - ptr to the start of the rrs section
 *              num_rrs - number of RRs found in this section
 * Output:      rrs_len - # of bytes found in this section.
 * 
 */
static void _debug_print_pkt_rrs
  (guchar * pkt_start, guchar * pkt, int num_rrs, int *rrs_len)
{
  int i;
  int rr_len = 0;

  *rrs_len = 0;

  for (i = 0; i < num_rrs; i++) {
    _debug_print_pkt_rr (pkt_start, pkt, &rr_len);
    *rrs_len += rr_len;
    pkt += rr_len;
  }
}				/* end _debug_print_pkt_rrs () */

/*
 * dns_debug_print_pkt - print out a packet in human readable format.
 *
 * See RFC 1035 section 4 for the description of the packet format.
 *
 *  Inputs:     pkt - ptr to the start of the raw packet buffer
 *              pktsize - size of the packet in bytes
 */
void
dns_debug_print_pkt (guchar * pkt, int pktsize)
{
  int qdcount = 0,		/* # of entries in question section */
    ancount = 0,		/* # of Resource Records (RR) in the answer section */
    nscount = 0,		/* # of RRs in the authority section */
    arcount = 0,		/* # of RRs in the additional section */
    qd_size = 0,		/* # of bytes in question section */
    an_size = 0,		/* # of bytes in answer section */
    ns_size = 0,		/* # of bytes in authority section */
    ar_size = 0,		/* # of bytes in additional section */
    offset = 0;

  printf ("Header\n");

  _debug_print_pkt_header (pkt, &qdcount, &ancount, &nscount, &arcount);

  /* The header is 12 bytes (fixed) - too bad they didn't put some kind
     of identifier in the DNS packet so that we could stop here if the header
     didn't match.
   */
  offset += 12;
  if (offset > pktsize) {
    printf ("  ***INVALID PACKET: %d has exceeded packet size of %d***\n",
	    offset, pktsize);
    return;
  }
  printf ("Question Section (offset 0x%X, %d entries)\n", offset, qdcount);
  _debug_print_pkt_question (pkt + offset, qdcount, &qd_size);

  offset += qd_size;
  if (offset > pktsize) {
    printf ("  ***INVALID PACKET: %d has exceeded packet size of %d***\n",
	    offset, pktsize);
    return;
  }
  printf ("Answer Section (offset 0x%X,  %d entries) \n", offset, ancount);
  _debug_print_pkt_rrs (pkt, pkt + offset, ancount, &an_size);

  offset += an_size;
  if (offset > pktsize) {
    printf ("  ***INVALID PACKET: %d has exceeded packet size of %d***\n",
	    offset, pktsize);
    return;
  }
  printf ("Authority Section (offset 0x%X, %d entries)\n", offset, nscount);
  _debug_print_pkt_rrs (pkt, pkt + offset, nscount, &ns_size);

  offset += ns_size;
  if (offset > pktsize) {
    printf ("  ***INVALID PACKET: %d has exceeded packet size of %d***\n",
	    offset, pktsize);
    return;
  }
  printf ("Additional Section (offset 0x%X, %d entries)\n", offset, arcount);
  _debug_print_pkt_rrs (pkt, pkt + offset, arcount, &ar_size);

}				/* end dns_debug_print_pkt () */

/*
 * dns_debug_init () - when dents is compiled w/ debugging, no functions
 * in this source module are referenced.  This means they are not linked
 * into the executable on linux, so I added this dummy -EZA
 */
int
dns_debug_init ()
{
  /* this is a dummy function */
  return 0;
}
