/* dns_types.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/*
 * dns_types.c
 *
 * Handles all dealings with the various DNS RR types. This includes both
 * conversion from and to text representation's and internal format
 *
 * See dns_types.h for a list of instructions on how to add a new
 * RR type to DENTS.
 *
 * TODO: An unpack that doesn't rely on having dnsdbrrec (for module
 *         writers other than stddb)
 *       Hide the type dnsrrtype and provide access routines? 
 *       Add support for the rest of the QTYPES
 */

/*****************************************************************************
 *				Include Files
 ****************************************************************************/

/* DENTS support headers */
#include "common.h"
#include "addr_tools.h"

/* DENTS DNS headers */
#include "dns.h"

/*****************************************************************************
 *				Macro Definitions
 ****************************************************************************/

/*
 * Quick macro, uses macro concatenation and stringification 
 * for building an array of type dnsrrtyperec below.
 */
#define RRTYPE(x, na)	{na, DT_##x, {DB_RB_NIL, DB_RB_NIL, DB_RB_NIL, DB_RB_BLACK, NULL}, #x}

#ifndef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif

#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

/*****************************************************************************
 *				Type Definitions
 ****************************************************************************/

/* Array that has Quick lookup for all of the different types */
static dnsrrtyperec rrtypes[] = {
  RRTYPE (A, 1),
  RRTYPE (AAAA, 1),
  RRTYPE (NS, 1),
  RRTYPE (PTR, 1),
  RRTYPE (CNAME, 1),
  RRTYPE (MX, 2),
  RRTYPE (HINFO, 2),
  RRTYPE (TXT, 1),
  RRTYPE (SOA, 7),
  RRTYPE (WKS, 0),
  RRTYPE (SRV, 0),
  RRTYPE (NULL, 0),
  RRTYPE (MINFO, 2),
  RRTYPE (RP, 2),
  RRTYPE (AFSDB, 2),
  RRTYPE (X25, 0),
  RRTYPE (ISDN, 2),
  RRTYPE (RT, 0),
  RRTYPE (NSAP, 1),
  RRTYPE (NSAP_PTR, 0),
  RRTYPE (SIG, 0),
  RRTYPE (KEY, 0),
  RRTYPE (PX, 0),
  RRTYPE (GPOS, 0),
  RRTYPE (LOC, 0),
  RRTYPE (NXT, 0),
  RRTYPE (EID, 0),
  RRTYPE (NIMLOC, 0),
  RRTYPE (ATMA, 0),
  RRTYPE (NAPTR, 0),
  RRTYPE (UINFO, 0),
  RRTYPE (UID, 0),
  RRTYPE (GID, 0),
  RRTYPE (UNSPEC, 0),
  RRTYPE (MD, -1),
  RRTYPE (MF, -1),
  RRTYPE (MB, -1),
  RRTYPE (MG, -1),
  RRTYPE (MR, -1),
};

/* We use a RB tree for the names (for speed/simplicity) */
static dbrb dns_types_namedb;
/* We use a simple array for the RR types (for speed/simplicity) */
static dnsrrtyperec *dns_types_idlist[255];

/*****************************************************************************
 *****************************************************************************
 *				Exported Functions
 *****************************************************************************
 ****************************************************************************/

/*
 * dns_types_unpack
 *
 * Takes the internal format we store everything in and converts it to
 * ASCII suitable for debugging and writing zone files
 *
 * Inputs:      rr - a resource record in the internal format.
 *              buf - buffer for ASCII represnetation of RR to unpack
 *              buflen - length of RDATA value to unpack
 * Outputs:     buf - buffer filled ASCII representation of unpacked RR
 *
 * Returns:     0 on success 
 *              -1 on failure
 */
int
dns_types_unpack (dnsdbrrrec * rr, guchar * buf, int buflen)
{
  guint16 tmp16;
  guint32 tmp32;
  /* unused ??
     guint8  tmpip[4];
     guint8  tmpip6[16]; */
  char tmpbuf[256];
  int len, bufpos;

  /* FIXME: This still isn't thread safe */
  /* This is used for caching from the previous call */
  static guchar dns_tu_prevname[DNS_MAXNAMELENGTH + 1];
  static dbnoderec *dns_tu_prevnode = NULL;

  /* We do a little bit of caching, it improves performance a little */
  if (rr->parent != dns_tu_prevnode) {
    dns_db_getname (rr->parent, buf, buflen);
    dns_tu_prevnode = rr->parent;
    strncpyzt (dns_tu_prevname, buf, DNS_MAXNAMELENGTH);
  } else
    strncpyzt (buf, dns_tu_prevname, buflen);

  strncatzt (buf, "\t", buflen);

  /* Throw in the class */
  switch (rr->class) {
  case DC_IN:
    strncatzt (buf, "IN\t", buflen);
    break;
  case DC_CS:
    strncatzt (buf, "CS\t", buflen);
    break;
  case DC_CH:
    strncatzt (buf, "CH\t", buflen);
    break;
  case DC_HS:
    strncatzt (buf, "HS\t", buflen);
    break;
  default:
    return (-1);
  }

  /* Throw in the type */
  switch (rr->type) {
  case DT_A:
    strncatzt (buf, "A\t", buflen);
    if (at_inet4_ntop (rr->data, (char *) &tmpbuf, sizeof (tmpbuf)) == NULL)
      return -1;
    strncatzt (buf, tmpbuf, buflen);
    break;
  case DT_AAAA:
    strncatzt (buf, "AAAA\t", buflen);
    if (at_inet6_ntop (rr->data, (char *) tmpbuf, sizeof (tmpbuf)) == NULL)
      return -1;
    strncatzt (buf, tmpbuf, buflen);
    break;
  case DT_PTR:
    strncatzt (buf, "PTR\t", buflen);
    strncatzt (buf, rr->data, buflen);
    break;
  case DT_CNAME:
    strncatzt (buf, "CNAME\t", buflen);
    strncatzt (buf, rr->data, buflen);
    break;
  case DT_TXT:
    strncatzt (buf, "TXT\t\"", buflen);
    strncatzt (buf, rr->data, buflen);
    strncatzt (buf, "\"", buflen);
    break;
  case DT_NS:
    strncatzt (buf, "NS\t", buflen);
    strncatzt (buf, rr->data, buflen);
    break;
  case DT_MX:
    strncatzt (buf, "MX\t", buflen);
    memcpy (&tmp16, rr->data, sizeof (tmp16));
    len = strlen (buf);
    g_snprintf (buf + len, buflen - len, "%d %s", tmp16,
		rr->data + sizeof (tmp16));
    break;

  case DT_SOA:
    strncatzt (buf, "SOA\t", buflen);
    strncatzt (buf, rr->data, buflen);
    bufpos = strlen (rr->data);
    strncatzt (buf, rr->data + bufpos, buflen);
    bufpos += strlen (rr->data + bufpos);

    memcpy (&tmp32, rr->data + bufpos, sizeof (tmp32));

    len = strlen (buf);
    g_snprintf (buf + len, buflen - len, "(\n\t%d", tmp32);
    bufpos += sizeof (tmp32);

    len = strlen (buf);
    g_snprintf (buf + len, buflen - len, "\n\t%d", tmp32);
    bufpos += sizeof (tmp32);

    len = strlen (buf);
    g_snprintf (buf + len, buflen - len, "\n\t%d", tmp32);
    bufpos += sizeof (tmp32);

    len = strlen (buf);
    g_snprintf (buf + len, buflen - len, "\n\t%d", tmp32);
    bufpos += sizeof (tmp32);

    len = strlen (buf);
    g_snprintf (buf + len, buflen - len, "\n\t%d )", tmp32);
    bufpos += sizeof (tmp32);
    break;
  case DT_AFSDB:
    strncatzt (buf, "AFSDB\t", buflen);
    memcpy (&tmp16, rr->data, sizeof (tmp16));
    len = strlen (buf);
    g_snprintf (buf + len, buflen - len, "%d %s", tmp16,
		rr->data + sizeof (tmp16));
    break;
  default:
    return (-1);
  }

  return (0);
}				/* end dns_types_unpack() */

/*
 * dns_types_pack
 *
 * Takes ASCII arguments and converts it to the internal format we store
 * everything in.
 *
 * Inputs:      rrtype - non-null return value from dns_types_findrrtype_XXX
 *              argv - an array of strings to convert
 *              numargs - number of arguments in argv[] array
 *              defaultdomain - domain name to use
 *              rrbuf - buffer with space to place 'packed' rdata value in 
 *                   internal format.
 * Outputs:     rrbuf - buffer is filled with 'packed' rdata value in 
 *                   internal format.
 * Returns:     number of bytes copied into rrbuf on success
 *              -1 on failure.
 *
 * Example:     Here is an example that packs an MX record for "foo.com." 
 *              (the line in a master file would read:
 *
 *              foo.com.     IN MX   10 smtp.foo.com.
 * 
 *              The parameters in this example are hard coded - you 
 *              obviously won't do this in real code! 
 *
 * {
 *   char * argv[2];
 *   guchar rdata_buf[DNS_MAXRDATALENGTH];
 *   dnsrrtyperec * rrtype;
 *
 *   rrtype = dns_types_findrrtype_by_label("MX");
 *   // OR rrtype = dns_types_findrrtype_by_type(DT_MX);
 *
 *   if (rrtype == NULL) {
 *      err_warn("OOPS: type MX is not supported by dents");
 *      return;
 *   }
 *
 *   // encode RDATA for an MX record
 *   argv[0]="10";
 *   argv[1]="smtp";
 *
 *   if (dns_types_pack(rrtype, argv, 2, "foo.com.", rdata_buf) != 0) {
 *      err_warn("OOPS: couldn't pack RDATA for RR type MX");
 *      return;
 *   }
 *
 *  // now rdata contains a value suitable for passing as 
 *  // the 'data' arg to dns_build_add()
 *  dns_build_add(build, DNS_BUILD_ANSWER, "myhost.foo.com", 3600, DC_IN,
 *      DT_MX, rdata_buf);
 * }
 *
 */
int
dns_types_pack (dnsrrtyperec * rrtype, guchar * argv[], int numargs,
		guchar * defaultdomain, guchar * rrbuf)
{
#define ROUTINE "dns_types_pack()"

  guint16 tmp16;
  guint32 tmp32;
  guint8 tmpip4[4];
  guint8 tmpip6[16];
  int buflen = 0, curarg = 0;

  assert (rrtype);

  /* Is this an obsoleted or unsupported type? */
  if (rrtype->numargs < 0) {
    err_warn ("%s: Obsoleted RR type %s", ROUTINE, rrtype->label);
    return -1;
  } else if (rrtype->numargs == 0) {
    err_warn ("%s: Unsupported RR type %s", ROUTINE, rrtype->label);
    return -1;
  } else if (rrtype->numargs != numargs) {
    err_warn
      ("%s: wrong number of arguments for RR type %s. Passed %d expected %d\n",
       ROUTINE, rrtype->label ? rrtype->label : "<UNKNOWN>", numargs,
       rrtype->numargs);
    return -1;
  }

  /*
   *  When entering this snippet of code, rrtype is pointing to the RR type.
   *  We have NOT checked any of the arguments to make sure they make
   *  sense.
   */

  /* In the case of mod_stddb - 
   *  We have already made sure that there are enough arguments for this RR
   *  type. 
   */

  /* FIXME - we should check all of the domain names to make sure they */
  /*  contain only legal characters */

  switch (rrtype->qtype) {
  case DT_A:
    if (at_inet4_pton (argv[curarg++], tmpip4) == NULL)
      return -1;
    memcpy (rrbuf, &tmpip4, sizeof (tmpip4));
    buflen = 4;
    break;

  case DT_AAAA:
    if (at_inet6_pton (argv[curarg++], tmpip6) == NULL)
      return -1;
    memcpy (rrbuf, &tmpip6, sizeof (tmpip6));
    buflen = 16;
    break;

    /* RR's with a single <domain-name> rdata element */
  case DT_NS:
  case DT_MD:
  case DT_MF:
  case DT_CNAME:
  case DT_MB:
  case DT_MG:
  case DT_MR:
  case DT_PTR:
    dns_qualify_domain (rrbuf, argv[curarg++], defaultdomain);
    buflen = strlen (rrbuf) + 1;
    break;

    /* RRs that consist of a single <character-string> */
    /* RR's that return a single <hex-string> */
  case DT_TXT:
  case DT_NSAP:
    buflen = strlen (argv[curarg]) + 1;
    memcpy (rrbuf, argv[curarg], buflen);
    curarg++;

    break;

    /* RR's that have 2 <character-string> fields in RDATA  */
  case DT_HINFO:
  case DT_ISDN:
    {
      int arg1len, arg2len;

      arg1len = strlen (argv[curarg]) + 1;
      memcpy (rrbuf, argv[curarg], arg1len);
      curarg++;

      arg2len = strlen (argv[curarg]) + 1;
      memcpy (rrbuf + arg1len, argv[curarg], arg2len);
      curarg++;

      buflen = arg1len + arg2len;
    }
    break;

    /* RR's that have 2 <domain-name> fields in RDATA */
  case DT_RP:
  case DT_MINFO:
    {
      int arg1len, arg2len;

      dns_qualify_domain (rrbuf, argv[curarg++], defaultdomain);
      arg1len = strlen (rrbuf) + 1;

      dns_qualify_domain (rrbuf + arg1len, argv[curarg++], defaultdomain);
      arg2len = strlen (rrbuf + arg1len) + 1;

      buflen = arg1len + arg2len;
    }

    break;

    /* RRs that contain a 16 bit integer followed by a <domain-name> */
  case DT_MX:

    /* FIXME - we should check this value for legitness */
    tmp16 = atoi (argv[curarg++]);
    memcpy (rrbuf, (void *) &tmp16, sizeof (guint16));
    buflen = sizeof (guint16);

    /* dns_qualify_domain limits it to DNS_MAXNAMELENGTH which is */
    /*  smaller than DNS_MAXRRDATALENGTH-2, so we're alright with */
    /*  buffer overflows */
    dns_qualify_domain (rrbuf + buflen, argv[curarg++], defaultdomain);
    buflen += strlen (rrbuf + buflen) + 1;
    break;

  case DT_SOA:
    dns_qualify_domain (rrbuf, argv[curarg++], defaultdomain);
    buflen = strlen (rrbuf) + 1;

    dns_qualify_domain (rrbuf + buflen, argv[curarg++], defaultdomain);
    buflen += strlen (rrbuf + buflen) + 1;

    /* FIXME - we should really use something like strtoul or something */
    /*  similar */
    /* FIXME - we should check these values to make sure they're sane */
    /* FIXME - we should handle 1D meaning 1 Day (86400 secs), etc */

    /* Serial number */
    tmp32 = atoi (argv[curarg++]);
    memcpy (rrbuf + buflen, (void *) &tmp32, sizeof (guint32));
    buflen += sizeof (guint32);

    /* Refresh time */
    tmp32 = atoi (argv[curarg++]);
    memcpy (rrbuf + buflen, (void *) &tmp32, sizeof (guint32));
    buflen += sizeof (guint32);

    /* Retry time */
    tmp32 = atoi (argv[curarg++]);
    memcpy (rrbuf + buflen, (void *) &tmp32, sizeof (guint32));
    buflen += sizeof (guint32);

    /* Expire time */
    tmp32 = atoi (argv[curarg++]);
    memcpy (rrbuf + buflen, (void *) &tmp32, sizeof (guint32));
    buflen += sizeof (guint32);

    /* Minimum time */
    tmp32 = atoi (argv[curarg++]);
    memcpy (rrbuf + buflen, (void *) &tmp32, sizeof (guint32));
    buflen += sizeof (guint32);
    break;
  default:
    DPRINTF (("%s: Type %d not supported (yet)\n", ROUTINE, rrtype->qtype));
    return (-1);
  }

  return (buflen);

#undef ROUTINE
}				/* end dns_types_pack() */

/*
 * dns_types_findclass
 *
 * Given an ASCII string, (such as "IN", "CH", "HS", and "CS") 
 * tries to determine the class (the Dents DC_XXX value). If it isn't a
 * known value, it will return 0.
 *
 */
guint8
dns_types_findclass (guchar * label)
{
  /* Don't need anything complex, there's only 4 possible classes */
  /*  and it's most likely IN anyway. So we order the checks for */
  /*  the common case (and hopefully speed) */

  if (!stricmp (label, "IN"))	/* Internet */
    return (DC_IN);

  if (!stricmp (label, "CH"))	/* Chaos */
    return (DC_CH);

  if (!stricmp (label, "HS"))	/* Hesiod */
    return (DC_HS);

  if (!stricmp (label, "CS"))	/* CSNET (Obsolete - rfc1035) */
    return (DC_CS);

  return (0);
}

/*
 * dns_types_findrrtype_by_label
 *
 * Given an ASCII label, will try to find the dnsrrtyperec for the
 * associated qtype
 *
 * Inputs:
 *     An ascii string representing a TYPE label for a Resource Record.
 *     e.g.  "A" for an address RR or "PTR" for a domain name pointer
 * Outputs: 
 *     None
 * Returns:
 *     non-null pointer to a dnsrrtyperec * for the specified type.
 *        Do not modify this value or the contents of the pointer!
 *     NULL on failure.
 *  
 */
dnsrrtyperec *
dns_types_findrrtype_by_label (guchar * label)
{
  return (db_rb_find (&dns_types_namedb, &label));
}

/*
 * dns_types_findrrtype_by_type
 *
 * Given a label, will try to find the dnsrrtyperec for the
 * associated qtype
 * 
 * Inputs: 
 *    type - a valid DT_XXX macro definition
 * Outputs:
 *    None
 * Returns:
 *     non-null pointer to a dnsrrtyperec * for the specified type.
 *        Do not modify this value or the contents of the pointer!
 *     NULL on failure.
 * 
 */
dnsrrtyperec *
dns_types_findrrtype_by_type (guint8 type)
{
  /* NOTE, we don't have to do a bounds check on 'type' because 
     dns_types_idlist is an array with 256 entries and we are 
     using an 8 bit index.) 
   */
  return (dns_types_idlist[type]);
}

/*
 * dns_types_init
 *
 *  Called by the server when first started up to generate RB tree for
 *  all types for fast lookup and an array to lookup by the integer type 
 *  index.
 */
void
dns_types_init (void)
{
  int numtypes = sizeof (rrtypes) / sizeof (dnsrrtyperec);
  int i;

  /* Initialize the rb tree for string lookups */
  db_rb_create (&dns_types_namedb, db_rb_cmp_pstri);

  /* Initialize gbl_rrtype_ref with a set of NULL pointers */
  memset (dns_types_idlist, 0, sizeof (dns_types_idlist));

  for (i = 0; i < numtypes; i++) {
    /* Insert into the rb tree for quick string lookups */
    db_rb_insert (&dns_types_namedb, &rrtypes[i].labelnode, &rrtypes[i]);

    /* Insert into the gbl_rrtype_ref for quick int lookups */
    dns_types_idlist[rrtypes[i].qtype] = &rrtypes[i];
  }

}				/* end dns_types_init() */

/*****************************************************************************
 *****************************************************************************
 * 			Convenience pack routines
 *                    For the 'causal' module writer
 *
 * The following routines pack all RRs in the same format as dns_types_pack()
 * above, but they are 'strongly typed' functions.  They also accept a
 * buildrec as input, so that you can just call this function to build
 * the RDATA portion and not worry about allocating memory.
 *
 * if 'buildrec' is populated, memory for rdata_ptr is allocated with
 * dns_build_data_alloc()
 *
 * if 'buildrec'is NULL and '*rdata_ptr' is NON-NULL, then memory is expected
 * to have already been allocated (or static.)  This can be tricky! 
 * the memory has to persist after your 'answer' function returns (it cannot
 * just be an automatic variable declared inside a function.)
 *
 * if 'buildrec' is NULL and '*rdata_ptr' is NULL, memory for rdata_ptr 
 * is allocated with malloc().  The caller is responsible for calling 'free()' 
 * on this memory
 *
 * default_domain - the domain name to append to unqualified domain names
 * in the RDATA section.  (that is,. if a domain-name does not end with a '.' 
 * character.) 
 * 
 *****************************************************************************
 *****************************************************************************/

/*
 * _pack_alloc - internal routine to alloc rdata for dns_types_pack_ routines
 *
 * Inputs: 
 *      buildrec - a handle to a dnsbuildrec structure for passing to
 *         dns_build_alloc().  This memory will persist until the packet
 *         is built and will be freed when the query/reply is complete.
 *         If you wish for the memory to persist after this, pass NULL
 *         for this value.
 *      rdata_ptr - Address of a pointer to place the rdata.
 *         The value of the pointer should be initialized to NULL if you
 *         wish this routine to dynamically allocate space with 
 *         dns_build_alloc() (if buildrec is passed) or malloc() (if buildrec
 *         is set to NULL)
 *      rdata_size - number of bytes to allocate.  Must be a value > 0 and
 *         less than DNS_MAXRRDATALENGTH.
 * Ouptuts:
 *      None.
 * Returns:
 *      NULL on failure 
 *          if rdata_size == 0 or rdata_size > DNS_MAXRRDATALENGTH.
 *          error allocating memory with malloc().
 *      non-null value pointing to allocated memory
 */
static guchar *
_pack_alloc (dnsbuildrec * buildrec, guchar * rdata_ptr, guint32 rdata_size)
{

  /* we should do SOME kind of range checking! */
  if (rdata_size == 0) {
    err_warn ("dns_types_pack_XXX: rdata_size is zero.\n");
    return NULL;
  }

  if (rdata_size > DNS_MAXRRDATALENGTH) {
    err_warn
      ("dns_types_pack_XXX: rdata_size is too big (%u bytes. Max %u)\n",
       rdata_size, DNS_MAXRRDATALENGTH);
    return NULL;
  }

  if (buildrec)
    return dns_build_alloc_data (buildrec, rdata_size);

  if (rdata_ptr)
    return rdata_ptr;

  return (guchar *) malloc (rdata_size);

}				/* end _pack_alloc() */

/*
 * dns_types_pack_A - pack the RDATA for a DT_A type RR
 * 
 */
int
dns_types_pack_A (dnsbuildrec * buildrec, guchar ** rdata_ptr,
		  guchar * ipv4_addr)
{
  int rdata_size = 4;
  guchar *string_args[1];

  assert (ipv4_addr);
  assert (rdata_ptr);

  *rdata_ptr = _pack_alloc (buildrec, *rdata_ptr, rdata_size);

  if (!(*rdata_ptr)) {
    return -1;
  }

  string_args[0] = ipv4_addr;
  return dns_types_pack (dns_types_findrrtype_by_type (DT_A), string_args, 1,
			 "", *rdata_ptr);

  return rdata_size;

}				/* end dns_types_pack_A() */

/*
 * dns_types_pack_AAAA - pack the RDATA for a DT_AAAA type RR
 * 
 */
int
dns_types_pack_AAAA (dnsbuildrec * buildrec, guchar ** rdata_ptr,
		     guchar * ipv6_addr)
{
  int rdata_size = 16;
  guchar *string_args[1];

  assert (ipv6_addr);
  assert (rdata_ptr);

  *rdata_ptr = _pack_alloc (buildrec, *rdata_ptr, rdata_size);

  if (!(*rdata_ptr)) {
    return -1;
  }

  string_args[0] = ipv6_addr;
  return dns_types_pack (dns_types_findrrtype_by_type (DT_AAAA), string_args,
			 1, "", *rdata_ptr);

  return rdata_size;

}				/* end dns_types_pack_AAAA() */

/*
 * dns_types_pack_1_domain_name - pack the RDATA for an RR that takes 1 domain
 * 
 */
int
dns_types_pack_1_domain_name (dnsbuildrec * buildrec, guchar ** rdata_ptr,
			      guchar * default_domain, guchar * domain_name)
{
  int rdata_size = 0;
  guchar tmp_domain[DNS_MAXNAMELENGTH];

  assert (rdata_ptr);

  dns_qualify_domain (tmp_domain, domain_name, default_domain);
  rdata_size = strlen (tmp_domain) + 1;
  *rdata_ptr = _pack_alloc (buildrec, *rdata_ptr, rdata_size);
  if (!(*rdata_ptr)) {
    return -1;
  }

  memcpy (*rdata_ptr, tmp_domain, rdata_size);

  return rdata_size;
}				/* end dns_types_pack_1_domain_name() */

/*
 * dns_types_pack_2_domain_names - pack RDATA for an RR that takes 2 domains
 * 
 */
int
dns_types_pack_2_domain_names (dnsbuildrec * buildrec,
			       guchar ** rdata_ptr,
			       guchar * default_domain,
			       guchar * domain_name1, guchar * domain_name2)
{
  int rdata_size = 0, domain1_len, domain2_len;
  guchar tmp_domain1[DNS_MAXNAMELENGTH], tmp_domain2[DNS_MAXNAMELENGTH];

  assert (rdata_ptr);
  assert (default_domain);
  assert (domain_name1);
  assert (domain_name2);

  dns_qualify_domain (tmp_domain1, domain_name1, default_domain);
  dns_qualify_domain (tmp_domain2, domain_name2, default_domain);
  domain1_len = strlen (tmp_domain1) + 1;
  domain2_len = strlen (tmp_domain2) + 1;
  rdata_size = domain1_len + domain2_len;
  *rdata_ptr = _pack_alloc (buildrec, *rdata_ptr, rdata_size);
  if (!(*rdata_ptr)) {
    return -1;
  }

  memcpy (*rdata_ptr, tmp_domain1, domain1_len);
  memcpy ((*rdata_ptr) + domain1_len, tmp_domain2, domain2_len);

  return rdata_size;
}				/* end dns_types_pack_2_domain_names() */

/*
 * dns_types_pack_1_character_string - pack RDATA for an RR that takes 1 string
 *
 */
int
dns_types_pack_1_character_string (dnsbuildrec * buildrec,
				   guchar ** rdata_ptr,
				   guchar * character_string)
{
  int rdata_size = 0;

  assert (rdata_ptr);
  assert (character_string);

  rdata_size = MAX (strlen (character_string) + 1, DNS_MAXNAMELENGTH);
  *rdata_ptr = _pack_alloc (buildrec, *rdata_ptr, rdata_size);
  if (!(*rdata_ptr)) {
    return -1;
  }

  strncpyzt (*rdata_ptr, character_string, rdata_size);

  return rdata_size;

}				/* end dns_types_pack_1_character_string () */

/*
 * dns_types_pack_2_character_strings - pack RDATA for RRs that take 2 strings
 *
 */
int
dns_types_pack_2_character_strings (dnsbuildrec * buildrec,
				    guchar ** rdata_ptr,
				    guchar * character_string1,
				    guchar * character_string2)
{
  int rdata_size = 0, string1_len, string2_len;

  assert (rdata_ptr);
  assert (character_string1);
  assert (character_string2);

  string1_len = MIN (strlen (character_string1) + 1, DNS_MAXNAMELENGTH);
  string2_len = MIN (strlen (character_string2) + 1, DNS_MAXNAMELENGTH);
  rdata_size = string1_len + string2_len;
  *rdata_ptr = _pack_alloc (buildrec, *rdata_ptr, rdata_size);
  if (!(*rdata_ptr)) {
    return -1;
  }

  strncpyzt (*rdata_ptr, character_string1, string1_len);
  strncpyzt ((*rdata_ptr) + string1_len, character_string2, string2_len);

  return rdata_size;
}				/* end dns_types_pack_2_character_strings() */

/*
 * dns_types_pack_uint16_domain - pack RDATA for RRs with an int & domain name
 */
int
dns_types_pack_uint16_domain (dnsbuildrec * buildrec, guchar ** rdata_ptr,
			      guint16 int_arg,
			      guchar * default_domain, guchar * domain_name)
{
  int rdata_size, domain_len;
  guchar tmp_domain[DNS_MAXNAMELENGTH];

  dns_qualify_domain (tmp_domain, domain_name, default_domain);
  domain_len = strlen (tmp_domain) + 1;
  rdata_size = domain_len + sizeof (guint16);
  *rdata_ptr = _pack_alloc (buildrec, *rdata_ptr, rdata_size);
  if (!(*rdata_ptr)) {
    return -1;
  }

  memcpy ((*rdata_ptr), &int_arg, sizeof (guint16));
  memcpy ((*rdata_ptr) + sizeof (guint16), tmp_domain, domain_len);

  return rdata_size;
}				/* end dns_types_pack_uint16_domain() */

/*
 * dns_types_pack_SOA - pack rdata for the SOA RR
 */
int
dns_types_pack_SOA (dnsbuildrec * buildrec, guchar ** rdata_ptr,
		    guchar * default_domain, guchar * mname,
		    guchar * rname, guint32 serial,
		    gint32 refresh, gint32 retry, gint32 expire,
		    guint32 minimum)
{
  int rdata_size = 0, domain1_len, domain2_len, offset;
  guchar tmp_domain1[DNS_MAXNAMELENGTH], tmp_domain2[DNS_MAXNAMELENGTH];

  assert (rdata_ptr);
  assert (default_domain);
  assert (rname);
  assert (mname);

  dns_qualify_domain (tmp_domain1, mname, default_domain);
  dns_qualify_domain (tmp_domain2, rname, default_domain);
  domain1_len = strlen (tmp_domain1) + 1;
  domain2_len = strlen (tmp_domain2) + 1;
  rdata_size = domain1_len + domain2_len + sizeof (guint32) * 5;
  *rdata_ptr = _pack_alloc (buildrec, *rdata_ptr, rdata_size);
  if (!(*rdata_ptr)) {
    return -1;
  }

  offset = 0;
  memcpy ((*rdata_ptr) + offset, tmp_domain1, domain1_len);
  offset += domain1_len;
  memcpy ((*rdata_ptr) + offset, tmp_domain2, domain2_len);
  offset += domain2_len;
  memcpy ((*rdata_ptr) + offset, &serial, sizeof (guint32));
  offset += sizeof (guint32);
  memcpy ((*rdata_ptr) + offset, &refresh, sizeof (gint32));
  offset += sizeof (guint32);
  memcpy ((*rdata_ptr) + offset, &retry, sizeof (gint32));
  offset += sizeof (guint32);
  memcpy ((*rdata_ptr) + offset, &expire, sizeof (gint32));
  offset += sizeof (guint32);
  memcpy ((*rdata_ptr) + offset, &minimum, sizeof (guint32));

  return rdata_size;

}				/* end dns_types_pack_SOA() */
