/* err.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif



/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/*
 * err.c
 *
 *  Whenever an error occurs, we call one of these functions. They handle
 * syslogging, exiting, etc.
 *
 * FIXME: we should allow configurable syslogging, etc. ala BIND 8.1
 *        (there is a command line option to specify which facility
 *         to log to.  Is that what you meant? -EZA)
 * 
 */
#include <stdarg.h>
#include <stdio.h>
#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif /* HAVE_SYSLOG_H */
#include <sys/types.h>
#include "common.h"

#define MAX_FORMATLENGTH 4096

/*
 * err_fatal
 *
 * Generates a fatal error. Something which cannot be corrected. It will
 * syslog, etc the message then exit out.  If the server is compiled with
 * debugging on, the error will be printed to stdout.
 *
 * Inputs:
 *    format - printf() style format string
 *    ...    - arguments passed in the same manner as printf()
 * Outputs: None.
 */
void
err_fatal (char *format, ...)
{
  va_list ap;
  guchar fmt[MAX_FORMATLENGTH + 1];

  va_start (ap, format);

  g_snprintf ((char *)fmt, MAX_FORMATLENGTH, "FATAL: %s", format);
#ifndef DENTS_DEBUG
  vsyslog (LOG_INFO, fmt, ap);
#else
  printf ("err_fatal: ");
  vprintf (fmt, ap);
  printf ("\n");
#endif

  va_end (ap);

  fflush (stdout);
  fflush (stderr);
#ifdef DENTS_DEBUG
  printf ("abort()'ing\n");
  abort ();
#else
  exit (1);
#endif
}

/*
 * err_warn
 *
 * Generates a warning. If there's a minor error in a zone file, someone
 * trying to attack the system using one of the numerous BIND bugs.  These
 * errors go to UNIX syslog at the WARNING level under normal operation
 * (to stderr if the server is compiled with debugging on)
 *
 * Inputs:
 *    format - printf() style format string
 *    ...    - arguments passed in the same manner as printf()
 * Outputs: None.
 * 
 */
void
err_warn (char *format, ...)
{
  va_list ap;
  guchar fmt[MAX_FORMATLENGTH + 1];

  va_start (ap, format);

  g_snprintf ((char *)fmt, MAX_FORMATLENGTH, "Warning: %s", format);
#ifndef DENTS_DEBUG
  vsyslog (LOG_WARNING, fmt, ap);
#else
  printf ("err_warn: ");
  vprintf (fmt, ap);
  printf ("\n");
#endif

  va_end (ap);
}

/*
 * err_info
 *
 * Generates an informational message. If there's multiple queries in a 
 * packet, etc.  These errors go to UNIX syslog at the INFO level under
 * normal operation (to stderr if the server is compiled with debugging on.)
 *
 * Inputs:
 *    format - printf() style format string
 *    ...    - arguments passed in the same manner as printf()
 * Outputs: None.
 *
 */
void
err_info (char *format, ...)
{
  va_list ap;

  va_start (ap, format);

#ifndef DENTS_DEBUG
  vsyslog (LOG_INFO, format, ap);
#else
  printf ("err_info: ");
  vprintf (format, ap);
  printf ("\n");
#endif

  va_end (ap);
}
