/* link.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/*
 * link.c
 *
 *  Implements some generic routines for managing linked list's. Also
 * implements a generic block allocator for performance
 *
 * FIXME: Finer locking?
 * 
 * See link.h for function prototypes.
 *
 * See test driver in ../test/test_link.c
 *
 */
#include "common.h"
#include "link.h"

/*
 * link_add
 *
 *  Add's to a linked list
 */
void
link_add (linkedlist * link, llentry * entry)
{
  LOCK ("link_add", link->lock);

  /* Setup the previous and the next in the list */
  entry->prev = link->end;
  entry->next = NULL;

  /* Are we first for this list? */
  if (!link->begin)
    link->begin = entry;

  /* Then add ourselves at the end */
  if (link->end)
    ((llentry *) link->end)->next = entry;

  link->end = entry;

  /* Increment the usage */
  link->num++;

  UNLOCK ("link_add", link->lock);
}

/*
 * block_allocatemorefree
 *
 *  Allocate's some more memory for the linked list
 */
static void
block_allocatemorefree (blocklist * block)
{
  guchar *p;
  guint16 i;

  if (!block->numalloc) {
    err_warn
      ("block_allocatemorefree: block @ %p has 0 numalloc, forcing to 1",
       block);
    block->numalloc = 1;
  }

  /* Allocate some memory */
  if ((p = (guchar *) malloc (block->structsize * block->numalloc)) == NULL)
    err_fatal ("block_allocatemorefree: malloc: %m");

  /* Clear them all out */
  memset (p, 0, (block->structsize * block->numalloc));

  /* Go through all of them adding it to the list */
  for (i = 0; i < block->numalloc; i++, p += block->structsize)
    link_add (&block->free, (llentry *) p);
}

/*
 * block_alloc
 *
 *  Allocate's a structure
 */
static void *
block_alloc (blocklist * block)
{
  llentry *e;

  LOCK ("block_alloc", block->lock);

  /* Do we have any free? */
  if (!block->free.num)
    block_allocatemorefree (block);

  /* Grab a structure */
  e = block->free.begin;
  block->free.begin = e->next;

  /* Update the counters */
  block->free.num--;

  if (!block->free.num)
    block->free.end = block->free.begin = NULL;

  UNLOCK ("block_alloc", block->lock);

  return (e);
}

/*
 * block_allocandaddtolink
 *
 *  Allocate's a structure (if needed) and then add's it to the linked list
 */
void *
block_allocandaddtolink (blocklist * block, linkedlist * link)
{
  llentry *e;

  /* Allocate the structure */
  e = block_alloc (block);

  /* Add it */
  link_add (link, e);

  return (e);
}

/*
 * link_delete
 *
 *  Delete an entry from the linked list
 */
void
link_delete (linkedlist * link, llentry * entry)
{
  LOCK ("link_delete", link->lock);

  /* Close the gap */
  if (entry->prev)
    entry->prev->next = entry->next;
  else
    link->begin = entry->next;

  if (entry->next)
    entry->next->prev = entry->prev;
  else
    link->end = entry->prev;

  link->num--;

  if (!link->num)
    link->begin = link->end = NULL;

  UNLOCK ("link_delete", link->lock);
}

/*
 * block_delete
 *
 *  Returns a block back from the used pool to the free pool
 */
void
block_delete (blocklist * block, llentry * entry)
{
  /* Delete it */
  link_delete (&block->used, entry);

  /* And throw this entry into the free list */
  link_add (&block->free, entry);
}

/*
 * link_reset
 *
 *  Reset the linked list. This is dangerous to do on a list being used
 * as all of the entries will be leaked out
 */
void
link_reset (linkedlist * link, guint16 offset)
{
#ifdef WITH_PTREAD
  pthread_mutex_init (&link->lock, NULL);
#endif

  link->begin = link->end = NULL;
  link->num = 0;
  link->offset = offset;
}

/*
 * block_setup
 *
 *  Set's up a linked list. Allocates some memory, sets up the rest of the
 * variables
 */
void
block_setup (blocklist * block, guint16 structsize, guint16 numalloc)
{
#ifdef WITH_PTREAD
  pthread_mutex_init (&block->lock, NULL);
#endif

  /* Setup the allocation information.  Always alloc to allign this
     structure to an even boundry (for architectures like the Alpha)
   */
  block->structsize = structsize + structsize % sizeof (void *);
  block->numalloc = numalloc;

  /* Reset the linked list's */
  link_reset (&block->used, 0);
  link_reset (&block->free, 0);
}				/* end block_setup() */
