/* test_addr_tools.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/*
 * test_addr_tools.c - test suite for the addr_tools.c support file
 * 
 */

/*****************************************************************************
 *				Include Files
 *****************************************************************************/

#include <stdio.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "common.h"
#include "addr_tools.h"

/*****************************************************************************
 *				Type Definitions
 *****************************************************************************/
struct inet4_pton_test_str
{
  guchar input[30];		/* only 16 bytes are needed for legit addrs.
				   Added some extra for negative tests */
  guchar result[4];
};

struct inet4_ntop_test_str
{
  guchar input[4];
  guchar result[16];
};

struct inet6_pton_test_str
{
  guchar input[48];		/* only 40 bytes are needed for legit addrs.
				   Added some extra for negative tests */
  guchar result[16];
};

struct inet6_ntop_test_str
{
  guchar input[16];
  guchar result[40];
};

/*****************************************************************************
 *				Global Data
 *****************************************************************************/

/* tests for at_inet4_pton() these should all convert sucessfully */
static struct inet4_pton_test_str inet4_pton_tests_good[] = {
  {"255.255.255.255", {255, 255, 255, 255}},
  {"0.0.0.0", {0, 0, 0, 0}},
  {"123.45.67.89", {123, 45, 67, 89}},
  {"98.76.54.32", {98, 76, 54, 32}},
};

/* tests for at_inet4_pton() these should all fail conversion */
static struct inet4_pton_test_str inet4_pton_tests_bad[] = {
  {"255.255.255", ""},		/*too few octets     */
  {"255.255.255.255.255", ""},	/*too many octets    */
  {"256.255.255.255", ""},	/*octet out of range */
  {"255.256.255.255", ""},	/*octet out of range */
  {"255.255.256.255", ""},	/*octet out of range */
  {"255.255.255.256", ""},	/*octet out of range */
  {"-1.0.0.0", ""},		/*octet out of range */
  {"0.-1.0.0", ""},		/*octet out of range */
  {"0.0.-1.0", ""},		/*octet out of range */
  {"0.0.0.-1", ""},		/*octet out of range */
  {"1234.0.0.0", ""},		/*octet out of range */
  {"0.1234.0.0", ""},		/*octet out of range */
  {"0.1234.0.0", ""},		/*octet out of range */
  {"0.0.0.1234", ""},		/*octet out of range */
  {"301.0.0.0", ""},		/*octet out of range */
  {"0.301.0.0", ""},		/*octet out of range */
  {"0.0.301.0", ""},		/*octet out of range */
  {"0.0.0.301", ""},		/*octet out of range */
  {"98,76,54,32", ""},		/*invalid separator  */
  {"98.O.45.67", ""},		/*invalid character  */
  {"25.25.25.25-1", ""},	/*a bizzare syntax!  */
  {"fe25.25.25.25", ""},	/*invalid characters  */
  {"25.fe25.25.25", ""},	/*invalid characters  */
  {"25.25.fe25.25", ""},	/*invalid characters  */
  {"25.25.25.fe25", ""},	/*invalid characters  */
  {".25.25.25", ""},		/*blank octets        */
  {"25..25.25", ""},		/*blank octets        */
  {"25.25..25", ""},		/*blank octets        */
  {"25..25.25..25", ""},	/*too many separators */
  {"25.25..25.25", ""},		/*too many separators */
  {"25.25.25..25", ""},		/*too many separators */

  {"25.25.25.25.25", ""},	/*too many octets    */

};

static struct inet4_ntop_test_str inet4_ntop_tests[] = {
  {{255, 255, 255, 255}, "255.255.255.255"},
  {{0, 0, 0, 0}, "0.0.0.0"},
  {{1, 1, 1, 1}, "1.1.1.1"},
  {{123, 45, 67, 89}, "123.45.67.89"},
  {{98, 76, 54, 32}, "98.76.54.32"}
};

static struct inet6_pton_test_str inet6_pton_tests_good[] = {
  {"FEDC:BA98:7654:3210:FEDC:BA98:7654:3210",
   {0xFE, 0xDC, 0xBA, 0X98, 0x76, 0x54, 0x32, 0x10,
    0xFE, 0xDC, 0xBA, 0x98, 0x76, 0x54, 0x32, 0x10}},

  {"1080:0:0:0:8:800:200C:417A",
   {0x10, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x08, 0x08, 0x00, 0x20, 0x0C, 0x41, 0x7A}},

  {"FF01:0:0:0:0:0:0:101",
   {0xFF, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01}},

  {"0:0:0:0:0:0:0:1",
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01}},

  {"0:0:0:0:0:0:0:0",
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}},

  {"1080::8:800:200C:417A",
   {0x10, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x08, 0x08, 0x00, 0x20, 0x0C, 0x41, 0x7A}},

  {"FF01::101",
   {0xFF, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01}},

  {"::1",
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01}},

  {"::",
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}},

  {"FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF",
   {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}},

};

/* Tests for at_inet6_pton(). The following tests should fail */
static struct inet6_pton_test_str inet6_pton_tests_bad[] = {

  /* too few bits */
  {"FEDC:BA98:7654:3210:FEDC:BA98:7654", ""},

  /* too few bits */
  {"", ""},

  /* too many :'s */
  {":::", ""},

  /* too many instances of :: */
  {"FEDC:BA98::3210:FEDC::BA98:7654", ""},

  /* Invalid Character */
  {"FEDC:BA98:7654:321G:FEDC:BA98:7654:3210", ""},

  /* Imbedded white space */
  {"FEDC:BA98:7654:321G FEDC:BA98:7654:3210", ""},

  /* too many bits */
  {"FEDC:BA98:7654:3210:FEDC:BA98:7654:3210:1234", ""},

  /* leading whitespace */
  {" EDC:BA98:7654:3210:FEDC:BA98:7654:3210", ""},

  /* trailing whitespace */
  {"FEDC:BA98:7654:3210:FEDC:BA98:7654:321 ", ""},

  /* too many bits */
  {"F:E:D:3210:FEDCBA98:BA98:7654:3210", ""},

  /* too many bits */
  {"FEDC:3:4:3210:FEDC:BA98:7654:3210:1234", ""},

};

static struct inet6_ntop_test_str inet6_ntop_tests[] = {
  {{0xFE, 0xDC, 0xBA, 0X98, 0x76, 0x54, 0x32, 0x10,
    0xFE, 0xDC, 0xBA, 0x98, 0x76, 0x54, 0x32, 0x10},
   "FEDC:BA98:7654:3210:FEDC:BA98:7654:3210"},

  {{0x10, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x08, 0x08, 0x00, 0x20, 0x0C, 0x41, 0x7A},
   "1080::8:800:200C:417A"},

  {{0xFF, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01},
   "FF01::101"},

  {
   {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01},
   "::1"},

  {{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
   "::"},

  {{0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF},
   "FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF"},
};

/*****************************************************************************
 *				Function Definitions
 *****************************************************************************/

int
main (int argc, char *argv[])
{
  guchar inet6_string[40], inet4_string[16];
  guchar *char_result;
  guint8 inet6_addr[16], inet4_addr[4];
  guint8 *int8_result;
  int i, num_tests;
  int error_count = 0;

  /* test at_inet4_pton() for good input values */
  num_tests =
    sizeof (inet4_pton_tests_good) / sizeof (struct inet4_pton_test_str);

  for (i = 0; i < num_tests; i++) {
    memset (inet4_addr, 0, sizeof (inet4_addr));;
    int8_result = at_inet4_pton (inet4_pton_tests_good[i].input, inet4_addr);
    if (!int8_result) {
      fprintf (stderr,
	       "FATAL: at_inet4_pton() Failed to convert inet4 address %s\n",
	       inet4_pton_tests_good[i].input);
      error_count++;
      continue;
    }

    if (memcmp (int8_result, inet4_pton_tests_good[i].result, 4)) {
      fprintf (stderr,
	       "FATAL: at_inet4_pton() Conversion of %s did not match expected results\n",
	       inet4_pton_tests_good[i].input);
      error_count++;
      continue;
    }
  }

  /* test at_inet4_pton() for bad input values */
  num_tests =
    sizeof (inet4_pton_tests_bad) / sizeof (struct inet4_pton_test_str);
  for (i = 0; i < num_tests; i++) {
    memset (inet4_addr, 0, sizeof (inet4_addr));;
    int8_result = at_inet4_pton (inet4_pton_tests_bad[i].input, inet4_addr);
    if (int8_result) {
      fprintf (stderr,
	       "FATAL: at_inet4_pton() Converted invalid inet4 address %s w/o error\n",
	       inet4_pton_tests_bad[i].input);
      error_count++;
      continue;
    }
  }

  /* test at_inet4_ntop() */
  num_tests = sizeof (inet4_ntop_tests) / sizeof (struct inet4_ntop_test_str);

  for (i = 0; i < num_tests; i++) {
    memset (inet4_string, 0, sizeof (inet4_string));

    char_result = at_inet4_ntop (inet4_ntop_tests[i].input, inet4_string,
				 sizeof (inet4_string));
    if (!char_result) {
      fprintf (stderr,
	       "FATAL: at_inet4_ntop() Failed to convert inet4 address %d.%d.%d.%d to string\n",
	       inet4_ntop_tests[i].input[0], inet4_ntop_tests[i].input[1],
	       inet4_ntop_tests[i].input[2], inet4_ntop_tests[i].input[3]);
      error_count++;
      continue;
    }

    if (strcmp (char_result, inet4_ntop_tests[i].result)) {
      fprintf (stderr,
	       "FATAL: at_inet4_ntop() Conversion of %d.%d.%d.%d did not match expected results '%s' != '%s'\n",
	       inet4_ntop_tests[i].input[0], inet4_ntop_tests[i].input[1],
	       inet4_ntop_tests[i].input[2], inet4_ntop_tests[i].input[3],
	       char_result, inet4_ntop_tests[i].result);
      error_count++;
      continue;
    }
  }

  /* test at_inet4_ntop(): pass a string that is too small */
  memset (inet4_addr, 255, 4);
  if ((char_result = at_inet4_ntop (inet4_addr, inet4_string, 15)) != NULL) {

    fprintf (stderr,
	     "FATAL: at_inet4_ntop() Allowed a string that was too small\n");
    error_count++;
  }

  /* size_t argument is unsigned
     if ((char_result = at_inet4_ntop(inet4_addr, inet4_string, -1)) != NULL) {
     fprintf (stderr, "FATAL: at_inet4_ntop() Allowed a string that was too small\n");
     error_count++;
     }
   */

  if ((char_result = at_inet4_ntop (inet4_addr, inet4_string, 0)) != NULL) {
    fprintf (stderr,
	     "FATAL: at_inet4_ntop() Allowed a string that was too small\n");
    error_count++;
  }

  /*

     These tests are for the most part just taking the examples from rfc2373
     -EZA

     (begin quote)

     FEDC:BA98:7654:3210:FEDC:BA98:7654:3210

     1080:0:0:0:8:800:200C:417A

     Note that it is not necessary to write the leading zeros in an
     individual field, but there must be at least one numeral in every
     field (except for the case described in 2.).

     2. Due to some methods of allocating certain styles of IPv6
     addresses, it will be common for addresses to contain long strings
     of zero bits.  In order to make writing addresses containing zero
     bits easier a special syntax is available to compress the zeros.
     The use of "::" indicates multiple groups of 16-bits of zeros.
     The "::" can only appear once in an address.  The "::" can also be
     used to compress the leading and/or trailing zeros in an address.

     For example the following addresses:

     1080:0:0:0:8:800:200C:417A  a unicast address
     FF01:0:0:0:0:0:0:101        a multicast address
     0:0:0:0:0:0:0:1             the loopback address
     0:0:0:0:0:0:0:0             the unspecified addresses

     may be represented as:

     1080::8:800:200C:417A       a unicast address
     FF01::101                   a multicast address
     ::1                         the loopback address
     ::                          the unspecified addresses
   */

  /* test at_inet6_pton() for good input values */
  num_tests =
    sizeof (inet6_pton_tests_good) / sizeof (struct inet6_pton_test_str);

  for (i = 0; i < num_tests; i++) {
    memset (inet6_addr, 0, sizeof (inet6_addr));;
    int8_result = at_inet6_pton (inet6_pton_tests_good[i].input, inet6_addr);
    if (!int8_result) {
      fprintf (stderr,
	       "FATAL: at_inet6_pton() Failed to convert inet6 address %s\n",
	       inet6_pton_tests_good[i].input);
      error_count++;
      continue;
    }

    if (memcmp (int8_result, inet6_pton_tests_good[i].result, 16)) {
      fprintf (stderr,
	       "FATAL: at_inet6_pton() Conversion of %s did not match expected results\n",
	       inet6_pton_tests_good[i].input);
      error_count++;
      continue;
    }
  }

  /* test at_inet6_pton() for bad input values */
  num_tests =
    sizeof (inet6_pton_tests_bad) / sizeof (struct inet6_pton_test_str);
  for (i = 0; i < num_tests; i++) {
    memset (inet6_addr, 0, sizeof (inet6_addr));;
    int8_result = at_inet6_pton (inet6_pton_tests_bad[i].input, inet6_addr);
    if (int8_result) {
      fprintf (stderr,
	       "FATAL: at_inet6_pton() Converted invalid inet6 address %s w/o error\n",
	       inet6_pton_tests_bad[i].input);
      error_count++;
      continue;
    }
  }

  /* test at_inet6_ntop() */
  num_tests = sizeof (inet6_ntop_tests) / sizeof (struct inet6_ntop_test_str);

  for (i = 0; i < num_tests; i++) {
    memset (inet6_string, 0, sizeof (inet6_string));

    char_result = at_inet6_ntop (inet6_ntop_tests[i].input, inet6_string,
				 sizeof (inet6_string));
    if (!char_result) {
      fprintf (stderr,
	       "FATAL: at_inet6_ntop() Failed to convert inet6 address %0X%0X:%0X%0X:%0X%0X:%0X%0X:%0X%0X:%0X%0X:%0X%0X:%0X%0X to string\n",
	       inet6_ntop_tests[i].input[0], inet6_ntop_tests[i].input[1],
	       inet6_ntop_tests[i].input[2], inet6_ntop_tests[i].input[3],
	       inet6_ntop_tests[i].input[4], inet6_ntop_tests[i].input[5],
	       inet6_ntop_tests[i].input[6], inet6_ntop_tests[i].input[7],
	       inet6_ntop_tests[i].input[8], inet6_ntop_tests[i].input[9],
	       inet6_ntop_tests[i].input[10], inet6_ntop_tests[i].input[11],
	       inet6_ntop_tests[i].input[12], inet6_ntop_tests[i].input[13],
	       inet6_ntop_tests[i].input[14], inet6_ntop_tests[i].input[15]);

      error_count++;
      continue;
    }

    if (strcmp (char_result, inet6_ntop_tests[i].result)) {
      fprintf (stderr,
	       "FATAL: at_inet6_ntop() Conversion of %0X%0X:%0X%0X:%0X%0X:%0X%0X:%0X%0X:%0X%0X:%0X%0X:%0X%0X did not match expected results '%s' != '%s'\n",
	       inet6_ntop_tests[i].input[0], inet6_ntop_tests[i].input[1],
	       inet6_ntop_tests[i].input[2], inet6_ntop_tests[i].input[3],
	       inet6_ntop_tests[i].input[4], inet6_ntop_tests[i].input[5],
	       inet6_ntop_tests[i].input[6], inet6_ntop_tests[i].input[7],
	       inet6_ntop_tests[i].input[8], inet6_ntop_tests[i].input[9],
	       inet6_ntop_tests[i].input[10], inet6_ntop_tests[i].input[11],
	       inet6_ntop_tests[i].input[12], inet6_ntop_tests[i].input[13],
	       inet6_ntop_tests[i].input[14], inet6_ntop_tests[i].input[15],
	       char_result, inet6_ntop_tests[i].result);

      error_count++;
      continue;
    }
  }

  /* test at_inet6_ntop(): pass a string that is too small */
  memset (inet6_addr, 255, 16);
  if ((char_result = at_inet6_ntop (inet6_addr, inet6_string, 39)) != NULL) {

    fprintf (stderr,
	     "FATAL: at_inet6_ntop() Allowed a string that was too small\n");
    error_count++;
  }

  /* size_t argument is unsigned
     if ((char_result = at_inet6_ntop(inet6_addr, inet6_string, -1))!= NULL) {

     fprintf (stderr, "FATAL: at_inet6_ntop() Allowed a string that was too small\n");
     error_count++;
     }
   */

  if ((char_result = at_inet6_ntop (inet6_addr, inet6_string, 0)) != NULL) {

    fprintf (stderr,
	     "FATAL: at_inet6_ntop() Allowed a string that was too small\n");
    error_count++;
  }

  if (error_count) {
    fprintf (stderr, "Test failed with %d errors\n", error_count);
    fflush (stderr);
    exit (1);
  }

  /* test passed! */
  return 0;

}				/* end main () */
