/* test_link.c - XXCOMMENTXX

   Dents - a threaded modular dns-server.
   Copyright (C) 1999  Dents Team  (http://www.dents.org)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version. This program is distributed in the
   hope that it will be useful, but WITHOUT ANY WARRANTY; without even
   the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Add system headers here */
#include <glib.h>

/* Add dents headers here */
#include "../include/common.h"

/* test_link - checks functions from link.c for consistency */

#include <stdarg.h>
#include "link.h"

#define NUMOPS 100		/* # of operations to perform on each list */

typedef struct
{
  llentry ll;
  int data;
}
teststruct;

int
main (int argc, char *argv[])
{
  int errorcount = 0;
  const unsigned int verbose = (argc > 1);

  /* Declare a separate list to hold entries            */
  linkedlist queuehead = LINK_INITIALIZER (teststruct, ll);

  /* Declare and initalize a pool for getting entries   */
  blocklist blocklisthead = BLOCK_INITIALIZER (teststruct, 10);

  /* for iterating through the lists */
  teststruct *mylink_ptr, *next_ptr;

  /* loop control variable */
  int i;

  /* Add some entries to the queue.  You have to allocate these
     yourself with malloc() or get memory from one of the blocklist
     routines.  */
  for (i = 100; i < 100 + NUMOPS; i++) {
    mylink_ptr = (teststruct *) malloc (sizeof (teststruct));
    if (NULL == mylink_ptr) {
      printf ("NULL == malloc(%d)\n", sizeof (teststruct));
      errorcount++;
    }
    mylink_ptr->data = i;
    link_add (&queuehead, &mylink_ptr->ll);
  }

  for (mylink_ptr = link_first (&queuehead), i = 100;
       mylink_ptr && i < 100 + NUMOPS;
       mylink_ptr = link_next (mylink_ptr, mylink_ptr->ll), i++) {
    if (verbose || (i != mylink_ptr->data)) {
      printf ("list contains %d, this should be %d\n", mylink_ptr->data, i);
      if (i != mylink_ptr->data) {
	errorcount++;
      }
    }
  }
  if ((i != 100 + NUMOPS) || mylink_ptr) {
    printf ("wrong amount of elements in new list\n");
    errorcount++;
  }

  /* delete some members from the list */
  for (i = 100, next_ptr = mylink_ptr = link_first (&queuehead);
       i < 100 + NUMOPS && next_ptr; i++, mylink_ptr = next_ptr) {
    next_ptr = link_next (mylink_ptr, mylink_ptr->ll);
    if (i % 2) {
      if (verbose || (i != mylink_ptr->data)) {
	printf ("tried to delete %d from list, deleted %d\n",
		mylink_ptr->data, i);
	if (i != mylink_ptr->data) {
	  errorcount++;
	}
      }
      link_delete (&queuehead, &mylink_ptr->ll);
      /* Help keep dents beautiful. Free your litter */
      free (mylink_ptr);
    }
  }

  /* print the list */
  for (i = 100, mylink_ptr = link_first (&queuehead);
       i < 100 + NUMOPS && mylink_ptr;
       i += 2, mylink_ptr = link_next (mylink_ptr, mylink_ptr->ll)) {
    if (verbose || (i != mylink_ptr->data)) {
      printf ("list contains %d, expected %d\n", mylink_ptr->data, i);
      if (i != mylink_ptr->data) {
	errorcount++;
      }
    }
  }
  if (mylink_ptr || (i != NUMOPS + 100)) {
    printf ("wrong amount of elements in partially deleted list\n");
    errorcount++;
  }

  /* Cleanup the entire list */
  while ((mylink_ptr = link_first (&queuehead))) {
    link_delete (&queuehead, &mylink_ptr->ll);
    free (mylink_ptr);
  }

  /* print the list */
  for (mylink_ptr = link_first (&queuehead);
       mylink_ptr; mylink_ptr = link_next (mylink_ptr, mylink_ptr->ll)) {
    printf ("listelement %d shouldn't have survived\n", mylink_ptr->data);
    errorcount++;
  }

  if (verbose) {
    printf ("End of linkedlist example (%d error(s))\n", errorcount);
  }

  /* Demonstration of the blocklist routines */

  /* Create a list and add NUMOPS elements to it */
  for (i = 0; i < NUMOPS; i++) {
    mylink_ptr = (teststruct *) block_allocandadd (&blocklisthead);
    mylink_ptr->data = i;
  }

  /* print the list */
  for (i = 0, mylink_ptr = link_first (&blocklisthead.used);
       i < NUMOPS && mylink_ptr;
       i++, mylink_ptr = link_next (mylink_ptr, mylink_ptr->ll)) {
    if (verbose || (i != mylink_ptr->data)) {
      printf ("blocklist contains %d, this should be %d\n", mylink_ptr->data,
	      i);
      if (i != mylink_ptr->data) {
	errorcount++;
      }
    }
  }
  if ((i != NUMOPS) || mylink_ptr) {
    printf ("wrong amount of elements in new blocklist\n");
    errorcount++;
  }

  /* delete half of the members from the list */
  for (i = 0, next_ptr = mylink_ptr = link_first (&blocklisthead.used);
       i < NUMOPS && next_ptr; i++, mylink_ptr = next_ptr) {
    next_ptr = link_next (mylink_ptr, mylink_ptr->ll);
    if (i % 2) {
      if (verbose || (i != mylink_ptr->data)) {
	printf ("tried to delete %d from blocklist, deleted %d\n",
		mylink_ptr->data, i);
	if (i != mylink_ptr->data) {
	  errorcount++;
	}
      }
      block_delete (&blocklisthead, &mylink_ptr->ll);
      /* Don't try to cleanup the pointer with free().  Dents recycles
         these pointers and since the pointers are allocated a chunk
         at a time, it'll likely cause a segfault.  */
    }
  }

  /* print the list */
  for (i = 0, mylink_ptr = link_first (&blocklisthead.used);
       i < NUMOPS && mylink_ptr;
       i += 2, mylink_ptr = link_next (mylink_ptr, mylink_ptr->ll)) {
    if (verbose || (i != mylink_ptr->data)) {
      printf ("blocklist contains %d, expected %d\n", mylink_ptr->data, i);
      if (i != mylink_ptr->data) {
	errorcount++;
      }
    }
  }
  if ((i != NUMOPS) || mylink_ptr) {
    printf ("wrong amount of elements in partially deleted blocklist\n");
    errorcount++;
  }

  /* Cleanup the entire list */
  while ((mylink_ptr = link_first (&blocklisthead.used))) {
    block_delete (&blocklisthead, &mylink_ptr->ll);
  }

  /* print the _empty_ list */
  for (mylink_ptr = link_first (&blocklisthead.used);
       mylink_ptr; mylink_ptr = link_next (mylink_ptr, mylink_ptr->ll)) {
    printf ("blockelement %d shouldn't have survived\n", mylink_ptr->data);
    errorcount++;
  }

  if (verbose) {
    printf ("End of blocklist example (%d error(s))\n", errorcount);
  }

  /* More sophisticated example.  Demonstration of the linkedlist
     together with blocklist routines to allocate memory */

  /* add some entries to the queue. Use the blockalloc */
  for (i = 200; i < 200 + NUMOPS; i++) {
    mylink_ptr = (teststruct *) block_allocandaddtolink
      (&blocklisthead, &queuehead);
    mylink_ptr->data = i;
  }

  /* print the list */
  for (i = 200, mylink_ptr = link_first (&queuehead);
       i < 200 + NUMOPS && mylink_ptr;
       i++, mylink_ptr = link_next (mylink_ptr, mylink_ptr->ll)) {
    if (verbose || (i != mylink_ptr->data)) {
      printf ("hybridlist contains %d, this should be %d\n", mylink_ptr->data,
	      i);
      if (i != mylink_ptr->data) {
	errorcount++;
      }
    }
  }
  if ((i != 200 + NUMOPS) || mylink_ptr) {
    printf ("wrong amount of elements in partially new hybridlist\n");
    errorcount++;
  }

  /* delete some members from the list */
  for (i = 200, next_ptr = mylink_ptr = link_first (&queuehead);
       next_ptr && i < 200 + NUMOPS; i++, mylink_ptr = next_ptr) {
    next_ptr = link_next (mylink_ptr, mylink_ptr->ll);
    if (i % 2) {
      if (verbose || (i != mylink_ptr->data)) {
	printf ("hybridlist contains %d, this should be %d\n",
		mylink_ptr->data, i);
	if (i != mylink_ptr->data) {
	  errorcount++;
	}
      }
      block_delete (&blocklisthead, &mylink_ptr->ll);
    }
  }

  /* print the list */
  for (i = 200, mylink_ptr = link_first (&queuehead);
       i < 200 + NUMOPS && mylink_ptr;
       i += 2, mylink_ptr = link_next (mylink_ptr, mylink_ptr->ll)) {
    if (verbose || (i != mylink_ptr->data)) {
      printf ("hybridlist contains %d, expected %d\n", mylink_ptr->data, i);
      if (i != mylink_ptr->data) {
	errorcount++;
      }
    }
  }
  if ((i != 200 + NUMOPS) || mylink_ptr) {
    printf
      ("wrong amount of elements in partially partially deleted hybridlist\n");
    errorcount++;
  }

  /* Cleanup the entire list */
  while ((mylink_ptr = link_first (&blocklisthead.used))) {
    block_delete (&blocklisthead, &mylink_ptr->ll);
  }

  /* print the list */
  for (mylink_ptr = link_first (&blocklisthead.used);
       mylink_ptr; mylink_ptr = link_next (mylink_ptr, mylink_ptr->ll)) {
    printf ("blockelement %d shouldn't have survived\n", mylink_ptr->data);
    errorcount++;
  }

  if (verbose) {
    printf ("End of hybrid example (%d error(s))\n", errorcount);
  }

  return (0 != errorcount);
}				/* end main () */
